/***************************************************************************
                          lutgenerator.cpp  -  description
                             -------------------
    begin                : Fri Sep 26 10:24:00 CEST 2014
    copyright            : (C) 2014 by Nicola Doebelin
    email                : ndoebelin@gmx.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "math.h"
#include "lutgenerator.h"
#include <QDebug>

#ifndef M_PI
    #define M_PI acos(-1.0)
#endif

LutGenerator::LutGenerator(int s, bool t)
{
    size = s;
    transparent = t;
}

/*
 * calculates various lookup tables.
 *
 * Note: in case of grayscale map, the lut will only be of size "size",
 * because the dmin = black, and dmax = white
 * all other maps will have size "2*size", because 0.0 = black, and one
 * full color range extens to the negative range, and one to the positive range
 */
QMap<QString, colorMaps::Lut> LutGenerator::getLuts()
{
    lutList.clear();

    // smooth
    lutList.insert("01 Grayscale", lutGrayScale(1, 0, 0));
    lutList.insert("02 Blue-Red", lutBlueRed());
    lutList.insert("03 Bright Glow", lutBrightGlow());
    lutList.insert("04 Dark Glow", lutDarkGlow());
    lutList.insert("05 False Colors", lutFalseColor(1, 0, 0));

    // contour
    lutList.insert("06 Gray Contour", lutGrayScale(16, 1, 0));
    lutList.insert("07 False Color Contour", lutFalseColor(16, 1, 0));

    // contour lines
    lutList.insert("08 Coarse Contour Lines", lutGrayScale(16, 2, 1));
    lutList.insert("09 Coarse Contour Lines Bold", lutGrayScale(16, 2, 2));
    lutList.insert("10 Fine Contour Lines", lutGrayScale(8, 2, 1));
    lutList.insert("11 Fine Contour Lines Bold", lutGrayScale(8, 2, 2));

    // contour color and lines
    lutList.insert("12 Gray with Contour Lines", lutGrayScale(16, 1, 1));
    lutList.insert("13 Gray with Contour Lines Bold", lutGrayScale(16, 1, 2));
    lutList.insert("14 False Color with Contour Lines", lutFalseColor(16, 1, 1));
    lutList.insert("15 False Color with Contour Lines Bold", lutFalseColor(16, 1, 2));

    // colored contour lines
    lutList.insert("16 Colored Contour Lines", lutFalseColor(16, 3, 1));
    lutList.insert("17 Colored Contour Lines Bold", lutFalseColor(16, 3, 2));

    return lutList;
}

colorMaps::Lut LutGenerator::lutGrayScale(int div, int type, int con)
{
    colorMaps::Lut lut("Grayscale", type, con);
    int sz = size / div;

    for (int i = 0; i < sz; ++i) {
        int g = int(255.0 * float(i)/float(sz));
        lut.data.append(qRgb(g, g, g));
    }

    return lut;
}

colorMaps::Lut LutGenerator::lutBlueRed()
{
    colorMaps::Lut lut("Blue-Red", 0, 0);

    // negative range
    for (int i = 0; i <= size; ++i) {
        float f = float(size - i)/float(size);
        int r = int(255.0 * f);
        int g = int(255.0 * f);
        int b = int(255.0 * sqrt(f));
        lut.data.append(qRgb(r, g, b));
    }

    // positive range
    for (int i = 1; i <= size; ++i) {
        float f = float(i)/float(size);
        int r = int(255.0 * sqrt(f));
        int g = int(255.0 * f);
        int b = int(255.0 * f);
        lut.data.append(qRgb(r, g, b));
    }

    return lut;
}

colorMaps::Lut LutGenerator::lutBrightGlow()
{
    colorMaps::Lut lut("Bright Glow", 0, 0);

    // negative range
    for (int i = 0; i <= size; ++i) {
        float f = float(size - i)/float(size);
        float r = 255.0 * 1.0 * f >= 255.0 ? 255.0 : 255.0 * 1.0 * f;
        float g = 255.0 * 2.0 * f >= 255.0 ? 255.0 : 255.0 * 2.0 * f;
        float b = 255.0 * 3.0 * f >= 255.0 ? 255.0 : 255.0 * 3.0 * f;
        lut.data.append(qRgb(int(r), int(g), int(b)));
    }

    // positive range
    for (int i = 1; i <= size; ++i) {
        float f = float(i)/float(size);
        float r = 255.0 * 3.0 * f >= 255.0 ? 255.0 : 255.0 * 3.0 * f;
        float g = 255.0 * 2.0 * f >= 255.0 ? 255.0 : 255.0 * 2.0 * f;
        float b = 255.0 * 1.0 * f >= 255.0 ? 255.0 : 255.0 * 1.0 * f;
        lut.data.append(qRgb(int(r), int(g), int(b)));
    }

    return lut;
}

colorMaps::Lut LutGenerator::lutDarkGlow()
{
    colorMaps::Lut lut("Dark Glow", 0, 0);

    // negative range
    for (int i = 0; i <= size; ++i) {
        float f = float(size - i)/float(size) * float(M_PI/2.0);
        // float r = 0.0;
        float g = 255.0 * (1.0 - cos(f));
        float b = 255.0 * sin(f);
        lut.data.append(qRgb(0, int(g), int(b)));
    }

    // positive range
    for (int i = 1; i <= size; ++i) {
        float f = float(i)/float(size) * float(M_PI/2.0);
        float r = 255.0 * sin(f);
        float g = 255.0 * pow(float(sin(f)), float(2.0));
        float b = 255.0 * (1.0 - cos(f));
        lut.data.append(qRgb(int(r), int(g), int(b)));
    }

    return lut;
}

colorMaps::Lut LutGenerator::lutFalseColor(int div, int type, int con)
{
    colorMaps::Lut lut("False Colors", type, con);
    int sz = size / div;

    int r = 0;
    int g = 0;
    int b = 0;

    for (int i = 0; i < sz; ++i) {
        float f = float(i) / float(sz - 1);
        r = int(f * 1024.0 - 512.0);
        g = f < 0.5 ? int(f * 1024.0) : int(1024.0 - f * 1024.0);
        b = int(512.0 - f * 1024.0);

        r = r > 255 ? 255 : r;
        g = g > 255 ? 255 : g;
        b = b > 255 ? 255 : b;

        r = r < 0 ? 0 : r;
        g = g < 0 ? 0 : g;
        b = b < 0 ? 0 : b;

        lut.data.append(qRgb(r, g, b));
    }

    return lut;
}
