/***************************************************************************
                          functions.h  -  description
                             -------------------
    begin                : Mon Jun 17, 2019
    copyright            : (C) 2019 by Nicola Doebelin
    email                : ndoebelin@gmx.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef FUNCTIONS_H
#define FUNCTIONS_H

#include "structs.h"
#include <QObject>
#include <QColor>

#if defined XRDIO
 #define XRDIO_EXPORT Q_DECL_EXPORT
#else
 #define XRDIO_EXPORT Q_DECL_IMPORT
#endif

namespace global {

class XRDIO_EXPORT Functions
{
public:
    Functions();

    static QColor colorToDarkMode(const QColor &);

    /*
     * expects d and wl in angstrom
     * returns t in degrees
     */
    static double dToTwoTheta(double d, double wl);

    /*
     * expects wl in angstrom and t in degrees
     * returns d in angstrom
     */
    static double twoThetaToD(double t, double wl);

    /*
     * a fuzzy compare method that is more tolerant to certain rounding errors
     */
    static bool fuzzyCompare(double, double);

    /*
     * returns true if the coordinate is special, i.e. a round number such
     * as 0, 1/2, 5/6 etc.
     */
    static bool isSpecialCoordinate(double);

    /*
     * returns the 2theta offset due to BGMN displacement parameters EPS1, EPS2, EPS3
     */
    static double angularCorrection(double tt, double eps1 = 0.0, double eps2 = 0.0, double eps3 = 0.0);

    /*
     * shifts fractional coordinates into the unit cell and replaces 0.3333 with 1/3 etc
     */
    static void normalizeCoordinates(double &x, double &y, double &z);

    /*
     * calculates the d spacing in the unit of the provided cell constants
     */
    static double dSpacing(double _a, double _b, double _c, double _al, double _be, double _ga, int _h, int _k, int _l);

    /*
     * calculates the volume of the unit cell in the dimensions of the provided cell contstants
     */
    static double cellVolume(double _a, double _b, double _c, double _al, double _be, double _ga);

    static double bisoFromUiso(double uiso);
    static double bisoFromBeta(double a, double b, double c,
                        double alpha, double beta, double gamma,
                        double b11, double b22, double b33,
                        double b12, double b13, double b23);
    static double bisoFromBaniso(double alpha, double beta, double gamma,
                          double b11, double b22, double b33,
                          double b12, double b13, double b23);
    static double bisoFromUaniso(double alpha, double beta, double gamma,
                          double u11, double u22, double u33,
                          double u12, double u13, double u23);

    static double wavelengthToEnergy(double w);
    static double energyToWavelength(double e);

    static QList<global::CharWaveLength> getAllWavelengths();

    static double lpFactor(double d, double wl);

    /*
     * return tick mark positions
     */
    static QVector<double> scaleAxis1(double, double, int);
    static QVector<double> scaleAxis2(double, double, int);
    static QVector<double> scaleAxisLog(double, double, int);
    static QVector<double> scaleAxisD(double, double, double, int, int);
};

} // end of namespace global

#endif // FUNCTIONS_H
