/***************************************************************************
                          philipsudfimport.cpp  -  description
                             -------------------
    begin                : Thu May 21, 2013
    copyright            : (C) 2013 by Nicola Doebelin
    email                : ndoebelin@gmx.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "philipsudfimport.h"

PhilipsUdfImport::PhilipsUdfImport(QObject *parent) :
    GenericImport(parent)
{
    extens << "udf";
    descr = QLatin1String("Philips UDF scan");
}

bool PhilipsUdfImport::isSupported(const QByteArray &ba)
{
    // check if the first line is of format "param,value,/"
    QString line = QString(ba).left(QString(ba).indexOf("\n")).trimmed();

    QRegExp rx("^(?:[\\w\\s]+,)+/$");
    if (rx.indexIn(line) >= 0) {
        return true;
    }

    return false;
}

int PhilipsUdfImport::load( const QString &file, QVector<Scan> &scanHeap, bool minimal )
{
    Q_UNUSED(minimal);

    qDebug() << QString("PhilipsUdfImport::load(): Loading file %1").arg(file);
    QStringList content(BgmnFileIO::readTextFileLines(file));

    int rs = content.indexOf(QRegExp("^RawScan.*"));

    if (rs < 0) {
        qDebug() << QString("PhilipsUdfImport::load(): No valid data found in file %1").arg(file);
        return -1;
    }

    QVector<double> vec_a;       // angle
    QVector<double> vec_i;       // intensity

    Scan scan(file, QColor(), 1);
    scan.setSourceFileName(file);

    QRegExp rxSampleIdent("(^SampleIdent,\\s*)(.*)(,/$)");
    QRegExp rxComment("(^Title1,\\s*)(.*)(,/$)");
    QRegExp rxWaveLength("(^LabdaAlpha1,\\s*)(.*)(,/$)");
    QRegExp rxNamedWaveLength("(^Anode,\\s*)(.*)(,/$)");
    QRegExp rxDataAngleRange("(^DataAngleRange,\\s*)(\\d+\\.?\\d*)(,\\s*)(\\d+\\.?\\d*)(,/$)");
    QRegExp rxScanStepSize("(^ScanStepSize,\\s*)(\\d+\\.?\\d*)(,/$)");
    QRegExp rxScanStepTime("(^ScanStepTime,\\s*)(\\d+\\.?\\d*)(,/$)");

    if (content.indexOf(rxSampleIdent) > -1) {
        scan.setName(rxSampleIdent.cap(2));
    }

    if (content.indexOf(rxComment) > -1) {
        scan.setComment(rxComment.cap(2));
    }

    if (content.indexOf(rxNamedWaveLength) > -1) {
        scan.setNamedWaveLength(rxNamedWaveLength.cap(2));
    }

    if (content.indexOf(rxWaveLength) > -1) {
        scan.setWaveLength(rxWaveLength.cap(2).toFloat());
    }

    bool ok;

    // extract start angle
    double ang;
    if (content.indexOf(rxDataAngleRange) > -1) {
        ang = rxDataAngleRange.cap(2).toDouble(&ok);
    } else {
        ok = false;
    }

    // check if start angle was extracted correctly
    if (!ok) {
        qDebug() << QString("PhilipsUdfImport::load(): Could not find start angle in file %1").arg(file);
        return -1;
    }

    // extract step size
    double stepSize = 0.0;
    if (content.indexOf(rxScanStepSize) > -1) {
        stepSize = rxScanStepSize.cap(2).toDouble(&ok);
    } else {
        ok = false;
    }

    // extract step size
    double stepTime = -1.0;
    if (content.indexOf(rxScanStepTime) > -1) {
        stepTime = rxScanStepTime.cap(2).toDouble(&ok);
    }



    // check if step size was extracted correctly
    if (!ok) {
        qDebug() << QString("PhilipsUdfImport::load(): Could not find step size in file %1").arg(file);
        return -1;
    }

    QRegExp sep(",\\s+");
    ++rs;

    for (int k = rs; k < content.size(); ++k) {
        QStringList val = content.at(k).split(sep);

        for (int j = 0; j < val.size(); ++j) {
            bool ok;
            double f = val.at(j).toDouble(&ok);

            if (ok) {
                vec_a.push_back(ang);
                vec_i.push_back(f);
                ang += stepSize;
            }
        }
    }

    scan.setDataAng(vec_a);
    scan.setDataInt(vec_i);
    scan.setTimePerStep(stepTime);
    scan.setTypes(Scan::XY | Scan::MEASURED);
    scanHeap.push_back(scan);

    return scanHeap.size();
}
