/***************************************************************************
                          reynoldspltimport.cpp  -  description
                             -------------------
    begin                : Thu May 21, 2013
    copyright            : (C) 2013 by Nicola Doebelin
    email                : ndoebelin@gmx.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "reynoldspltimport.h"

ReynoldsPltImport::ReynoldsPltImport(QObject *parent) :
    GenericImport(parent)
{
    extens << "plt";
    descr = QLatin1String("Reynolds PLT data");
}

bool ReynoldsPltImport::isSupported(const QByteArray &ba)
{
    QStringList header = QString(ba).split(QRegularExpression("\\r\\n?|\\n"));

    // we need at least 10 lines to reach the data block
    if (header.size() < 10) {
        return false;
    }

    int h = header.at(5).toInt() + 10; // line 5 describes the number of additional header lines

    // we want enough lines to read some data lines
    if (header.size() <= h) {
        return false;
    }

    QRegExp rx("^\\s+\\d+\\.\\d+\\s*$");

    for (int i = h; i < header.size(); ++i) {
        int t = rx.indexIn(header.at(i));
        if (t < 0) {
            return false;
        }
    }

    return true;
}

int ReynoldsPltImport::load( const QString &file, QVector<Scan> &scanHeap, bool minimal )
{
    Q_UNUSED(minimal);

    qDebug() << QString("ReynoldsPltImport::load(): Loading file %1").arg(file);
    QStringList content(BgmnFileIO::readTextFileLines(file));

    if (content.size() < 11) {
        qDebug() << QString("ReynoldsPltImport::load(): No valid data found in file %1").arg(file);
        return -1;
    }

    QVector<double> vec_a;       // angle
    QVector<double> vec_i;       // intensity

    Scan scan(content.at(0), QColor(), 1);
    scan.setSourceFileName(file);
    scan.setComment(QString("%1, %2").arg(content.at(1)).arg(content.at(2)));
    scan.setWaveLength(content.at(3).toDouble());

    bool ok = false;
    int auxLines = content.at(5).toInt(&ok);
    if (!ok) {
        auxLines = 0;
    }

    double startAng = content.at(7 + auxLines).toDouble();
    double endAng = content.at(8 + auxLines).toDouble();
    double stepSize = content.at(9 + auxLines).toDouble();

    Q_UNUSED(endAng);

    double ang = startAng;

    for (int j = 10 + auxLines; j < content.size(); ++j) {
        vec_a.push_back(ang);
        vec_i.push_back(content.at(j).toDouble());
        ang += stepSize;
    }

    scan.setDataAng(vec_a);
    scan.setDataInt(vec_i);
    scan.setTypes(Scan::XY | Scan::MEASURED);
    scanHeap.push_back(scan);

    return scanHeap.size();
}
