/***************************************************************************
                          seifertvalimport.cpp  -  description
                             -------------------
    begin                : Thu Mar 14, 2013
    copyright            : (C) 2013 by Nicola Doebelin
    email                : ndoebelin@gmx.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "seifertvalimport.h"

SeifertValImport::SeifertValImport(QObject *parent) :
    GenericImport(parent)
{
    extens << "val";
    descr = QLatin1String("Seifert/FPM VAL scan");

    nMode = 0;
    nScans = 0;
    nWavelength = "";
}

bool SeifertValImport::isSupported(const QByteArray &ba)
{
    if (QString(ba).left(7) == "FilePar") {
        return true;
    }

    return false;
}

int SeifertValImport::load( const QString &file, QVector<Scan> &scanHeap, bool minimal )
{
    Q_UNUSED(minimal);

    qDebug() << QString("SeifertValImport::load(): Loading file %1").arg(file);

    QFile ff(file);

    if (!ff.open(QIODevice::ReadOnly | QIODevice::Text)) {
        qDebug() << QString("SeifertValImport::load(): Could not read file %1").arg(file);
        return -1;
    }

    QStringList content(QString::fromLatin1(ff.readAll()).split(QRegularExpression("\\r\\n|\\n")));
    ff.close();

    if (content.size() < 11) {
        qDebug() << QString("SeifertValImport::load(): No valid data found in file %1").arg(file);
        return -1;
    }

    nMode = content.at(3).toInt();
    nWavelength = content.at(6);
    nScans = content.at(10).toInt();

    QRegularExpression rxBerPar("^BerPar\\d+$");
    QRegularExpression rxIntens("^Intens\\d+$");
    QRegularExpression rxTps("^[TC]$");

    QList<int> beginningBerPar;
    QList<int> beginningIntens;
    QList<int> beginningTps;
    QStringList typeTps;

    // find the indexes of all "BerPar[nn]" lines
    int idx = content.indexOf(rxBerPar, 0);

    while (idx > -1) {
        beginningBerPar.append(idx);
        idx = content.indexOf(rxBerPar, idx+1);
    }

    // find the indexes of all "Intens[nn]" lines
    idx = content.indexOf(rxIntens, 0);

    while (idx > -1) {
        beginningIntens.append(idx);
        idx = content.indexOf(rxIntens, idx+1);
    }

    // find the indexes of all "T" or "C" lines ("Vorwahlart")
    idx = content.indexOf(rxTps, 0);

    while (idx > -1) {
        beginningTps.append(idx);
        typeTps.append(content.at(idx));
        idx = content.indexOf(rxTps, idx+1);
    }

    // the number of scans we can actually read is the lowest of the following 3
    // (in theory they should be identical, but exceeding a list would cause a crash,
    // so better make sure we only iterate over the shortest list)
    int n = qMin(beginningIntens.size(), beginningBerPar.size());

    for (int i = 0; i < n; ++i) {
        // we will count a fixed number from here, so check if the stringlist is long enough
        if (content.size() <= beginningBerPar.at(i) + 11) {
            qDebug() << QString("SeifertValImport::load(): Incomplete scan found in file %1").arg(file);
            break;
        }

        double tps = -1.0;

        double startAng = content.at(beginningBerPar.at(i) + 2).toDouble();
        double endAng = content.at(beginningBerPar.at(i) + 3).toDouble();
        double stepSize = content.at(beginningBerPar.at(i) + 4).toDouble();
        int points = content.at(beginningBerPar.at(i) + 8).toInt();

        if (i < qMin(beginningTps.size(), typeTps.size())) {
            if (typeTps.at(i) == "T") {
                tps = content.at(beginningTps.at(i) + 1).toDouble();
            }

            // not sure what "C" means, but BGMN takes the intensity as read from the file
            if (typeTps.at(i) == "C") {
                tps = -1.0;
            }
        }

        if (startAng * endAng * stepSize * double(points) == 0.0) {
            qDebug() << QString("SeifertValImport::load(): Invalid values found in file %1").arg(file);
            break;
        }

        if (content.size() < beginningIntens.at(i) + points) {
            qDebug() << QString("SeifertValImport::load(): Invalid number of values found in file %1").arg(file);
            break;
        }

        QVector<double> vec_a;       // angle
        QVector<double> vec_i;       // intensity

        double ang = startAng;

        for (int j = beginningIntens.at(i) + 1; j < points; ++j) {
            if (tps > 0.0) vec_i.push_back(content.at(j).toDouble() * tps);
            else           vec_i.push_back(content.at(j).toDouble());
            vec_a.push_back(ang);
            ang += stepSize;
        }

        Scan scan(content.at(1), QColor(), 1);
        scan.setSourceFileName(file);
        scan.setDataAng(vec_a);
        scan.setDataInt(vec_i);
        scan.setTimePerStep(tps);
        scan.setNamedWaveLength(nWavelength);
        scan.setTypes(Scan::XY | Scan::MEASURED);
        scanHeap.push_back(scan);
    }

    return n;
}
