/***************************************************************************
                          bgmnparparser.cpp  -  description
                             -------------------
    begin                : Oct 26 18:20:07 CEST 2013
    copyright            : (C) 2013 by Nicola Doebelin
    email                : ndoebelin@gmx.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "bgmnparparser.h"
#include "bgmnfileio.h"
#include <math.h>
#include <QFile>
#include <QTextStream>
#include <QtConcurrent>
#include <QDebug>

#ifndef M_PI
    #define M_PI acos(-1.0)
#endif

BgmnParParser::BgmnParParser()
{
    settings = SettingsManager::getInstance();
}

BgmnParParser::BgmnParParser(const QString &s, bool *ok)
{
    settings = SettingsManager::getInstance();
    *ok = load(s);
}

bool BgmnParParser::load(const QString &s)
{
    if (!QFile::exists(s)) {
        qDebug() << QString("BgmnParParser::load(): File %1 doesnt exist").arg(s);
        return false;
    }

    QList<QByteArray> baLst = BgmnFileIO::readBinaryFileLines(s);
    if (baLst.size() < 2) return false;

    QList<QByteArray>::const_iterator itStart = baLst.constBegin();
    QList<QByteArray>::const_iterator itEnd   = baLst.constEnd();
    ++itStart; // skip title line
    --itEnd;   // skip background coefficients

    QFuture<QMap<QString, QVector<Hkl> > > f = QtConcurrent::mappedReduced(itStart, itEnd, parseDataLine, sortByPhase, QtConcurrent::OrderedReduce | QtConcurrent::SequentialReduce);
    f.waitForFinished();
    _data = f.result();

    return (_data.size() > 0);
}

QVector<Hkl> BgmnParParser::parseDataLine(const QByteArray &l)
{
    if (l.size() < 33) return QVector<Hkl>();

    // hard coded positions
    double intensity = l.mid(2, 14).trimmed().toDouble();
    double dinverted = l.mid(17, 10).trimmed().toDouble();
    double xpos      = qFuzzyIsNull(dinverted) ? 0.0 : 1.0 / dinverted;

    double b1        = 0.0;
    double b2        = 0.0;
    double gsum      = 1.0;
    QString phase;

    const QRegularExpression rxProfile("([234]).{27}([\\d\\.]+)?\\s+([\\d\\.]+)?");
    const QRegularExpression rxPhase("(?:GSUM=([\\d\\.]+)\\s+)?(?:PHASE=([^ ]+))");
    const QRegularExpression rxHkl("H=\\d+\\s+(?:TEXTUR=([\\d\\.]+)\\s+)?(-?\\d+ -?\\d+ -?\\d+)");

    QRegularExpressionMatch rm = rxProfile.match(l);
    if (rm.hasMatch()) {
        int rp = rm.captured(1).toInt();
        if (rp > 2) b1 = rm.captured(2).toDouble();
        if (rp > 3) b2 = rm.captured(3).toDouble();
    }

    QVector<Hkl> vec;
    int idx = 27;
    rm = rxPhase.match(l, idx);

    if (rm.hasMatch()) {
        if (!rm.captured(1).isEmpty()) gsum = rm.captured(1).toDouble();
        phase = rm.captured(2);
        idx = rm.capturedEnd(0) + 1;
    }

    rm = rxHkl.match(l, idx);

    while (rm.hasMatch()) {
        double tex = rm.captured(1).isEmpty() ? 1.0 : rm.captured(1).toDouble();
        QString hkl = rm.captured(2);

        vec.append(Hkl(xpos, hkl, 0, phase, QColor(), intensity * gsum, tex, 1, b1, b2));

        idx = rm.capturedEnd(0) + 1;
        rm = rxHkl.match(l, idx);
    }

    return vec;
}

void BgmnParParser::sortByPhase(QMap<QString, QVector<Hkl> > &map, const QVector<Hkl> &vhkl)
{
    for (int i = 0; i < vhkl.size(); ++i) {
        if (!vhkl.at(i).isEmpty()) map[vhkl.at(i).phase()].append(vhkl.at(i));
    }
}

QVector<Hkl> BgmnParParser::getReflections(const QString &pname)
{
    if (pname.isEmpty()) {
        QVector<Hkl> vec;
        QMapIterator<QString, QVector<Hkl> > it(_data);

        while (it.hasNext()) {
            it.next();
            vec.append(it.value());
        }

        return vec;
    }

    return _data.value(pname);
}

const QMap<QString, QVector<Hkl> > * BgmnParParser::reflections()
{
    return &_data;
}

QVector<double> BgmnParParser::getBackgroundCoefficients(const QString &f)
{
    QVector<double> vec;
    if (!QFile::exists(f)) return vec;

    QByteArray ba;
    BgmnFileIO::readBinaryFile(f, ba);

    int idxLast = ba.lastIndexOf('\n');

    if (idxLast < 0) return vec;

    QList<QByteArray> lastLine = ba.mid(idxLast).split(' ');

    for (int i = 1; i < lastLine.size(); ++i) {
        vec.append(lastLine.at(i).toDouble());
    }

    return vec;
}
