/***************************************************************************
                          savparser.cpp  -  description
                             -------------------
    begin                : Fri Feb 03 11:00:00 CEST 2012
    copyright            : (C) 2012 by Nicola Doebelin
    email                : ndoebelin@gmx.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <QTextStream>
#include <QDir>
#include <QFile>
#include <QFileInfo>
#include <QRegularExpression>
#include <QThread>
#include <QDebug>

#include "bgmnsavparser.h"
#include "bgmnstrparser.h"
#include "bgmnfileio.h"

/*
 * receive the content as a string. the name and location of the
 * control file name must also be known
 */
BgmnSavParser::BgmnSavParser(const QString &str, const QString &name)
{
    settings = SettingsManager::getInstance();

    qprefixRel = settings->value("bgmnProject/quantGoalRelPrefix", "Qrel").toString();
    sumvarRel  = settings->value("bgmnProject/quantSumRelPrefix",  "sumrel").toString();
    qprefixAbs = settings->value("bgmnProject/quantGoalAbsPrefix", "Qabs").toString();
    sumvarAbs  = settings->value("bgmnProject/quantSumAbsPrefix",  "sumabs").toString();

    fileName = name;
    setContent(str);
}

/*
 * read the content from file "file"
 */
BgmnSavParser::BgmnSavParser(const QString &file, bool *ok)
{
    settings = SettingsManager::getInstance();

    qprefixRel = settings->value("bgmnProject/quantGoalRelPrefix", "Qrel").toString();
    sumvarRel  = settings->value("bgmnProject/quantSumRelPrefix",  "sumrel").toString();
    qprefixAbs = settings->value("bgmnProject/quantGoalAbsPrefix", "Qabs").toString();
    sumvarAbs  = settings->value("bgmnProject/quantSumAbsPrefix",  "sumabs").toString();

    bool b = loadFile(file);
    if (ok) *ok = b;
}

bool BgmnSavParser::loadFile(const QString &file)
{
    fileName = file;
    QString s =BgmnFileIO::readTextFile(file);

    if (!s.isEmpty()) {
        setContent(s);
        return true;
    }

    return false;
}

void BgmnSavParser::setContent(const QString &s)
{
    rxFileNames = "([^$\\n\\r]*)";
    content = s.split(QRegularExpression("\\r\\n?|\\n"));
}

void BgmnSavParser::setSampleId(const QString &s)
{
    for (int i = 0; i < content.size(); ++i) {
        if (content.at(i).contains(QRegularExpression("SampleID[:=]?.*"))) {
            content[i] = QString("% SampleID: %1").arg(s);
            return;
        }
    }

    content.prepend(QString("% SampleID: %1").arg(s));
}

void BgmnSavParser::setDeviceFile(const QString &s)
{
    if (content.isEmpty()) {
        qDebug() << "SavParser::setDeviceFile failed: no savFile loaded";
        return;
    }

    QFileInfo fi(s);
    QRegularExpression rx(QString("\\bVERZERR=%1").arg(rxFileNames));

    int i = content.indexOf(rx, 0);

    if ((i >= 0) && (i < content.size())) {
        content[i] = "VERZERR=" + fi.fileName();
    }
}

void BgmnSavParser::addAmorphous(const QString &s)
{
    addStructureFiles(QStringList(s), false, false);
}

/*
 * Rewrites the STRUC[n] and GOAL[n] sections
 */
void BgmnSavParser::addStructureFiles(const QStringList &s, bool goals, bool multi)
{
    if (content.isEmpty()) {
        qDebug() << "BgmnSavParser::addStructureFiles(): failed, no savFile loaded";
        return;
    }

    // read existing STRUC files
    QStringList strFiles = getPhases();

    // the stringlist s contains full path plus file names, but here we only
    // want to append the file name without the path
    for (int i = 0; i < s.size(); ++i) {
        QFileInfo fi(s.at(i));
        strFiles.append(fi.fileName());
    }

    if (!multi) {
        strFiles.removeDuplicates();
    }

    // now clear all STRUC lines from the file, store the position where they
    // were located
    int l = clearPhases();

    // no previous STRUC lines? Try to locate "% Phases", else just append
    // the phases to the end of the file content
    if (l < 0) {
        if (content.contains("% Phases")) {
            l = content.indexOf("% Phases") + 1;
        } else {
            l = content.size();
        }
    }

    // write back the STRUC files
    for (int i = 0; i < strFiles.size(); ++i) {
        content.insert(l, QString("STRUC[%1]=%2").arg(i + 1).arg(strFiles.at(i)));
        ++l;
    }

    // now write the new GOAL[n] section
    if (goals) {
        writeGoals();
    }
}

/*
 * Accepts the *.str file name and removes the corresponding STRUC[n]=... entry
 * returns the number of removed structure files
 *
 * s = structure file name as specified in STRUC[]=
 * d = project directory
 * goals = manage goals flag
 *
 * also removes all STRUCOUT[n], SimpleSTRUCOUT[n], RESOUT[n], FCFOUT[n],
 * and PDBOUT[n] files
 *
 * this function does not delete the STR files. must be done
 * by the caller.
 */
int BgmnSavParser::removeStructureFile(const QString &s, const QString &d, bool goals)
{
    if (goals) {
        QString istd = internalStandard();

        if (!istd.isEmpty()) {
            BgmnStrParser strParser(d + "/" + s);

            if (strParser.getQuantGoal() == istd) {
                // if the phase to be removed is also declared as
                // internal standard, switch back to no internal standard
                unsetInternalStandard();
            }
        }
    }

    QRegularExpression rxPhase(QString("STRUC\\[(\\d+)\\]=%1").arg(s));
    QRegularExpressionMatch m;
    int indexToRemove = -1;

    // find the line at which rxStruc matches
    int i = content.indexOf(rxPhase, 0);

    // no match? return
    if (i < 0) return 0;

    // match the regular expression
    m = rxPhase.match(content.at(i));

    // extract n
    if (m.hasMatch()) {
        indexToRemove = m.captured(1).toInt();
    } else {
        return 0;
    }

    QRegularExpression rxStruc(QString("STRUC\\[%1\\]=.+").arg(indexToRemove));
    QRegularExpression rxStrOut(QString("STRUCOUT\\[%1\\]=.+").arg(indexToRemove));
    QRegularExpression rxSstrOut(QString("SimpleSTRUCOUT\\[%1\\]=.+").arg(indexToRemove));
    QRegularExpression rxResOut(QString("RESOUT\\[%1\\]=.+").arg(indexToRemove));
    QRegularExpression rxFcfOut(QString("FCFOUT\\[%1\\]=.+").arg(indexToRemove));
    QRegularExpression rxPdbOut(QString("PDBOUT\\[%1\\]=.+").arg(indexToRemove));

    // remove all lines with n=indexToRemove
    if (content.indexOf(rxStruc,   0) >= 0) content.removeAt(content.indexOf(rxStruc));
    if (content.indexOf(rxStrOut,  0) >= 0) content.removeAt(content.indexOf(rxStrOut));
    if (content.indexOf(rxSstrOut, 0) >= 0) content.removeAt(content.indexOf(rxSstrOut));
    if (content.indexOf(rxResOut,  0) >= 0) content.removeAt(content.indexOf(rxResOut));
    if (content.indexOf(rxFcfOut,  0) >= 0) content.removeAt(content.indexOf(rxFcfOut));
    if (content.indexOf(rxPdbOut,  0) >= 0) content.removeAt(content.indexOf(rxPdbOut));

    QRegularExpression rxLut("STRUC\\[(\\d+)\\]=.+");
    QList<int> lut;

    for (int i = 0; i < content.size(); ++i) {
        m = rxLut.match(content.at(i));
        if (m.hasMatch()) lut.append(m.captured(1).toInt());
    }

    // renumber all entries with [n]
    for (int i = 0; i < lut.size(); ++i) {
        content.replaceInStrings(QString("STRUC[%1]=").arg(lut.at(i)), QString("STRUC[%1]=").arg(i+1), Qt::CaseSensitive);
        content.replaceInStrings(QString("STRUCOUT[%1]=").arg(lut.at(i)), QString("STRUCOUT[%1]=").arg(i+1), Qt::CaseSensitive);
        content.replaceInStrings(QString("SimpleSTRUCOUT[%1]=").arg(lut.at(i)), QString("SimpleSTRUCOUT[%1]=").arg(i+1), Qt::CaseSensitive);
        content.replaceInStrings(QString("RESOUT[%1]=").arg(lut.at(i)), QString("RESOUT[%1]=").arg(i+1), Qt::CaseSensitive);
        content.replaceInStrings(QString("FCFOUT[%1]=").arg(lut.at(i)), QString("FCFOUT[%1]=").arg(i+1), Qt::CaseSensitive);
        content.replaceInStrings(QString("PDBOUT[%1]=").arg(lut.at(i)), QString("PDBOUT[%1]=").arg(i+1), Qt::CaseSensitive);
    }

    // update the goals if requested
    if (goals) {
        writeGoals();
    }

    return 1;
}

/*
 * returns the sampleID string
 */
QString BgmnSavParser::sampleId() const
{
    QRegularExpression rx("SampleID[:=]?\\s*(.*)");
    QRegularExpressionMatch rm;

    for (int i = 0; i < content.size(); ++i) {
        if (content.at(i).contains(rx, &rm)) return rm.captured(1).trimmed();
    }

    return QString();
}

/*
 * returns the LAMBDA= string
 */
QString BgmnSavParser::lamFile() const
{
    QRegularExpression rx("LAMBDA=(\\S+)");
    QRegularExpressionMatch rm;

    for (int i = 0; i < content.size(); ++i) {
        if (content.at(i).contains(rx, &rm)) return rm.captured(1);
    }

    return QString();
}

/*
 * returns the WMIN value as a double
 */
double BgmnSavParser::getWmin() const
{
    QRegularExpression rx("^\\s*(?!%)\\s*WMIN=(\\d\\.?\\d*)");
    QRegularExpressionMatch rm;

    for (int i = 0; i < content.size(); ++i) {
        if (content.at(i).contains(rx, &rm)) return rm.captured(1).toDouble();
    }

    return -1.0;
}

/*
 * returns the WMAX value as a double
 */
double BgmnSavParser::getWmax() const
{
    QRegularExpression rx("^\\s*(?!%)\\s*WMAX=(\\d\\.?\\d*)");
    QRegularExpressionMatch rm;

    for (int i = 0; i < content.size(); ++i) {
        if (content.at(i).contains(rx, &rm)) return rm.captured(1).toDouble();
    }

    return -1.0;
}

/*
 * returns the SYNCHROTRON value as a double
 */
double BgmnSavParser::getSynchrotron() const
{
    QRegularExpression rx("^\\s*(?!%)\\s*SYNCHROTRON=(\\d\\.?\\d*)");
    QRegularExpressionMatch rm;

    for (int i = 0; i < content.size(); ++i) {
        if (content.at(i).contains(rx, &rm)) return rm.captured(1).toDouble();
    }

    return -1.0;
}
/*
 * returns all STRUC[n] files.
 */
QStringList BgmnSavParser::getPhases() const
{
    return getAllMatches(QRegularExpression("STRUC\\[\\d+\\]=(.+\\.(str|STR))"));
}

QStringList BgmnSavParser::getStrucOut() const
{
    return getAllMatches(QRegularExpression("STRUCOUT\\[\\d+\\]=(.+\\.(str|STR))"));
}

QStringList BgmnSavParser::getSimpleStrucOut() const
{
    return getAllMatches(QRegularExpression("SimpleSTRUCOUT\\[\\d+\\]=(.+\\.(str|STR))"));
}

QStringList BgmnSavParser::getResOut() const
{
    return getAllMatches(QRegularExpression("RESOUT\\[\\d+\\]=(.+\\.(res|RES))"));
}

QStringList BgmnSavParser::getFcfOut() const
{
    return getAllMatches(QRegularExpression("FCFOUT\\[\\d+\\]=(.+\\.(fcf|FCF))"));
}

QStringList BgmnSavParser::getPdbOut() const
{
    return getAllMatches(QRegularExpression("PDBOUT\\[\\d+\\]=(.+\\.(pdb|PDB))"));
}

/*
 * returns all lines matching rx. Returned lines are trimmed.
 */
QStringList BgmnSavParser::getAllMatches(const QRegularExpression &rx, int m) const
{
    QRegularExpressionMatch match;
    QStringList lst;

    for (int i = 0; i < content.size(); ++i) {
        match = rx.match(content.at(i));
        // we found a matching line
        if (match.hasMatch()) {
            lst.append(match.captured(m).trimmed());
        }
    }

    return lst;
}


/*
 * returns the number of lines containing STRUC[n]=s
 */
int BgmnSavParser::hasPhase(const QString &s) const
{
    // convert all strings ::toLower(), to remain case insensitive
    QRegularExpression rx(QString("struc\\[\\d+\\]=%1").arg(s.toLower()));
    QRegularExpressionMatch match;
    int n = 0;

    for (int i = 0; i < content.size(); ++i) {
        match = rx.match(content.at(i).toLower());
        // we found a STRUC[n]= line
        if (match.hasMatch()) {
            ++n;
        }
    }

    return n;
}

/*
 * removes all STRUC[n]=xxx.str lines from the content and returns the line
 * of the first entry. does not remove STRUCOUT, SimpleSTRUCOUT etc.
 */
int BgmnSavParser::clearPhases()
{
    QRegularExpression rxStruc("STRUC\\[\\d+\\]=.+\\.(?:str|STR)");

    int i = content.indexOf(rxStruc);

    while (content.indexOf(rxStruc) >= 0) {
        content.removeAt(content.indexOf(rxStruc));
    }

    return i;
}

/*
 * writes the file to disk using 'fileName'
 */
bool BgmnSavParser::saveFile()
{
    return BgmnFileIO::writeTextFile(fileName, content.join("\n"));
}

/*
 * returns the entire content as a string
 */
QString BgmnSavParser::getContent() const
{
    return content.join("\n");
}

/*
 * constructs a block of GOAL phase quantities and appends it to content.
 * Removes the previous GOAL block from the sav file content first.
 */
void BgmnSavParser::writeGoals()
{
    bool addBothNewFormat = settings->value("bgmnProject/quantGoalsBoth", false).toBool();
    int formatVersion = settings->value("bgmnProject/quantGoalFormat", 0).toInt();

    qprefixRel = settings->value("bgmnProject/quantGoalRelPrefix", "Q").toString();
    sumvarRel  = settings->value("bgmnProject/quantSumRelPrefix",  "sum").toString();
    qprefixAbs = settings->value("bgmnProject/quantGoalAbsPrefix", "Qabs").toString();
    sumvarAbs  = settings->value("bgmnProject/quantSumAbsPrefix",  "sumabs").toString();

    // buffer parameters we need later on
    QStringList phases(getAllPhaseNames());
    QStringList nonQuantGoals(getNonQuantificationGoals());
    QStringList listd(getAllMatches(QRegularExpression("^ISTD=(\\S+)")));
    QStringList listdq(getAllMatches(QRegularExpression("^ISTDQ=(\\d+\\.?\\d*)")));

    QString istd  = listd.isEmpty()  ? QString() : listd.first();
    QString istdq = listdq.isEmpty() ? QString() : listdq.first();
    bool setIstd = !(istd.isEmpty() || istdq.isEmpty());

    removeSumsAndGoals();

    QStringList goalblock;
    int n = 0;

    if (formatVersion == 0) {
        if (setIstd) n = createQuantGoalAbsOldFormat(phases, istd, goalblock);
        else         n = createQuantGoalRelOldFormat(phases, goalblock);
    }

    if (formatVersion == 1) {
        if (setIstd) n = createQuantGoalAbsNewFormat(phases, istd, addBothNewFormat, goalblock);
        else         n = createQuantGoalRelNewFormat(phases, goalblock);
    }

    ++n;

    // append the non-managed goals
    for (int i = 0; i < nonQuantGoals.size(); ++i) {
        goalblock.append(QString("GOAL[%1]=%2").arg(n).arg(nonQuantGoals.at(i)));
        ++n;
    }

    content.append(QString());
    content.append(goalblock);
}

void BgmnSavParser::removeSumsAndGoals()
{
    // remove all 'sum=', 'sumrel=', and 'sumabs=' lines
    removeAllLines(QRegularExpression("^sum=\\S+"));
    removeAllLines(QRegularExpression(QString("^%1=\\S+").arg(sumvarRel)));
    removeAllLines(QRegularExpression(QString("^%1=\\S+").arg(sumvarAbs)));

    // remove all 'QPhase=', 'QrelPhase=', and 'QabsPhase=' lines
    removeAllLines(QRegularExpression("^Q[^=]+=\\S+"));
    removeAllLines(QRegularExpression(QString("^%1[^=]+=[^\\/]+\\/%2").arg(qprefixRel).arg(sumvarRel)));
    removeAllLines(QRegularExpression(QString("^%1[^=]+=[^\\/]+\\/%2").arg(qprefixAbs).arg(sumvarAbs)));

    // remove all 'Amorph=' and 'QabsAmorph=' lines
    removeAllLines(QRegularExpression("^Amorph=\\S+"));
    removeAllLines(QRegularExpression(QString("^%1Amorph=\\S+").arg(qprefixAbs)));

    // remove all 'GOAL[n]=' lines
    removeAllLines(QRegularExpression("^GOAL\\[\\d+\\]=.+$"));

    // remove all empty lines at the end of the document
    while ((content.last().simplified().isEmpty()) && content.size()) {
        content.removeLast();
    }
}

/*
 * appends the goal code to goalblock, and returns the number of GOALs
 */
int BgmnSavParser::createQuantGoalRelOldFormat(const QStringList &phases, QStringList &goalblock)
{
    bool qPercent = settings->value("bgmnProject/quantGoals100percent", false).toBool();

    goalblock.append(QString("sum=%1").arg(phases.join("+")));

    for (int i = 0; i < phases.size(); ++i) {
        goalblock.append(QString("GOAL[%1]=%2%3/sum").arg(i+1).arg(qPercent ? "100*" : QString()).arg(phases.at(i)));
    }

    return phases.size();
}

/*
 * appends the goal code to goalblock, and returns the number of GOALs
 */
int BgmnSavParser::createQuantGoalAbsOldFormat(const QStringList &phases, const QString &istd, QStringList &goalblock)
{
    bool qPercent = settings->value("bgmnProject/quantGoals100percent", false).toBool();

    QStringList amline;

    if (qPercent) {
        goalblock.append(QString("sum=ISTD*(100-ISTDQ)/ISTDQ"));
    } else {
        goalblock.append(QString("sum=ISTD*(1-ISTDQ)/ISTDQ"));
    }

    for (int i = 0; i < phases.size(); ++i) {
        if (phases.at(i) == istd) continue;
        amline.append(phases.at(i));
    }

    if (qPercent) {
        goalblock.append(QString("Amorph=100-100*(%1)/sum").arg(amline.join("+")));
    } else {
        goalblock.append(QString("Amorph=1-(%1)/sum").arg(amline.join("+")));
    }

    goalblock.append(QString());

    int n = 1;

    for (int i = 0; i < phases.size(); ++i) {
        if (phases.at(i) != istd) {
            goalblock.append(QString("GOAL[%1]=%2%3/sum").arg(n).arg(qPercent ? "100*" : QString()).arg(phases.at(i)));
            ++n;
        }
    }

    goalblock.append(QString("GOAL[%1]=Amorph").arg(phases.size()));

    return phases.size();
}

/*
 * appends the goal code to goalblock, and returns the number of GOALs
 */
int BgmnSavParser::createQuantGoalRelNewFormat(const QStringList &phases, QStringList &goalblock)
{
    bool qPercent = settings->value("bgmnProject/quantGoals100percent", false).toBool();

    goalblock.append(QString("%1=%2").arg(sumvarRel).arg(phases.join("+")));

    for (int i = 0; i < phases.size(); ++i) {
        goalblock.append(QString("%1%2=%3%2/%4").arg(qprefixRel).arg(phases.at(i)).arg(qPercent ? "100*" : QString()).arg(sumvarRel));
    }

    goalblock.append(QString());

    for (int i = 0; i < phases.size(); ++i) {
        goalblock.append(QString("GOAL[%1]=%2%3").arg(i+1).arg(qprefixRel).arg(phases.at(i)));
    }

    return phases.size();
}

/*
 * appends the goal code to goalblock, and returns the number of GOALs.
 * if 'addBoth = true', relative and absolute GOALs will be created
 */
int BgmnSavParser::createQuantGoalAbsNewFormat(const QStringList &phases, const QString &istd, bool addBoth, QStringList &goalblock)
{
    bool qPercent = settings->value("bgmnProject/quantGoals100percent", false).toBool();

    QStringList amline;
    int n = 0;

    if (addBoth) {
        n = createQuantGoalRelNewFormat(phases, goalblock);
        goalblock.append(QString());
    }

    if (qPercent) {
        goalblock.append(QString("%1=ISTD*(100-ISTDQ)/ISTDQ").arg(sumvarAbs));
    } else {
        goalblock.append(QString("%1=ISTD*(1-ISTDQ)/ISTDQ").arg(sumvarAbs));
    }

    for (int i = 0; i < phases.size(); ++i) {
        if (phases.at(i) == istd) continue;
        goalblock.append(QString("%1%2=%3%2/%4").arg(qprefixAbs).arg(phases.at(i)).arg(qPercent ? "100*" : QString()).arg(sumvarAbs));
        amline.append(phases.at(i));
    }

    if (qPercent) {
        goalblock.append(QString("%1Amorph=100-100*(%2)/%3").arg(qprefixAbs).arg(amline.join("+")).arg(sumvarAbs));
    } else {
        goalblock.append(QString("%1Amorph=1-(%2)/%3").arg(qprefixAbs).arg(amline.join("+")).arg(sumvarAbs));
    }

    goalblock.append(QString());
    ++n;

    for (int i = 0; i < phases.size(); ++i) {
        if (phases.at(i) != istd) {
            goalblock.append(QString("GOAL[%1]=%2%3").arg(n).arg(qprefixAbs).arg(phases.at(i)));
            ++n;
        }
    }

    goalblock.append(QString("GOAL[%1]=%2Amorph").arg(n).arg(qprefixAbs));

    return n;
}

int BgmnSavParser::removeAllLines(const QRegularExpression &rx)
{
    int n = 0;

    while (content.indexOf(rx) > -1) {
        content.removeAt(content.indexOf(rx));
        ++n;
    }

    return n;
}

QStringList BgmnSavParser::getAllPhaseNames() const
{
    QStringList phases;
    QFileInfo fi(fileName);
    QStringList strFiles = getAllMatches(QRegularExpression("^STRUC\\[\\d+\\]=(.+)\\s*"));

    for (int i = 0; i < strFiles.size(); ++i) {
        BgmnStrParser strParser(fi.absolutePath() + "/" + strFiles.at(i));
        QString qGoal = strParser.getQuantGoal();

        if (!qGoal.isEmpty()) {
            phases.append(strParser.getQuantGoal());
        }
    }

    return phases;
}

QStringList BgmnSavParser::getNonQuantificationGoals() const
{
    QStringList goals(getAllMatches(QRegularExpression("^GOAL\\[\\d+\\]=([^=]+)")));

    if (goals.isEmpty()) return QStringList();

    QStringList pattern;
    pattern << QString("^Amorph");
    pattern << QString("^[^\\/]+\\/sum(?:=\\d+\\.?\\d*)?");
    pattern << QString("^%1[^\\/]+(?:=\\d+\\.?\\d*)?").arg(qprefixRel);
    pattern << QString("^%2[^\\/]+(?:=\\d+\\.?\\d*)?").arg(qprefixAbs);

    for (int p = 0; p < pattern.size(); ++p) {
        QRegularExpression rx(pattern.at(p));

        while (goals.indexOf(rx) >= 0) {
            goals.removeAt(goals.indexOf(rx));
        }
    }

    return goals;
}

void BgmnSavParser::renumberGoals()
{
    int n = 1;
    QRegularExpression rx("^GOAL\\[\\d+\\]=(.+)$");
    QRegularExpressionMatch rm;

    for (int i = 0; i < content.size(); ++i) {
        rm = rx.match(content.at(i));

        if (rm.hasMatch()) {
            content[i] = QString("GOAL[%1]=%2").arg(n).arg(rm.captured(1));
            ++n;
        }
    }
}

void BgmnSavParser::setValFile(const QStringList &l)
{
    qDebug() << QString("BgmnSavParser::setValFile(): adding VAL files %1").arg(l.join(";"));

    if (content.isEmpty()) {
        qDebug() << "BgmnSavParser::setValFile(): Failed, no savFile loaded";
        return;
    }

    int n = clearValFiles();

    if (!l.size()) {
        content.insert(n, QString("VAL[1]="));
    } else {
        for (int j = 0; j < l.size(); ++j) {
            content.insert(n+j, QString("VAL[%1]=%2").arg(j+1).arg(l.at(j)));
        }
    }
}

/*
 * clears all VAL[n]= entries, returns the line number of the last one
 */
int BgmnSavParser::clearValFiles()
{
    QRegularExpression rx(QString("VAL\\[\\d+\\]=%1").arg(rxFileNames));

    int i = content.indexOf(rx, 0);
    int n = 0;

    while (i > -1) {
        content.removeAt(i);
        n = i;
        i = content.indexOf(rx, i);
    }

    return n;
}

int BgmnSavParser::generateValFiles(const QString &loadedGraphFile, const QStringList &nativeGraphFormats, bool convert, int n)
{
    QFileInfo fi(loadedGraphFile);

    if(!convert) {
        qDebug() << QString("BgmnSavParser::generateValFileNames(): Conversion not required, using %1 as VAL file").arg(fi.filePath());
        setValFile(QStringList(fi.filePath()));
        return 1;
    }

    if (nativeGraphFormats.contains(fi.suffix().toLower())) {
        qDebug() << QString("BgmnSavParser::generateValFileNames(): File format is native, using %1 as VAL file").arg(fi.filePath());
        setValFile(QStringList(fi.filePath()));
        return 1;
    }

    if (fi.suffix().toLower() == "dia") {
        QString valFileName(QString("%1.xy").arg(fi.completeBaseName()));
        qDebug() << QString("BgmnSavParser::generateValFileNames(): Creating VAL file name %1").arg(valFileName);
        setValFile(QStringList(valFileName));
        return 1;
    }

    if (n == 1) {
        QString valFileName(QString("%1.xy").arg(fi.completeBaseName()));
        qDebug() << QString("BgmnSavParser::generateValFileNames(): Creating VAL file name %1").arg(valFileName);
        setValFile(QStringList(valFileName));
        return 1;
    }

    QStringList valOut;
    int digits = int(log10(double(n))) + 1;

    for (int i = 0; i < n; ++i) {
        QString valFileName(QString("%1-%2.xy").arg(fi.completeBaseName()).arg(i, digits, 10, QLatin1Char('0')));
        qDebug() << QString("BgmnSavParser::generateValFileNames(): Creating VAL file name %1").arg(valFileName);
        valOut.append(valFileName);
    }

    setValFile(valOut);
    return n;
}

void BgmnSavParser::setOutputFile(const QString &s)
{
    qDebug() << QString("BgmnSavParser::setOutputFile(): setting OUT file to %1").arg(s);
    QRegularExpression rx(QString("OUTPUT=%1").arg(rxFileNames));

    int i = content.indexOf(rx, 0);

    if (i > -1) {
        content[i] = QString("OUTPUT=%1").arg(s);
    } else {
        content.append(QStringLiteral("% Peak list output"));
        content.append(QString("OUTPUT=%1").arg(s));
    }
}

void BgmnSavParser::setListFile(const QString &s)
{
    qDebug() << QString("BgmnSavParser::setListFile(): setting LST file to %1").arg(s);
    QRegularExpression rx(QString("LIST=%1").arg(rxFileNames));

    int i = content.indexOf(rx, 0);

    if (i > -1) {
        content[i] = QString("LIST=%1").arg(s);
    } else {
        content.append(QStringLiteral("% Results list output"));
        content.append(QString("LIST=%1").arg(s));
    }
}

void BgmnSavParser::setDiagramFile(const QString &s)
{
    qDebug() << QString("BgmnSavParser::setDiagramFile(): setting DIA file to %1").arg(s);

    // note: We also capture DIAGRAM= (with only 1 M), because it is an error likely to occur from non-german speaking users
    QRegularExpression rx(QString("DIAGRAM+=%1").arg(rxFileNames));

    int i = content.indexOf(rx, 0);

    if (i > -1) {
        content[i] = QString("DIAGRAMM=%1").arg(s);
    } else {
        content.append(QStringLiteral("% Diagram output"));
        content.append(QString("DIAGRAMM=%1").arg(s));
    }
}


void BgmnSavParser::setUntFile(const QString &s)
{
    qDebug() << QString("BgmnSavParser::setUntFile(): setting UNT file to %1").arg(s);

    QRegularExpression rx(QString("UNT=%1").arg(rxFileNames));

    int i = content.indexOf(rx, 0);

    if (i > -1) {
        content[i] = QString("UNT=%1").arg(s);
    } else {
        content.append(QStringLiteral("% Measured Background file"));
        content.append(QString("UNT=%1").arg(s));
    }
}

/*
 * structure output files *.res, *.fcf etc should always have a file name
 * of format <phase>-<projectBaseName>.ext in order to avoid overwriting
 * by other projects. Here we change the <projectBaseName> to a new
 * name.
 */
void BgmnSavParser::setStrucOutBaseName(const QString &s)
{
    qDebug() << QString("BgmnSavParser::setStrucOutBaseName(): setting Structure output file basename to %1").arg(s);
    QRegularExpression rxs("STRUC\\[(\\d+)\\]=(.+)\\.(str|STR)");
    QRegularExpression rxo("(STRUCOUT|SimpleSTRUCOUT|RESOUT|FCFOUT|PDBOUT)\\[(\\d+)\\]=(?:.+)\\.(str|res|fcf|pdb|STR|RES|FCF|PDB)");

    // we have to re-create all structure output file names, because
    // there is no way to parse the previous ones correctly.
    // first we gather STRUC[n]=<phase>.str in a map <n, phase>

    QMap<QString, QString> nameMap;
    for (int i = 0; i < content.size(); ++i) {
        QRegularExpressionMatch match = rxs.match(content.at(i));
        if (match.hasMatch()) {
            nameMap.insert(match.captured(1), match.captured(2));
        }
    }

    // now we re-create the file names for STRUCOUT etc.

    for (int i = 0; i < content.size(); ++i) {
        QRegularExpressionMatch match = rxo.match(content.at(i));
        if (match.hasMatch()) {
            QString key = match.captured(1);
            QString num = match.captured(2);
            QString pha = nameMap.value(num);
            QString ext = match.captured(3);

            if (key == "SimpleSTRUCOUT") {
                pha += "-simple";
            }

            content[i] = QString("%1[%2]=%3-%4.%5").arg(key).arg(num).arg(pha).arg(s).arg(ext);
        }
    }
}

/*
 * defines the variables ISTD and ISTDQ (Profex-specific) and
 * changes the goal calculation to use phase p as an internal
 * standard with quantity q
 */
void BgmnSavParser::setInternalStandard(const QString &p, double q)
{
    bool qPercent = settings->value("bgmnProject/quantGoals100percent", false).toBool();

    // first we unset an existing internal standard, because the
    // phase name has to be added to the regular GOALs
    unsetInternalStandard();

    QRegularExpression rxi("^ISTD=(\\S+)");
    QRegularExpression rxq("^ISTDQ=(\\d\\.\\d*)");
    QRegularExpression rxp(QString("^GOAL\\[\\d*\\]=%1.*").arg(p));

    // if there is a GOAL entry for phase p, we need to remove it
    if (content.indexOf(rxp) >= 0) content.removeAt(content.indexOf(rxp));

    // rewrite or append the ISTD tag
    if (content.indexOf(rxi) < 0) {
        content.append(QString("ISTD=%1").arg(p));
    } else {
        content.replaceInStrings(rxi, QString("ISTD=%1").arg(p));
    }

    // rewrite or append the ISTDQ tag
    if (content.indexOf(rxq) < 0) {
        if (qPercent) {
            content.append(QString("ISTDQ=%1").arg(100.0 * q, 0, 'f', 2));
        } else {
            content.append(QString("ISTDQ=%1").arg(q, 0, 'f', 4));
        }
    } else {
        if (qPercent) {
            content.replaceInStrings(rxq, QString("ISTDQ=%1").arg(100.0 * q, 0, 'f', 2));
        } else {
            content.replaceInStrings(rxq, QString("ISTDQ=%1").arg(q, 0, 'f', 4));
        }
    }

    // call rewriting of sum and GOAL lines
    writeGoals();
}

void BgmnSavParser::unsetInternalStandard()
{
    QRegularExpression rxi("^ISTD=(\\S+)");
    QRegularExpression rxq("^ISTDQ=(\\d+\\.?\\d*)");
    QRegularExpression rxa("^Amorph=\\S*");

    bool rewrite = false;

    // check if ISTD tag exists, if yes, remove it and flag rewriting of the goals.
    // rewriting goals will add the ISTD=phase phase back to the normal goals
    if (content.indexOf(rxi) >= 0) {
        content.removeAt(content.indexOf(rxi));
        rewrite = true;
    }

    // check if ISTDQ tag exists, if yes, remove it and flag rewriting of the goals
    if (content.indexOf(rxq) >= 0) {
        content.removeAt(content.indexOf(rxq));
        rewrite = true;
    }

    // check if Amorph tag exists, if yes, remove it and flag rewriting of the goals
    if (content.indexOf(rxa) >= 0) {
        content.removeAt(content.indexOf(rxa));
        rewrite = true;
    }

    if (rewrite) writeGoals();
}

QString BgmnSavParser::internalStandard() const
{
    QRegularExpression rxi("^ISTD=(\\S+)");

    int n = content.indexOf(rxi);

    if (n >= 0) {
        QRegularExpressionMatch rmi = rxi.match(content.at(n));
        if (rmi.hasMatch()) return rmi.captured(1);
    }

    return QString();
}

double BgmnSavParser::internalStandardQuantity() const
{
    QRegularExpression rxq("^ISTDQ=(\\d\\.\\d*)");

    int n = content.indexOf(rxq);

    if (n >= 0) {
        QRegularExpressionMatch rmq = rxq.match(content.at(n));
        if (rmq.hasMatch()) return rmq.captured(1).toDouble();
    }

    return -1.0;
}

QString BgmnSavParser::deviceFile() const
{
    QString s = getParameterValue(QRegularExpression(QString("VERZERR=%1").arg(rxFileNames)), 1);
    return s.trimmed();
}

QStringList BgmnSavParser::valFile() const
{
    QRegularExpression rx(QString("VAL\\[\\d+\\]=%1").arg(rxFileNames));
    QRegularExpressionMatch match;
    QStringList lst;

    for (int i = 0; i < content.size(); ++i) {
        match = rx.match(content.at(i));
        // we found a VAL[n]= line
        if (match.hasMatch()) {
            QString s = match.captured(1).trimmed();
            if (!s.isEmpty()) {
                lst.append(s);
            }
        }
    }

    return lst;
}

QString BgmnSavParser::outputFile() const
{
    QString s = getParameterValue(QRegularExpression(QString("OUTPUT=%1").arg(rxFileNames)), 1);
    // qDebug() << QString("BgmnSavParser::outputFile(): %1").arg(s);
    return s.trimmed();
}

QString BgmnSavParser::listFile() const
{
    QString s = getParameterValue(QRegularExpression(QString("LIST=%1").arg(rxFileNames)), 1);
    // qDebug() << QString("BgmnSavParser::listFile(): %1").arg(s);
    return s.trimmed();
}

QString BgmnSavParser::diagramFile() const
{
    QString s = getParameterValue(QRegularExpression(QString("DIAGRAM+(\\[\\d+\\])?=%1").arg(rxFileNames)), 2);
    // qDebug() << QString("BgmnSavParser::diagramFile(): %1").arg(s);
    return s.trimmed();
}

QString BgmnSavParser::untFile() const
{
    QString s = getParameterValue(QRegularExpression(QString("UNT=%1").arg(rxFileNames)), 1);
    // qDebug() << QString("BgmnSavParser::untFile(): %1").arg(s);
    return s.trimmed();
}

QString BgmnSavParser::untcFile() const
{
    QString s = getParameterValue(QRegularExpression(QString("UNTC=%1").arg(rxFileNames)), 1);
    // qDebug() << QString("BgmnSavParser::untcFile(): %1").arg(s);
    return s.trimmed();
}

/*
 * Searches and returns a string defined in a regexp. c defines the capture of the
 * regexp to be returned
 */
QString BgmnSavParser::getParameterValue(const QRegularExpression &rx, int c) const
{
    QRegularExpressionMatch match;
    for (int i = 0; i < content.size(); ++i) {
        match = rx.match(content.at(i));
        if (match.hasMatch()) {
            return match.captured(c).trimmed();
        }
    }

    return QString();
}

/*
 * use c <= 0 to autodetect the number of cores
 */
void BgmnSavParser::setNumberOfThreads(int c)
{
    QRegularExpression rx("(?:NTHREADS=)(\\d+)");

    int n = 4; // this is the fallback value

    if (c > 0) {
        n = c;
    } else {
        // autodetect
        int a = QThread::idealThreadCount();
        if (a > 0) {
            n = a;
        } // else use the fallback
    }

    int l = content.indexOf(rx, 0);
    if (l > -1) {
        content.replace(l, QString("NTHREADS=%1").arg(n));
    } else {
        content.append(QString("NTHREADS=%1").arg(n));
    }
}

/*
 * returns the value of NTHREAD, or -1 if not set
 */
int BgmnSavParser::numberOfThreads() const
{
    return getParameterValue(QRegularExpression("(?:[^\\%]*NTHREADS=)(\\d+)"), 1).toInt();
}

/*
 * returns a string of the current sav file, but without any input
 * or output file names, to be used as a template SAV file
 */
QString BgmnSavParser::createTemplate() const
{
    QRegularExpression rxClear("(VERZERR|DIAGRAMM|VAL\\[\\d+\\]|LIST|OUTPUT)=.*");
    QStringList output;

    bool hasVal = false;

    for (int i = 0; i < content.size(); ++i) {
        QRegularExpressionMatch match = rxClear.match(content.at(i));
        if (match.hasMatch()) {
            if (match.captured(1).contains(QRegularExpression("VAL\\[\\d+\\]"))) {
                if (!hasVal) {
                    // this is the first val line. we will append it (number hardcoded to 1),
                    // and set the flag to true to ignore any further val lines
                    output << QStringLiteral("VAL[1]=");
                    hasVal = true;
                    continue;
                } else {
                    // a val line was already read before. We only need the first one, so ignore this one
                    continue;
                }
            }

            // only add the first capture, but not the output file name
            output << QString("%1=").arg(match.captured(1));
            continue;
        }

        // no matches, so it is a line that will be used without modification
        output << content.at(i);
    }

    return output.join("\n");
}

double BgmnSavParser::limitOfDetection() const
{
    bool ok;
    double d = getParameterValue(QRegularExpression("^LOD=(\\d\\.?\\d*)"), 1).toDouble(&ok);
    if (ok) return d;
    return -1.0;
}

double BgmnSavParser::limitOfQuantification() const
{
    bool ok;
    double d = getParameterValue(QRegularExpression("^LOQ=(\\d\\.?\\d*)"), 1).toDouble(&ok);
    if (ok) return d;
    return -1.0;
}

double BgmnSavParser::minimumEsd() const
{
    bool ok;
    double d = getParameterValue(QRegularExpression("^MINESD=(\\d\\.?\\d*)"), 1).toDouble(&ok);
    if (ok) return d;
    return -1.0;
}
