/***************************************************************************
                          cifparser2.cpp  -  description
                             -------------------
    begin                : Wed Feb 03 20:30:00 CEST 2021
    copyright            : (C) 2021 by Nicola Doebelin
    email                : ndoebelin@gmx.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "cifparser2.h"
#include "bgmnfileio.h"
#include "functions.h"

#include <math.h>
#include <QRegularExpression>

#ifndef M_PI
    #define M_PI acos(-1.0)
#endif

CifParser2::CifParser2(const QString &f)
{
    cifFile = QFileInfo(f);
    cifStr = BgmnFileIO::readTextFileLines(f);
    QMap<QString, QStringList> dataStructure = parseCif(cifStr);
    toStructure(dataStructure);
}

QString CifParser2::cifString() const
{
    return cifStr.join("\n");
}

CrystalStructure CifParser2::getCrystalStructure(bool *ok) const
{
    if (ok) *ok = true;
    return cStructure;
}

QMap<QString, QStringList> CifParser2::parseCif(const QStringList &l)
{
    QMap<QString, QStringList> data;

    QStringList contentNoComments = stripComments(l);
    QStringList tokens = tokenize(contentNoComments);

    data = parseTokens(tokens);
    return data;
}

QStringList CifParser2::stripComments(const QStringList &l)
{
    QStringList nc;

    for (int i = 0; i < l.size(); ++i) {
        if (!l.at(i).contains("#")) {
            nc.append(l.at(i));
        } else {
            int n = l.at(i).indexOf("#");
            if (n > 0) nc.append(l.at(i).left(n));
        }
    }

    return nc;
}

QStringList CifParser2::tokenize(const QStringList &l)
{
    QRegularExpression rx("((?<=')(.*?)(?=')|(?<= |\n|^)([^ ']+?)(?= |\n|$))");
    QStringList tokens;

    for (int i = 0; i < l.size(); ++i) {
        if (l.at(i).left(1) == ";") {
            tokens.append(multiLineToken(l, i));
        } else {
            QRegularExpressionMatchIterator rmi = rx.globalMatch(l.at(i));

            while (rmi.hasNext()) {
                QRegularExpressionMatch rm = rmi.next();
                tokens.append(rm.captured(1));
            }
        }
    }

    return tokens;
}

QString CifParser2::multiLineToken(const QStringList &l, int &n)
{
    QString tk = l.at(n).mid(1);

    int i = n + 1;

    while ((l.at(i).left(1) != ";") && (i < l.size())) {
        tk += " " + l.at(i);
        ++i;
    }

    n = i;
    return tk.trimmed();
}

QMap<QString, QStringList> CifParser2::parseTokens(const QStringList &l)
{
    QMap<QString, QStringList> data;

    int i = 0;
    while (i < l.size()) {
        if (l.at(i).left(5) == "data_") {
            data["data"].append(l.at(i).mid(5));
            ++i;
        } else if (l.at(i).left(1) == "_") {
            i = parseParameter(data, l, i);
        } else if (l.at(i).left(5) == "loop_") {
            i = parseLoop(data, l, i);
        } else {
            ++i;
        }

        if (i < 0) break;
    }

    return data;
}

void CifParser2::toStructure(const QMap<QString, QStringList> &m)
{
    cStructure.setName(getPhaseName(m));
    cStructure.unitCell().setAxisUnit(NM);

    cStructure.setAuxInfo("Reference", getDatabaseCode(m));
    cStructure.setAuxInfo("Formula", getFormula(m));

    cStructure.unitCell().setItNumber(getITnum(m));
    cStructure.unitCell().setSpaceGroupHMCif(getHMsymb(m));

    double cell_a  = getCellA(m);
    double cell_b  = getCellB(m);
    double cell_c  = getCellC(m);
    double cell_al = getCellAlpha(m);
    double cell_be = getCellBeta(m);
    double cell_ga = getCellGamma(m);

    if (cell_a  > 0.0) cStructure.unitCell().setA(cell_a);
    if (cell_b  > 0.0) cStructure.unitCell().setB(cell_b);
    if (cell_c  > 0.0) cStructure.unitCell().setC(cell_c);
    if (cell_al > 0.0) cStructure.unitCell().setAlpha(cell_al);
    if (cell_be > 0.0) cStructure.unitCell().setBeta(cell_be);
    if (cell_ga > 0.0) cStructure.unitCell().setGamma(cell_ga);

    cStructure.unitCell().fixCell();

    QMap<QString, CrystalAtom> atms = getAtomList(m);
    QMapIterator<QString, CrystalAtom> it(atms);

    while (it.hasNext()) {
        it.next();
        cStructure.addAtom(it.value());
    }

    cStructure.unitCell().setSymmetryOperations(getSymOps(m));
}

int CifParser2::parseParameter(QMap<QString, QStringList> &m, const QStringList &l, int i)
{
    if (i >= l.size() - 1) return -1;
    m[l.at(i)].append(l.at(i+1));
    return i+1;
}

int CifParser2::parseLoop(QMap<QString, QStringList> &m, const QStringList &l, int i)
{
    QStringList header;
    int n = i+1;

    while (l.at(n).left(1) == "_") {
        if (n >= l.size()) break;
        header.append(l.at(n));
        ++n;
    }

    int k = 0;

    // the loop ends when we detect "loop_", something stating with "_", or the end of the list
    while (n < l.size()) {
        if (l.at(n) == "loop_") return n;
        if (l.at(n).left(1) == "_") return n;

        m[header.at(k)].append(l.at(n));
        ++n;
        ++k;
        if (k >= header.size()) k = 0;
    }

    return n+1;
}

QStringList CifParser2::getParameter(const QString &name, const QMap<QString, QStringList> &m)
{
    // this guarantees that we can always access the QStringList::first() element of the returned
    // list without having to check for the size each time.
    return m.value(name, QStringList(ERRSTR));
}


QString CifParser2::stripStdDev(const QString &s)
{
    // used to extract values, considering standard deviations (d.ddd(d))
    QRegularExpression rxV("([\\d\\.Ee\\+-]+)(?:\\(\\d*\\))");
    QRegularExpressionMatch match;

    match = rxV.match(s);

    if (match.hasMatch()) {
        return match.captured(1);
    }

    return s;
}

QString CifParser2::getDatabaseCode(const QMap<QString, QStringList> &m)
{
    QString dbName;
    QString dbRec;
    QRegularExpression rx("_?([A-Za-z]+)?_database_code_?([A-Za-z]+)?");
    QRegularExpressionMatch match;

    QMapIterator<QString, QStringList> it(m);
    while (it.hasNext()) {
        it.next();
        match = rx.match(it.key());

        if (match.hasMatch()) {
            if (!match.captured(1).isEmpty()) dbName = match.captured(1);
            if (!match.captured(2).isEmpty()) dbName = match.captured(2);
            dbRec = it.value().first();
            return dbName + "_" + dbRec.replace(QRegularExpression("\\s+"), "_");
        }
    }

    return QString();
}

QString CifParser2::getPhaseName(const QMap<QString, QStringList> &m)
{
    QString out = getParameter("_chemical_name_mineral", m).first();

    if (out == ERRSTR) out = getParameter("_chemical_name_common", m).first();
    if (out == ERRSTR) out = getParameter("_chemical_name", m).first();
    if (out == ERRSTR) out = getParameter("_chemical_name_systematic", m).first();

    // here we can have invalid strings, for example "_chemical_name_mineral ' ?'"
    // after replacing invalid symbols, the string will be empty
    out = out.replace(QRegularExpression("[^A-Za-z0-9]"), QString());

    if ((out == ERRSTR) || out.isEmpty()) {
        out = cifFile.completeBaseName();
        out = out.replace(QRegularExpression("[^A-Za-z0-9]"), QString());
    }

    return out;
}

QString CifParser2::getFormula(const QMap<QString, QStringList> &m)
{
    QString out = getParameter("_chemical_formula_sum", m).first();

    if (out == ERRSTR) out = getParameter("_chemical_formula_structural", m).first();
    if (out == ERRSTR) out = getParameter("_chemical_formula_moiety", m).first();
    if (out == ERRSTR) out = getParameter("_chemical_formula_analytical", m).first();

    return out.replace(QRegularExpression("\\s+"), "_");
}

int CifParser2::getITnum(const QMap<QString, QStringList> &m)
{
    QString out = getParameter("_space_group_IT_number", m).first();                 // new tag name
    if (out == ERRSTR) out = getParameter("_symmetry_Int_Tables_number", m).first(); // old tag name
    return out.toInt();
}

QString CifParser2::getHMsymb(const QMap<QString, QStringList> &m)
{
    QString out = getParameter("_space_group_name_H-M_alt", m).first();                 // new tag name
    if (out == ERRSTR) out = getParameter("_symmetry_space_group_name_H-M", m).first(); // old tag name

    // the cif dictionary allows some additions to the HM symbol, e.g. info in () or :origin
    // we need to clip these
    QRegularExpression rx("^([A-Za-z0-9_\\s/-]+)");
    QRegularExpressionMatch rm = rx.match(out);

    if (rm.hasMatch()) return rm.captured(1).trimmed();
    return out;
}

QList<QStringList> CifParser2::getSymOps(const QMap<QString, QStringList> &m)
{
    QList<QStringList> out;

    QStringList l;
    if (m.contains("_space_group_symop_operation_xyz")) {
        l = m.value("_space_group_symop_operation_xyz");
    } else if (m.contains("_symmetry_equiv_pos_as_xyz")) {
        l = m.value("_symmetry_equiv_pos_as_xyz");
    }

    for (int i = 0; i < l.size(); ++i) {
        QStringList s = l.at(i).split(QRegularExpression("[, ]+"));
        if (s.size() == 3) out.append(s);
    }

    return out;
}

double CifParser2::getCellA(const QMap<QString, QStringList> &m)
{
    return getCellParam(m, "_cell_length_a") / 10.0;
}

double CifParser2::getCellB(const QMap<QString, QStringList> &m)
{
    return getCellParam(m, "_cell_length_b") / 10.0;
}

double CifParser2::getCellC(const QMap<QString, QStringList> &m)
{
    return getCellParam(m, "_cell_length_c") / 10.0;
}

double CifParser2::getCellAlpha(const QMap<QString, QStringList> &m)
{
    return getCellParam(m, "_cell_angle_alpha");
}

double CifParser2::getCellBeta(const QMap<QString, QStringList> &m)
{
    return getCellParam(m, "_cell_angle_beta");
}

double CifParser2::getCellGamma(const QMap<QString, QStringList> &m)
{
    return getCellParam(m, "_cell_angle_gamma");
}

double CifParser2::getCellParam(const QMap<QString, QStringList> &m, const QString &s)
{
    QString val = getParameter(s, m).first();

    if (val == ERRSTR) {
        qDebug() << QString("CifParser2::getCellParam(): Parameter %1 not found").arg(s);
        return -1.0;
    }

    return stripStdDev(val).toDouble();
}

QMap<QString, CrystalAtom> CifParser2::getAtomList(const QMap<QString, QStringList> &m)
{
    QMap<QString, CrystalAtom> mout;
    QMap<QString, QMap<QString, QString> > data = getAtomDataStructure(m);
    QStringList atLabels = data.value("_atom_site_label").keys();

    for (int i = 0; i < atLabels.size(); ++i) {
        QString l = atLabels.at(i);

        if (getAtomIsDummy(data, l)) continue;

        double x = -1.0;
        double y = -1.0;
        double z = -1.0;
        if (!getAtomCoordinates(data, l, x, y, z)) continue;

        double biso = 0.0;
        if (!getAtomBiso(data, l, biso)) biso = 0.0;

        double occ = 1.0;
        if (!getAtomOccupancy(data, l, occ)) occ = 1.0;

        CrystalAtom atom(l, x, y, z, occ, biso);

        QString el = l;
        if (getAtomType(data, l, el)) atom.setElement(el);

        QString wyck;
        int mult;
        if (getAtomWyckoffAndMultiplicity(data, l, wyck, mult)) {
            atom.setWyckoff(wyck);
            atom.setMultiplicity(mult);
        }

        mout[l] = atom;
    }

    return mout;
}

/*
 * for controlled access, we create the following structure first:
 *
 * QMap<QString(_cif_tag), QMap<QString(_atom_site_label), QString(_cif_value)> >
 *
 * This helps accessing the data because not all atom sites have all cif tags.
 */
QMap<QString, QMap<QString, QString> > CifParser2::getAtomDataStructure(const QMap<QString, QStringList> &m)
{
    QMap<QString, QMap<QString, QString> > data;

    QStringList laSiteLabel                = m.value("_atom_site_label", QStringList());
    QStringList laSiteTypeSymbol           = m.value("_atom_site_type_symbol", QStringList());
    QStringList laSiteWyckoffSymbol        = m.value("_atom_site_Wyckoff_symbol", QStringList());
    QStringList laSiteSymmetryMultiplicity = m.value("_atom_site_symmetry_multiplicity", QStringList());
    QStringList laSiteOccupancy            = m.value("_atom_site_occupancy", QStringList());
    QStringList laSiteFractX               = m.value("_atom_site_fract_x", QStringList());
    QStringList laSiteFractY               = m.value("_atom_site_fract_y", QStringList());
    QStringList laSiteFractZ               = m.value("_atom_site_fract_z", QStringList());
    QStringList laSiteBiso                 = m.value("_atom_site_B_iso_or_equiv", QStringList());
    QStringList laSiteUiso                 = m.value("_atom_site_U_iso_or_equiv", QStringList());
    QStringList laSiteCalcFlag             = m.value("_atom_site_calc_flag", QStringList());

    // these are essential
    if (laSiteLabel.isEmpty() || laSiteFractX.isEmpty() || laSiteFractY.isEmpty() || laSiteFractZ.isEmpty()) {
        qDebug() << QString("CifParser2::getAtomDataStructure(): Essential parameters are missing. Exiting.");
        return data;
    }

    QStringList laAnisoLabel  = m.value("_atom_site_aniso_label", QStringList());
    QStringList laAnisoB11    = m.value("_atom_site_aniso_B_11", QStringList());
    QStringList laAnisoB12    = m.value("_atom_site_aniso_B_12", QStringList());
    QStringList laAnisoB13    = m.value("_atom_site_aniso_B_13", QStringList());
    QStringList laAnisoB22    = m.value("_atom_site_aniso_B_22", QStringList());
    QStringList laAnisoB23    = m.value("_atom_site_aniso_B_23", QStringList());
    QStringList laAnisoB33    = m.value("_atom_site_aniso_B_33", QStringList());

    QStringList laAnisoU11    = m.value("_atom_site_aniso_U_11", QStringList());
    QStringList laAnisoU12    = m.value("_atom_site_aniso_U_12", QStringList());
    QStringList laAnisoU13    = m.value("_atom_site_aniso_U_13", QStringList());
    QStringList laAnisoU22    = m.value("_atom_site_aniso_U_22", QStringList());
    QStringList laAnisoU23    = m.value("_atom_site_aniso_U_23", QStringList());
    QStringList laAnisoU33    = m.value("_atom_site_aniso_U_33", QStringList());

    for (int i = 0; i < laSiteLabel.size(); ++i) {
        data["_atom_site_label"][laSiteLabel.at(i)] = laSiteLabel.at(i);
        if (i < laSiteTypeSymbol.size()) data["_atom_site_type_symbol"][laSiteLabel.at(i)] = laSiteTypeSymbol.at(i);
        if (i < laSiteWyckoffSymbol.size()) data["_atom_site_Wyckoff_symbol"][laSiteLabel.at(i)] = laSiteWyckoffSymbol.at(i);
        if (i < laSiteSymmetryMultiplicity.size()) data["_atom_site_symmetry_multiplicity"][laSiteLabel.at(i)] = laSiteSymmetryMultiplicity.at(i);
        if (i < laSiteOccupancy.size()) data["_atom_site_occupancy"][laSiteLabel.at(i)] = laSiteOccupancy.at(i);
        if (i < laSiteFractX.size()) data["_atom_site_fract_x"][laSiteLabel.at(i)] = laSiteFractX.at(i);
        if (i < laSiteFractY.size()) data["_atom_site_fract_y"][laSiteLabel.at(i)] = laSiteFractY.at(i);
        if (i < laSiteFractZ.size()) data["_atom_site_fract_z"][laSiteLabel.at(i)] = laSiteFractZ.at(i);
        if (i < laSiteBiso.size()) data["_atom_site_B_iso_or_equiv"][laSiteLabel.at(i)] = laSiteBiso.at(i);
        if (i < laSiteUiso.size()) data["_atom_site_U_iso_or_equiv"][laSiteLabel.at(i)] = laSiteUiso.at(i);
        if (i < laSiteCalcFlag.size()) data["_atom_site_calc_flag"][laSiteLabel.at(i)] = laSiteCalcFlag.at(i);
    }

    for (int i = 0; i < laAnisoLabel.size(); ++i) {
        if (i < laAnisoB11.size()) data["_atom_site_aniso_B_11"][laSiteLabel.at(i)] = laAnisoB11.at(i);
        if (i < laAnisoB12.size()) data["_atom_site_aniso_B_12"][laSiteLabel.at(i)] = laAnisoB12.at(i);
        if (i < laAnisoB13.size()) data["_atom_site_aniso_B_13"][laSiteLabel.at(i)] = laAnisoB13.at(i);
        if (i < laAnisoB22.size()) data["_atom_site_aniso_B_22"][laSiteLabel.at(i)] = laAnisoB22.at(i);
        if (i < laAnisoB23.size()) data["_atom_site_aniso_B_23"][laSiteLabel.at(i)] = laAnisoB23.at(i);
        if (i < laAnisoB33.size()) data["_atom_site_aniso_B_33"][laSiteLabel.at(i)] = laAnisoB33.at(i);

        if (i < laAnisoU11.size()) data["_atom_site_aniso_U_11"][laSiteLabel.at(i)] = laAnisoU11.at(i);
        if (i < laAnisoU12.size()) data["_atom_site_aniso_U_12"][laSiteLabel.at(i)] = laAnisoU12.at(i);
        if (i < laAnisoU13.size()) data["_atom_site_aniso_U_13"][laSiteLabel.at(i)] = laAnisoU13.at(i);
        if (i < laAnisoU22.size()) data["_atom_site_aniso_U_22"][laSiteLabel.at(i)] = laAnisoU22.at(i);
        if (i < laAnisoU23.size()) data["_atom_site_aniso_U_23"][laSiteLabel.at(i)] = laAnisoU23.at(i);
        if (i < laAnisoU33.size()) data["_atom_site_aniso_U_33"][laSiteLabel.at(i)] = laAnisoU33.at(i);
    }

    return data;
}

bool CifParser2::getAtomIsDummy(const QMap<QString, QMap<QString, QString> > &d, const QString &l)
{
    if (!d["_atom_site_calc_flag"].contains(l)) return false;
    return d["_atom_site_calc_flag"][l] == "dum";
}

bool CifParser2::getAtomCoordinates(const QMap<QString, QMap<QString, QString> > &d, const QString &l, double &x, double &y, double &z)
{
    if (!d["_atom_site_fract_x"].contains(l)) return false;
    if (!d["_atom_site_fract_y"].contains(l)) return false;
    if (!d["_atom_site_fract_z"].contains(l)) return false;

    bool ok;
    x = stripStdDev(d["_atom_site_fract_x"][l]).toDouble(&ok);
    if (!ok) return false;
    y = stripStdDev(d["_atom_site_fract_y"][l]).toDouble(&ok);
    if (!ok) return false;
    z = stripStdDev(d["_atom_site_fract_z"][l]).toDouble(&ok);
    if (!ok) return false;

    global::Functions::normalizeCoordinates(x, y, z);
    return true;
}

bool CifParser2::getAtomType(const QMap<QString, QMap<QString, QString> > &d, const QString &l, QString &t)
{
    if (!d["_atom_site_type_symbol"].contains(l)) return false;

    // we store the oxidation state in conventional format "At2+".
    // It is converted to BGMN format "At+2" elsewhere when needed
    QRegularExpression rx("([A-Za-z]+)([\\+-]?)(\\d*)([\\+-]?)");
    QRegularExpressionMatch rm = rx.match(d["_atom_site_type_symbol"][l]);

    if (rm.hasMatch()) {
        // 2, 3, and 4 can be empty. Either 2 or 4 must be empty, therefore
        // there will only be one sign (4 or 2) after the digit (3)
        t = rm.captured(1)+ rm.captured(3)+ rm.captured(4) + rm.captured(2);
        return true;
    }

    return false;
}

bool CifParser2::getAtomBiso(const QMap<QString, QMap<QString, QString> > &d, const QString &l, double &b)
{
    bool ok;

    if (d["_atom_site_B_iso_or_equiv"].contains(l)) {
        b = stripStdDev(d["_atom_site_B_iso_or_equiv"][l]).toDouble(&ok);
        return ok;
    }

    if (d["_atom_site_U_iso_or_equiv"].contains(l)) {
        double u = stripStdDev(d["_atom_site_U_iso_or_equiv"][l]).toDouble(&ok);
        b = global::Functions::bisoFromUiso(u);
        return ok;
    }

    if (d["_atom_site_aniso_B_11"].contains(l)) {
        double b11 = stripStdDev(d["_atom_site_aniso_B_11"][l]).toDouble(&ok);
        if (!ok) return false;
        double b12 = stripStdDev(d["_atom_site_aniso_B_12"][l]).toDouble(&ok);
        if (!ok) return false;
        double b13 = stripStdDev(d["_atom_site_aniso_B_13"][l]).toDouble(&ok);
        if (!ok) return false;
        double b22 = stripStdDev(d["_atom_site_aniso_B_22"][l]).toDouble(&ok);
        if (!ok) return false;
        double b23 = stripStdDev(d["_atom_site_aniso_B_23"][l]).toDouble(&ok);
        if (!ok) return false;
        double b33 = stripStdDev(d["_atom_site_aniso_B_33"][l]).toDouble(&ok);
        if (!ok) return false;

        // warning: angles may be wrong because we have not yet determined the crystal system
        b = global::Functions::bisoFromBaniso(cStructure.unitCell().alpha(), cStructure.unitCell().beta(), cStructure.unitCell().gamma(),
                                              b11, b22, b33, b12, b13, b23);
        return true;
    }

    if (d["_atom_site_aniso_U_11"].contains(l)) {
        double u11 = stripStdDev(d["_atom_site_aniso_U_11"][l]).toDouble(&ok);
        if (!ok) return false;
        double u12 = stripStdDev(d["_atom_site_aniso_U_12"][l]).toDouble(&ok);
        if (!ok) return false;
        double u13 = stripStdDev(d["_atom_site_aniso_U_13"][l]).toDouble(&ok);
        if (!ok) return false;
        double u22 = stripStdDev(d["_atom_site_aniso_U_22"][l]).toDouble(&ok);
        if (!ok) return false;
        double u23 = stripStdDev(d["_atom_site_aniso_U_23"][l]).toDouble(&ok);
        if (!ok) return false;
        double u33 = stripStdDev(d["_atom_site_aniso_U_33"][l]).toDouble(&ok);
        if (!ok) return false;

        b = global::Functions::bisoFromUaniso(cStructure.unitCell().alpha(), cStructure.unitCell().beta(), cStructure.unitCell().gamma(),
                                              u11, u22, u33, u12, u13, u23);
        return true;
    }

    return false;
}

bool CifParser2::getAtomOccupancy(const QMap<QString, QMap<QString, QString> > &d, const QString &l, double &o)
{
    if (!d["_atom_site_occupancy"].contains(l)) return false;

    bool ok;
    o = stripStdDev(d["_atom_site_occupancy"][l]).toDouble(&ok);
    if (!ok) o = 1.0;
    return true;
}

bool CifParser2::getAtomWyckoffAndMultiplicity(const QMap<QString, QMap<QString, QString> > &d, const QString &l, QString &w, int &m)
{
    if (!d["_atom_site_Wyckoff_symbol"].contains(l)) return false;

    QRegularExpression rx("(\\d*)([a-z]+)");
    QRegularExpressionMatch rm = rx.match(d["_atom_site_Wyckoff_symbol"][l]);

    if (rm.hasMatch()) {
        bool ok;
        w = rm.captured(2);
        m = rm.captured(1).toInt(&ok);
        if (!ok) m = 1;
        return true;
    }

    return false;
}

