/***************************************************************************
                          settingsmanager.cpp  -  description
                             -------------------
    begin                : Tue Mar 15 13:10:07 CET 2016
    copyright            : (C) 2016 by Nicola Doebelin
    email                : ndoebelin@gmx.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "settingsmanager.h"
#include "structs.h"
#include <QtMath>
#include <QStandardPaths>
#include <QGuiApplication>
#include <QPalette>
#include <QDir>

bool SettingsManager::instanceFlag = false;
bool SettingsManager::needsSync = false;
SettingsManager* SettingsManager::instance = nullptr;

SettingsManager::SettingsManager()
{
    settings.setDefaultFormat(QSettings::NativeFormat);
}

SettingsManager::~SettingsManager()
{
    instanceFlag = false;
    if (needsSync) sync();
}

SettingsManager* SettingsManager::getInstance()
{
    if (!instanceFlag) {
        instance = new SettingsManager();
        instanceFlag = true;
        return instance;
    } else {
        return instance;
    }
}

void SettingsManager::destroy()
{
    if (instance) delete instance;
}

void SettingsManager::initialize()
{
    settingsData.clear();

    QStringList keys = settings.allKeys();

    for (int i = 0; i < keys.size(); ++i) {
        settingsData.insert(keys.at(i), settings.value(keys.at(i), QVariant()));
    }

    needsSync = false;
}

void SettingsManager::sync()
{
    QMap<QString, QVariant>::ConstIterator it = settingsData.constBegin();

    while (it != settingsData.constEnd()) {
        settings.setValue(it.key(), it.value());
        ++it;
    }

    settings.sync();
    needsSync = false;
}

void SettingsManager::setValue(const QString &key, const QVariant &value)
{
    settingsData.insert(key, value);
    needsSync = true;
}

QVariant SettingsManager::value(const QString &key, const QVariant &value) const
{
    return settingsData.value(key, value);
}

QSettings::Format SettingsManager::format() const
{
    return settings.format();
}

QString SettingsManager::fileName() const
{
    return settings.fileName();
}

QStringList SettingsManager::allKeys() const
{
    return settings.allKeys();
}

QColor SettingsManager::getRandomColor(int n) const
{
    double steps = M_E + 3.0; // steps between colors. use a non-periodic value
                               // so the colors don't repeat after one revolution
                               // in h space
    double yellowDamp = 0.5;  // dampens v in the yellow range to avoid very bright yellow hues
                         // value = 0: no dampening, yellow is bright
                         // value = 1: maximum dampening, yellow turns black
    int offset = 4;      // change this to adjust the first color in h space. Avoid color number 4 to be
                         // dark blue, as it would be very similar to the hard-coded blue for background
                         // curves

    // h goes in circles through hue space in steps of n/steps.
    //   steps is a non-int value so h is shifted a little bit with each revolution
    // v is a sine curve that is darkest ("value") in the yellow and green area,
    //   and brightest (1.0) in blue and violet area
    double h = double(n + offset) / steps - qFloor(double(n + offset) / steps);
    double s = 1.0;
    double v = 1.0 - yellowDamp * 0.5 * (1.0 + qSin(M_PI * (2.0*h - 1.0/6.0)));
    double a = 1.0;

    QColor col;
    col.setHsvF(h, s, v, a);

    return col;
}

QColor SettingsManager::toFillColor(const QColor &c) const
{
    QColor fcol;

    if (isDarkMode()) {
        fcol.setHsl(c.hslHue(), c.hslSaturation() / 3, c.lightness() / 3);
    } else {
        fcol.setHsl(c.hslHue(), c.hslSaturation() / 3, (c.lightness() + 768) / 4);
    }

    return fcol;
}

QString SettingsManager::getAppDataLocation() const
{
    QString s = QStandardPaths::writableLocation(QStandardPaths::AppDataLocation);
    return settings.value("config/hklBufferFile", QDir::fromNativeSeparators(s)).toString();
}

QString SettingsManager::getTempLocation() const
{
    QString s = settings.value("config/tempFileLocation", QStandardPaths::writableLocation(QStandardPaths::TempLocation)).toString();
    return QDir::fromNativeSeparators(s);
}

int SettingsManager::getSyntaxHighlightingMode() const
{
    int hlmode = settings.value("config/editorSyntaxHighlighting", -1).toInt();

    if (hlmode < 0) {
        if (isDarkMode()) {
            // assuming dark mode
            return 1;
        } else {
            // assuming light mode
            return 0;
        }
    }

    return hlmode;
}

bool SettingsManager::isDarkMode() const
{
    return (QGuiApplication::palette().color(QPalette::Base).value() < 150);
}

double SettingsManager::defaultWavelength()
{
    return settings.value("config/defaultWl", 1.54056).toDouble();
}
