/***************************************************************************
                          instrumentraytracer.cpp  -  description
                             -------------------
    begin                : Sat Jul 25 11:00:00 CEST 2020
    copyright            : (C) 2020 by Nicola Doebelin
    email                : ndoebelin@gmx.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "instrumentraytracer.h"
#include "../libXrdIO/bgmnfileio.h"
#include <QRegularExpressionMatch>
#include <QCoreApplication>
#include <QtMath>
#include <QJSEngine>
#include <limits>

InstrumentRayTracer::InstrumentRayTracer(QObject *parent)
    : QObject(parent)
{
    settings = SettingsManager::getInstance();
    progressBar = nullptr;
    outputEditor = nullptr;

    initProcesses();
}

InstrumentRayTracer::~InstrumentRayTracer()
{
    if (progressBar) progressBar = nullptr;
    if (outputEditor) outputEditor = nullptr;
    if (geomet) delete geomet;
    if (makegeq) delete makegeq;
}

void InstrumentRayTracer::initProcesses()
{
    geomet  = new ProcessHandler(this);
    makegeq = new ProcessHandler(this);

    connect(geomet, SIGNAL(pollOutput()), this, SLOT(pollOutput()));
    connect(geomet, SIGNAL(allComplete()), this, SLOT(geometComplete()));
    connect(geomet, SIGNAL(aborted()), this, SLOT(processAborted()));

    connect(makegeq, SIGNAL(pollOutput()), this, SLOT(pollOutput()));
    connect(makegeq, SIGNAL(allComplete()), this, SLOT(makegeqComplete()));
    connect(makegeq, SIGNAL(aborted()), this, SLOT(processAborted()));
}

void InstrumentRayTracer::setProgressBar(QProgressBar *p)
{
    progressBar = p;
}

void InstrumentRayTracer::setOutputEditor(QPlainTextEdit *e)
{
    outputEditor = e;
}

void InstrumentRayTracer::setConfigFile(const QString &cont, const QString &fn)
{
    savFileContent = cont;
    savFileName = fn;
}

void InstrumentRayTracer::runProcess()
{
    if (savFileContent.isEmpty() || savFileName.isEmpty()) {
        outputEditor->setPlainText(tr("No sav file available. Exiting."));
        return;
    }

    geometExec = settings->value("bgmnProject/geometExec", "").toString();
    makeGeqExec = settings->value("bgmnProject/makegeqExec", "").toString();

    progressRanges = getRanges();
    int totalProgress = int((progressRanges.ttEnd - progressRanges.ttStart) + (progressRanges.wmax - progressRanges.wmin));
    progressBar->setMaximum(totalProgress);

    geomet->init(savFileName, geometExec);
    abort = false;

    if (geomet->run()) {
        qDebug() << QString("InstrumentRayTracer::runProcess(): Process started: %1 %2 %3")
                    .arg(geomet->getExec())
                    .arg(geomet->getFile())
                    .arg(geomet->getArguments().join(" "));
    } else {
        qDebug() << QString("InstrumentRayTracer::runProcess(): Process failed: %1 %2 %3")
                    .arg(geomet->getExec())
                    .arg(geomet->getFile())
                    .arg(geomet->getArguments().join(" "));
    }
}

void InstrumentRayTracer::pollOutput()
{
    QString output;
    int currentStep = 0;

    if (geomet->isRunning()) {
        output = geomet->readOutput();
        currentStep = int(getCurrentAngle(output) - progressRanges.ttStart);
    }

    if (makegeq->isRunning()) {
        output = makegeq->readOutput();
        currentStep = int((progressRanges.ttEnd - progressRanges.ttStart) + (getCurrentAngle(output) - progressRanges.wmin));
    }

    outputEditor->appendPlainText(output.trimmed());
    progressBar->setValue(currentStep);
    qApp->processEvents();
}

void InstrumentRayTracer::processAborted()
{
    abort = true;
    if (geomet->isRunning()) geomet->abort();
    if (makegeq->isRunning()) makegeq->abort();

    progressBar->setValue(progressBar->maximum());
    emit processComplete();
}

void InstrumentRayTracer::geometComplete()
{
    if (abort) return;

    applyDistortionCorrection();

    makegeq->init(savFileName, makeGeqExec);

    if (makegeq->run()) {
        qDebug() << QString("InstrumentRayTracer::geometComplete(): Process started: %1 %2 %3")
                    .arg(makegeq->getExec())
                    .arg(makegeq->getFile())
                    .arg(makegeq->getArguments().join(" "));
    } else {
        qDebug() << QString("InstrumentRayTracer::geometComplete(): Process failed: %1 %2 %3")
                    .arg(makegeq->getExec())
                    .arg(makegeq->getFile())
                    .arg(makegeq->getArguments().join(" "));
    }
}

void InstrumentRayTracer::makegeqComplete()
{
    if (abort) return;

    progressBar->setValue(progressBar->maximum());
    emit processComplete();
}


TtRange InstrumentRayTracer::getRanges()
{
    QRegularExpression rxSteps("zweiTheta\\[\\d+\\]=(-?\\d+\\.?\\d*)");
    QRegularExpression rxWmin("WMIN=(-?\\d+\\.?\\d*)");
    QRegularExpression rxWmax("WMAX=(-?\\d+\\.?\\d*)");

    QRegularExpressionMatch rmWmin = rxWmin.match(savFileContent);
    QRegularExpressionMatch rmWmax = rxWmax.match(savFileContent);
    QRegularExpressionMatchIterator riSteps = rxSteps.globalMatch(savFileContent);

    QList<double> steps;

    while (riSteps.hasNext()) {
        QRegularExpressionMatch rmSteps = riSteps.next();
        if (rmSteps.hasMatch()) {
            steps.append(rmSteps.captured(1).toDouble());
        }
    }

    qSort(steps);

    TtRange range;
    range.steps   = steps.size();
    range.ttStart = steps.first();
    range.ttEnd   = steps.last();

    range.wmin    = rmWmin.hasMatch() ? rmWmin.captured(1).toDouble() : 0.0;
    range.wmax    = rmWmax.hasMatch() ? rmWmax.captured(1).toDouble() : 1.0;

    return range;
}


double InstrumentRayTracer::getCurrentAngle(const QString &output)
{
    double d = 0.0;

    QRegularExpression rx("^zweiTheta=(\\d+\\.\\d+)\\s+");
    QRegularExpressionMatchIterator ri = rx.globalMatch(output);
    QRegularExpressionMatch rm;

    while (ri.hasNext()) {
        rm = ri.next();

        if (rm.hasMatch()) {
            d = rm.captured(1).toDouble();
        }
    }

    return d;
}

void InstrumentRayTracer::applyDistortionCorrection()
{
    QString ztStretch = getSavVariable("ZTSTRETCH=(.+)", QString()).trimmed();

    if (ztStretch.isEmpty()) return;
    ztStretch = bgmnMathToJS(ztStretch);

    QString fGer(getGerName());
    QStringList gerContent = BgmnFileIO::readTextFileLines(fGer);

    QRegularExpression rxTt("\\s*THETA=(\\d+\\.?\\d*)\\s+N=\\d+");
    QRegularExpression rxVal("\\s*(-?\\d+\\.?\\d*)\\s+(-?\\d+\\.?\\d*)\\s+(-?\\d+\\.?\\d*)");

    QJSEngine jsEngine;
    QJSValue jsValue = jsEngine.evaluate(QString("(function(zweiTheta) {return %1;} )").arg(ztStretch));
    outputEditor->appendPlainText(QString("Scaling peak width with f(zweiTheta) = %1").arg(ztStretch));

    double ztScale = 1.0;

    for (int i = 0; i < gerContent.size(); ++i) {
        QRegularExpressionMatch rmTt = rxTt.match(gerContent.at(i));
        QRegularExpressionMatch rmVal = rxVal.match(gerContent.at(i));

        if (rmTt.hasMatch()) {
            double theta = rmTt.captured(1).toDouble();

            QJSValueList jsArgs;
            jsArgs << 2.0 * theta;

            ztScale = jsValue.call(jsArgs).toNumber();

            outputEditor->appendPlainText(QString("Stretch factor at theta=%1 is %2").arg(theta).arg(ztScale, 0, 'f', 6));

            continue;
        }

        if (rmVal.hasMatch()) {
            double g = rmVal.captured(1).toDouble();
            double e = rmVal.captured(2).toDouble();
            double q = rmVal.captured(3).toDouble();

            gerContent[i] = QString("%1 %2 %3").arg(g, 0, 'f', 5).arg(e * ztScale, 0, 'f', 7).arg(q * ztScale, 0, 'f', 8);
        }
    }

    BgmnFileIO::writeTextFile(fGer, gerContent.join("\n"));
}

QString InstrumentRayTracer::getGerName()
{
    QRegularExpression rxGer("VERZERR=(.+\\.(ger|GER))");
    QRegularExpressionMatch rmGer = rxGer.match(savFileContent);

    if (!rmGer.hasMatch()) return QString();

    QFileInfo fiSav(savFileName);
    QFileInfo fiGer(fiSav.absolutePath() + "/" + rmGer.captured(1));
    qDebug() << QString("EditInstrumentDialog::getGerName(): returning file %1").arg(fiGer.absoluteFilePath());
    return fiGer.absoluteFilePath();
}

QString InstrumentRayTracer::getSavVariable(const QString &s, const QString &def)
{
    QRegularExpression rx(s);
    QRegularExpressionMatch rm = rx.match(savFileContent);

    if (rm.hasMatch()) return rm.captured(1);
    return def;
}

QString InstrumentRayTracer::bgmnMathToJS(const QString &istr)
{
    QString ostr(istr);

    QStringList bgPatterns;
    QStringList jsPatterns;

    bgPatterns << "abs" << "sqrt" << "sin";
    jsPatterns << "Math.abs" << "Math.sqrt" << "Math.sin";

    bgPatterns << "cos" << "tan" << "asin" << "acos";
    jsPatterns << "Math.cos" << "Math.tan" << "Math.asin" << "Math.acos";

    bgPatterns << "atan" << "log" << "exp" << "power";
    jsPatterns << "Math.atan" << "Math.log" << "Math.exp" << "Math.pow";

    bgPatterns << "min" << "max" << "pi";
    jsPatterns << "Math.min" << "Math.max" << "Math.PI";

    for (int i = 0; i < qMin(bgPatterns.size(), jsPatterns.size()); ++i) {
        QRegularExpression rx(QString("(?<!Math\\.)%1(?=[^a-zA-Z0-9])").arg(bgPatterns.at(i)));
        ostr.replace(rx, jsPatterns.at(i));
    }

    return ostr;
}

