/***************************************************************************
                          electrondensitymapdialog.cpp  -  description
                             -------------------
    begin                : Fri May 31 15:00:00 CEST 2019
    copyright            : (C) 2019 by Nicola Doebelin
    email                : ndoebelin@gmx.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <QFileDialog>
#include "electrondensitymapdialog.h"
#include "ui_electrondensitymapdialog.h"
#include "threadfouriersynth.h"
#include "threadrenderemap.h"
#include "../libXrdIO/colorMaps/lutgenerator.h"
#include "emapimporthandler.h"
#include "emapsettingsdialog.h"
#include "../libXrdIO/structs.h"
#include "math.h"

ElectronDensityMapDialog::ElectronDensityMapDialog(QWidget *parent) :
    QDialog(parent),
    ui(new Ui::ElectronDensityMapDialog)
{
    ui->setupUi(this);
    settings = SettingsManager::getInstance();

    cell.a = 1.0;
    cell.b = 1.0;
    cell.c = 1.0;
    cell.alpha = 90.0;
    cell.beta  = 90.0;
    cell.gamma = 90.0;
    cell.volume = 1.0;

    dmin = -1.0;
    dmax = 1.0;
    activeThreads = 0;

    LutGenerator lgen(256, true);
    lutList = lgen.getLuts();
    pdlg = nullptr;

    ui->labelStatus->setText(QString());

    setAcceptDrops(true);

    initConnections();
    initGuiElements();
    initSettings();
}

ElectronDensityMapDialog::~ElectronDensityMapDialog()
{
    saveSettings();
    delete ui;
}

void ElectronDensityMapDialog::initGuiElements()
{
    ui->comboBoxSynthesis->addItem("Fcalc");
    ui->comboBoxSynthesis->addItem("Fobs");
    ui->comboBoxSynthesis->addItem("Fobs - Fcalc");

    ui->comboBoxProjection->addItem("AB");
    ui->comboBoxProjection->addItem("AC");
    ui->comboBoxProjection->addItem("BC");

    ui->comboBoxColorMap->addItems(lutList.keys());

    ui->doubleSpinBoxRange->setMinimum(0.01);
    ui->doubleSpinBoxRange->setMaximum(99.9);
    ui->doubleSpinBoxRange->setSingleStep(0.1);
    ui->doubleSpinBoxRange->setDecimals(2);
    ui->doubleSpinBoxRange->setValue(dmax);

    ui->drawingWidget2D->showDefaultText();
}

void ElectronDensityMapDialog::initSettings()
{
    restoreGeometry(settings->value("emapDialog/geometry", QByteArray()).toByteArray());

    nCpus = settings->value("config/nThreads", QThread::idealThreadCount()).toInt();
    nCpus = nCpus == 0 ? QThread::idealThreadCount() : nCpus;

    ui->comboBoxSynthesis->setCurrentIndex(settings->value("emapDialog/fourierType", 1).toInt());

    QString lut = settings->value("emapDialog/lookupTable", QString()).toString();
    if (lut.isEmpty()) {
        ui->comboBoxColorMap->setCurrentIndex(0);
        ui->scaleWidget2D->setLut(lutList.value(ui->comboBoxColorMap->currentText()));
    } else {
        ui->comboBoxColorMap->setCurrentText(lut);
        ui->scaleWidget2D->setLut(lutList.value(lut));
    }

    ui->drawingWidget2D->setImageSize(settings->value("emapDialog/imageHeight", 768).toInt());

    projection = static_cast<projection_t>(settings->value("emapDialog/projectionPlane", 0).toInt());
    ui->comboBoxProjection->setCurrentIndex(static_cast<int>(projection));
    ui->drawingWidget2D->setProjection(projection);

    bool daxes = settings->value("emapDialog/drawAxes", true).toBool();
    ui->checkBoxDrawAxes->setChecked(daxes);
    ui->drawingWidget2D->toggleAxes(daxes);

    bool datms = settings->value("emapDialog/drawAtoms", true).toBool();
    ui->checkBoxDrawAtms->setChecked(datms);
    ui->drawingWidget2D->toggleAtoms(datms);

    bool datlbls = settings->value("emapDialog/drawAtomLabels", true).toBool();
    ui->checkBoxDrawAtomLabels->setChecked(datlbls);
    ui->drawingWidget2D->toggleAtomLabels(datlbls);

    bool preRender = settings->value("emapDialog/prerender", true).toBool();
    ui->checkBoxPreRender->setChecked(preRender);

    ui->splitter->setStretchFactor(0, 1);
    ui->splitter->setStretchFactor(1, 5);
    ui->splitter->restoreState(settings->value("emapDialog/splitter", QByteArray()).toByteArray());
}

void ElectronDensityMapDialog::initConnections()
{
    connect(ui->checkBoxDrawAtms, SIGNAL(toggled(bool)), ui->drawingWidget2D, SLOT(toggleAtoms(bool)));
    connect(ui->checkBoxDrawAtomLabels, SIGNAL(toggled(bool)), ui->drawingWidget2D, SLOT(toggleAtomLabels(bool)));
    connect(ui->checkBoxDrawAxes, SIGNAL(toggled(bool)), ui->drawingWidget2D, SLOT(toggleAxes(bool)));

    connect(ui->drawingWidget2D, SIGNAL(coordinates(float,float,float)), this, SLOT(coordinates(float,float,float)));

    connect(ui->comboBoxProjection, SIGNAL(currentIndexChanged(int)), this, SLOT(setProjection(int)));

    connect(ui->sliderLevel, SIGNAL(valueChanged(int)), this, SLOT(currentLevelChanged(int)));
    connect(ui->doubleSpinBoxRange, SIGNAL(valueChanged(double)), this, SLOT(dmaxChanged(double)));

    connect(ui->toolButtonLoad, SIGNAL(clicked(bool)), this, SLOT(fileLoad()));
    connect(ui->toolButtonCalculate, SIGNAL(clicked(bool)), this, SLOT(calculateMap()));
    connect(ui->toolButtonRender, SIGNAL(clicked(bool)), this, SLOT(renderMapImage()));
    connect(ui->toolButtonSaveCurrent, SIGNAL(clicked(bool)), this, SLOT(fileSaveAs()));
    connect(ui->toolButtonSaveAll, SIGNAL(clicked(bool)), this, SLOT(fileSaveAllAs()));

    connect(ui->comboBoxColorMap, SIGNAL(currentIndexChanged(QString)), this, SLOT(lutChanged(QString)));
    connect(ui->toolButtonSettings, SIGNAL(clicked()), this, SLOT(settingsDialog()));
}

void ElectronDensityMapDialog::saveSettings()
{
    settings->setValue("emapDialog/geometry", saveGeometry());

    settings->setValue("emapDialog/fourierType", ui->comboBoxSynthesis->currentIndex());
    settings->setValue("emapDialog/lookupTable", ui->comboBoxColorMap->currentText());

    settings->setValue("emapDialog/projectionPlane", static_cast<int>(projection));
    settings->setValue("emapDialog/drawAxes", ui->checkBoxDrawAxes->isChecked());
    settings->setValue("emapDialog/drawAtoms", ui->checkBoxDrawAtms->isChecked());
    settings->setValue("emapDialog/drawAtomLabels", ui->checkBoxDrawAtomLabels->isChecked());
    settings->setValue("emapDialog/splitter", ui->splitter->saveState());
    settings->setValue("emapDialog/prerender", ui->checkBoxPreRender->isChecked());
}

/*
 * drag and drop events to load fcf files
 */
void ElectronDensityMapDialog::dragEnterEvent(QDragEnterEvent *e)
{
    if (e->mimeData()->hasFormat("text/plain")) {
        e->acceptProposedAction();
    }
}

void ElectronDensityMapDialog::dropEvent(QDropEvent *e)
{
    QList<QUrl> urls = e->mimeData()->urls();
    if (urls.size()) {
        setFileName(urls.at(0).toLocalFile());
        e->acceptProposedAction();
    }
}

void ElectronDensityMapDialog::fileLoad()
{
    QString s(QFileDialog::getOpenFileName(this,
                                           tr("FCF file"),
                                           workingDir,
                                           "FCF file (*.fcf *.FCF);;FOU file (*.fou *.FOU)"));

    if (!QFileInfo::exists(s)) {
        qDebug() << QString("MainWindow::fileLoad(): File does not exist: %1").arg(hklFileName.absoluteFilePath());
        return;
    }

    setFileName(s);
}

void ElectronDensityMapDialog::fileSaveAs()
{
    QString fn = QFileDialog::getSaveFileName(this, tr("Save image as..."), workingDir, "PNG (*.png *.PNG)");
    if (!fn.isEmpty()) ui->drawingWidget2D->save(fn);
}


/* (GUI slot)
 *
 * saves all images in the image stack to disk.
 *
void ElectronDensityMapDialog::fileSaveAllAs()
{
    // get the base name and format for the output files
    QString fn(QFileDialog::getSaveFileName(this, tr("Save All images as..."), workingDir, "PNG (*.png *.PNG)"));

    if (fn.isEmpty()) {
        return;
    }

    // block the GUI
    setGuiState(false);

    // set up the progress dialog
    if (!pdlg) {
        pdlg = new QProgressDialog(this);
        connect(pdlg, SIGNAL(canceled()), this, SLOT(abortCalculation()));
    }

    pdlg->reset();
    pdlg->setMinimumDuration(0);
    pdlg->setMaximum(ui->sliderLevel->maximum());
    pdlg->setLabelText("Saving buffered images to disk");
    pdlg->show();
    qApp->processEvents();

    // dissect the file name into path, basename, and suffix
    QFileInfo fi(fn);
    QString path = fi.absolutePath();
    QString bn = fi.completeBaseName();
    QString ext = fi.suffix().isEmpty() ? "png" : fi.suffix();

    // loop over all levels and try to get the pre-rendered image. If successful, save it, if not, store
    // the level number for rendering
    for (int i = 0; i <= ui->sliderLevel->maximum(); ++i) {
        QString fname = QString("%1/%2_%3.%4").arg(path).arg(bn).arg(i, 4, 10, QChar('0')).arg(ext);

        qDebug() << QString("Rendering and saving image %1").arg(fname);

        QImage img(ui->drawingWidget2D->getImage(i));
        if (!img.isNull()) img.save(fname);

        pdlg->setValue(i);
        qApp->processEvents();
    }

    // hide the progress dialog and unblock the GUI
    pdlg->hide();
    pdlg->reset();
    setGuiState(true);
}
*/

void ElectronDensityMapDialog::fileSaveAllAs()
{
    // get the base name and format for the output files
    QString fn(QFileDialog::getSaveFileName(this, tr("Save All images as..."), workingDir, "PNG (*.png *.PNG)"));

    if (fn.isEmpty()) {
        return;
    }

    // block the GUI
    setGuiState(false);

    // set up the progress dialog
    if (!pdlg) {
        pdlg = new QProgressDialog(this);
        connect(pdlg, SIGNAL(canceled()), this, SLOT(abortCalculation()));
    }

    pdlg->reset();
    pdlg->setMinimumDuration(0);
    pdlg->setMaximum(ui->sliderLevel->maximum());
    pdlg->setLabelText("Saving buffered images to disk");
    pdlg->show();
    qApp->processEvents();

    // dissect the file name into path, basename, and suffix
    QFileInfo fi(fn);
    QString path = fi.absolutePath();
    QString bn = fi.completeBaseName();
    QString ext = fi.suffix().isEmpty() ? "png" : fi.suffix();

    // loop over all levels and try to get the pre-rendered image. If successful, save it, if not, store
    // the level number for rendering
    for (int i = 0; i <= ui->sliderLevel->maximum(); ++i) {
        QString fname = QString("%1/%2_%3.%4").arg(path).arg(bn).arg(i, 4, 10, QChar('0')).arg(ext);

        qDebug() << QString("Rendering and saving image %1").arg(fname);

        ui->drawingWidget2D->drawLevel(i, true);
        ui->drawingWidget2D->save(fname);

        pdlg->setValue(i);
        qApp->processEvents();
    }

    // hide the progress dialog and unblock the GUI
    pdlg->hide();
    pdlg->reset();
    setGuiState(true);
}

/*
 * set new file names for input files, either from a QFileDialog, or from drag&drop events,
 * or from a command line argument. Determines all necessary file names and calls
 * ::calculateMap()
 */
void ElectronDensityMapDialog::setFileName(const QString &s)
{
    hklFileName = s;
    workingDir = hklFileName.absolutePath();
    QString bName = workingDir + "/" + hklFileName.completeBaseName();

    QString mode = "UNKNOWN";
    if (hklFileName.suffix().toLower() == "fcf") mode = "BGMN";
    if (hklFileName.suffix().toLower() == "fou") mode = "FP";

    // bgmn output files
    if (mode == "BGMN") {
        cellFileName.setFile(bName + ".pdb"); // try this first
        cellFileName.setFile(bName + ".res"); // but prefer this
    }

    // fullprof output files
    if (mode == "FP") {
        cellFileName.setFile(bName + ".hkl"); // but prefer this
    }

    calculateMap();
}

/* (GUI slot)
 *
 * starts the calculation of the electron density map.
 * the fcf and res files must have been set before.
 *
 * This slot can also be called directly from the GUI to
 * repeat the calculation. If no file names have been set,
 * it will call ::fileLoad() first. Else it will just
 * re-read the files specified before.
 */
void ElectronDensityMapDialog::calculateMap()
{
    // no file name yet? Call slot fileLoad()
    if (hklFileName.fileName().isEmpty()) {
        fileLoad();
        return;
    }

    EmapImportHandler iHandler;
    atoms.clear();

    // read the data files from disk, return if something goes wrong
    if (hklFileName.suffix().toLower() == "fcf") {
        if (!iHandler.readFcf(hklFileName.absoluteFilePath(), cell)) return;

        if (cellFileName.suffix().toLower() == "res") {
            if (!iHandler.readRes(cellFileName.absoluteFilePath(), cell, atoms)) return;
        }

        if (cellFileName.suffix().toLower() == "pdb") {
            if (!iHandler.readPdb(cellFileName.absoluteFilePath(), cell)) return;
        }
    }

    // read the data files from disk, return if something goes wrong
    if (hklFileName.suffix().toLower() == "fou") {
        if (!iHandler.readFou(hklFileName.absoluteFilePath(), cell)) return;
        if (!iHandler.readHkl(cellFileName.absoluteFilePath(), cell)) return;
    }

    // check if there is hkl and unit cell data available
    if (!cell.hkl.size()) {
        qDebug() << QString("MainWindow::calculateMap(): No hkl data available");
        return;
    }

    if ((cell.a <= 0.0) || (cell.b <= 0.0) || (cell.c <= 0.0)) {
        qDebug() << QString("MainWindow::calculateMap(): No unit cell data available");
        return;
    }

    double res = settings->value("emapDialog/levels", 0.25).toDouble();

    levelsX = int(cell.a / res);
    levelsY = int(cell.b / res);
    levelsZ = int(cell.c / res);

    // create the transformation matrix from the unit cell
    mf2c = createF2CMatrix(cell, projection);

    // disable the GUI
    setGuiState(false);

    // set the size of the vector array to the number of levels, filled with empty vectors
    vmap.clear();
    for (int i = 0; i < levelsZ; ++i) {
        vmap.append(QVector<QVector<float> >());
    }

    // reset some thread counting variables
    nextZLevel = 0;
    activeThreads = 0;
    abort = false;

    // reset the minimum and maximum electron values
    dmax = dmin = 0.0;

    // start the timer for profiling
    startTime.start();

    // prepare and show the progress dialog
    if (!pdlg) {
        pdlg = new QProgressDialog(this);
        connect(pdlg, SIGNAL(canceled()), this, SLOT(abortCalculation()));
    }

    pdlg->setMinimumDuration(0);
    pdlg->setMaximum(levelsZ);
    pdlg->setValue(0);
    pdlg->setLabelText(QString("Computing electron density map"));
    pdlg->show();

    // start computation threads
    for (int i = 0; i < qMin(nCpus, levelsZ); ++i) {
        startFourierSynth();
    }
}

/*
 * creates a new thread to process one level (nextZLevel)
 */
void ElectronDensityMapDialog::startFourierSynth()
{
    // create a new thread
    ThreadFourierSynth *fsThread = new ThreadFourierSynth(this);

    // advance the progress dialog
    if (pdlg) pdlg->setValue(nextZLevel);

    // set up connections
    connect(fsThread, SIGNAL(newDmax(float)), this, SLOT(newDmax(float)));
    connect(fsThread, SIGNAL(newDmin(float)), this, SLOT(newDmin(float)));
    connect(fsThread, SIGNAL(finished()), this, SLOT(fourierSyntThreadCompleted()));
    connect(fsThread, SIGNAL(finished()), fsThread, SLOT(deleteLater()));

    // set up the thread parameters
    fsThread->setMode(ui->comboBoxSynthesis->currentIndex()); // 0 = Fcalc, 1 = Fobs, 2 = Fobs - Fcalc
    fsThread->setResolutionLevel(levelsX, levelsY, levelsZ, nextZLevel);
    fsThread->setVoxelMap(&vmap);
    fsThread->setHklList(cell.hkl);
    fsThread->setVolume(cell.volume);

    // run the thread
    QMutex mutex;
    QMutexLocker locker(&mutex);
    ++activeThreads;
    ++nextZLevel;
    fsThread->start();
}

/*
 * A fourier synthesis thread has terminated.
 * Checks if:
 *
 * - abort was called
 * - more levels have to be processed
 * - the electron density array is complete
 */
void ElectronDensityMapDialog::fourierSyntThreadCompleted()
{
    // reduce active thread count
    --activeThreads;

    // check if abort was clicked
    if (abort) {
        // check if all threads terminated
        if (!activeThreads) {
            // if yes, hide the progress dialog
            pdlg->hide();
            qDebug() << QString("All threads terminated");
        }

        // enable the gui and exit
        setGuiState(true);
        return;
    }

    // if there are levels left to process, start a new thread
    if (nextZLevel < levelsZ) {
        startFourierSynth();
    } else {
        // threads for all levels have been started, so we don't start any new ones
        // check if they all have terminated
        if (!activeThreads) {
            // all threads terminated
            qDebug() << QString("%1 seconds needed for fourier synthesis").arg(double(startTime.elapsed())/1000.0, 0, 'f', 2);

            // hide the progress dialog
            if (pdlg) {
                pdlg->hide();
                pdlg->reset();
            }

            // update and enable the GUI
            ui->doubleSpinBoxRange->setValue(double(dmax));
            ui->doubleSpinBoxRange->setSingleStep(double(dmax)/50.0);
            renderMapImage();
        }
    }
}

void ElectronDensityMapDialog::renderMapImage()
{
    // send all parameters to the scale widget
    ui->scaleWidget2D->setRange(dmin, dmax);
    ui->scaleWidget2D->setLut(lutList.value(ui->comboBoxColorMap->currentText()));

    // data available?
    if (!vmap.size()) {
        qDebug() << QString("MainWindow::renderMapImage(): EMap is empty, exiting.");
        setGuiState(true);
        return;
    }

    // in case the projection changed, update the gui
    int levels = 0;
    if ((projection == AB_PLANE) || (projection == C_AXIS)) levels = levelsZ - 1;
    if ((projection == AC_PLANE) || (projection == B_AXIS)) levels = levelsY - 1;
    if ((projection == BC_PLANE) || (projection == A_AXIS)) levels = levelsX - 1;

    ui->sliderLevel->setMaximum(levels);
    ui->sliderLevel->setValue(qMin(ui->sliderLevel->value(), ui->sliderLevel->maximum()));

    // block the GUI
    setGuiState(false);

    bool preRender = ui->checkBoxPreRender->isChecked();

    // send all parameters to the drawing widget
    ui->drawingWidget2D->setLut(lutList.value(ui->comboBoxColorMap->currentText()), dmax, dmin);
    ui->drawingWidget2D->setProjection(projection);
    ui->drawingWidget2D->setVmap(&vmap);
    ui->drawingWidget2D->setImageSize(settings->value("emapDialog/imageHeight", 768).toInt());
    ui->drawingWidget2D->setSuperSamplingFactor(settings->value("emapDialog/superSampling", 1).toInt());
    ui->drawingWidget2D->setInterpolation(settings->value("emapDialog/interpolation", 1).toInt());
    ui->drawingWidget2D->setMatrix(mf2c);
    ui->drawingWidget2D->setNCpus(nCpus);
    ui->drawingWidget2D->initLevel(ui->sliderLevel->value());
    ui->drawingWidget2D->setAtomList(atoms);

    // pre-render all images if requested
    if (preRender) {
        pdlg->setLabelText(QString("Rendering images"));
        pdlg->setMaximum(levels);
        pdlg->setValue(0);
        pdlg->show();

        for (int i = 0; i <= levels; ++i) {
            if (abort) break;
            ui->drawingWidget2D->preRenderImage(i, i == 0);
            pdlg->setValue(i);
            qApp->processEvents();
        }

        pdlg->hide();
    }

    ui->drawingWidget2D->drawLevel(ui->sliderLevel->value(), false);
    ui->drawingWidget2D->zoomToContent();
    setGuiState(true);
}

void ElectronDensityMapDialog::abortCalculation()
{
    abort = true;

    if (pdlg) {
        pdlg->setLabelText(QString("Waiting for threads to finish..."));
    }

    qDebug() << QString("Aborting...");
}

/*
 * a new maximum electron density was reported by one of the
 * fourier synth threads. Check if it is greater than the previous
 * one and if yes, store it in a global variable
 */
void ElectronDensityMapDialog::newDmax(float f)
{
    dmax = qMax(dmax, f);
}

/*
 * a new minimum electron density was reported by one of the
 * fourier synth threads. Check if it is smaller than the previous
 * one and if yes, store it in a global variable
 */
void ElectronDensityMapDialog::newDmin(float f)
{
    dmin = qMin(dmin, f);
}

void ElectronDensityMapDialog::dmaxChanged(double d)
{
    dmax = float(d);
    dmin = float(-d);
}

void ElectronDensityMapDialog::currentLevelChanged(int i)
{
    if (vmap.size()) {
        ui->drawingWidget2D->drawLevel(i, true);
    }
}

void ElectronDensityMapDialog::setProjection(int i)
{
    projection = static_cast<projection_t>(i);
    mf2c = createF2CMatrix(cell, projection);
}

void ElectronDensityMapDialog::lutChanged(QString s)
{
    ui->scaleWidget2D->setLut(lutList.value(s));
}

/*
 * matrix adapted from:
 * http://www.ruppweb.org/Xray/tutorial/Coordinate%20system%20transformation.htm
 */
QMatrix4x4 ElectronDensityMapDialog::createF2CMatrix(const UnitCell &uc, projection_t prj)
{
    float data[16] = {0.0, 0.0, 0.0, 0.0,
                      0.0, 0.0, 0.0, 0.0,
                      0.0, 0.0, 0.0, 0.0,
                      0.0, 0.0, 0.0, 1.0};

    // cosines
    float ca = uc.alpha == 90.0 ? 0.0 : cos(uc.alpha * M_PI / 180.0);
    float cb = uc.beta  == 90.0 ? 0.0 : cos(uc.beta  * M_PI / 180.0);
    float cc = uc.gamma == 90.0 ? 0.0 : cos(uc.gamma * M_PI / 180.0);

    // sines
    float sa = uc.alpha == 90.0 ? 1.0 : sin(uc.alpha * M_PI / 180.0);
    float sb = uc.beta  == 90.0 ? 1.0 : sin(uc.beta  * M_PI / 180.0);
    float sc = uc.gamma == 90.0 ? 1.0 : sin(uc.gamma * M_PI / 180.0);

    float v = uc.a * uc.b * uc.c * sqrt(1.0 - pow(ca, float(2.0)) - pow(cb, float(2.0)) - pow(cc, float(2.0))
                   + 2.0 * ca * cb * cc);

    // reciprocal axes
    float ra = uc.b * uc.c * sa / v;
    float rb = uc.a * uc.c * sb / v;
    float rc = uc.a * uc.b * sc / v;

    // reciprocal angle cosines
    float rca = (cb * cc - ca)/(sb * sc);
    float rcb = (ca * cc - cb)/(sa * sc);
    float rcc = (ca * cb - cc)/(sa * sb);
    Q_UNUSED(rcc);

    // write the matrix elements
    if (prj == AB_PLANE) {
        data[0] = uc.a;
        data[1] = uc.b * cc;
        data[2] = uc.c * cb;
        data[5] = -uc.b * sc; // negative, to keep the origin bottom-left instead of top-left
        data[6] = -uc.c * rca * sb;
        data[10] = 1.0 / rc;
    }

    if (prj == AC_PLANE) {
        data[0] = uc.a;
        data[1] = uc.c * cb;
        data[2] = uc.b * cc;
        data[5] = -uc.c * sb; // negative, to keep the origin bottom-left instead of top-left
        data[6] = -uc.b * rca * sc;
        data[10] = 1.0 / rb;
    }

    if (prj == BC_PLANE) {
        data[0] = uc.b;
        data[1] = uc.c * ca;
        data[2] = uc.a * cc;
        data[5] = -uc.c * sa; // negative, to keep the origin bottom-left instead of top-left
        data[6] = -uc.a * rcb * sc;
        data[10] = 1.0 / ra;
    }

    QMatrix4x4 m(data);
    return m;
}

void ElectronDensityMapDialog::coordinates(float x, float y, float z)
{
    ui->labelStatus->setText(QString("x=%1 y=%2 z=%3").arg(x, 0, 'f', 4).arg(y, 0, 'f', 4).arg(z, 0, 'f', 4));
}

/*
 * b = false: disables the GUI during long calculations
 * b = true:  enables the GUI
 */
void ElectronDensityMapDialog::setGuiState(bool b)
{
    ui->toolButtonCalculate->setEnabled(b);
    ui->toolButtonLoad->setEnabled(b);
    ui->toolButtonRender->setEnabled(b);
    ui->toolButtonSaveCurrent->setEnabled(b);
    ui->groupBox->setEnabled(b);

    if (!b) {
        qApp->setOverrideCursor(Qt::WaitCursor);
    } else {
        // a bit ugly, but the only way to make sure all
        // override cursors are reverted
        while (qApp->overrideCursor()) qApp->restoreOverrideCursor();
    }
}

void ElectronDensityMapDialog::mousePressEvent(QMouseEvent *e)
{
    e->ignore();
}

void ElectronDensityMapDialog::mergeImagesAverage()
{
    QString fn = QFileDialog::getSaveFileName(this, tr("Save merged image as..."), workingDir, "PNG (*.png *.PNG)");
    if (fn.isEmpty()) return;

    bool stretch = true;

    int max = ui->sliderLevel->maximum();
    QProgressDialog pdlg(tr("Merging images..."), tr("Cancel"), 0, max + 3, this);
    pdlg.setMinimumDuration(0);
    pdlg.show();
    qApp->processEvents();

    QImage frame(ui->drawingWidget2D->renderImage(0));

    QVector< QVector<double> > channel_r(frame.height(), QVector<double>(frame.width(), 0.0));
    QVector< QVector<double> > channel_g(frame.height(), QVector<double>(frame.width(), 0.0));
    QVector< QVector<double> > channel_b(frame.height(), QVector<double>(frame.width(), 0.0));
    QVector< QVector<double> > channel_a(frame.height(), QVector<double>(frame.width(), 0.0));

    for (int i = 0; i < max; ++i) {
        QImage img(ui->drawingWidget2D->renderImage(i));

        for (int h = 0; h < qMin(img.height(), frame.height()); ++h) {
            for (int w = 0; w < qMin(img.width(), frame.width()); ++w) {
                QRgb pixel = img.pixel(w, h);
                channel_r[h][w] += (double)qRed(pixel)   / double(max*255.0);
                channel_g[h][w] += (double)qGreen(pixel) / double(max*255.0);
                channel_b[h][w] += (double)qBlue(pixel)  / double(max*255.0);
                channel_a[h][w] += (double)qAlpha(pixel) / double(max*255.0);
            }
        }

        pdlg.setValue(i + 1);

        if (pdlg.wasCanceled()) {
            pdlg.close();
            return;
        }

        qApp->processEvents();
    }

    double dmin = 1.0;
    double dmax = 0.0;

    if (stretch) {
        qDebug() << QString("Stretching image");
        for (int h = 0; h < frame.height(); ++h) {
            for (int w = 0; w < frame.width(); ++w) {
                if (qFuzzyIsNull(channel_a[h][w])) continue;

                dmin = qMin(dmin, channel_r[h][w]);
                dmin = qMin(dmin, channel_g[h][w]);
                dmin = qMin(dmin, channel_b[h][w]);

                dmax = qMax(dmax, channel_r[h][w]);
                dmax = qMax(dmax, channel_g[h][w]);
                dmax = qMax(dmax, channel_b[h][w]);
            }
        }
    }

    pdlg.setValue(max + 2);
    pdlg.setLabelText(tr("Composing output image..."));
    qApp->processEvents();

    qDebug() << QString("Composing output image");
    QImage out(frame.width(), frame.height(), QImage::Format_RGBA8888);

    for (int h = 0; h < qMin(out.height(), frame.height()); ++h) {
        for (int w = 0; w < qMin(out.width(), frame.width()); ++w) {
            QRgb rgb = qRgba(int((channel_r[h][w] - dmin) * (255.0 / (dmax - dmin))),
                             int((channel_g[h][w] - dmin) * (255.0 / (dmax - dmin))),
                             int((channel_b[h][w] - dmin) * (255.0 / (dmax - dmin))),
                             int((channel_a[h][w] * 255.0)));
            out.setPixel(w, h, rgb);
        }
    }

    pdlg.setLabelText(tr("Saving file..."));
    pdlg.setValue(max + 3);
    qApp->processEvents();

    out.save(fn);

    qDebug() << QString("Done. File saved as %1").arg(fn);

    pdlg.close();
}

void ElectronDensityMapDialog::mergeImagesMax()
{
    QString fn = QFileDialog::getSaveFileName(this, tr("Save merged image as..."), workingDir, "PNG (*.png *.PNG)");
    if (fn.isEmpty()) return;

    int max = ui->sliderLevel->maximum();
    QProgressDialog pdlg(tr("Merging images..."), tr("Cancel"), 0, max + 1, this);
    pdlg.setMinimumDuration(0);
    pdlg.show();
    qApp->processEvents();

    QImage frame(ui->drawingWidget2D->renderImage(0));

    for (int i = 0; i < max; ++i) {
        qDebug() << QString("Rendering image %1").arg(i);

        QImage img(ui->drawingWidget2D->renderImage(i));

        for (int h = 0; h < qMin(img.height(), frame.height()); ++h) {
            for (int w = 0; w < qMin(img.width(), frame.width()); ++w) {
                QRgb pixelCur = img.pixel(w, h);
                QRgb pixelMax = frame.pixel(w, h);
                pixelMax = qRgba(qMax(qRed(pixelCur),   qRed(pixelMax)),
                                 qMax(qGreen(pixelCur), qGreen(pixelMax)),
                                 qMax(qBlue(pixelCur),  qBlue(pixelMax)),
                                 qMax(qAlpha(pixelCur), qAlpha(pixelMax)));
                frame.setPixel(w, h, pixelMax);
            }
        }

        pdlg.setValue( i + 1);

        if (pdlg.wasCanceled()) {
            pdlg.close();
            return;
        }

        qApp->processEvents();
    }

    pdlg.setLabelText(tr("Saving file..."));
    frame.save(fn);

    qDebug() << QString("Done. File saved as %1").arg(fn);

    pdlg.close();
}

void ElectronDensityMapDialog::settingsDialog()
{
    EmapSettingsDialog *sdlg = new EmapSettingsDialog(this);

    if (sdlg->exec() == QDialog::Accepted) {

    }

    delete sdlg;
}
