/***************************************************************************
                          mainwindow.cpp  -  description
                             -------------------
    begin                : Thu Apr 14 18:00:00 CEST 2011
    copyright            : (C) 2011 by Nicola Doebelin
    email                : ndoebelin@gmx.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "mainwindow.h"
#include "projectWidget/projectwidget.h"
#include "projectWidget/fpprojectwidget.h"
#include "projectWidget/bgmnprojectwidget.h"
#include "exportgraphdialog.h"
#include "strucImportDialog/strucimportdialog.h"
#include "zoomrangedialog.h"
#include "projectselectdialog.h"
#include "newinstrumentdialog.h"
#include "helpaboutdialog.h"
#include "learnprofiledialog.h"
#include "projectWidget/bgmnbackendconfig.h"
#include "../libXrdIO/structs.h"
#include "cifexportdialog.h"
#include "projectWidget/bgmnindexingerrordialog.h"
#include "ui_mainwindow.h"
#include "../quazip/JlCompress.h"
#include "wavelengthselectdialog.h"
#include "3rdparty/qsimpleupdater/include/QSimpleUpdater.h"
#include "tools/browsereferencestructuredialog/browsereferencestructuresdialog.h"

#include <QFileDialog>
#include <QInputDialog>
#include <QMessageBox>
#include <QProgressDialog>
#include <QDir>
#include <QDesktopServices>
#include <QMapIterator>
#include <QStandardPaths>
#include <QWidgetAction>
#include <QThread>
#include <QElapsedTimer>
#include <QIcon>

MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow)
{
    bool devel = false;
    settings = SettingsManager::getInstance();
    refStrManager = BgmnRefStructureManager::getInstance();
    QGuiApplication::setAttribute(Qt::AA_UseHighDpiPixmaps);

    // set the program version
    version = QString("%1.%2.%3").arg(VERSION_MAJOR).arg(VERSION_MINOR).arg(VERSION_BUILD);

    if (devel) {
        QLocale loc(QLocale::C);
        QDateTime dt = loc.toDateTime(QString(__DATE__).simplified(), "MMM dd yyyy");
        version += QString("-rc%1").arg(dt.toString("yyMMdd"));
    }

    // print some info
    qDebug() << QString("Profex version %1").arg(version);
    qDebug() << QString("Running executable %1").arg(qApp->applicationFilePath());
    qDebug() << QString("Qt compile version %1").arg(QT_VERSION_STR);
    qDebug() << QString("Qt runtime version %2").arg(qVersion());

    settings->initialize();

    // create dock widgets
    setupGui();

    // temporarily deactivate some actions in stable releases
    // until they are in a usable condition
    if (!devel) {
        ui->actionLearn_Profile->setVisible(false);
        ui->menu_Process_DIA_Files->setVisible(false);
    }

#ifdef Q_OS_LINUX
    ui->actionCheck_for_Updates->setVisible(false);
#endif

#ifndef Q_OS_WIN
    ui->actionExport_to_Excel->setVisible(false);
    ui->actionEdit_Excel_Export->setVisible(false);
#endif

    // create some dialogs as null pointers for lazy loading
    preferencesDlg = nullptr;
    spdlg = nullptr;
    asfdlg = nullptr;
    stidlg = nullptr;
    instrdlg = nullptr;
    mouseHelpDialog = nullptr;
    purifyScansDlg = nullptr;
    simScansDlg = nullptr;
    formatResultsDlg = nullptr;
    indexProgressDialog = nullptr;
    abscoeffdlg = nullptr;
    bgmninsprofdlg = nullptr;
    smdlg = nullptr;
    blDialog = nullptr;
    sSmoothDlg = nullptr;
    electMapDialog = nullptr;
    searchInFilesDlg = nullptr;
    elScatDataDlg = nullptr;
    refstrDlg = nullptr;
    scanTraceDlg = nullptr;
    waterFallPlotDlg = nullptr;
    importHandler = new ImportHandler(this);

    // load the settings
    initSettings();
    initConnections();

    toggleProjectMenusEnabled(false);

    restoreState(settings->value("window/state", QByteArray()).toByteArray(), 0);
    restoreGeometry(settings->value("window/geometry", QByteArray()).toByteArray());

    qDebug() << QString("App data location: %1").arg(settings->getAppDataLocation());
    qDebug() << QString("Temp location: %1").arg(settings->getTempLocation());
}

void MainWindow::postShowInitialization()
{
    int idx = settings->value("config/indexHkl", 0).toInt();

    if (idx == -1) {
        qDebug() << QString("MainWindow::postShowInitialization(): Initial indexing of reference structures requested by the user. Running it now, but setting the preference to false.");
        settings->setValue("config/indexHkl", 0);
        indexBgmnHkl();
    } else if (idx == 1) {
        qDebug() << QString("MainWindow::postShowInitialization(): Indexing of reference structures requested.");
        indexBgmnHkl();
    } else {
        qDebug() << QString("MainWindow::postShowInitialization(): Indexing of reference structures skipped.");
    }

    restoreOpenProjects();

    QDateTime lastUpdCheck = settings->value("config/updateLastCheck", QDateTime()).toDateTime();

    if (!lastUpdCheck.isValid()) {
        lastUpdCheck = QDateTime::currentDateTime();
        settings->setValue("config/updateLastCheck", lastUpdCheck);
    }

    QString updFreq = settings->value("config/updateFrequency", "n").toString();

    int updFreqN = -1; // updFreq == "s"
    Q_UNUSED(updFreqN);
    if (updFreq == "d") updFreqN = 0;
    if (updFreq == "w") updFreqN = 7;
    if (updFreq == "m") updFreqN = 31;
    if (updFreq == "n") updFreqN = std::numeric_limits<int>::max();

    //if (lastUpdCheck.daysTo(QDateTime::currentDateTime()) > updFreqN) {
    //    checkUpdates();
    //}
}

MainWindow::~MainWindow()
{
    qDebug() << QString("MainWindow::~MainWindow(): Destructor called");

    // properly delete all projects and related widgets
    while (widgetStack->count() > 1) {
        ProjectWidget *pw = static_cast<ProjectWidget*>(widgetStack->widget(1));

        if (pw) {
            dwScans->removeWidget(pw->scanList);
            dwOutput->removeWidget(pw->refOutput);
            dwChemistry->removeWidget(pw->chemOutput);
            dwPeakIntegration->removeWidget(pw->peakIntegrationWidget);
            dwConvergence->removeWidget(pw->convDisplay);
            dwPeakList->removeWidget(pw->peakListWidget);
            dwSearchMatch->removeWidget(pw->searchMatchWidget);
            dwResultsTree->removeWidget(pw->getResultsTree());
            dwPeakFit->removeWidget(pw->peakFitWidget);

            projectToolbarStack->removeWidget(pw->toolbarWidget);
            widgetStack->removeWidget(pw);
            delete pw;
        }
    }

    twProjects->clear();

    // delete global dialogs
    // check if lazy-loaded dialogs were created
    if (preferencesDlg)      delete preferencesDlg;
    if (stidlg)              delete stidlg;
    if (spdlg)               delete spdlg;
    if (asfdlg)              delete asfdlg;
    if (instrdlg)            delete instrdlg;
    if (mouseHelpDialog)     delete mouseHelpDialog;
    if (indexProgressDialog) delete indexProgressDialog;
    if (abscoeffdlg)         delete abscoeffdlg;
    if (bgmninsprofdlg)      delete bgmninsprofdlg;
    if (smdlg)               delete smdlg;
    if (blDialog)            delete blDialog;
    if (sSmoothDlg)          delete sSmoothDlg;
    if (purifyScansDlg)      delete purifyScansDlg;
    if (simScansDlg)         delete simScansDlg;
    if (formatResultsDlg)    delete formatResultsDlg;
    if (electMapDialog)      delete electMapDialog;
    if (elScatDataDlg)       delete elScatDataDlg;
    if (refstrDlg)           delete refstrDlg;
    if (presetMenuMgr)       delete presetMenuMgr;
    if (scanTraceDlg)        delete scanTraceDlg;
    if (waterFallPlotDlg)    delete waterFallPlotDlg;

    delete importHandler;
    delete menuRecentGraphs;
    delete menuRecentText;
    delete menuPresets;
    delete menuTextBlocks;
    delete menuLocationsStructures;
    delete menuLocationsDevices;
    delete menuLocationsPresets;
    if (logLocationAction) delete logLocationAction;

    delete ui;
    qDebug() << QString("Mainwindow::~MainWindow(): UI deleted");

    // delete the singletons
    refStrManager->destroy();
    settings->destroy();

    qDebug() << QString("Mainwindow::~MainWindow(): Exiting... good bye!");
}

bool MainWindow::doShowMaximized()
{
    return settings->value("window/maximized", false).toBool();
}

void MainWindow::changeEvent(QEvent *e)
{
    QMainWindow::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
        ui->retranslateUi(this);
        break;
    default:
        break;
    }
}

void MainWindow::closeEvent(QCloseEvent *e)
{
    qDebug() << QString("MainWindow::closeEvent(): Saving settings...");

    saveSettings();
    settings->sync();

    qDebug() << QString("Mainwindow::closeEvent(): Settings saved");

    // note: We have to delete the sysTrayIcon when closing the main window. Else
    // it keeps on running after closing the main window on some systems (KDE5),
    // the destructor is never called, and the event loop never exits.
    if (sysTrayIcon) delete sysTrayIcon;
    e->accept();
}

/*
 * Creates dock widgets, toolbar, and status bar
 */
void MainWindow::setupGui()
{
    // the icon theme setting must be read here already
    QString iconTheme = settings->value("config/iconTheme", QString()).toString();

    if (iconTheme.isEmpty()) {
        if (settings->isDarkMode()) {
            // assuming dark gui mode
            iconTheme = QString("profex-dark");
        } else {
            // assuming light gui mode
            iconTheme = QString("profex-light-colored");
        }
    }

    QIcon::setThemeName(iconTheme);

    setDockNestingEnabled(true);

    // construct the main GUI
    ui->setupUi(this);
    setWindowTitle(QString("Profex %1").arg(version));
    widgetStack = new QStackedWidget(this);
    setCentralWidget(widgetStack);
    setAcceptDrops(true);

    backdropWidget = new BackdropWidget;
    widgetStack->addWidget(backdropWidget);

    // create the systray icon
    sysTrayIcon = new QSystemTrayIcon(QIcon(":/icons/profex4.png"), this);
    sysTrayIcon->setToolTip(tr("Idle"));
    sysTrayIcon->show();

    // create widgetstack in project toolbar
    projectToolbarStack = new QStackedWidget(Q_NULLPTR);
    referenceToolBar = addToolBar(tr("Reference Structure Toolbar"));
    referenceToolBar->setObjectName("tbProject");
    referenceToolBar->addWidget(projectToolbarStack);

    // Project list
    dwProjects = new QDockWidget(tr("Projects"), this);
    dwProjects->setObjectName("dwProjects");
    twProjects = new ProjectsTreeWidget(this);
    twProjects->setSelectionMode(QAbstractItemView::ExtendedSelection);
    twProjects->setContextMenuPolicy(Qt::CustomContextMenu);
    twProjects->setColumnCount(2);
    QStringList header;
    header << tr("Name") << tr("Status");
    twProjects->setHeaderLabels(header);
    dwProjects->setWidget(twProjects);
    addDockWidget(Qt::LeftDockWidgetArea, dwProjects);
    ui->menu_Window->addAction(dwProjects->toggleViewAction());

    // Scans list
    dwScans = new ProjectWidgetDock(tr("Plot Options"), this);
    dwScans->setObjectName("dwScans");
    addDockWidget(Qt::LeftDockWidgetArea, dwScans);
    ui->menu_Window->addAction(dwScans->toggleViewAction());
    dwScans->toggleVisibility(true);

    tabifyDockWidget(dwProjects, dwScans);
    dwProjects->raise();

    // Refinement log output
    dwOutput = new ProjectWidgetDock(tr("Refinement Protocol"), this);
    dwOutput->setObjectName("dwOutput");
    addDockWidget(Qt::BottomDockWidgetArea, dwOutput);
    ui->menu_Window->addAction(dwOutput->toggleViewAction());
    dwOutput->toggleVisibility(true);

    // Chemistry output
    dwChemistry = new ProjectWidgetDock(tr("Chemical Composition"), this);
    dwChemistry->setObjectName("dwChemistry");
    addDockWidget(Qt::BottomDockWidgetArea, dwChemistry);
    ui->menu_Window->addAction(dwChemistry->toggleViewAction());
    dwChemistry->close();
    dwChemistry->toggleVisibility(false);

    // Context help
    dwContextHelp = new QDockWidget(tr("Context Help"), this);
    dwContextHelp->setObjectName("dwContextHelp");
    chelp = new ContextHelpDisplay(dwContextHelp);
    dwContextHelp->setWidget(chelp);
    addDockWidget(Qt::RightDockWidgetArea, dwContextHelp);
    ui->menu_Window->addAction(dwContextHelp->toggleViewAction());
    dwContextHelp->close();

    // Convergence graph
    dwConvergence = new ProjectWidgetDock(tr("Convergence Progress"), this);
    dwConvergence->setObjectName("dwConvergence");
    addDockWidget(Qt::LeftDockWidgetArea, dwConvergence);
    ui->menu_Window->addAction(dwConvergence->toggleViewAction());
    dwConvergence->toggleVisibility(true);

    // Peak Integration tables
    dwPeakIntegration = new ProjectWidgetDock(tr("Peak Integrals"), this);
    dwPeakIntegration->setObjectName("dwPeakIntegration");
    addDockWidget(Qt::BottomDockWidgetArea, dwPeakIntegration);
    ui->menu_Window->addAction(dwPeakIntegration->toggleViewAction());
    dwPeakIntegration->close();
    dwPeakIntegration->toggleVisibility(false);

    // Peak list widget
    dwPeakList = new ProjectWidgetDock(tr("Peak List"), this);
    dwPeakList->setObjectName("dwPeakList");
    addDockWidget(Qt::RightDockWidgetArea, dwPeakList);
    ui->menu_Window->addAction(dwPeakList->toggleViewAction());
    dwPeakList->close();
    dwPeakList->toggleVisibility(false);

    // Search/Match widget
    dwSearchMatch = new ProjectWidgetDock(tr("Search/Match Phases"), this);
    dwSearchMatch->setObjectName("dwSearchMatch");
    addDockWidget(Qt::RightDockWidgetArea, dwSearchMatch);
    ui->menu_Window->addAction(dwSearchMatch->toggleViewAction());
    dwSearchMatch->close();
    dwSearchMatch->toggleVisibility(false);
    ui->actionRun_Search_Match->setEnabled(false);
    connect(dwSearchMatch, SIGNAL(visibilityChanged(bool)), ui->actionRun_Search_Match, SLOT(setEnabled(bool)));

    // results tree widget
    dwResultsTree = new ProjectWidgetDock(tr("Refined Parameters"), this);
    dwResultsTree->setObjectName("dwResultsTree");
    addDockWidget(Qt::BottomDockWidgetArea, dwResultsTree);
    ui->menu_Window->addAction(dwResultsTree->toggleViewAction());
    dwResultsTree->toggleVisibility(true);

    dwPeakFit = new ProjectWidgetDock(tr("Peak Fitting"), this);
    dwPeakFit->setObjectName("dwPeakFit");
    addDockWidget(Qt::RightDockWidgetArea, dwPeakFit);
    ui->menu_Window->addAction(dwPeakFit->toggleViewAction());
    dwPeakFit->close();
    dwPeakFit->toggleVisibility(false);

    tabifyDockWidget(dwProjects, dwScans);
    tabifyDockWidget(dwResultsTree, dwChemistry);
    tabifyDockWidget(dwChemistry, dwPeakIntegration);
    tabifyDockWidget(dwPeakList, dwSearchMatch);
    dwProjects->raise();
    dwResultsTree->raise();

    // create widgets in status bar
    statusLambda = new QLabel(statusBar());
    statusAng = new QLabel(statusBar());
    statusInt = new QLabel(statusBar());
    statusD = new QLabel(statusBar());
    statusCursor = new QLabel(statusBar());
    statusFileName = new QLabel(statusBar());
    statusOpenProjects = new QLabel(statusBar());

    statusLambda->setFrameStyle(QFrame::StyledPanel | QFrame::Sunken);
    statusAng->setFrameStyle(QFrame::StyledPanel | QFrame::Sunken);
    statusInt->setFrameStyle(QFrame::StyledPanel | QFrame::Sunken);
    statusD->setFrameStyle(QFrame::StyledPanel | QFrame::Sunken);
    statusCursor->setFrameStyle(QFrame::StyledPanel | QFrame::Sunken);
    statusFileName->setFrameStyle(QFrame::StyledPanel | QFrame::Sunken);
    statusOpenProjects->setFrameStyle(QFrame::StyledPanel | QFrame::Sunken);

    statusLambda->setMinimumWidth(statusLambda->fontMetrics().boundingRect(QString(tr("%1 = m.mmmmmm").arg(global::lambda))).width());
    statusAng->setMinimumWidth(statusAng->fontMetrics().boundingRect(QString(tr("2%1 = mmmmmmmm").arg(global::theta))).width());
    statusInt->setMinimumWidth(statusInt->fontMetrics().boundingRect(QString(tr("I = mmmmmmmmmmm"))).width());
    statusD->setMinimumWidth(statusD->fontMetrics().boundingRect(QString(tr("d = mmmmmmmm"))).width());
    statusCursor->setMinimumWidth(statusCursor->fontMetrics().boundingRect(QString(tr("Line: mmmm, Column: mmm"))).width());
    statusFileName->setMinimumWidth(1);
    statusOpenProjects->setMinimumWidth(statusOpenProjects->fontMetrics().boundingRect(QString(tr("999 Projects"))).width());

    statusBar()->addPermanentWidget(statusOpenProjects, 0);
    statusBar()->addPermanentWidget(statusLambda, 0);
    statusBar()->addPermanentWidget(statusAng, 0);
    statusBar()->addPermanentWidget(statusInt, 0);
    statusBar()->addPermanentWidget(statusD, 0);
    statusBar()->addPermanentWidget(statusCursor, 0);
    statusBar()->addWidget(statusFileName, 0);

    menuRecentGraphs = new QMenu(QString(tr("Recent Graph Files")), this);
    menuRecentText = new QMenu(QString(tr("Recent Text Files")), this);
    ui->menu_File->insertMenu(ui->actionPrint, menuRecentGraphs);
    ui->menu_File->insertMenu(ui->actionPrint, menuRecentText);
    ui->menu_File->insertSeparator(ui->actionPrint);

    menuPresets = new QMenu(QString(tr("Refinement Presets")), this);
    menuPresets->setIcon(QIcon::fromTheme("profex-preset"));
    ui->menuProject->insertMenu(ui->actionCreatePreset, menuPresets);

    presetMenuMgr = new PresetMenuManager;
    presetMenuMgr->setMenu(menuPresets);
    presetMenuMgr->update();
    connect(presetMenuMgr, SIGNAL(sigApplyPreset(QString)), this, SLOT(applyPreset(QString)));

    menuLocationsStructures = new QMenu(tr("Structure Repositories"), this);
    menuLocationsDevices    = new QMenu(tr("Device Repositories"), this);
    menuLocationsPresets    = new QMenu(tr("Preset Repositories"), this);
    ui->menu_Locations->addMenu(menuLocationsStructures);
    ui->menu_Locations->addMenu(menuLocationsDevices);
    ui->menu_Locations->addMenu(menuLocationsPresets);

#ifndef Q_OS_LINUX
    logLocationAction = new QAction(tr("Log File"), this);
    logLocationAction->setData("LOGFILE");
    ui->menu_Locations->addSeparator();
    ui->menu_Locations->addAction(logLocationAction);
#else
    logLocationAction = nullptr;
#endif

    // create a tool button in the projectToolBar showing the menuPresets menu
    QToolButton *presetButton = new QToolButton(this);
    presetButton->setToolTip(tr("Create the control file from a refinement preset"));
    presetButton->setText(tr("Presets"));
    presetButton->setMenu(menuPresets);
    presetButton->setPopupMode(QToolButton::InstantPopup);
    presetButton->setIcon(QIcon::fromTheme("profex-preset"));
    ui->projectToolBar->addWidget(presetButton);
    menuButtons.append(presetButton);

    menuTextBlocks = new QMenu(QString(tr("Insert Text Block")), this);
    ui->menu_Edit->insertMenu(ui->actionRevertControlFile, menuTextBlocks);

    // create a tool button in the projectToolBar showing the textBlocks menu
    QToolButton *textBlockButton = new QToolButton(this);
    textBlockButton->setText(tr("Text blocks"));
    textBlockButton->setToolTip(tr("Insert text block at cursor position"));
    textBlockButton->setMenu(menuTextBlocks);
    textBlockButton->setPopupMode(QToolButton::InstantPopup);
    textBlockButton->setIcon(QIcon::fromTheme("profex-text-block"));
    ui->projectToolBar->addWidget(textBlockButton);
    menuButtons.append(textBlockButton);
}

/*
 * initialize the program's settings
 */
void MainWindow::initSettings()
{
    if (settings->format() == QSettings::NativeFormat)  qDebug() << QString("MainWindow::initSettings(): Settings format Native");
    if (settings->format() == QSettings::IniFormat)     qDebug() << QString("MainWindow::initSettings(): Settings format INI");
    if (settings->format() == QSettings::InvalidFormat) qDebug() << QString("MainWindow::initSettings(): Settings format Invalid");
    qDebug() << QString("MainWindow::initSettings(): Reading settings from %1").arg(settings->fileName());

    twProjects->header()->restoreState(settings->value("config/projectsListHeader", QByteArray()).toByteArray());

    // prepare some settings, depending on whether or not an older version was installed before
    if (settings->value("bgmnProject/quantGoalFormat", -1).toInt() == -1) {
        if (hasOldPreviousInstallation(3, 12, 0)) {
            settings->setValue("bgmnProject/quantGoalFormat", 0);
        } else {
            settings->setValue("bgmnProject/quantGoalFormat", 1);
        }
    }

    // try to read new format of reported goals strings
    QStringList gGoals = settings->value("bgmnProject/reportedGlobalGoals", QString()).toStringList();

    // if empty, check if there is a string in the old format
    if (gGoals.isEmpty()) {
        gGoals = settings->value("bgmnProject/includedGlobalGoals", QString()).toStringList();

        // if an old settings string was found, make sure it contains some mandatory entries and
        // immediately write it back in the new format
        if (gGoals.size()) {
            if (!gGoals.contains(QString("^Q\\S+$"))) gGoals.append(QString("^Q\\S+$"));
            if (!gGoals.contains(QString("^Qabs\\S+$"))) gGoals.append(QString("^Qabs\\S+$"));
            settings->setValue("bgmnProject/reportedGlobalGoals", gGoals);
        } else { // if also empty, get the default strings
            gGoals = global::defaultBgmnGlobalGoals.split("\n");
        }
    }

    workingDir = settings->value("config/workingdir", QDir::homePath()).toString();
    lastOpenTextFilter = settings->value("config/lastOpenTextFilter", "").toString();
    lastOpenGraphFilter = settings->value("config/lastOpenGraphFilter", "").toString();
    lastOpenProjectFilter = settings->value("config/lastOpenProjectFilter", "").toString();

    recentFilesGraph = settings->value("config/recentFilesGraph", QStringList()).toStringList();
    recentFilesText = settings->value("config/recentFilesText", QStringList()).toStringList();
    setupRecentFiles();

    QMap<QString, QVariant> textBlocksFallback;
    for (int i = 0; i < 2*global::defaultTextBlockNumber; i += 2) {
        textBlocksFallback[global::defaultTextBlocks[i]] = global::defaultTextBlocks[i+1];
    }

    initTextBlocks(settings->value("config/textBlocks", textBlocksFallback).toMap());

    // default file extensions associated with the different project types
    QStringList fpFext;
    QStringList bgmnFext;

    fpFext << "prf" << "pcr" << "sum";
    bgmnFext << "dia" << "sav" << "lst" << "par" << "val";

    fpFext = settings->value("fpProject/fileExtensions", fpFext).toStringList();
    bgmnFext = settings->value("bgmnProject/fileExtensions", bgmnFext).toStringList();

    for (int i = 0; i < fpFext.size(); ++i) {
        fExtProjectType.insert(fpFext.at(i), 0);
    }

    for (int i = 0; i < bgmnFext.size(); ++i) {
        fExtProjectType.insert(bgmnFext.at(i), 1);
    }

    // no project has indexed the hkl files yet, so we have to initialize the "config/sessionHklIndexed"
    // flag to false. It is correct to *write* this settings value here. It is an initialization.
    settings->setValue("config/sessionHklIndexed", false);

    // initialize the reference structure manager
    refStrManager->init();
    hklBufferDb = refStrManager->getHklBufferFileName();
    connect(refStrManager, SIGNAL(signalUpdateComplete(QStringList)), this, SLOT(indexBgmnHklComplete(QStringList)));

    // read some more settings and try to guess reasonable defaults
    BgmnBackendConfig bgCfg;
    bool hasValidConfig = bgCfg.checkExecutables();
    bool hasValidRepos  = bgCfg.checkRepositories();

    if (!hasValidConfig || !hasValidRepos) {
        QString msgTitle(tr("BGMN Backend Configuration"));
        QString msgText(tr("No valid BGMN configuration found. Would you like to run auto-configuration?"));

        int autoConf = QMessageBox::question(this, msgTitle, msgText, QMessageBox::Yes, QMessageBox::No);

        if (autoConf == QMessageBox::Yes) {
            if (!hasValidConfig) hasValidConfig = bgCfg.guessExecutables();
            if (!hasValidRepos)  hasValidRepos  = bgCfg.guessRepositories();

            if (hasValidConfig && hasValidRepos) {
                QFileInfo fiBgmnEx(settings->value("bgmnProject/bgmnExec", QString()).toStringList().first());

                msgText = QString(tr("Auto-configuration was successful.\n\n"
                                     "Would you like to index the structure file repository now?\n"
                                     "This will take up to several minuts."));

                int autoIndex = QMessageBox::question(this, msgTitle, msgText, QMessageBox::Yes, QMessageBox::No);

                if (autoIndex == QMessageBox::Yes) {
                    settings->setValue("config/indexHkl", -1);
                }
            } else {
                msgText = QString(tr("Auto-detection of %1%2%3 failed. Please verify the configuration in the following dialog.")
                                  .arg(!hasValidConfig ? "BGMN" : "")
                                  .arg((!hasValidConfig && !hasValidRepos) ? " and " : "")
                                  .arg(!hasValidRepos ? "repositories" : ""));
                QMessageBox::warning(this, msgTitle, msgText);

                if (!hasValidConfig) {
                    preferences("Backend Configuration"); // if config or both failed
                } else if (!hasValidRepos) {
                    preferences("Repositories");          // if only repos failed
                }
            }
        }
    }

    ui->actionLegend->setChecked(settings->value("graph/drawLegend", true).toBool());
    ui->actionHkl_Indices->setChecked(settings->value("graph/drawHkl", true).toBool());
    ui->actionShow_Hide_Phase_Patterns->setChecked(settings->value("graph/phaseVisibility", true).toBool());
    updateLocationsMenu();

    Qt::ToolButtonStyle toolStyle = static_cast<Qt::ToolButtonStyle>(settings->value("window/toolButtonLayout", Qt::ToolButtonIconOnly).toInt());

    ui->mainToolBar->setToolButtonStyle(toolStyle);
    ui->projectToolBar->setToolButtonStyle(toolStyle);
    ui->refinementToolBar->setToolButtonStyle(toolStyle);
    referenceToolBar->setToolButtonStyle(toolStyle);

    for (int i = 0; i < menuButtons.size(); ++i) {
        menuButtons.at(i)->setToolButtonStyle(toolStyle);
    }

    // clear old data from the spacegroup.dat dialog. It will be re-loaded from the settings whe
    // the dialog is opened the next time. We don't reload it here, because it would delay
    // program start even if we don't need the dialog at all
    if (spdlg) spdlg->clear();

    chelp->parseHelpFiles();
}

/*
   store the program's settings
*/
void MainWindow::saveSettings()
{
        settings->setValue("window/state", saveState(0));
        settings->setValue("window/geometry", saveGeometry());
        settings->setValue("window/size", size());
        settings->setValue("window/maximized", isMaximized());
        settings->setValue("config/projectsListHeader", twProjects->header()->saveState());

        settings->setValue("config/workingdir", workingDir);
        settings->setValue("config/lastOpenTextFilter", lastOpenTextFilter);
        settings->setValue("config/lastOpenGraphFilter", lastOpenGraphFilter);
        settings->setValue("config/lastOpenProjectFilter", lastOpenProjectFilter);

        QStringList rfGraph;
        QStringList rfText;

        if (recentFilesGraph.size()) {
            rfGraph = recentFilesGraph.mid(qMax(0, recentFilesGraph.length() - 10),
                                           qMin(recentFilesGraph.length(), 10));
        }

        if (recentFilesText.size()) {
            rfText = recentFilesText.mid(qMax(0, recentFilesText.length() - 10),
                                         qMin(recentFilesText.length(), 10));
        }

        settings->setValue("config/recentFilesGraph", rfGraph);
        settings->setValue("config/recentFilesText", rfText);

        settings->setValue("graph/drawLegend", ui->actionLegend->isChecked());
        settings->setValue("graph/drawHkl", ui->actionHkl_Indices->isChecked());

        // collect all open graph files, so we can restore the open projects
        // next time.
        QStringList openGraphFiles;
        QList<ProjectWidget *> openProjects = getAllProjects();
        for (int i = 0; i < openProjects.size(); ++i) {
            openGraphFiles.append(openProjects.at(i)->getGraphFile());
        }

        settings->setValue("config/openProjects", openGraphFiles);
        settings->setValue("config/ProfexVersion", version);
}

/*
 * initialize the Mainwindow's connections
 */
void MainWindow::initConnections()
{
    connect(twProjects, SIGNAL(currentItemChanged(QTreeWidgetItem*,QTreeWidgetItem*)), this, SLOT(setCurrentProject(QTreeWidgetItem*,QTreeWidgetItem*)));
    connect(twProjects, SIGNAL(customContextMenuRequested(QPoint)), this, SLOT(projectListContextMenu(QPoint)));
    connect(menuRecentGraphs, SIGNAL(triggered(QAction*)), this, SLOT(openRecentGraph(QAction*)));
    connect(menuRecentText, SIGNAL(triggered(QAction*)), this, SLOT(openRecentText(QAction*)));
    connect(menuTextBlocks, SIGNAL(triggered(QAction*)), this, SLOT(applyTextBlock(QAction*)));
    connect(menuLocationsStructures, SIGNAL(triggered(QAction*)), this, SLOT(openLocation(QAction*)));
    connect(menuLocationsDevices,    SIGNAL(triggered(QAction*)), this, SLOT(openLocation(QAction*)));
    connect(menuLocationsPresets,    SIGNAL(triggered(QAction*)), this, SLOT(openLocation(QAction*)));
    if (logLocationAction) connect(logLocationAction, SIGNAL(triggered()), this, SLOT(openLogFileLocation()));
}

void MainWindow::checkUpdates()
{
    bool useSsl = settings->value("config/updateSsl", false).toBool();

    QString url(useSsl
                ? "https://www.profex-xrd.org/wp-content/uploads/update-ssl.json"
                : "http://www.profex-xrd.org/wp-content/uploads/update.json");
    QString ptf;

#ifdef Q_OS_WIN64
        ptf = "win64";
#else
    #ifdef Q_OS_WIN32
        ptf = "win32";
    #endif
#endif
#ifdef Q_OS_MACOS
    #ifdef Q_PROCESSOR_X86_32
        ptf = "osx_x32";
    #else
        #ifdef Q_PROCESSOR_X86_64
            ptf = "osx_x64";
        #endif
    #endif
#endif

    if (ptf.isEmpty()) return;

    QSimpleUpdater *upd = QSimpleUpdater::getInstance();
    upd->setDownloaderEnabled(url, true);
    upd->setNotifyOnFinish(url, false);
    upd->setNotifyOnUpdate(url, true);
    upd->setModuleName(url, "Profex");
    upd->setModuleVersion(url, version);
    upd->setPlatformKey(url, ptf);
    upd->checkForUpdates(url);

    settings->setValue("config/updateLastCheck", QDateTime::currentDateTime());
}

void MainWindow::restoreOpenProjects()
{
    if (settings->value("config/restoreProjects", false).toBool()) {
        QStringList openProjects = settings->value("config/openProjects", QStringList()).toStringList();

        int npr = openProjects.size();

        if (npr > 20) {
            QMessageBox msgBox;
            msgBox.setText(QString(tr("%1 projects will be restored. This may take a while.\nHow do you want to proceed?")).arg(npr));
            QPushButton *allButton = msgBox.addButton(tr("Load all"), QMessageBox::AcceptRole);
            QPushButton *tenButton = msgBox.addButton(tr("Load first 10"), QMessageBox::RejectRole);
            msgBox.addButton(tr("Skip loading"), QMessageBox::DestructiveRole);

            msgBox.exec();

            if (msgBox.clickedButton() == allButton) {
                loadGraphFiles(openProjects, QString());
                qDebug() << QString("MainWindow::restoreOpenProjects(): %1 projects should be restored, user confirmed loading all").arg(npr);
            } else if (msgBox.clickedButton() == tenButton) {
                loadGraphFiles(openProjects.mid(0, 10), QString());
                qDebug() << QString("MainWindow::restoreOpenProjects(): %1 projects should be restored, but user clicked first 10 only").arg(npr);
            } else {
                qDebug() << QString("MainWindow::restoreOpenProjects(): %1 projects should be restored, but user clicked none").arg(npr);
            }
        } else {
            qDebug() << QString("MainWindow::restoreOpenProjects(): Restoring %1 projects").arg(openProjects.size());
            loadGraphFiles(openProjects, QString());
        }
    } else {
        qDebug() << QString("MainWindow::restoreOpenProjects(): Restoring not requested in settings.");
    }
}

/*
 * populates the Edit->TextBlocks menu with actins
 */
void MainWindow::initTextBlocks(const QMap<QString, QVariant> &blocks)
{
    menuTextBlocks->clear();

    QMap<QString, QVariant>::ConstIterator it = blocks.constBegin();

    while (it != blocks.constEnd()) {
        QAction *action = new QAction(it.key(), this);
        action->setData(it.value());
        menuTextBlocks->addAction(action);
        ++it;
    }

}

/*
 * slot to open preferences dialog
 */
void MainWindow::preferences()
{
    preferences(QString());
}

/*
 * opens the preferences dialog, raises page with title "str". Use
 * empty string to raise the first page. Str is the text shown in the
 * tree widget item of the preferences dialog's navigation list.
 */
void MainWindow::preferences(const QString &str)
{
    if (!preferencesDlg) {
        preferencesDlg = new PreferencesDialog(this);
        connect(preferencesDlg, SIGNAL(sigApplySettings()), this, SLOT(applyPreferences()));
    } else {
        preferencesDlg->updateSettings();
    }

    if (!str.isEmpty()) {
        preferencesDlg->raisePage(str);
    }

    // apply settings in real time
    if (preferencesDlg->exec() == QDialog::Accepted) {
        applyPreferences();
    }
}

void MainWindow::applyPreferences()
{
    // write settings to disk. It would not be necessary here (will be done in the
    // destructor anyway), but if the program crashes, the settings applied here
    // will be persistent
    settings->sync();

    // settings for the main window
    Qt::ToolButtonStyle tbStyle = static_cast<Qt::ToolButtonStyle>(settings->value("window/toolButtonLayout", 4).toInt());
    ui->mainToolBar->setToolButtonStyle(tbStyle);
    ui->projectToolBar->setToolButtonStyle(tbStyle);
    ui->refinementToolBar->setToolButtonStyle(tbStyle);
    referenceToolBar->setToolButtonStyle(tbStyle);

    for (int i = 0; i < menuButtons.size(); ++i) {
        menuButtons.at(i)->setToolButtonStyle(tbStyle);
    }

    chelp->parseHelpFiles();
    updateLocationsMenu();

    QString newIcons = settings->value("config/iconTheme", QString()).toString();

    if (newIcons != QIcon::themeName()) {
        QIcon::setThemeName(newIcons);
        update();
    }

    // settings for projects
    for (int i = 0; i < twProjects->topLevelItemCount(); ++i) {
        ProjectWidget *pw = getProjectFromUid(twProjects->topLevelItem(i)->data(0, Qt::UserRole).toUuid());

        if (pw) {
            pw->initSettings();
            qDebug() << QString("MainWindow::preferences(): Updating settings of project %1").arg(pw->baseName());
        } else {
            qDebug() << QString("MainWindow::preferences(): Project %1 is not a valid project").arg(i);
        }
    }
}

void MainWindow::updateLocationsMenu()
{
    QStringList strDirs = settings->value("bgmnProject/structureDatabase", QStringList()).toStringList();
    QStringList devDirs = settings->value("bgmnProject/deviceDatabase", QStringList()).toStringList();
    QStringList prsDirs = settings->value("bgmnProject/presetDirectory", QStringList()).toStringList();

    QList<QAction *> lStr = menuLocationsStructures->actions();
    QList<QAction *> lDev = menuLocationsDevices->actions();
    QList<QAction *> lPrs = menuLocationsPresets->actions();

    menuLocationsStructures->clear();
    menuLocationsDevices->clear();
    menuLocationsPresets->clear();

    while (!lStr.isEmpty()) {
        QAction *act = lStr.takeFirst();
        if (act) delete act;
    }

    while (!lDev.isEmpty()) {
        QAction *act = lDev.takeFirst();
        if (act) delete act;
    }

    while (!lPrs.isEmpty()) {
        QAction *act = lPrs.takeFirst();
        if (act) delete act;
    }

    for (int i = 0; i < strDirs.size(); ++i) {
        QAction *act = new QAction(strDirs.at(i), this);
        act->setData(strDirs.at(i));
        menuLocationsStructures->addAction(act);
    }

    for (int i = 0; i < devDirs.size(); ++i) {
        QAction *act = new QAction(devDirs.at(i), this);
        act->setData(devDirs.at(i));
        menuLocationsDevices->addAction(act);
    }

    for (int i = 0; i < prsDirs.size(); ++i) {
        QAction *act = new QAction(prsDirs.at(i), this);
        act->setData(prsDirs.at(i));
        menuLocationsPresets->addAction(act);
    }
}

void MainWindow::openLocation(QAction *act)
{
    QUrl url = QUrl::fromLocalFile(act->data().toString());

    if (QDesktopServices::openUrl(url)) {
        qDebug() << QString("MainWindow::openLocation(): Opening %1").arg(url.toDisplayString());
    } else {
        qDebug() << QString("MainWindow::openLocation(): Opening %1 failed").arg(url.toDisplayString());
    }
}

void MainWindow::openLogFileLocation()
{
    QFileInfo fi(logDest);
    QUrl url = QUrl::fromLocalFile(fi.absolutePath());

    if (QDesktopServices::openUrl(url)) {
        qDebug() << QString("MainWindow::openLogFileLocation(): Opening %1").arg(url.toDisplayString());
    } else {
        qDebug() << QString("MainWindow::openLogFileLocation(): Opening %1 failed").arg(url.toDisplayString());
    }
}

/*
 * raises tab t on all open projects
 */
void MainWindow::raiseTab(int t)
{
    QString s = getCurrentProject()->tabText(t);

    for (int i = 0; i < twProjects->topLevelItemCount(); ++i) {
        ProjectWidget *pw = getProjectFromUid(twProjects->topLevelItem(i)->data(0, Qt::UserRole).toUuid());
        if (pw) pw->raiseTab(s);
    }
}

void MainWindow::projectListContextMenu(QPoint p)
{
    QTreeWidgetItem *it = twProjects->itemAt(p);
    if (!it) return;

    while (it->parent()) it = it->parent();

    QUuid uid = it->data(0, Qt::UserRole).toUuid();

    ProjectWidget *pw = getProjectFromUid(uid);
    if (!pw) return;

    QMenu plContextMenu(this);
    plContextMenu.addAction(QIcon::fromTheme("profex-run"), tr("Start refinement"), pw, SLOT(runRefinement()));
    plContextMenu.addAction(QIcon::fromTheme("profex-document-close"), tr("Close selected projects"), this, SLOT(fileCloseCurrentProject()));
    plContextMenu.addSeparator();
    plContextMenu.addAction(QIcon::fromTheme("profex-save-as"), tr("Export project data"), this, SLOT(exportProjectInfo()));
    plContextMenu.addSeparator();
    plContextMenu.addAction(tr("Expand all"), twProjects, SLOT(expandAll()));
    plContextMenu.addAction(tr("Collapse all"), twProjects, SLOT(collapseAll()));
    plContextMenu.exec(twProjects->mapToGlobal(p));
}

/*
 * this function accepts a list of files and tries to determine if the files
 * are text or graph files. It is used to parse the command line and drag&drop
 * events
 */
void MainWindow::loadFileList(const QStringList &l)
{
    QStringList graphExtensions(importHandler->extensions());

    QStringList gfiles;
    QStringList tfiles;

    for (int i = 0; i < l.size(); ++i) {
        QFileInfo f(l.at(i));
        if (f.exists()) {
            if (graphExtensions.contains(f.suffix().toLower())) {
                gfiles.append(l.at(i));
            } else {
                tfiles.append(l.at(i));
            }
        }
    }

    if (l.size()) {
        // filter is an empty string, because files of different formats
        // may be loaded
        loadGraphFiles(gfiles, QString());
        loadTextFiles(tfiles);
    }
}

void MainWindow::dragEnterEvent(QDragEnterEvent *e)
{
    if (e->mimeData()->hasFormat("text/uri-list")) {
        e->acceptProposedAction();
    }
}

void MainWindow::dropEvent(QDropEvent *e)
{
    QList<QUrl> u = e->mimeData()->urls();
    QStringList l;

    for (int i = 0; i < u.size(); ++i) {
        l.append(u.at(i).toLocalFile());
    }

    if (!l.size()) return;

    if (e->keyboardModifiers().testFlag(Qt::ControlModifier)) {
        insertGraphFiles(l, QString());
        return;
    }

    if (e->keyboardModifiers().testFlag(Qt::NoModifier)) {
        loadFileList(l);
        return;
    }
}

/*
 * creates a new project
 */
void MainWindow::fileNewProject()
{
    /* TODO */
}

/*
 * Open a text file.
 */
void MainWindow::fileOpenText()
{
    QStringList filters;
    filters << tr("BGMN Control File (*.SAV *.sav)");
    filters << tr("Fullprof Control File (*.PCR *.pcr)");
    filters << tr("Structure File (*.STR *.str)");
    filters << tr("Results Summary (*.SUM *.sum)");
    filters << tr("Results List (*.LST *.lst)");
    filters << tr("Output File (*.OUT *.out)");
    filters << tr("All Files (*.*)");

    QFileDialog fdlg(this, tr("Open File"), workingDir);
    fdlg.setAcceptMode(QFileDialog::AcceptOpen);
    fdlg.setNameFilters(filters);
    fdlg.setFileMode(QFileDialog::ExistingFiles);
    fdlg.setOption(QFileDialog::HideNameFilterDetails, false);
    fdlg.selectNameFilter(lastOpenTextFilter);

    if (!lastOpenTextFilter.isEmpty()) {
        fdlg.selectNameFilter(lastOpenTextFilter);
    }

    if (fdlg.exec()) {
        QStringList fn = fdlg.selectedFiles();
        lastOpenTextFilter = fdlg.selectedNameFilter();
        loadTextFiles(fn);

        recentFilesText += fn;
        setupRecentFiles();
    }
}

/*
 * open a raw scan file
 */
void MainWindow::fileOpenGraph()
{
    QStringList filters;
    filters = importHandler->rawFilters();

    QFileDialog fdlg(this, tr("Open Raw Data File"), workingDir);
    fdlg.setAcceptMode(QFileDialog::AcceptOpen);
    fdlg.setNameFilters(filters);
    fdlg.setFileMode(QFileDialog::ExistingFiles);
    fdlg.setOption(QFileDialog::HideNameFilterDetails, false);

    if (!lastOpenGraphFilter.isEmpty()) {
        fdlg.selectNameFilter(lastOpenGraphFilter);
    }

    if (fdlg.exec()) {
        QStringList fn = fdlg.selectedFiles();
        lastOpenGraphFilter = fdlg.selectedNameFilter();
        qDebug() << QString("MainWindow::fileOpenGraph(): Last Open Graph Filter was: %1").arg(lastOpenGraphFilter);
        qDebug() << QString("MainWindow::fileOpenGraph(): Loading file: %1").arg(fn.join(", "));
        loadGraphFiles(fn, lastOpenGraphFilter);

        recentFilesGraph += fn;
        setupRecentFiles();
    }
}

/*
 * open a refinement project
 */
void MainWindow::fileOpenProject()
{
    QStringList filters;
    filters = importHandler->projectFilters();

    QFileDialog fdlg(this, tr("Open Refinement Project"), workingDir);
    fdlg.setAcceptMode(QFileDialog::AcceptOpen);
    fdlg.setNameFilters(filters);
    fdlg.setFileMode(QFileDialog::ExistingFiles);
    fdlg.setOption(QFileDialog::HideNameFilterDetails, false);

    if (!lastOpenProjectFilter.isEmpty()) {
        fdlg.selectNameFilter(lastOpenProjectFilter);
    }

    if (fdlg.exec()) {
        QStringList fn = fdlg.selectedFiles();
        lastOpenProjectFilter = fdlg.selectedNameFilter();
        qDebug() << QString("MainWindow::fileOpenProject(): Last Open Project Filter was: %1").arg(lastOpenProjectFilter);
        loadGraphFiles(fn, lastOpenProjectFilter);

        recentFilesGraph += fn;
        setupRecentFiles();
    }
}

/*
 * loads text files from a QStringList into new project widgets
 * Profex will attempt to load all the other files, too. So it doesn't really matter
 * which one was selected. If the other files don't exist, they are just skipped. But this
 * code is handled in ProjectWidget, here we just pass the filenames on and register
 * the project in the treeWidget (twProjects).
 */
void MainWindow::loadTextFiles(const QStringList &fn)
{
    int defaultProjectType = settings->value("config/defaultProjectType", 1).toInt();

    QProgressDialog *progDlg = new QProgressDialog(tr("Loading Files"), tr("Cancel"), 0, fn.size(), this);

    QTreeWidgetItem *newCurrent = nullptr;

    // loop through the stringlist of file names
    for (int i = 0; i < fn.size(); ++i) {

        // advance the progress dialog and check if the abort button was pressed
        progDlg->setValue(i);

        if (progDlg->wasCanceled()) {
            progDlg->close();
            break;
        }

        qApp->processEvents();

        // determine file name, base name, directory, extension etc.
        QFileInfo fi(fn.at(i));
        QString fext = fi.suffix().toLower();
        QString bn = fi.completeBaseName();
        QString dr = fi.absolutePath();

        // special treatment for *.str files, because they have different basenames, but
        // belong to the current project
        if ((fext == "str") && (widgetStack->count() > 1)) {
            ProjectWidget *pw = getCurrentProject();
            if (pw) {
                // only load to existing project if the project type is bgmn
                if (pw->type().toLower() == "bgmn") {
                    pw->appendTextFile(QStringList(fn.at(i)));
                    // file loaded, go to next one
                    continue;
                }
            }
        }

        bool projectExists = false;

        // check if a project with the same basename and directory already exists. if yes,
        // don't create a new one, update the existing one instead.
        for (int j = 0; j < twProjects->topLevelItemCount(); ++j) {
            ProjectWidget *pw = getProjectFromUid(twProjects->topLevelItem(j)->data(0, Qt::UserRole).toUuid());
            if (!pw) continue;

            if (pw->isIdenticalTo(bn, dr)) {
                pw->appendTextFile(QStringList(fn.at(i)));
                projectExists = true;
                break;
            }
        }

        // text was loaded into an existing project
        if (projectExists) {
            // all done, go to the next file
            continue;
        }

        // identify the type of project to be created
        int type = fExtProjectType.value(fext, defaultProjectType);

        // create the project widget
        ProjectWidget *pw = createProject(type, fn.at(i));
        if (pw) {
            pw->load(fn.at(i), QString(), QString());
            if (!newCurrent) newCurrent = pw->projectSelectItem();
        }
    }

    twProjects->sortItems(0, Qt::AscendingOrder);

    // activate the first new project
    if (newCurrent) twProjects->setCurrentItem(newCurrent);

    progDlg->close();
    delete progDlg;
}

void MainWindow::loadGraphFiles(const QStringList &fn, const QString &filter)
{
    QElapsedTimer timer;
    timer.start();

    int defaultProjectType = settings->value("config/defaultProjectType", 1).toInt();

    QProgressDialog *progDlg = new QProgressDialog(tr("Loading Files..."), tr("Cancel"), 0, fn.size(), this);
    progDlg->setWindowTitle(tr("Open Projects"));
    progDlg->setAutoClose(false);

    // the default behaviour of the progress dialog doesn't seem to work reliably. Sometimes it is not shown
    // even though a large number of files is opened. So let's force it to show as soon as there are 3 or more
    // files to open.
    if (fn.size() > 2) progDlg->show();

    QTreeWidgetItem *newCurrent = nullptr;

    // get the uid of the file format. If filter is QString(), it will be determined
    // for each file separately below
    QString uid = importHandler->uidByFilter(filter);

    // loop through the list of filenames and create new projects
    for (int i = 0; i < fn.size(); ++i) {
        if (!QFile::exists(fn.at(i))) {
            qDebug() << QString("MainWindow::loadGraphFiles: File doesn't exist: %1").arg(fn.at(i));
            continue;
        }

        // get this file's uid if no filter was provided
        if (filter.isNull()) {
            uid = importHandler->uidByFileName(fn.at(i));
        }

        qDebug() << QString("MainWindow::loadGraphFiles: Loading file %1 with uid = %2").arg(fn.at(i)).arg(uid);

        if (!importHandler->isSupported(fn.at(i), uid)) {
            progDlg->close();
            QMessageBox::information(this, tr("Unknown file format"), tr("No import filter found for file:\n%1").arg(fn.at(i)));
            qDebug() << QString("MainWindow::loadGraphFiles: Unknown file format, aborting: %1").arg(fn.at(i));
            break;
        }

        progDlg->setValue(i);

        if (progDlg->wasCanceled()) {
            break;
        }

        qApp->processEvents();

        // extract the file extension and project type to be created (BGMN or FP)
        QFileInfo fi(fn.at(i));
        QString fext = fi.suffix().toLower();

        int type = fExtProjectType.value(fext, defaultProjectType);

        ProjectWidget *pw = createProject(type, fn.at(i));

        if (pw) {
            pw->load(QString(), fn.at(i), uid);
            statusWavelength(pw->wavelength());

            if (!newCurrent) {
                newCurrent = pw->projectSelectItem();
            }
        } else {
            qDebug() << QString("MainWindow::loadGraphFiles: Could not create project of type %1 for file extension %2").arg(type).arg(fext);
        }
    }

    twProjects->sortItems(0, Qt::AscendingOrder);

    if (newCurrent) {
        twProjects->setCurrentItem(newCurrent);
        setCurrentProject(newCurrent, nullptr);
    }

    progDlg->close();
    delete progDlg;

    qDebug() << QString("MainWindow::loadGraphFiles(): Needed %1 ms to open %2 files").arg(timer.elapsed()).arg(fn.size());
}

void MainWindow::fileInsertGraphFile()
{
    // set up the file dialog
    QStringList filters;
    filters = importHandler->filters();
    filters << tr("All Files (*.*)");

    QFileDialog fdlg(this, tr("Import Graph File"), workingDir);
    fdlg.setAcceptMode(QFileDialog::AcceptOpen);
    fdlg.setNameFilters(filters);
    fdlg.setFileMode(QFileDialog::ExistingFiles);
    fdlg.setOption(QFileDialog::HideNameFilterDetails, false);

    if (!lastOpenGraphFilter.isEmpty()) {
        fdlg.selectNameFilter(lastOpenGraphFilter);
    }

    // sent graph files to project widget
    if (fdlg.exec()) {
        QStringList fn = fdlg.selectedFiles();

        lastOpenGraphFilter = fdlg.selectedNameFilter();
        insertGraphFiles(fn, lastOpenGraphFilter);
    }
}

void MainWindow::insertGraphFiles(const QStringList &files, const QString &filter)
{
    QStringList fn = files;
    ProjectWidget *pw = getCurrentProject();
    qDebug() << QString("MainWindow::insertGraphFiles(): Inserting files %1").arg(files.join("; "));

    if (!pw) {
        // apparently no existing projectWidget available yet. So lets create a new one
        // from the first file name, and then proceed as planned
        QStringList firstfn(fn.takeFirst());
        loadGraphFiles(firstfn, filter);

        // try again to get a pointer to the new projectWidget
        pw = getCurrentProject();

        // still not ok? -> Exit
        if (!pw) {
            qDebug() << QString("MainWindow::insertGraphFiles(): Could not create project. Exiting.");
            return;
        } else {
            qDebug() << QString("MainWindow::insertGraphFiles(): New project was created.");
        }
    }

    QStringList err;

    for (int i = 0; i < fn.size(); ++i) {
        QString uid;
        if (filter.isEmpty()) uid = importHandler->uidByFileName(fn.at(i));
        else                  uid = importHandler->uidByFilter(filter);

        err.append(pw->insertGraph(QStringList(fn.at(i)), uid, true));
        recentFilesGraph.append(fn.at(i));
    }

    setupRecentFiles();

    if (err.size()) {
        qDebug() << QString("MainWindow::insertGraphFiles(): The following files could not be opened: %1").arg(err.join("; "));
        QMessageBox::information(this, tr("Unknown file format"),
                                 tr("The following files could not be opened:\n%1").arg(err.join("\n")));
    }
}

/*
 * creates a new projectWidget of type "type"
 */
ProjectWidget * MainWindow::createProject(int type, const QString &file)
{
    if ((type < 0) || (type > 1)) {
        qDebug() << QString("MainWindow::createProject(): Type %1 is invalid").arg(type);
        return nullptr;
    }

    QFileInfo fi(file);
    workingDir = fi.absolutePath();

    // create the project
    ProjectWidget *pw = nullptr;

    switch (type) {
        case 0:  pw = new FpProjectWidget(this); break;
        case 1:  pw = new BgmnProjectWidget(this); break;
        default: pw = new BgmnProjectWidget(this); break;
    }

    twProjects->addTopLevelItem(pw->projectSelectItem());

    widgetStack->addWidget(pw);
    projectToolbarStack->addWidget(pw->toolbarWidget);

    dwScans->addWidget(pw->scanList);
    dwOutput->addWidget(pw->refOutput);
    dwChemistry->addWidget(pw->chemOutput);
    dwConvergence->addWidget(pw->convDisplay);
    dwPeakIntegration->addWidget(pw->peakIntegrationWidget);
    dwPeakList->addWidget(pw->peakListWidget);
    dwSearchMatch->addWidget(pw->searchMatchWidget);
    dwResultsTree->addWidget(pw->getResultsTree());
    dwPeakFit->addWidget(pw->peakFitWidget);

    initProjectConnections(pw);
    updateNumberOfOpenProject();
    toggleProjectMenusEnabled(true);

    return pw;
}

/*
   initializes connections between a projectWidget and this. Always call this function
   after creating a new projectWidget.
*/
void MainWindow::initProjectConnections(ProjectWidget *pw)
{
    connect(pw, SIGNAL(openNewProject(QStringList)), this, SLOT(loadTextFiles(QStringList)));
    connect(pw, SIGNAL(updateProjectName(ProjectWidget*, QString)), this, SLOT(updateProjectName(ProjectWidget*, QString)));
    connect(pw, SIGNAL(updateStatus(ProjectWidget*, QString)), this, SLOT(updateStatus(ProjectWidget*, QString)));
    connect(pw, SIGNAL(setCoordinates(double,double,double)), this, SLOT(coordinates(double,double,double)));
    connect(pw, SIGNAL(completed(ProjectWidget*)), this, SLOT(refinementCompleted(ProjectWidget*)));
    connect(pw, SIGNAL(cursorPosition(int,int)), this, SLOT(editorCursorChanged(int,int)));
    connect(pw, SIGNAL(updateWavelength(double)), this, SLOT(statusWavelength(double)));
    connect(pw, SIGNAL(currentTabChanged(ProjectWidget*,int)), this, SLOT(projectTabChanged(ProjectWidget*,int)));
    connect(pw, SIGNAL(referenceStructureFavoritesToggled(bool)), this, SLOT(refStrFavoritesToggled(bool)));
    connect(pw, SIGNAL(sigAllPeakIntegrals(QList<global::PeakRange>)), this, SLOT(applyPeakIntegralsToAll(QList<global::PeakRange>)));
    connect(pw, SIGNAL(openGraphFileInNewProject(QStringList, QString)), this, SLOT(loadGraphFiles(QStringList,QString)));
    connect(pw, SIGNAL(tabBarDoubleClicked(int)), this, SLOT(raiseTab(int)));
    connect(pw, SIGNAL(sigAllCurveFits()), this, SLOT(applyCurveFits()));
    connect(pw, SIGNAL(sigUpdatePresetMenu()), presetMenuMgr, SLOT(update()));
}

/*
   Close all selected projects.
*/
void MainWindow::fileCloseCurrentProject()
{
    if (!twProjects->topLevelItemCount()) {
        return;
    }

    QList<QTreeWidgetItem *> selection = twProjects->selectedItems();
    if (selection.isEmpty()) selection.append(twProjects->currentItem());

    QVector<int> indices;
    for (int i = 0; i < selection.size(); ++i) {
        indices.append(twProjects->indexOfTopLevelItem(selection.at(i)));
    }

    std::sort(indices.begin(), indices.end());

    while (indices.size()) {
        closeProject(indices.takeLast(), true);
    }
}

/*
 * closes all open projects. Checks whether some are running
 */
void MainWindow::fileCloseAllProjects()
{
    if (hasRunningProjects()) {
        QMessageBox::information(this, tr("Close all projects"), tr("Please stop all running refinements first."));
        return;
    }

    qApp->setOverrideCursor(Qt::WaitCursor);

    twProjects->setCurrentItem(twProjects->topLevelItem(0));

    QProgressDialog *progDlg = new QProgressDialog(tr("Closing projects"), tr("Cancel"), 0, twProjects->topLevelItemCount(), this);
    progDlg->setWindowTitle(tr("Closing all projects"));

    while (twProjects->topLevelItemCount()) {
        closeProject(twProjects->topLevelItemCount() - 1, false);

        progDlg->setValue(progDlg->value() + 1);
        qApp->processEvents();

        if (progDlg->wasCanceled()) break;
    }

    delete progDlg;
    qApp->restoreOverrideCursor();
}

void MainWindow::closeProject(int n, bool update)
{
    QTreeWidgetItem *it = twProjects->topLevelItem(n);
    if (!it) return; // probably because n is out of index

    ProjectWidget *pw = getProjectFromUid(it->data(0, Qt::UserRole).toUuid());
    if (!pw) return; // something fishy!?! Exit safely.

    if (pw->isRunning()) {
        if (QMessageBox::question(this, "Closing Project",
                                  "Refinement is running.\nDo you want to close it anyway?",
                                  QMessageBox::Yes, QMessageBox::No) == QMessageBox::No) {
            return;
        }

        pw->abort();
    }

    statusFileName->setText(QString());

    // if the currently selected item is going to be removed,
    // we must set the current item to the new position first
    if (update) {
        if (n == 0) {
            // the current Item was at pos 0, so lets select the one at pos 1,
            // but first make sure there is another project at pos 1.
            if (twProjects->topLevelItemCount() > 1) {
                twProjects->setCurrentItem(twProjects->topLevelItem(1));
                setCurrentProject(twProjects->currentItem(), Q_NULLPTR);
            }
        } else {
            // the current item was at a pos > 0, so lets select the previous one
            twProjects->setCurrentItem(twProjects->topLevelItem(n - 1));
            setCurrentProject(twProjects->currentItem(), Q_NULLPTR);
        }
    }

    // remove all stack widgets of the closed project
    // TODO: handle unsafed data
    dwScans->removeWidget(pw->scanList);
    dwOutput->removeWidget(pw->refOutput);
    dwChemistry->removeWidget(pw->chemOutput);
    dwConvergence->removeWidget(pw->convDisplay);
    dwPeakIntegration->removeWidget(pw->peakIntegrationWidget);
    dwPeakList->removeWidget(pw->peakListWidget);
    dwSearchMatch->removeWidget(pw->searchMatchWidget);
    dwResultsTree->removeWidget(pw->getResultsTree());
    dwPeakFit->removeWidget(pw->peakFitWidget);

    projectToolbarStack->removeWidget(pw->toolbarWidget);
    widgetStack->removeWidget(pw);
    delete pw;

    // now we can safely delete the previously selected toplevelitem
    delete twProjects->takeTopLevelItem(n);

    if (!twProjects->topLevelItemCount()) {
        toggleProjectMenusEnabled(false);
        clearToolsDialogs();

        statusLambda->setText(QString());
        statusAng->setText(QString());
        statusInt->setText(QString());
        statusD->setText(QString());
        statusCursor->setText(QString());
        statusFileName->setText(QString());
    }

    updateNumberOfOpenProject();
}

/*
   another project was selected in the tree widget. Raise the corresponding page
   in the widget stack and enable/disable menu actions
*/
void MainWindow::setCurrentProject(QTreeWidgetItem *cur, QTreeWidgetItem *prev)
{
    if (!cur)        return;
    if (cur == prev) return;

    // safely get a pointer to the current toplevelitem
    QTreeWidgetItem *it = cur;
    while (it->parent()) it = it->parent();
    if (!it) return;

    QUuid uid = it->data(0, Qt::UserRole).toUuid();
    setCurrentProject(uid);
}

void MainWindow::setCurrentProject(const QUuid &uid)
{
    ProjectWidget *pw = getProjectFromUid(uid);
    if (!pw) return;

    bool ok = true;

    // check if all widgets we want to raise are present
    if (widgetStack->indexOf(pw) < 0)                               ok = false;
    if (projectToolbarStack->indexOf(pw->toolbarWidget) < 0)        ok = false;
    if (dwScans->indexOf(pw->scanList) < 0)                         ok = false;
    if (dwOutput->indexOf(pw->refOutput) < 0)                       ok = false;
    if (dwChemistry->indexOf(pw->chemOutput) < 0)                   ok = false;
    if (dwConvergence->indexOf(pw->convDisplay) < 0)                ok = false;
    if (dwPeakIntegration->indexOf(pw->peakIntegrationWidget) < 0)  ok = false;
    if (dwPeakList->indexOf(pw->peakListWidget) < 0)                ok = false;
    if (dwSearchMatch->indexOf(pw->searchMatchWidget) < 0)          ok = false;
    if (dwResultsTree->indexOf(pw->getResultsTree()) < 0)           ok = false;
    if (dwPeakFit->indexOf(pw->peakFitWidget) < 0)                  ok = false;

    if (!ok) return;

    widgetStack->setCurrentWidget(pw);
    projectToolbarStack->setCurrentWidget(pw->toolbarWidget);

    dwScans->setCurrentWidget(pw->scanList);
    dwOutput->setCurrentWidget(pw->refOutput);
    dwChemistry->setCurrentWidget(pw->chemOutput);
    dwConvergence->setCurrentWidget(pw->convDisplay);
    dwPeakIntegration->setCurrentWidget(pw->peakIntegrationWidget);
    dwPeakList->setCurrentWidget(pw->peakListWidget);
    dwSearchMatch->setCurrentWidget(pw->searchMatchWidget);
    dwResultsTree->setCurrentWidget(pw->getResultsTree());
    dwPeakFit->setCurrentWidget(pw->peakFitWidget);

    statusWavelength(pw->wavelength());
    statusFileName->setText(QDir::toNativeSeparators(pw->currentFileName()));
    updateToolsDialogs();
}

/*
 * execute the refinement for the current projectWidget
 */
void MainWindow::runRefinement()
{
    ProjectWidget *pw = getCurrentProject();
    if (!pw) return;

    if (!pw->isRunning()) {
        sysTrayIcon->setIcon(QIcon(":/icons/profex4-active.png"));
        sysTrayIcon->setToolTip(tr("Running..."));
        pw->runRefinement();
        dwOutput->raise();
    }
}

/*
 *  start batch refinement
 */
void MainWindow::runBatchRefinement()
{
    if (hasScheduledProjects()) {
        appendToRunningBatch();
    } else {
        startNewBatch();
    }
}

/*
 * shows a project select dialog and starts a new batch refinement
 */
void MainWindow::startNewBatch()
{
    int nparallel = settings->value("config/batchParallelJobs", 1).toInt();

    QList<ProjectWidget*> plst = getProjectSelection(tr("Batch refinement"), false, true, 0);

    if (!plst.size()) return;

    for (int i = 0; i < plst.size(); ++i) {
        plst.at(i)->setStatus(SCHEDULED);
        plst.at(i)->setScheduledForBatch(true);
    }

    sysTrayIcon->setIcon(QIcon(":/icons/profex-active.png"));
    sysTrayIcon->setToolTip(QString(tr("Processing %1 projects").arg(plst.size())));

    sysTrayIcon->showMessage(tr("Profex"),
                             QString(tr("Running batch refinement: Processing %1 projects").arg(plst.size())),
                             QSystemTrayIcon::Information,
                             5000);

    if (ui->actionFollow_Active_Refinement->isChecked()) {
        twProjects->setCurrentItem(plst.at(0)->projectSelectItem());
        setCurrentProject(twProjects->currentItem(), nullptr);
    }

    for (int i = 0; i < nparallel; ++i) {
        if (i >= plst.size()) break;
        plst.at(i)->runRefinement();
    }

    dwOutput->raise();
}

/*
 * appends all selected projects to a running batch. Does not
 * start a new batch refinement, just changes the status to SCHEDULED
 */
void MainWindow::appendToRunningBatch()
{
    QList<ProjectWidget *> projects = getAllProjects();

    for (int i = 0; i < projects.count(); ++i) {
        // skip projects that are already scheduled for a batch or running
        if ((projects.at(i)->status() == RUNNING) || (projects.at(i)->status() == SCHEDULED)) continue;

        if (projects.at(i)->projectSelectItem()->isSelected()) {
            projects.at(i)->setStatus(SCHEDULED);
            projects.at(i)->setScheduledForBatch(true);
        }
    }
}

void MainWindow::runSearchMatch()
{
    ProjectWidget *pw = getCurrentProject();
    if (pw) pw->runSearchMatch();
}

void MainWindow::runPeakDetection()
{
    ProjectWidget *pw = getCurrentProject();
    if (!pw) return;

    if (!pw->isRunning()) {
        sysTrayIcon->setIcon(QIcon(":/icons/profex-active.png"));
        sysTrayIcon->setToolTip(tr("Running..."));
        pw->runPeakDetection();
    }
}

void MainWindow::save()
{
    ProjectWidget *pw = getCurrentProject();
    if (pw) pw->saveCurrent();
}

void MainWindow::saveAs()
{
    ProjectWidget *pw = getCurrentProject();
    if (pw) pw->saveCurrentAs();
}

void MainWindow::saveAll()
{
    qApp->setOverrideCursor(Qt::WaitCursor);

    QList<ProjectWidget *> lst = getAllProjects();
    for (int i = 0; i < lst.size(); ++i) {
        lst.at(i)->saveAll();
    }

    qApp->restoreOverrideCursor();
}

void MainWindow::revertControlFile()
{
    ProjectWidget *pw = getCurrentProject();
    if (pw) pw->revertControl();
}

/*
 * if a batch is running, it aborts / unschedules all projects that are part of the batch.
 * if no batch is running, it just aborts all running projects
 */
void MainWindow::abortAllRefinements()
{
    if (hasScheduledProjects() > 0) {
        if (QMessageBox::question(this, tr("Abort batch processing"), tr("Do you want to abort <b>all</b> running refinements?")) != QMessageBox::Yes) {
            return;
        }
    }

    sysTrayIcon->setIcon(QIcon(":/icons/profex4.png"));
    sysTrayIcon->showMessage(tr("Profex"), tr("Refinement aborted"), QSystemTrayIcon::Information, 5000);
    sysTrayIcon->setToolTip(tr("Idle"));

    QList<ProjectWidget *> projects = getAllProjects();

    for (int i = 0; i < projects.count(); ++i) {
        projects.at(i)->setScheduledForBatch(false);

        if (projects.at(i)->isRunning()) {
            projects.at(i)->abort();
        } else {
            projects.at(i)->setStatus(IDLE);
        }
    }
}

/*
 * The effect of this function will depend on the status of the current project:
 * - RUNNING:  project will be aborted.
 *
 *             * if pw->isScheduledForBatch() == true, all other scheduled projects
 *               will be unscheduled first to stop the batch
 *
 *             * if pw->isScheduledForBatch() == false, nothing other than aborting pw
 *               will be done
 *
 * - SCHEDULED: the project will be unscheduled but the batch will continue
 */
void MainWindow::abortCurrentRefinement()
{
    ProjectWidget *pw = getCurrentProject();
    if (!pw) return;


    if (pw->isRunning()) {
        if (pw->isScheduledForBatch()) {
            // it is part of a batch, unschedule all other projects first, to avoid automatically
            // launching the next one when pw is aborted
            QList<ProjectWidget *> pwlist = getScheduledProjects();

            for (int i = 0; i < pwlist.size(); ++i) {
                if (!pwlist.at(i)->isRunning()) {
                    pwlist.at(i)->setStatus(IDLE);
                    pwlist.at(i)->setScheduledForBatch(false);
                }
            }
        }

        // now stop the running project
        pw->abort();
        pw->setScheduledForBatch(false);
    } else {
        pw->setScheduledForBatch(false);
        pw->setStatus(IDLE);
    }
}

void MainWindow::updateProjectName(ProjectWidget *pw, QString s)
{
    pw->projectSelectItem()->setText(0, s);
}

void MainWindow::updateStatus(ProjectWidget *pw, QString s)
{
    pw->projectSelectItem()->setText(1, s);
}

/*
 * this slot is called by all projects when they are done with the refinement. If we are in batch mode,
 * we call the next project, otherwise do nothing.
 */
void MainWindow::refinementCompleted(ProjectWidget *pw)
{
    QList<ProjectWidget *> pwlst = getScheduledProjects();

    if (!pwlst.size()) {
        qDebug() << QStringLiteral("MainWindow::refinementCompleted(): No scheduled projects left.");
        pw->setScheduledForBatch(false);

        sysTrayIcon->setIcon(QIcon(":/icons/profex4.png"));
        sysTrayIcon->setToolTip(tr("Refinement Complete"));
        sysTrayIcon->showMessage(tr("Profex"), tr("Refinement Complete"), QSystemTrayIcon::Information, 5000);

        return;
    } else {
        // completed project was not part of a batch
        if (!pw->isScheduledForBatch()) {
            qDebug() << QStringLiteral("MainWindow::refinementCompleted(): Project was not part of a batch. Doing nothing.");
            return;
        }
    }

    // the status variable was already changed to COMPLETED by the project itself
    pw->setScheduledForBatch(false);

    while (pwlst.size()) {
        ProjectWidget *pwnext = pwlst.takeFirst();

        if (!pwnext->isRunning()) {
            qDebug() << QString("MainWindow::refinementCompleted(): Starting next scheduled project %1.").arg(pw->getGraphFile());
            sysTrayIcon->showMessage(tr("Profex"), QString(tr("%1 projects left to refine")).arg(pwlst.size()+1), QSystemTrayIcon::Information, 5000);
            sysTrayIcon->setToolTip(QString(tr("%1 projects left to refine")).arg(pwlst.size()));

            if (ui->actionFollow_Active_Refinement->isChecked()) {
                twProjects->setCurrentItem(pwnext->projectSelectItem());
                setCurrentProject(twProjects->currentItem(), Q_NULLPTR);
            }

            pwnext->runRefinement();
            return;
        }
    }

    // no scheduled projects left
    sysTrayIcon->setIcon(QIcon(":/icons/profex4.png"));
    sysTrayIcon->setToolTip(tr("Refinement Complete"));
    sysTrayIcon->showMessage(tr("Profex"), tr("Refinement Complete"), QSystemTrayIcon::Information, 5000);
}

void MainWindow::coordinates(double x, double y, double d)
{
    statusAng->setText(QString(tr("2%1 = %2%3")).arg(global::theta).arg(x, 8, 'f', 3).arg(global::degree));
    statusInt->setText(QString(tr("I = %1 cts")).arg(y, 10, 'f', 3));
    statusD->setText(QString(tr("d = %1 %2")).arg(d, 6, 'f', 3).arg(global::angstrom));
}

void MainWindow::statusWavelength(double d)
{
    statusLambda->setText(QString(tr("%1 = %2 %3")).arg(global::lambda).arg(d, 0, 'f', 5).arg(global::angstrom));
}

void MainWindow::editorCursorChanged(int l, int c)
{
    statusCursor->setText(QString(tr("Line: %1, Column: %2")).arg(l).arg(c));
}

void MainWindow::addRemovePhase()
{
    ProjectWidget *pw = getCurrentProject();
    if (pw) pw->addRemovePhase();
}

void MainWindow::exportResults()
{
    QList<ProjectWidget *> pl = getProjectSelection(tr("Export local GOALs"), false, false);
    if (!pl.size()) return;

    QString eFile = QFileDialog::getSaveFileName(this, tr("Export results"), workingDir, tr("CSV File (*.CSV *.csv)"));
    if (eFile.isEmpty()) return;

    qApp->setOverrideCursor(Qt::WaitCursor);

    QList<DataTable> data;

    for (int i = 0; i < pl.size(); ++i) {
        data.append(DataTable(pl.at(i)->getResultsCsv()));
    }

    // open the file for writing
    QFile f(eFile);
    if (!f.open(QIODevice::WriteOnly | QIODevice::Text)) {
        qDebug() << QString("MainWindow::exportResults(): Could not write to file %1").arg(eFile);
        qApp->restoreOverrideCursor();
        return;
    }

    // write content to file
    QTextStream out(&f);
    out << buildGoalsCsvTable(data);
    f.close();

    qApp->restoreOverrideCursor();
}

QString MainWindow::buildGoalsCsvTable(const QList<DataTable> &data)
{
    if (data.isEmpty()) return QString();

    QStringList globalHeader(data.first().header());

    // if other tables contain more columns, append the header cells to globalHeader
    for (int i = 1; i < data.size(); ++i) {
        QStringList dHeader = data.at(i).header();
        for (int j = 0; j < dHeader.size(); ++j) {
            if (!globalHeader.contains(dHeader.at(j))) globalHeader.append(dHeader.at(j));
        }
    }

    QString output(globalHeader.join(";"));
    output += "\n";

    // loop over all tables
    for (int t = 0; t < data.size(); ++t) {
        // loop over all lines in table t
        for (int r = 0; r < data.at(t).rows(); ++r) {
            // loop over all header cells c in globalHeaders, and query the
            // cell value for c in table t, row r
            QStringList globalLine;
            for (int c = 0; c < globalHeader.size(); ++c) {
                globalLine.append(data.at(t).value(globalHeader.at(c), r));
            }

            output += globalLine.join(";") + QStringLiteral("\n");
        }
    }

    return output;
}

void MainWindow::exportQuantities()
{
    QList<ProjectWidget *> pl = getProjectSelection(tr("Export global GOALs"), false, false);
    if (!pl.size()) return;

    QString eFile = QFileDialog::getSaveFileName(this, tr("Global GOALs output file"), workingDir, tr("CSV File (*.CSV *.csv)"));
    if (eFile.isEmpty()) return;

    qApp->setOverrideCursor(Qt::WaitCursor);

    QList<DataTable> data;

    for (int i = 0; i < pl.size(); ++i) {
        data.append(DataTable(pl.at(i)->getQuantitiesCsv()));
    }

    // open the file for writing
    QFile f(eFile);
    if (!f.open(QIODevice::WriteOnly | QIODevice::Text)) {
        qDebug() << QString("MainWindow::exportQuantities(): Could not write to file %1").arg(eFile);
        qApp->restoreOverrideCursor();
        return;
    }

    // write content to file
    QTextStream out(&f);
    out << buildGoalsCsvTable(data);
    f.close();

    qApp->restoreOverrideCursor();
}

void MainWindow::exportChemistry()
{
    QList<ProjectWidget *> pl = getProjectSelection(tr("Export chemical composition"), false, false);
    if (!pl.size()) return;

    QString eFile = QFileDialog::getSaveFileName(this, tr("Export to file"), workingDir, tr("CSV File (*.CSV *.csv)"));
    if (eFile.isEmpty()) return;

    qApp->setOverrideCursor(Qt::WaitCursor);

    QStringList cont;
    for (int i = 0; i < pl.size(); ++i) {
        cont.append(pl.at(i)->getChemistry(i == 0 ? true : false));
    }

    if (!BgmnFileIO::writeTextFile(eFile, cont.join("\n"))) {
        qDebug() << QString("MainWindow::exportChemistry(): Could not write to file %1").arg(eFile);
    }

    qApp->restoreOverrideCursor();
}

void MainWindow::exportPeakIntegrals()
{
    QList<ProjectWidget *> pl = getProjectSelection(tr("Export Peak Integrals"), false, true);
    if (!pl.size()) return;

    QString eFile = QFileDialog::getSaveFileName(this, tr("Peak integral table"), workingDir, tr("CSV File (*.CSV *.csv)"));
    if (eFile.isEmpty()) return;

    qApp->setOverrideCursor(Qt::WaitCursor);

    QStringList cont;
    for (int i = 0; i < pl.size(); ++i) {
        cont.append(pl.at(i)->getPeakIntegralData());
    }

    if (!BgmnFileIO::writeTextFile(eFile, cont.join("\n"))) {
        qDebug() << QString("MainWindow::exportPeakIntegrals(): Could not write to file %1").arg(eFile);
    }

    qApp->restoreOverrideCursor();
}

void MainWindow::exportCurveFits()
{
    QList<ProjectWidget *> pl = getProjectSelection(tr("Export Curve Fit Results"), false, true);
    if (!pl.size()) return;

    QString eFile = QFileDialog::getSaveFileName(this, tr("Curve fit table"), workingDir, tr("CSV File (*.CSV *.csv)"));
    if (eFile.isEmpty()) return;

    qApp->setOverrideCursor(Qt::WaitCursor);

    QStringList cont;
    for (int i = 0; i < pl.size(); ++i) {
        cont.append(pl.at(i)->getCurveFitData());
    }

    if (!BgmnFileIO::writeTextFile(eFile, cont.join("\n"))) {
        qDebug() << QString("MainWindow::exportCurveFits(): Could not write to file %1").arg(eFile);
    }

    qApp->restoreOverrideCursor();
}

void MainWindow::exportCifFiles()
{
    QList<ProjectWidget *> pl = getProjectSelection(tr("Export CIF files"), false, false);

    CifExportDialog cdlg(this);
    if (cdlg.exec() == QDialog::Accepted) {
        QMap<QString, QVariant> auxData(cdlg.getAuxData());
        if (cdlg.getOutputMode() == SINGLE) exportSingleCif(pl, auxData);
        if (cdlg.getOutputMode() == PROJECT) exportProjectCif(pl, auxData);
        if (cdlg.getOutputMode() == GLOBAL) exportMergedCif(pl, auxData);
    }
}

void MainWindow::exportSingleCif(const QList<ProjectWidget *> &pl, const QMap<QString, QVariant> &auxData)
{
    qApp->setOverrideCursor(Qt::WaitCursor);

    for (int i = 0; i < pl.size(); ++i) {
        if (!pl.at(i)->isRunning()) {
            pl.at(i)->saveCifFiles(true, auxData);
        }
    }

    qApp->restoreOverrideCursor();
}

void MainWindow::exportProjectCif(const QList<ProjectWidget *> &pl, const QMap<QString, QVariant> &auxData)
{
    qApp->setOverrideCursor(Qt::WaitCursor);


    for (int i = 0; i < pl.size(); ++i) {
        QString fname(QDir::toNativeSeparators(pl.at(i)->workingDir() + "/" + pl.at(i)->baseName() + ".cif"));
        QList<phaseData> cifList;

        if (!pl.at(i)->isRunning()) {
            cifList.append(pl.at(i)->getCifData(auxData));
        }

        QString output;

        output += QStringLiteral("###############################################################################\n");
        output += QString(       "# Crystal structure data for project %1\n").arg(pl.at(i)->baseName());
        output += QString(       "# Created with Profex %1.%2.%3\n").arg(VERSION_MAJOR).arg(VERSION_MINOR).arg(VERSION_BUILD);
        output += QString(       "# Export date: %1\n").arg(QDateTime::currentDateTime().toString("MMMM dd, yyyy - hh:mm"));
        output += QStringLiteral("###############################################################################\n\n");
        output += QStringLiteral("data_global\n\n");
        output += QStringLiteral("_publ_contact_author_name       ? # Name of author for correspondence\n");
        output += QStringLiteral("_publ_contact_author_address      # Address of author for correspondence\n");
        output += QStringLiteral(";?\n;\n\n");

        for (int j = 0; j < cifList.size(); ++j) {
            output.append(cifList.at(j).data.toString() + "\n\n");
        }

        output += QStringLiteral("\n# The following lines are used to test the character set of files sent by\n");
        output += QStringLiteral("# network email or other means. They are not part of the CIF data set.\n");
        output += QStringLiteral("# abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789\n");
        output += QString("# !@#$%^&*()_+{}:\"~<>?|\\-=[];'`,./ \n");

        if (!BgmnFileIO::writeTextFile(fname, output)) {
            qDebug() << QString("MainWindow::exportMergedCif(): Could not write to file %1").arg(fname);
            qApp->restoreOverrideCursor();
            return;
        }

        pl.at(i)->sendMessageToOutputConsole(QString("Exported CIF file:\n%1\n").arg(fname));
    }

    qApp->restoreOverrideCursor();
}

void MainWindow::exportMergedCif(const QList<ProjectWidget *> &pl, const QMap<QString, QVariant> &auxData)
{
    QString fname(QFileDialog::getSaveFileName(this, tr("Export CIF File"), workingDir, tr("CIF File (*.CIF *.cif)")));

    qApp->setOverrideCursor(Qt::WaitCursor);

    QList<phaseData> cifList;

    for (int i = 0; i < pl.size(); ++i) {
        if (!pl.at(i)->isRunning()) {
            cifList.append(pl.at(i)->getCifData(auxData));
        }
    }

    QString output;

    output += QStringLiteral("###############################################################################\n");
    output += QStringLiteral("# Crystal structure data\n");
    output += QString(       "# Created with Profex %1.%2.%3\n").arg(VERSION_MAJOR).arg(VERSION_MINOR).arg(VERSION_BUILD);
    output += QString(       "# Export date: %1\n").arg(QDateTime::currentDateTime().toString("MMMM dd, yyyy - hh:mm"));
    output += QStringLiteral("###############################################################################\n\n");
    output += QStringLiteral("data_global\n\n");
    output += QStringLiteral("_publ_contact_author_name       ? # Name of author for correspondence\n");
    output += QStringLiteral("_publ_contact_author_address      # Address of author for correspondence\n");
    output += QStringLiteral(";?\n;\n\n");

    for (int i = 0; i < cifList.size(); ++i) {
        output.append(cifList.at(i).data.toString() + "\n\n");
    }

    output += QStringLiteral("\n# The following lines are used to test the character set of files sent by\n");
    output += QStringLiteral("# network email or other means. They are not part of the CIF data set.\n");
    output += QStringLiteral("# abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789\n");
    output += QString("# !@#$%^&*()_+{}:\"~<>?|\\-=[];'`,./ \n");

    if (!BgmnFileIO::writeTextFile(fname, output)) {
        qDebug() << QString("MainWindow::exportMergedCif(): Could not write to file %1").arg(fname);
        qApp->restoreOverrideCursor();
        return;
    }

    qApp->restoreOverrideCursor();
}

void MainWindow::exportCastepCellFiles()
{
    QList<ProjectWidget *> pl = getProjectSelection(tr("Export Castep CELL files"), false, false);

    for (int i = 0; i < pl.size(); ++i) {
        if (!pl.at(i)->isRunning()) {
            pl.at(i)->saveCastepCellFiles();
        }
    }
}

/*
 * exports information about all open projects to a CSV file
 */
void MainWindow::exportProjectInfo()
{
    QString file = QFileDialog::getSaveFileName(this, tr("Export project information"), workingDir, tr("CSV File (*.CSV *.csv)"));
    if (file.isEmpty()) return;

    qApp->setOverrideCursor(Qt::WaitCursor);

    QMap<QString, QStringList> data;
    QString sep = ";";
    int n = twProjects->topLevelItemCount();

    for (int i = 0; i < n; ++i) {
        QTreeWidgetItem *tit = twProjects->topLevelItem(i);
        data["Name"].append(tit->text(0));

        for (int j = 0; j < tit->childCount(); ++j) {
            QTreeWidgetItem *cit = tit->child(j);
            data[cit->text(0)].append(cit->text(1));
        }
    }

    QStringList keys = data.keys();
    QStringList out(keys.join(sep));

    for (int i = 0; i < n; ++i) {
        QStringList line;

        for (int j = 0; j < keys.size(); ++j) {
            line.append(i < data[keys.at(j)].size() ? data[keys.at(j)].at(i) : "");
        }

        out.append(line.join(sep));
    }


    if (!BgmnFileIO::writeTextFile(file, out.join("\n"))) {
        qDebug() << QString("MainWindow::exportProjectInfo(): Could not write to file %1").arg(file);
    }

    qApp->restoreOverrideCursor();
}

/*
 * scan batch conversion
 */
void MainWindow::fileScanBatchConversion()
{
    QStringList files;

    // if projects are loaded, show a project selection dialog and fill the
    // batch conversion dialog with all selected project scans
    if (countProjects(false, true) > 0) {
        QList<ProjectWidget *> pl = getProjectSelection(tr("Convert Scan Files"), false, true);
        if (!pl.size()) return;

        for (int i = 0; i < pl.size(); ++i) {
            files.append(pl.at(i)->getGraphFile());
        }
    } // else just show the empty batch dialog

    ExportGraphDialog *edlg = new ExportGraphDialog(this);
    edlg->setFiles(files);
    edlg->exec();
    delete edlg;
}

/*
 * imports a CIF or ICDD XML structure file
 */
void MainWindow::fileImportStructure()
{
    if (!stidlg) {
        stidlg = new StrucImportDialog(this);
        connect(stidlg, SIGNAL(emitRunIndexing()), this, SLOT(indexBgmnHkl()));
    }

    stidlg->show();
}

/*
 * copies the current control file to other projects. A dialog
 * will be shown to select the projects
 */
void MainWindow::applyControl()
{
    ProjectWidget *cpw = getCurrentProject();
    if (!cpw) return;

    // get the control file content of the current project
    QString cont = cpw->getControlFileContent();
    QFileInfoList sourceProjectFiles = cpw->getSharedInputProjectFilesList();

    // if nothing useful was returned, there is nothing to do.
    if (cont.isEmpty()) {
        QMessageBox::information(this, tr("No control file found"), tr("No control file found.\nCreate a control file first."));
        return;
    }

    QList<ProjectWidget *> pwl = getProjectSelection(tr("Apply Control File Content"), true, false);
    QProgressDialog pdlg(tr("Applying project files..."), tr("Abort"), 0, pwl.size(), this);
    pdlg.setWindowModality(Qt::WindowModal);

    // send the control file content to all selected projects
    qApp->setOverrideCursor(Qt::WaitCursor);

    bool forceOverwrite = false;
    for (int i = 0; i < pwl.size(); ++i) {
        pdlg.setValue(i);
        ProjectWidget *pw = pwl.at(i);

        // receive the control file content and adapt all file names
        pw->applyControlFileContent(cont);

        // gather shared files at the new destination
        forceOverwrite = pw->gatherSharedInputFiles(cpw->workingDir(), sourceProjectFiles, forceOverwrite);

        // if some files were present and the user decided not to overwrite them, bail out
        if (!forceOverwrite) break;
        if (pdlg.wasCanceled()) break;
    }

    pdlg.setValue(pwl.size());
    qApp->restoreOverrideCursor();
}

void MainWindow::applyCurveFits()
{
    ProjectWidget *cpw = getCurrentProject();
    if (!cpw) return;

    QList<global::CurveFitCurve> cfc = cpw->curveFitCurves();
    QList<global::CurveFitVariable> cfv = cpw->curveFitVariables();
    QMap<QString, QVariant> cft = cpw->curveFitControls();

    QList<ProjectWidget *> pwl = getProjectSelection(tr("Apply Curve Fits"), true, false);

    // send the control file content to all selected projects
    qApp->setOverrideCursor(Qt::WaitCursor);

    for (int i = 0; i < pwl.size(); ++i) {
        ProjectWidget *pw = pwl.at(i);
        if (pw) {
            pw->applyPeakFits(cft, cfv, cfc);
        }
    }

    qApp->restoreOverrideCursor();
}

void MainWindow::print()
{
    ProjectWidget *pw = getCurrentProject();
    if (pw) pw->print();
}

void MainWindow::printAllGraphs()
{
    QList<ProjectWidget *> pl = getProjectSelection(tr("Print Graphs"), false, true);
    if (!pl.size()) return;

    QPrinter printer(QPrinterInfo::defaultPrinter(), QPrinter::HighResolution);

    if (!printer.isValid()) {
        qDebug() << QString("MainWindow::printAllGraphs(): Invalid printer object, exiting.");
        return;
    }

    printer.setPageOrientation(QPageLayout::Landscape);
    printer.setOutputFormat(QPrinter::NativeFormat);

    // this works on Linux et al, but not on OS X and Windows
#ifdef Q_OS_UNIX
    #ifndef Q_OS_OSX
        QFileInfo fi(pl.first()->getGraphFile());
        printer.setOutputFileName(QDir::fromNativeSeparators(fi.absolutePath() + "/" + fi.completeBaseName() + ".pdf"));
    #endif
#endif

    QPrintDialog printDialog(&printer, this);

    if (printDialog.exec() != QDialog::Accepted) {
        return;
    }

    QPainter painter;
    qApp->setOverrideCursor(Qt::WaitCursor);
    painter.begin(&printer);

    while (pl.size()) {
        pl.takeFirst()->printGraphDirectly(printer, painter);

        if (pl.size()) {
            printer.newPage();
            painter.setClipping(false);
            painter.resetTransform();
        }
    }

    painter.end();
    qApp->restoreOverrideCursor();
}

void MainWindow::renderAllSvg()
{
    QList<ProjectWidget *> pl = getProjectSelection(tr("Save SVG Graph Files"), false, true);
    QStringList files;

    while (pl.size()) {
        files.append(pl.takeFirst()->renderSvgDirectly());
    }

    if (files.size()) {
        QMessageBox::information(this, tr("SVG Files Saved"), QString("SVG files saved:\n\n%1").arg(files.join("\n")));
    }
}

void MainWindow::helpAbout()
{
    HelpAboutDialog *hdlg = new HelpAboutDialog(this);
    hdlg->setVersion(version);
    hdlg->setLogDestination(logDest);
    hdlg->exec();
    delete hdlg;
}

/*
 * shows a help dialog with mouse actions on graphWindow
 */
void MainWindow::helpMouse()
{
    if (!mouseHelpDialog) {
        mouseHelpDialog = new NonModalInfoDialog(this);
        mouseHelpDialog->setTitle(QString(tr("Mouse and keyboard actions")));
        mouseHelpDialog->setText(global::helpTextMouseActions);
    }

    mouseHelpDialog->show();
}

/*
 *  launches the BGMN website in a browser
 */
void MainWindow::helpBgmnVariables()
{
    QDesktopServices::openUrl(QUrl("http://www.bgmn.de/variables.html", QUrl::TolerantMode));
}

void MainWindow::spacegrpDat()
{
    // lazy initialization
    if (spdlg == Q_NULLPTR) {
        spdlg = new SpacegroupDialog(this);
    }

    // only load the data from hd if it is necessary
    if (!spdlg->hasData()) {
        spdlg->setDirectory(settings->value("bgmnProject/bgmnExec", "").toString());
    }

    spdlg->show();
}

void MainWindow::afaparmDat()
{
    // lazy initialization
    if (asfdlg == Q_NULLPTR) {
        asfdlg = new AtomicScatteringFactorDialog(this);
    }

    asfdlg->show();
}

void MainWindow::editConfiguration()
{
    if (!instrdlg) {
        instrdlg = new EditInstrumentDialog(this);
    }

    instrdlg->resetDialog(true);
    instrdlg->show();
}

void MainWindow::editCurrentConfiguration()
{
    ProjectWidget *pw = getCurrentProject();
    if (!pw) return;

    QFileInfo fiInstrument(pw->getInstrumentConfigFile(true));
    QFileInfo fiControl(pw->getControlFileName());

    if (!fiInstrument.exists()) {
        QMessageBox::information(this, tr("Device file not found"),
                                 QString(tr("The device file was not found:\n\n%1"))
                                 .arg(fiInstrument.absoluteFilePath()));
        return;
    }

    if (!instrdlg) {
        instrdlg = new EditInstrumentDialog(this);
    }

    instrdlg->setWorkingDir(fiInstrument.absolutePath());
    instrdlg->setControlSavFile(fiControl.absoluteFilePath());
    instrdlg->openInstrumentSavFile(fiInstrument.absoluteFilePath());
    instrdlg->show();
}

void MainWindow::newConfiguration()
{
    NewInstrumentDialog *ndlg = new NewInstrumentDialog(this);
    ndlg->setWorkingDir(workingDir);

    if (!ndlg->openFile()) {
        delete ndlg;
        return;
    }

    if (ndlg->exec() == QDialog::Accepted) {
        QString savFile = ndlg->getSavFileName();

        if (QFile::exists(savFile)) {
            if (instrdlg == Q_NULLPTR) {
                instrdlg = new EditInstrumentDialog(this);
            }

            instrdlg->openInstrumentSavFile(savFile);
            instrdlg->show();
            instrdlg->raise();
        }
    }

    delete ndlg;
}

void MainWindow::showLegend(bool b)
{
    settings->setValue("graph/drawLegend", b);

    QList<ProjectWidget *> lst = getAllProjects();
    for (int i = 0; i < lst.size(); ++i) {
        lst.at(i)->setShowLegend(b);
    }
}

void MainWindow::showHkl(bool b)
{
    settings->setValue("graph/drawHkl", b);

    QList<ProjectWidget *> lst = getAllProjects();
    for (int i = 0; i < lst.size(); ++i) {
        lst.at(i)->setShowHkl(b);
    }
}

void MainWindow::contextHelp()
{
    ProjectWidget *pw = getCurrentProject();
    if (!pw) return;

    QStringList list = pw->contextHelp();

    if (list.size() > 1) {
        chelp->setKeyword(list.at(0), list.at(1));
    }
}

void MainWindow::undo()
{
    ProjectWidget *pw = getCurrentProject();
    if (pw) pw->undo();
}


void MainWindow::redo()
{
    ProjectWidget *pw = getCurrentProject();
    if (pw) pw->redo();
}

void MainWindow::editCut()
{
    ProjectWidget *pw = getCurrentProject();
    if (pw) pw->cut();
}

void MainWindow::editCopy()
{
    ProjectWidget *pw = getCurrentProject();
    if (pw) pw->copy();
}

void MainWindow::editPaste()
{
    ProjectWidget *pw = getCurrentProject();
    if (pw) pw->paste();
}

bool MainWindow::hasCurrentProject()
{
    if (!twProjects->currentItem()) return false;
    return true;
}

ProjectWidget * MainWindow::getProjectFromUid(const QUuid &u)
{
    for (int i = 1; i < widgetStack->count(); ++i) {
        if (widgetStack->widget(i)->inherits("ProjectWidget")) {
            ProjectWidget *pw = static_cast<ProjectWidget*>(widgetStack->widget(i));
            if (pw->uId() == u) return pw;
        }
    }

    return nullptr;
}

ProjectWidget * MainWindow::getCurrentProject()
{
    QTreeWidgetItem *it = twProjects->currentItem();
    if (!it) return nullptr;

    while (it->parent()) it = it->parent();

    return getProjectFromUid(it->data(0, Qt::UserRole).toUuid());
}

/*
 * returns a list of all ProjectWidgets
 */
QList<ProjectWidget *> MainWindow::getAllProjects()
{
    QList<ProjectWidget *> lst;

    for (int i = 1; i < widgetStack->count(); ++i) {
        ProjectWidget *pw = static_cast<ProjectWidget*>(widgetStack->widget(i));
        if (pw) lst.append(pw);
    }

    return lst;
}

/*
 * calls the project widget's ::setZoomRange() function
 */
void MainWindow::setZoomRange()
{
    ProjectWidget *pw = getCurrentProject();
    if (!pw) return;

    QVector<double> r = pw->getZoomRange();
    ZoomRangeDialog *zdlg = new ZoomRangeDialog(this);

    zdlg->setIdealValues(r[4], r[5], r[6], r[7]);
    zdlg->setValues(r[0], r[1], r[2], r[3]);

    if (zdlg->exec() == QDialog::Accepted) {
        zdlg->getValues(r[0], r[1], r[2], r[3]);
    } else {
        return;
    }

    QList<ProjectWidget*> pwlist = getProjectSelection(tr("Apply Zoom Ranges"), false, false);

    for (int i = 0; i < pwlist.size(); ++i) {
        qDebug() << QString("MainWindow::setZoomRange(): Setting zoom range for project %1: %2 %3 %4 %5").arg(pw->getControlFileName()).arg(r[0]).arg(r[1]).arg(r[2]).arg(r[3]);
        pwlist.at(i)->setZoomRange(r[0], r[1], r[2], r[3]);
    }

    delete zdlg;
}

void MainWindow::resetZoomRange()
{
    QList<ProjectWidget*> pwlist = getProjectSelection(tr("Reset Zoom Ranges"), false, false);

    for (int i = 0; i < pwlist.size(); ++i) {
        pwlist.at(i)->resetZoomRange();
    }

}

/*
 * opens all STR files referenced in a BGMN control file
 */
void MainWindow::fileOpenProjectStrFiles()
{
    ProjectWidget *pw = getCurrentProject();
    if (pw) pw->openProjectStrFiles();
}

/*
 * closes all STR files referenced in a BGMN control file
 */
void MainWindow::fileCloseProjectStrFiles()
{
    ProjectWidget *pw = getCurrentProject();
    if (pw) pw->closeProjectStrFiles();
}

/*
 * toggles the visibilities of all subphase patterns
 */
void MainWindow::togglePhasePatterns(bool b)
{
    settings->setValue("graph/phaseVisibility", b);
    ui->menu_Plot->setEnabled(b);

    QList<ProjectWidget *> lst = getAllProjects();

    for (int i = 0; i < lst.count(); ++i) {
        lst.at(i)->toggleSubScans(b);
    }
}

/*
 * returns all projects with status SCHEDULED
 */
QList<ProjectWidget *> MainWindow::getScheduledProjects()
{
    QList<ProjectWidget *> pwlist;

    for (int i = 0; i < twProjects->topLevelItemCount(); ++i) {
        QUuid uid = twProjects->topLevelItem(i)->data(0, Qt::UserRole).toUuid();
        ProjectWidget *pw = getProjectFromUid(uid);

        if (!pw) continue;

        if (pw->status() == SCHEDULED) pwlist.append(pw);
    }

    return pwlist;
}

/*
 * shows a dialog of all projects with checkboxes. Returns a list with checked projects
 *
 * - skipCurrent == false: all projects will be shown, including the current one
 * - skipCurrent == true: the current project will not be shown in the list
 * - ignoreType == false: only projects of the same type as the current one (BGMN, FP) will be shown
 * - ignoreType == true:  all project types will be shown (BGMN, FP)
 *
 * - precheck == 0: default behaviour, precheck selected projects, or all if none are selected
 * - precheck == 1: precheck the current project
 */
QList<ProjectWidget *> MainWindow::getProjectSelection(const QString &title, bool skipCurrent, bool ignoreType, int precheck)
{
    QList<ProjectWidget *> selPw;

    ProjectWidget *cpw = getCurrentProject();
    if (!cpw) return selPw;

    QList<QTreeWidgetItem*> selectedItems = twProjects->selectedItems();
    bool checkAll = selectedItems.size() > 1 ? false : true;

    QList<ProjectWidget*> allProjects;
    QList<QUuid> selectedUids;
    QUuid currentUid;

    // compile a map of all projects to show in the selection dialog
    for (int i = 0; i < twProjects->topLevelItemCount(); ++i) {
        ProjectWidget *pw = static_cast<ProjectItem*>(twProjects->topLevelItem(i))->projectWidget();

        if (pw) {
            if (pw == cpw) {
                if (skipCurrent) {
                    // we don't want to show the current project
                    continue;
                } else {
                    currentUid = pw->uId();
                }
            }

            if (pw->type() != cpw->type()) {
                // it's not the same type of project, so skip it if requested
                if (!ignoreType) continue;
            }

            allProjects.append(pw);

            if (precheck == 0) {
                if (pw->projectSelectItem()->isSelected() || checkAll) {
                    selectedUids.append(pw->uId());
                }
            }

            if (precheck == 1) {
                if (pw == cpw) selectedUids.append(pw->uId());
            }
        }
    }

    // create the selection dialog and show it
    ProjectSelectDialog *cpdlg = new ProjectSelectDialog(this);
    cpdlg->setWindowTitle(title);
    cpdlg->setFiles(allProjects, selectedUids);
    cpdlg->setCurrentUid(currentUid);

    // send the control file content to all selected projects
    if (cpdlg->exec() == QDialog::Accepted) {
        // the returned stringlist holds the selected projects' UIDs
        QList<QUuid> f = cpdlg->getChecked();

        for (int i = 0; i < f.count(); ++i) {
            ProjectWidget *pw = getProjectFromUid(f.at(i));
            if (pw) selPw.append(pw);
        }
    }

    delete cpdlg;
    return selPw;
}

/*
 * counts the number of projects
 *
 * - skipCurrent == false: all projects will be counted, including the current one
 * - skipCurrent == true: the current project will not be counted
 * - ignoreType == false: only projects of the same type as the current one (BGMN, FP) will be counted
 * - ignoreType == true:  all project types will be counted (BGMN, FP)
 */
int MainWindow::countProjects(bool skipCurrent, bool ignoreType)
{
    int n = 0;

    // get a pointer to the current widget, so we can skip it if necessary and
    // determine it's type
    ProjectWidget *cpw = getCurrentProject();
    if (!cpw) return n;

    QUuid currentUid;

    // loop over all projects
    QList<ProjectWidget *> lst = getAllProjects();

    for (int i = 0; i < lst.count(); ++i) {
        if (lst.at(i) == cpw) {
            // we don't want to count the current project
            if (skipCurrent) continue;
            else currentUid = lst.at(i)->uId();
        }

        if (lst.at(i)->type() != cpw->type()) {
            // it's not the same type of project, so skip it if requested
            if (!ignoreType) continue;
        }

        ++n;
    }

    return n;
}

/*
 * returns the number of currently running projects
 */
int MainWindow::hasRunningProjects()
{
    int n = 0;

    QList<ProjectWidget *> lst = getAllProjects();

    for (int i = 0; i < lst.count(); ++i) {
        if (lst.at(i)->status() == RUNNING) ++n;
    }

    return n;
}


/*
 * returns the number of currently running projects
 */
int MainWindow::hasScheduledProjects()
{
    int n = 0;

    QList<ProjectWidget *> lst = getAllProjects();

    for (int i = 0; i < lst.count(); ++i) {
        if (lst.at(i)->isScheduledForBatch()) ++n;
    }

    return n;
}

/*
 * Populates the menus:
 *  File -> Recent Graph Files
 *  File -> Recent Text Files
 */
void MainWindow::setupRecentFiles()
{
    // deleting all actions manually. QMenu::clear() will remove but not delete them,
    // because they are not owned by the menu. This is imporant and done
    // deliberately, because we must be able to re-arrange the actions (move
    // a clicked action to the top) without the actions being automatically
    // deleted by the menu when removed from the menu.
    QList<QAction *> lstG = menuRecentGraphs->actions();
    QList<QAction *> lstT = menuRecentText->actions();

    menuRecentGraphs->clear();
    menuRecentText->clear();

    while (!lstG.isEmpty()) {
        QAction *act = lstG.takeFirst();
        if (act) delete act;
    }

    while (!lstT.isEmpty()) {
        QAction *act = lstT.takeFirst();
        if (act) delete act;
    }

    // loop from the end of the stringList 10 positions down towards the beginning
    int szG = recentFilesGraph.size() - 1;
    for (int i = szG; i >= qMax(szG - 10, 0); --i) {
        QAction *act = new QAction(recentFilesGraph.at(i), this);
        act->setData(recentFilesGraph.at(i));
        menuRecentGraphs->addAction(act);
    }

    // loop from the end of the stringList 10 positions down towards the beginning
    int szT = recentFilesText.size() - 1;
    for (int i = szT; i >= qMax(szT - 10, 0); --i) {
        QAction *act = new QAction(recentFilesText.at(i), this);
        act->setData(recentFilesText.at(i));
        menuRecentText->addAction(act);
    }
}

/*
 * clicking a recent graph file menu entry.
 * the clicked action must be moved to the top of the menu
 */
void MainWindow::openRecentGraph(QAction *a)
{
    QList<QAction *> lst = menuRecentGraphs->actions();

    // find the action a in the list of actions
    int idx = lst.indexOf(a);

    // if it was found, move it to the beginning of the list
    // and re-populate the menu
    if ((idx > 0) && (idx < lst.size())) {
        lst.insert(0, lst.takeAt(idx));

        // this will not delete the actions, because they are not owned by the menu
        menuRecentGraphs->clear();

        // actions are still in the list, so we can add them back to the menu in the
        // new order
        menuRecentGraphs->addActions(lst);
    }

    loadGraphFiles(QStringList(a->data().toString()), QString());
    qDebug() << QString("MainWindow::openRecentGraph(): Opening recent file %1").arg(a->data().toString());
}

/*
 * clicking a recent text file menu entry.
 * the clicked action must be moved to the top of the menu
 */
void MainWindow::openRecentText(QAction *a)
{
    QList<QAction *> lst = menuRecentText->actions();

    // find the action a in the list of actions
    int idx = lst.indexOf(a);

    // if it was found, move it to the beginning of the list
    // and re-populate the menu
    if ((idx > 0) && (idx < lst.size())) {
        lst.insert(0, lst.takeAt(idx));

        // this will not delete the actions, because they are not owned by the menu
        menuRecentText->clear();

        // actions are still in the list, so we can add them back to the menu in the
        // new order
        menuRecentText->addActions(lst);
    }

    loadTextFiles(QStringList(a->data().toString()));
    qDebug() << QString("MainWindow::openRecentText(): Opening recent file %1").arg(a->data().toString());
}

void MainWindow::scanYOffsetUp()
{
    ProjectWidget *pw = getCurrentProject();
    if (pw) pw->scanYOffsetUp();
}

void MainWindow::scanYOffsetDown()
{
    ProjectWidget *pw = getCurrentProject();
    if (pw) pw->scanYOffsetDown();
}

void MainWindow::scanXOffsetLeft()
{
    ProjectWidget *pw = getCurrentProject();
    if (pw) pw->scanXOffsetLeft();
}

void MainWindow::scanXOffsetRight()
{
    ProjectWidget *pw = getCurrentProject();
    if (pw) pw->scanXOffsetRight();
}

void MainWindow::scanOffsetReset()
{
    ProjectWidget *pw = getCurrentProject();
    if (pw) pw->scanOffsetReset();
}

void MainWindow::removeScan()
{
    ProjectWidget *pw = getCurrentProject();
    if (pw) pw->removeCurrentScan();
}

void MainWindow::createTemplate()
{
    ProjectWidget *pw = getCurrentProject();
    if (pw) pw->createPreset();
}

void MainWindow::projectTabChanged(ProjectWidget *pw, int i)
{
    if (pw == getCurrentProject()) {
        statusFileName->setText(QDir::toNativeSeparators(pw->currentFileName()));
    }

    if (i) {
        if (hasScheduledProjects()) {
            ui->actionFollow_Active_Refinement->setChecked(false);
        }
    } else {
        statusCursor->setText(QString());
    }
}

/*
 * Saves a zip archive of the current project under a name created automatically
 */
void MainWindow::createZipArchive()
{
    ProjectWidget *pw = getCurrentProject();
    if (!pw) return;

    QDateTime dt = QDateTime::currentDateTime();
    QString zf = pw->workingDir() + "/" + pw->baseName() + "-" + dt.toString("yyyyMMdd-hhmmss");

    // add a number if the file name already exists
    if (QFile::exists(zf + ".zip")) {
        for (int i = 1; i < 999; ++i) {
            if (!QFile::exists(QString("%1-%2.zip").arg(zf).arg(i))) {
                zf = QString("%1-%2.zip").arg(zf).arg(i);
                break;
            }
        }
    } else {
        zf = zf + ".zip";
    }

    pw->createZipArchive(QDir::fromNativeSeparators(zf));
}

/*
 * Saves a zip archive of the current project under a user-given name
 */
void MainWindow::createZipArchiveAs()
{
    ProjectWidget *pw = getCurrentProject();
    if (!pw) return;

    QString zf = QFileDialog::getSaveFileName(this, tr("Save Compressed Archive"), pw->workingDir(), tr("Compressed Archive (*.zip *.ZIP)"));

    if (!zf.isEmpty()) {
        pw->createZipArchive(zf);
    }
}

void MainWindow::setInternalStandard()
{
    ProjectWidget *pw = getCurrentProject();
    if (pw) pw->setInternalStandard();
}

void MainWindow::unsetInternalStandard()
{
    ProjectWidget *pw = getCurrentProject();
    if (pw) pw->unsetInternalStandard();
}

void MainWindow::generateReport()
{
    ProjectWidget *pw = getCurrentProject();
    if (pw) pw->generateReport();
}

void MainWindow::learnProfile()
{
    LearnProfileDialog *lpdlg = new LearnProfileDialog(this);
    lpdlg->exec();
    delete lpdlg;
}

void MainWindow::instrumentPeakProfile()
{
    if (!bgmninsprofdlg) {
        bgmninsprofdlg = new BgmnInstrumentProfileDialog(this);
    }

    bgmninsprofdlg->show();
}

void MainWindow::applyTextBlock(QAction *action)
{
    ProjectWidget *pw = getCurrentProject();
    if (pw) pw->applyTextBlock(action->data().toString());
}

void MainWindow::searchReplace()
{
    ProjectWidget *pw = getCurrentProject();
    if (pw) pw->searchReplace();
}

void MainWindow::resetMarginColor()
{
    ProjectWidget *pw = getCurrentProject();
    if (pw) pw->resetMarginColor();
}

void MainWindow::updateNumberOfOpenProject()
{
    int n = twProjects->topLevelItemCount();
    statusOpenProjects->setText(QString("%1 Project%2").arg(n).arg(n == 1 ? "" : "s"));
}

void MainWindow::diaStripImpurities()
{
    if (!purifyScansDlg) {
        purifyScansDlg = new PurifyScans(this);
    }

    ProjectWidget *pw = getCurrentProject();

    if (pw) {
        if (pw->type() == "BGMN") {
            purifyScansDlg->loadFile(pw->getGraphFile());
        }
    }

    purifyScansDlg->show();
}

void MainWindow::diaSimulateCompositions()
{
    if (!simScansDlg) {
        simScansDlg = new SimulateScans(this);
    }

    simScansDlg->show();
}

void MainWindow::diaFormatResults()
{
    if (!formatResultsDlg) {
        formatResultsDlg = new FormatResults(this);
    }

    formatResultsDlg->show();
}

void MainWindow::refStrFavoritesToggled(bool b)
{
    QList<ProjectWidget *> lst = getAllProjects();

    for (int i = 0; i < lst.count(); ++i) {
        lst.at(i)->toggleRefStrFavorites(b);
    }
}

void MainWindow::indexReferenceStructures()
{
    indexBgmnHkl();
}

void MainWindow::indexBgmnHkl()
{
    QFileInfo fiBgmn(settings->value("bgmnProject/bgmnExec", QString()).toString());

    if (!fiBgmn.exists()) {
        qDebug() << QString("MainWindow::indexBgmnHkl(): BGMN executable not found. Exiting.");
        return;
    }

    if (!indexProgressDialog) {
        indexProgressDialog = new IndexProgressDialog(this);
        indexProgressDialog->setWindowTitle(tr("Indexing reference structures"));
        connect(refStrManager, SIGNAL(signalTotalFiles(int)), indexProgressDialog, SLOT(setMaximum(int)));
        connect(refStrManager, SIGNAL(signalCurrentFile(int)), indexProgressDialog, SLOT(setValue(int)));
        connect(refStrManager, SIGNAL(signalCurrentText(QString)), indexProgressDialog, SLOT(setLabelText(QString)));
        connect(refStrManager, SIGNAL(signalProcessOutput(const QString &)), indexProgressDialog, SLOT(appendOutputText(const QString &)));
        connect(indexProgressDialog, SIGNAL(canceled()), refStrManager, SLOT(cancel()));
    }

    indexProgressDialog->show();
    refStrManager->update();
}

void MainWindow::indexBgmnHklComplete(const QStringList &errFiles)
{
    indexProgressDialog->hide();
    indexProgressDialog->reset();

    // send list of indexed file names to all open BGMN projects
    QList<ProjectWidget *> lst = getAllProjects();

    for (int i = 0; i < lst.count(); ++i) {
        if (lst.at(i)->type() == "BGMN") lst.at(i)->setReferenceStructureFileList();
    }

    if (errFiles.size() > 1) {
        BgmnIndexingErrorDialog *idxErrDlg = new BgmnIndexingErrorDialog(this);
        idxErrDlg->setErrorString(errFiles);
        idxErrDlg->exec();
        delete idxErrDlg;
    }
}

void MainWindow::indexFpHkl()
{
    /* todo */
}

/*
 * shows the absorption coefficient dialog
 */
void MainWindow::absorptionCoefficients()
{
    if (!abscoeffdlg) {
        abscoeffdlg = new AbsorptionCoefficientDialog(this);
    }

    ProjectWidget *pw = getCurrentProject();
    if (pw) abscoeffdlg->setProjectData(pw->getResultsFile(), pw->wavelength());

    abscoeffdlg->show();
}

void MainWindow::applyPeakIntegralsToAll(QList<global::PeakRange> l)
{
    QList<ProjectWidget *> pl = getProjectSelection(tr("Apply Peak Integral Ranges"), true, true);
    if (!pl.size()) return;

    for (int i = 0; i < pl.size(); ++i) {
        pl.at(i)->applyPeakIntegralRanges(l);
    }
}

/*
 * returns true if an older version was found on the system
 */
bool MainWindow::hasOldPreviousInstallation(int maj, int min, int bld)
{
    // no window state of a previous installation found, so this must be a new installation
    QByteArray b = settings->value("window/state", QByteArray()).toByteArray();

    if (b.isEmpty()) {
        qDebug() << QString("MainWindow::hasOldPreviousInstallation(): No previous installation found.");
        return false;
    }

    QString v = settings->value("config/ProfexVersion", QString()).toString();

    // no version string found in settings, so there must have been a pre-3.12.0 version present
    if (v.isEmpty()) {
        qDebug() << QString("MainWindow::hasOldPreviousInstallation(): Pre-3.12.0 version was used on the system.");
        return true;
    }

    // evaluate the version string found in the settings, written by the last previous run
    QStringList l = v.split(QRegularExpression("[\\.-]"));

    if (l.size() < 3) {
        qDebug() << QString("MainWindow::hasOldPreviousInstallation(): Could not parse version string. Assuming older version.");
        return true;
    }

    // if the stored version is equal or newer than the present version, no older version was used last
    if ((l.at(0).toInt() >= maj) && (l.at(1).toInt() >= min) && (l.at(2).toInt() >= bld)) {
        qDebug() << QString("MainWindow::hasOldPreviousInstallation(): Recent version %1 was used on the system.").arg(v);
        return false;
    }

    // the stored version was older than the present version
    qDebug() << QString("MainWindow::hasOldPreviousInstallation(): Older version %1 was used on the system.").arg(v);
    return true;
}

void MainWindow::toggleProjectMenusEnabled(bool b)
{
    ui->menuProject->setEnabled(b);
    ui->menu_Results->setEnabled(b);
    ui->menu_Run->setEnabled(b);
    ui->menu_View->setEnabled(b);
    menuTextBlocks->setEnabled(b);

    ui->projectToolBar->setEnabled(b);
    ui->refinementToolBar->setEnabled(b);
    referenceToolBar->setEnabled(b);

    ui->actionSave->setEnabled(b);
    ui->actionSaveAs->setEnabled(b);
    ui->actionCloseProject->setEnabled(b);
    ui->actionRemove_Scan->setEnabled(b);
    ui->actionPrint->setEnabled(b);
    ui->actionPrint_All_Graphs->setEnabled(b);
    ui->actionExport_all_Graphs_to_SVG->setEnabled(b);
    ui->actionSave_all_files->setEnabled(b);
    ui->actionClose_All_Projects->setEnabled(b);
    ui->actionUndo->setEnabled(b);
    ui->actionRedo->setEnabled(b);
    ui->actionCut->setEnabled(b);
    ui->actionCopy->setEnabled(b);
    ui->actionPaste->setEnabled(b);
    ui->actionApply_Control->setEnabled(b);
    ui->actionRevertControlFile->setEnabled(b);
    ui->actionSearch_and_Replace->setEnabled(b);
    ui->actionScan_Math->setEnabled(b);
    ui->actionAdd_Base_Line->setEnabled(b);
    ui->actionSmooth_Scan->setEnabled(b);
    ui->actionEdit_Current_FPA_Configuration->setEnabled(b);
    ui->actionFind_in_Files->setEnabled(b);
    ui->actionProjectOpenFileManager->setEnabled(b);
}

void MainWindow::fileOpenProjectArchive()
{
    QFileInfo archive(QFileDialog::getOpenFileName(this, tr("Import Project Archive"), workingDir, tr("ZIP Archives (*.zip *.ZIP)")));
    qDebug() << QString("MainWindow::fileOpenProjectArchive(): Opening archive %1").arg(archive.absoluteFilePath());

    if (!archive.exists()) {
        qDebug() << QString("MainWindow::fileOpenProjectArchive(): File doesn't exist. Exiting.");
        return;
    }

    QString stdtmp = settings->getTempLocation();
    QDir tempdir(stdtmp + "/" + archive.completeBaseName());
    QDir finaldir(archive.absolutePath() + "/" + archive.completeBaseName());

    // check if the dirs already exists. If yes, add a increment counter to the dir path
    for (int i = 1; i < 10000; ++i) {
        if (tempdir.exists()) tempdir.setPath(stdtmp + "/" + archive.completeBaseName() + QString("-%1").arg(i));
        else break;
    }

    for (int i = 1; i < 10000; ++i) {
        if (finaldir.exists()) finaldir.setPath(archive.absolutePath() + "/" + archive.completeBaseName() + QString("-%1").arg(i));
        else break;
    }

    if (!finaldir.exists()) finaldir.mkpath(".");
    qDebug() << QString("MainWindow::fileOpenProjectArchive(): Extracting files to %1").arg(tempdir.path());

    // extract the archive to the temporary directory
    QStringList cfiles = JlCompress::getFileList(archive.absoluteFilePath());
    QStringList efiles = JlCompress::extractFiles(archive.absoluteFilePath(), cfiles, tempdir.path());
    QStringList ffiles;
    QStringList graphExtensions(importHandler->extensions());

    QStringList convExtensions;
    convExtensions << "sav" << "str" << "xy" << "val" << "ger" << "lam" << "dia" << "lst" << "par";

    // copy to final location, using functions converting line ends between Windows and Unix
    for (int i = 0; i < efiles.size(); ++i) {
        QFileInfo fi(efiles.at(i));
        QFileInfo fo(finaldir.absolutePath() + "/" + fi.fileName());

        if (convExtensions.contains(fi.suffix().toLower())) {
            qDebug() << QString("MainWindow::fileOpenProjectArchive(): Copying file %1 to %2, converting line endings if needed.").arg(fi.absoluteFilePath()).arg(fo.absoluteFilePath());
            BgmnFileIO::copyFile(fi.absoluteFilePath(), fo.absoluteFilePath());
        } else {
            qDebug() << QString("MainWindow::fileOpenProjectArchive(): Copying file %1 to %2, not converting line endings.").arg(fi.absoluteFilePath()).arg(fo.absoluteFilePath());
            QFile::copy(fi.absoluteFilePath(), fo.absoluteFilePath());
        }

        ffiles.append(fo.absoluteFilePath());
    }

    qDebug() << QString("MainWindow::fileOpenProjectArchive(): Removing temporary directory %1").arg(tempdir.path());
    tempdir.removeRecursively();

    // open the project
    for (int i = 0; i < ffiles.size(); ++i) {
        QFileInfo ffile(ffiles.at(i));

        qDebug() << QString("MainWindow::fileOpenProjectArchive(): Probing file %1").arg(ffile.absoluteFilePath());
        // check if the file extension is known and associated with a certain type of project
        if (fExtProjectType.contains(ffile.suffix().toLower())) {
            // check if it is a graph file extension
            if (graphExtensions.contains(ffile.suffix().toLower())) {
                // it's a graph file extension
                qDebug() << QString("MainWindow::fileOpenProjectArchive(): Loading as graph file %1").arg(ffile.absoluteFilePath());
                loadGraphFiles(QStringList(ffile.absoluteFilePath()), QString());
                return;
            } else {
                // must be a text file extension
                qDebug() << QString("MainWindow::fileOpenProjectArchive(): Loading as text file %1").arg(ffile.absoluteFilePath());
                loadTextFiles(QStringList(ffile.absoluteFilePath()));
                return;
            }
        }
    }
}

void MainWindow::projectOpenFileManager()
{
    ProjectWidget *pw = getCurrentProject();

    if (pw) {
        QUrl url = QUrl::fromLocalFile(pw->workingDir());

        if (QDesktopServices::openUrl(url)) {
            qDebug() << QString("MainWindow::projectOpenFileManager(): Opening %1").arg(url.toDisplayString());
        } else {
            qDebug() << QString("MainWindow::projectOpenFileManager(): Opening %1 failed").arg(url.toDisplayString());
        }
    }
}

void MainWindow::clearToolsDialogs()
{
    if (smdlg)      smdlg->clearProject();
    if (blDialog)   blDialog->clearProject();
    if (sSmoothDlg) sSmoothDlg->clearProject();
}

void MainWindow::updateToolsDialogs()
{
    updateScanMathDlg();
    updateBaseLineDlg();
    updateSmoothScanDlg();
}

void MainWindow::scanMath()
{
    if (!hasCurrentProject()) return;

    if (!smdlg) {
        smdlg = new ScanMathDialog(this);
    }

    smdlg->show();
    updateScanMathDlg();
}

void MainWindow::updateScanMathDlg()
{
    if (!smdlg) return;
    if (!smdlg->isVisible()) return;

    ProjectWidget *pw = getCurrentProject();
    if (pw) smdlg->setProject(pw);
    else    smdlg->clearProject();
}

void MainWindow::addBaseLine()
{
    if (!hasCurrentProject()) return;

    if (!blDialog) {
        blDialog = new BaseLineDialog(this);
    }

    blDialog->show();
    updateBaseLineDlg();
}

void MainWindow::updateBaseLineDlg()
{
    if (!blDialog) return;
    blDialog->clearTemporary(); // in the old project
    if (!blDialog->isVisible()) return;

    ProjectWidget *pw = getCurrentProject();
    if (pw) blDialog->setProject(pw);
    else    blDialog->clearProject();
}

void MainWindow::smoothScan()
{
    if (!hasCurrentProject()) return;

    if (!sSmoothDlg) {
        sSmoothDlg = new ScanSmoothDialog(this);
    }

    sSmoothDlg->show();
    updateSmoothScanDlg();
}

void MainWindow::updateSmoothScanDlg()
{
    if (!sSmoothDlg) return;
    sSmoothDlg->clearTemporary(); // in the old project
    if (!sSmoothDlg->isVisible()) return;

    ProjectWidget *pw = getCurrentProject();
    if (pw) sSmoothDlg->setProject(pw);
    else    sSmoothDlg->clearProject();
}

void MainWindow::applyPreset(QString s)
{
    QList<ProjectWidget *> pl = getProjectSelection(tr("Apply Preset"), false, false, 1);
    if (!pl.size()) return;

    QStringList errors;
    bool overwrite = false;

    QMessageBox msgBox;
    msgBox.setWindowTitle(tr("Overwrite project"));
    msgBox.setText(tr("The following project already has a control file."));
    msgBox.setIcon(QMessageBox::Question);
    msgBox.setStandardButtons(QMessageBox::YesToAll | QMessageBox::Yes | QMessageBox::No | QMessageBox::Cancel);
    msgBox.setDefaultButton(QMessageBox::No);

    QProgressDialog *prgDlg = new QProgressDialog(this);
    prgDlg->setMaximum(pl.size());
    prgDlg->setWindowTitle(tr("Applying Presets"));
    prgDlg->show();
    int prg = 0;

    for (int i = 0; i < pl.size(); ++i) {
        prgDlg->setLabelText(pl.at(i)->baseName());
        prgDlg->setValue(prg);
        if (prgDlg->wasCanceled()) break;

        int n = pl.at(i)->applyPreset(s, overwrite);

        QFileInfo gfi(pl.at(i)->getGraphFile());

        if (n < 0) { // a serious error occurred
            qDebug() << QString("MainWindow::applyPreset(): An error occurred when applying preset to project %1").arg(gfi.fileName());
            errors << gfi.fileName();
        }

        if (n == 0) { // project already has control file
            msgBox.setInformativeText(QString("Do you want to overwrite it?\n\n%1").arg(gfi.completeBaseName()));
            qApp->restoreOverrideCursor();
            int ovr = msgBox.exec();
            qApp->setOverrideCursor(Qt::WaitCursor);

            if (ovr == QMessageBox::YesToAll) {
                qDebug() << QString("MainWindow::applyPreset(): Overwrite project %1: User selection = yes to all").arg(gfi.fileName());
                overwrite = true;
                pl.at(i)->applyPreset(s, overwrite);
            }

            if (ovr == QMessageBox::Yes) {
                qDebug() << QString("MainWindow::applyPreset(): Overwrite project %1: User selection = yes").arg(gfi.fileName());
                pl.at(i)->applyPreset(s, true);
            }

            if (ovr == QMessageBox::No) {
                qDebug() << QString("MainWindow::applyPreset(): Overwrite project %1: User selection = no").arg(gfi.fileName());
            }

            if (ovr == QMessageBox::Cancel) {
                qDebug() << QString("MainWindow::applyPreset(): Overwrite project %1: User selection = cancel").arg(gfi.fileName());
                break;
            }
        }

        ++prg;
    }

    prgDlg->hide();
    delete prgDlg;

    qApp->restoreOverrideCursor();

    if (errors.size()) {
        QMessageBox::information(this, tr("Applying presets"),
                                 QString(tr("Applying the preset to the following projects failed.\n"
                                            "Please fix manually.\n\n%1")).arg(errors.join("\n")));
    }
}

void MainWindow::saveBatchScript()
{
    QList<ProjectWidget *> pl = getProjectSelection(tr("Save Batch Refinement Script"), false, false);
    if (!pl.size()) return;

    QStringList env;
    QString envUnixBash(tr("Mac / Linux (Bash)"));
    QString envWin(tr("Windows"));
    env << envUnixBash << envWin;

    bool ok;
    QString selectedEnv = QInputDialog::getItem(this, tr("Batch Refinement Script"), tr("Target operating system:"), env, 0, false, &ok);
    if (!ok) return;

    ProjectWidget *pw = getCurrentProject();
    QFileInfo exec(pw->type() == "BGMN" ? QDir::toNativeSeparators(settings->value("bgmnProject/bgmnExec", "").toString())
                                        : QDir::toNativeSeparators(settings->value("fpProject/fpExec", "").toString()));

    QString fileOut;
    QString strOut;

    if (selectedEnv == envUnixBash) {
        fileOut = QFileDialog::getSaveFileName(this, tr("Save Batch File"), workingDir, tr("Bash script (*.sh *.SH);;All files (*.*"));
        if (fileOut.isEmpty()) return;

        strOut += QString("#! /bin/bash\n\n");
        strOut += QString("export EFLECH=%1\nexport PATH=$PATH:%1\n").arg(exec.absolutePath());
        strOut += QString("EXEC=%1\n\n").arg(exec.absoluteFilePath());

        for (int i = 0; i < pl.size(); ++i) {
            pl.at(i)->saveAll();
            pl.at(i)->convertRawDataFile();
            QFileInfo fiCtrFile(pl.at(i)->getControlFileName());
            strOut += QString("echo refining project %1 of %2\n").arg(i+1).arg(pl.size());
            strOut += QString("cd %1\n").arg(fiCtrFile.absolutePath());
            strOut += QString("$EXEC %1\n\n").arg(fiCtrFile.fileName());
        }

        strOut += QString("echo All done.\n");
    }

    if (selectedEnv == envWin) {
        fileOut = QFileDialog::getSaveFileName(this, tr("Save Batch File"), workingDir, tr("CMD script (*.cmd *.CMD);;All files (*.*"));
        if (fileOut.isEmpty()) return;

        strOut += QString("@ECHO OFF\n\n");
        strOut += QString("setlocal\n");
        strOut += QString("SET \"PATH=%PATH%;%1\"\n").arg(exec.absolutePath());
        strOut += QString("SET \"EFLECH=%1\"\n").arg(exec.absolutePath());
        strOut += QString("SET \"EXEC=%1\"\n\n").arg(exec.absoluteFilePath());

        for (int i = 0; i < pl.size(); ++i) {
            pl.at(i)->saveAll();
            pl.at(i)->convertRawDataFile();
            QFileInfo fiCtrFile(pl.at(i)->getControlFileName());
            strOut += QString("ECHO refining project %1 of %2\n").arg(i+1).arg(pl.size());
            strOut += QString("cd /d \"%1\"\n").arg(fiCtrFile.absolutePath());
            strOut += QString("%EXEC% \"%1\"\n\n").arg(fiCtrFile.fileName());
        }

        strOut += QString("ECHO All done.\n");
        strOut += QString("PAUSE\n");
    }

    BgmnFileIO::writeTextFile(fileOut, strOut);
}

void MainWindow::editProjectStrFiles()
{
    ProjectWidget *pw = getCurrentProject();
    if (!pw) return;
    pw->editProjectStrFiles();
}

void MainWindow::electronDensityMap()
{
    QString pwd;
    ProjectWidget *pw = getCurrentProject();
    if (pw) pwd = pw->workingDir();

    if (!electMapDialog) {
        electMapDialog = new ElectronDensityMapDialog(this);
    }

    electMapDialog->setWorkingDir(pwd);
    electMapDialog->show();
}

void MainWindow::amorphousPeak()
{
    ProjectWidget *pw = getCurrentProject();
    if (pw) {
        if (pw->type() == "BGMN") pw->addAmorphousPeak();
    }
}

void MainWindow::editFindInFiles()
{
    if (!searchInFilesDlg) {
        searchInFilesDlg = new SearchInFilesDialog(this);
        connect(searchInFilesDlg, SIGNAL(runSearch()), this, SLOT(runFindInFiles()));
        connect(searchInFilesDlg, SIGNAL(projectFileSelected(QUuid,QString,int)), this, SLOT(raiseProjectFile(QUuid,QString,int)));
    }

    searchInFilesDlg->show();
}

void MainWindow::runFindInFiles()
{
    if (!searchInFilesDlg) return;
    QList<ProjectWidget*> pwList;

    if (searchInFilesDlg->getProjectSelection() == "cur") {
        pwList.append(getCurrentProject());
    } else {
        pwList.append(getAllProjects());
    }

    QString expr = searchInFilesDlg->getSearchExpression();

    for (int i = 0; i < pwList.size(); ++i) {
        QMap<QString, QMap<int, QStringList> > results = pwList.at(i)->searchInOpenFiles(searchInFilesDlg->getFileTypeSelection(), expr);
        searchInFilesDlg->appendResults(pwList.at(i)->baseName(), pwList.at(i)->uId(), results);
    }
}

void MainWindow::raiseProjectFile(QUuid uid, QString fn, int ln)
{
    ProjectWidget *pw = getProjectFromUid(uid);
    if (!pw) return;

    twProjects->setCurrentItem(pw->projectSelectItem());
    pw->raiseTab(fn, ln);
}

void MainWindow::exportToExcel()
{
    QList<ProjectWidget *> pl = getProjectSelection(tr("Export Results to Excel"));
    if (!pl.size()) return;

    for (int i = 0; i < pl.size(); ++i) {
        pl.at(i)->exportToExcel();
    }
}

void MainWindow::editExcelExport()
{
    ProjectWidget *pw = getCurrentProject();
    if (!pw) return;

    pw->editExcelExport();
}

void MainWindow::elementScatteringData()
{
    if (!elScatDataDlg) {
        elScatDataDlg = new ElementDataDialog(this);
    }

    elScatDataDlg->show();
}

void MainWindow::overrideProjectWavelength()
{
    double d = 0.154056;
    bool charact = true;

    WavelengthSelectDialog *wlsDlg = new WavelengthSelectDialog(this);

    if (wlsDlg->exec() != QDialog::Accepted) {
        delete wlsDlg;
        return;
    }

    d = wlsDlg->wavelength();
    charact = wlsDlg->isCharacteristic();
    delete wlsDlg;

    QList<ProjectWidget *> pl = getProjectSelection(tr("Set Wavelength"), false, true, 1);
    if (!pl.size()) return;

    for (int i = 0; i < pl.size(); ++i) {
        pl.at(i)->overrideProjectWavelength(10.0 * d, charact);
    }

    ProjectWidget *pw = getCurrentProject();
    if (pw) statusWavelength(pw->wavelength());
}

void MainWindow::toolsBrowseReferenceStructures()
{
    if (!refstrDlg) {
        refstrDlg = new BrowseReferenceStructuresDialog(this);
    }

    refstrDlg->show();
}

void MainWindow::traceScanFigures()
{
    if (!scanTraceDlg) {
        scanTraceDlg = new ScanTracerDialog(this);
    }

    scanTraceDlg->show();
}

void MainWindow::waterfallPlot()
{
    if (!waterFallPlotDlg) {
        waterFallPlotDlg = new WaterfallPlotDialog(this);
    }

    waterFallPlotDlg->show();
}

/* EOF */


