/***************************************************************************
                          bgmnpresethandler.cpp  -  description
                             -------------------
    begin                : Sat Jan 07 10:30:00 CEST 2017
    copyright            : (C) 2017 by Nicola Doebelin
    email                : ndoebelin@gmx.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "bgmnpresethandler.h"
#include "../libXrdIO/bgmnfileio.h"
#include "../libXrdIO/parser/bgmnsavparser.h"

#include <QFile>
#include <QFileInfo>
#include <QDir>
#include <QDateTime>
#include <QStringList>
#include <QDebug>
#include <QDomDocument>
#include <QMap>
#include <QMessageBox>

/*
 * Apply a refinement preset file (*.pfp)
 *
 * create a control file and copy all structure and device files
 * specified in the preset file of structure:
 *
 *  <preset>
 *      <device MD5hash="..." RelativePath="...">../Devices/cubix-ads-15mm.geq</device>
 *      <template MD5hash="..." RelativePath="...">../Devices/cubix-ads-15mm.tpl</template>
 *      <structures>
 *   		<structure MD5hash="..." RelativePath="...">../Structures/betaTCP.str</structure>
 *          <structure MD5hash="..." RelativePath="...">../Structures/hydroxylapatite.str</structure>
 *      </structures>
 *      <other>
 *          <textfile MD5hash="..." RelativePath="...">anyOtherFile.any</textfile>
 *          <textfile MD5hash="..." RelativePath="...">egBackground.xy</textfile>
 *          <binaryfile MD5hash="..." RelativePath="...">anyOtherFile.any</binaryfile>
 *      </other>
 *  </preset>
 *
 * relative or absolute paths can be used. Files specified in "other" tags
 * will just be copied to the destination. It could be a background file
 * for use with UNT=, or a description, or anything else.
 *
 * For <textfiles> BgmnProjectWidget::copyFile() will be used to convert line endings
 * correctly for platform independence.
 *
 * For <binaryfiles> QFile::copy() will be used
 *
 * Presets can also be used for baselines, peak integral ranges, and curve fits only, without
 * a refinement project. Such a preset would look as follows:
 *
 * <preset>
 *     <baseline Mode="SNIP" m="60" w="1"/>
 * </preset>
 *
 * or
 *
 *  <preset>
 *      <peakIntegrals>
 *          <peakIntegral Name="..." Start="..." End="..." Background="..."/>
 *          <peakIntegral Name="..." Start="..." End="..." Background="..."/>
 *      </peakIntegrals>
 *  </preset>
 *
 * or
 *
 *  <preset>
 *       <curveFit>
 *           <curveVariable LowerLimit="..." Name="..." UpperLimit="..." CheckState="..." Value="..."/>
 *           <curveVariable LowerLimit="..." Name="..." UpperLimit="..." CheckState="..." Value="..."/>
 *           <curveFunction Type="...">
 *               <ParameterNames>...;...;...</ParameterNames>
 *               <Variables>...;...;...</Variables>
 *           </curveFunction>
 *           <curveRange RangeEnd="..." RangeStart="..."/>
 *           <curveStepSize StepSizeCustom="..." StepSizeValue="..."/>
 *           <curveControls EpsX="..." MaxInt="..." DiffStep="..."/>
 *       </curveFit>
 *  </preset>
 *
 * or
 *
 * <preset>
 *      <excelWorkbook file="...">
 *          <excelCell WorkSheet=int Row=int Col=int Phase="..." Parameter="..."/>
 *      </excelWorkbook>
 * </preset>
 */

BgmnPresetHandler::BgmnPresetHandler(QObject *parent) :
    QObject(parent)
{
    fileTags << "device" << "template" << "structure" << "textfile" << "binaryfile";
    allTags = fileTags;
    allTags << "peakIntegral" << "baseline" << "curveVariable"
            << "curveFunction" << "curveControls" << "curveRange"
            << "curveStepSize" << "excelWorkbook" << "excelCell";
}

/********************************************************************************************
 *                                     Creating a preset                                    *
 ********************************************************************************************/

bool BgmnPresetHandler::createPreset(const QString &cfile,
                                     const QString &pfile,
                                     const QString &cfileContent,
                                     const QList<global::PeakRange> &pRanges,
                                     const QList<global::CurveFitCurve> &cFitsCurves,
                                     const QList<global::CurveFitVariable> &cFitVariables,
                                     const QMap<QString, QVariant> &cFitControls,
                                     const QList<global::AxObjectExcel> &axExcelControls,
                                     const QList<global::BaseLineSnip> &baseLines)
{
    QFileInfo fiControlFile(cfile);
    QFileInfo fiPresetFile(pfile);

    if (!createPresetDir(fiPresetFile.absolutePath())) {
        qDebug() << QString("BgmnPresetHandler::createPreset(): Could not create directory %1").arg(fiPresetFile.absolutePath());
        return false;
    }

    QList<BgmnPresetFile> srcFiles;
    QList<BgmnPresetFile> destFiles;

    if (!cfileContent.isEmpty()) {
        BgmnSavParser sparser(cfileContent, cfile);
        QStringList projectFiles;
        projectFiles << sparser.getPhases();
        projectFiles << allDevFiles(sparser.deviceFile(), fiControlFile, true);
        projectFiles << sparser.untFile();

        srcFiles = stringsToFileList(projectFiles, fiControlFile);
        destFiles = copyFilesToTemplate(srcFiles, fiPresetFile.absolutePath());

        destFiles << writeTemplateFile(fiPresetFile, sparser.createTemplate());
    }

    QDomDocument doc(createPresetXml(destFiles,
                                     fiPresetFile,
                                     pRanges,
                                     cFitsCurves,
                                     cFitVariables,
                                     cFitControls,
                                     axExcelControls,
                                     baseLines));

    if (!writePresetFile(doc, fiPresetFile)) {
        qDebug() << QString("BgmnPresetHandler::createPreset(): Could not write to file %1").arg(fiPresetFile.absoluteFilePath());
        return false;
    }

    return true;
}

bool BgmnPresetHandler::createPresetDir(const QString &s)
{
    QDir dir(s);
    if (dir.exists()) return true;

    return dir.mkpath(s);
}

QList<BgmnPresetFile> BgmnPresetHandler::stringsToFileList(const QStringList &strings, const QFileInfo &fiControlFile)
{
    QList<BgmnPresetFile> files;

    for (int i = 0; i < strings.size(); ++i) {
        BgmnPresetFile file;

        file.fileInfo = QFileInfo(strings.at(i));

        if (file.fileInfo.isRelative()) {
            file.relativePath = file.fileInfo.path();
            file.fileInfo.setFile(fiControlFile.absolutePath() + "/" + strings.at(i));
        }

        file.type = getFileType(file.fileInfo, fiControlFile);
        // warning: The MD5 hash may be calculated at the source location. If the file is copied with line
        // end conversion, the hash must be re-calculated at the destination!
        file.md5Hash = QString(BgmnFileIO::checksum(file.fileInfo.absoluteFilePath(), QCryptographicHash::Md5).toHex());

        qDebug() << QString("BgmnPresetHandler::stringsToFileList: Parsing file %1").arg(file.fileInfo.absoluteFilePath());
        qDebug() << QString("                                      Relative path %1").arg(file.relativePath);
        qDebug() << QString("                                      Type %1").arg(file.type);
        qDebug() << QString("                                      MD5 hash %1").arg(file.md5Hash);

        files.append(file);
    }

    return files;
}

FileType BgmnPresetHandler::getFileType(const QFileInfo &fi, const QFileInfo &fiControlFile)
{
    QStringList asciiFiles;
    QStringList binFiles;
    asciiFiles << "xy" << "csv" << "xyp" << "dat" << "asc" << "txt" << "xrdml"
               << "sav" << "lst" << "par" << "out" << "dia" << "ger"
               << "fcf" << "res" << "pdb";
    binFiles   << "raw" << "brml" << "rd" << "geq" << "jpg" << "png";

    if (fi.suffix().toLower() == "ger") return DEVICE_ASCII;
    if (fi.suffix().toLower() == "geq") return DEVICE_BINARY;
    if (fi.suffix().toLower() == "str") return STRUCTURE;

    if (fi.suffix().toLower() == "sav") {
        if (fi.completeBaseName() == fiControlFile.completeBaseName()) {
            return TEMPLATE;
        } else {
            return DEVICE_ASCII;
        }
    }

    if (asciiFiles.contains(fi.suffix().toLower())) return TEXT;
    if (binFiles.contains(fi.suffix().toLower())) return BINARY;

    // fallback
    qDebug() << QString("BgmnPresetHandler::getFileType: File %1 is of unknown file type. Falling back to %2").arg(fi.fileName()).arg(BINARY);
    return BINARY;
}

/*
 * returns a file list of destination files
 * Here we use QFile::copy for all file types, because we don't need any conversion of line endings
 * when creating presets.
 */
QList<BgmnPresetFile> BgmnPresetHandler::copyFilesToTemplate(const QList<BgmnPresetFile> &files, const QString &dest)
{
    qDebug() << QString("BgmnPresetHandler::copyFilesToTemplate: Copying %1 files to %2").arg(files.size()).arg(dest);

    QList<BgmnPresetFile> destFiles;

    for (int i = 0; i < files.size(); ++i) {
        BgmnPresetFile file = files.at(i);

        QString fsource(file.fileInfo.absoluteFilePath());
        QString fdest(dest + "/" + file.fileInfo.fileName());
        QFile::copy(fsource, fdest);

        file.fileInfo = QFileInfo(fdest);
        destFiles.append(file);

        qDebug() << QString("BgmnPresetHandler::copyFilesToTemplate: Copy %1 to %2").arg(fsource).arg(fdest);
    }

    return destFiles;
}

QStringList BgmnPresetHandler::allDevFiles(const QString &devFile, const QFileInfo &fiControlFile, bool relative)
{
    QStringList allFiles;

    QDir cdir(fiControlFile.absoluteDir());
    QFileInfo fidev(QFileInfo(cdir, devFile));

    QList<QFileInfo> searchFiles;
    searchFiles << QFileInfo(cdir, fidev.completeBaseName() + ".geq");
    searchFiles << QFileInfo(cdir, fidev.completeBaseName() + ".GEQ");
    searchFiles << QFileInfo(cdir, fidev.completeBaseName() + ".ger");
    searchFiles << QFileInfo(cdir, fidev.completeBaseName() + ".GER");
    searchFiles << QFileInfo(cdir, fidev.completeBaseName() + ".sav");
    searchFiles << QFileInfo(cdir, fidev.completeBaseName() + ".SAV");

    for (int i = 0; i < searchFiles.size(); ++i) {
        if (searchFiles.at(i).exists()) {
            if (relative) allFiles << cdir.relativeFilePath(searchFiles.at(i).absoluteFilePath());
            else          allFiles << searchFiles.at(i).absoluteFilePath();
        }
    }

    return allFiles;
}

BgmnPresetFile BgmnPresetHandler::writeTemplateFile(const QFileInfo &fiPresetFile, const QString &content)
{
    BgmnPresetFile file;
    QString tplFileName(fiPresetFile.absolutePath() + "/" + fiPresetFile.completeBaseName() + ".tpl");
    BgmnFileIO::writeTextFile(tplFileName, content);

    file.fileInfo = QFileInfo(tplFileName);
    file.md5Hash = QString(BgmnFileIO::checksum(tplFileName, QCryptographicHash::Md5).toHex());
    file.relativePath = QString();
    file.type = TEMPLATE;

    return file;
}

QDomDocument BgmnPresetHandler::createPresetXml(const QList<BgmnPresetFile> &files,
                                                const QFileInfo &fiPresetFile,
                                                const QList<PeakRange> &pRanges,
                                                const QList<CurveFitCurve> &cFitsCurves,
                                                const QList<CurveFitVariable> &cFitVariables,
                                                const QMap<QString, QVariant> &cFitControls,
                                                const QList<global::AxObjectExcel> &axExcelControls,
                                                const QList<BaseLineSnip> &baseLines)
{
    QDomDocument doc("ProfexPreset");

    QDomElement rootEl = doc.createElement("preset");
    rootEl.setAttribute("name", fiPresetFile.completeBaseName());
    rootEl.setAttribute("version", QDateTime::currentDateTime().toString("yyyyMMdd"));
    doc.appendChild(rootEl);

    bool addStrucs = false;
    bool addOther = false;
    bool addIntRanges = false;
    bool addCurveFits = false;
    bool addExcel = false;
    bool addBaseLine = false;

    QDomElement strucsEl = doc.createElement("structures");
    QDomElement otherEl  = doc.createElement("other");

    for (int i = 0; i < files.size(); ++i) {
        BgmnPresetFile file = files.at(i);

        if ((file.type == DEVICE_ASCII) || (file.type == DEVICE_BINARY)) {
            QDomElement el = doc.createElement("device");
            el.setAttribute("MD5hash", file.md5Hash);
            el.setAttribute("RelativePath", file.relativePath);
            QDomText txt = doc.createTextNode(file.fileInfo.fileName());
            el.appendChild(txt);
            rootEl.appendChild(el);
        }

        if (file.type == TEMPLATE) {
            QDomElement el = doc.createElement("template");
            el.setAttribute("MD5hash", file.md5Hash);
            el.setAttribute("RelativePath", file.relativePath);
            QDomText txt = doc.createTextNode(file.fileInfo.fileName());
            el.appendChild(txt);
            rootEl.appendChild(el);
        }

        if (file.type == STRUCTURE) {
            QDomElement el = doc.createElement("structure");
            el.setAttribute("MD5hash", file.md5Hash);
            el.setAttribute("RelativePath", file.relativePath);
            QDomText txt = doc.createTextNode(file.fileInfo.fileName());
            el.appendChild(txt);
            strucsEl.appendChild(el);
            addStrucs = true;
        }

        if (file.type == TEXT) {
            QDomElement el = doc.createElement("textfile");
            el.setAttribute("MD5hash", file.md5Hash);
            el.setAttribute("RelativePath", file.relativePath);
            QDomText txt = doc.createTextNode(file.fileInfo.fileName());
            el.appendChild(txt);
            otherEl.appendChild(el);
            addOther = true;
        }

        if (file.type == BINARY) {
            QDomElement el = doc.createElement("binaryfile");
            el.setAttribute("MD5hash", file.md5Hash);
            el.setAttribute("RelativePath", file.relativePath);
            QDomText txt = doc.createTextNode(file.fileInfo.fileName());
            el.appendChild(txt);
            otherEl.appendChild(el);
            addOther = true;
        }
    }

    QList<QDomElement> baseLineList;

    for (int i = 0; i < baseLines.size(); ++i) {
        QDomElement baselineEl = doc.createElement("baseline");
        baselineEl.setAttribute("Mode", "SNIP");
        baselineEl.setAttribute("m", baseLines.at(i).m);
        baselineEl.setAttribute("w", baseLines.at(i).mode);
        baseLineList.append(baselineEl);
        addBaseLine = true;
    }

    QDomElement prangeEl = doc.createElement("peakIntegrals");

    for (int i = 0; i < pRanges.size(); ++i) {
        QDomElement el = doc.createElement("peakIntegral");
        el.setAttribute("Name",  pRanges.at(i).name);
        el.setAttribute("Start", pRanges.at(i).start);
        el.setAttribute("End",   pRanges.at(i).end);
        el.setAttribute("Background", pRanges.at(i).background);
        prangeEl.appendChild(el);
        addIntRanges = true;
    }

    QDomElement cFitEl = doc.createElement("curveFit");

    for (int i = 0; i < cFitVariables.size(); ++i) {
        QDomElement el = doc.createElement("curveVariable");
        el.setAttribute("Name", cFitVariables.at(i).name);
        el.setAttribute("Value", cFitVariables.at(i).value);
        el.setAttribute("LowerLimit", cFitVariables.at(i).lowerLimit);
        el.setAttribute("UpperLimit", cFitVariables.at(i).upperLimit);
        el.setAttribute("CheckState", cFitVariables.at(i).checkState);
        cFitEl.appendChild(el);
    }

    for (int i = 0; i < cFitsCurves.size(); ++i) {
        QDomElement el = doc.createElement("curveFunction");
        el.setAttribute("Type", cFitsCurves.at(i).type);

        QDomElement elPnm = doc.createElement("ParameterNames");
        QDomElement elVar = doc.createElement("Variables");

        QDomText txPnm = doc.createTextNode(cFitsCurves.at(i).parameterNames.join(";"));
        QDomText txVar = doc.createTextNode(cFitsCurves.at(i).variables.join(";"));

        elPnm.appendChild(txPnm);
        elVar.appendChild(txVar);

        el.appendChild(elPnm);
        el.appendChild(elVar);
        cFitEl.appendChild(el);
        addCurveFits = true;
    }

    QDomElement elCfRange    = doc.createElement("curveRange");
    QDomElement elCfStepSize = doc.createElement("curveStepSize");
    QDomElement elCfControls = doc.createElement("curveControls");

    if (cFitControls.contains("RangeStart")) elCfRange.setAttribute("RangeStart", QString("%1").arg(cFitControls.value("RangeStart").toDouble()));
    if (cFitControls.contains("RangeEnd"))   elCfRange.setAttribute("RangeEnd",   QString("%1").arg(cFitControls.value("RangeEnd").toDouble()));

    if (cFitControls.contains("StepSizeValue"))  elCfStepSize.setAttribute("StepSizeValue",  QString("%1").arg(cFitControls.value("StepSizeValue").toDouble()));
    if (cFitControls.contains("StepSizeCustom")) elCfStepSize.setAttribute("StepSizeCustom", QString("%1").arg(cFitControls.value("StepSizeCustom").toBool() ? "1" : "0"));

    if (cFitControls.contains("MaxInt"))   elCfControls.setAttribute("MaxInt",   QString("%1").arg(cFitControls.value("MaxInt").toInt()));
    if (cFitControls.contains("EpsX"))     elCfControls.setAttribute("EpsX",     QString("%1").arg(cFitControls.value("EpsX").toDouble()));
    if (cFitControls.contains("DiffStep")) elCfControls.setAttribute("DiffStep", QString("%1").arg(cFitControls.value("DiffStep").toDouble()));

    if (cFitControls.size()) {
        cFitEl.appendChild(elCfRange);
        cFitEl.appendChild(elCfStepSize);
        cFitEl.appendChild(elCfControls);
    }

    QDomElement elAxExWorkbook = doc.createElement("excelWorkbook");

    if (axExcelControls.size()) {
        elAxExWorkbook.setAttribute("File", axExcelControls.first().file);

        for (int i = 0; i < axExcelControls.size(); ++i) {
            QDomElement elAxExCell = doc.createElement("excelCell");
            elAxExCell.setAttribute("Worksheet", axExcelControls.at(i).worksheet);
            elAxExCell.setAttribute("Row",       axExcelControls.at(i).row);
            elAxExCell.setAttribute("Column",    axExcelControls.at(i).col);
            elAxExCell.setAttribute("Phase",     axExcelControls.at(i).phase);
            elAxExCell.setAttribute("Parameter", axExcelControls.at(i).parameter);
            elAxExCell.setAttribute("Filter",    axExcelControls.at(i).filter);
            elAxExCell.setAttribute("Output",    axExcelControls.at(i).output);
            elAxExWorkbook.appendChild(elAxExCell);
        }

        addExcel = true;
    }

    if (addStrucs)     rootEl.appendChild(strucsEl);
    if (addOther)      rootEl.appendChild(otherEl);

    if (addBaseLine) {
        for (int i = 0; i < baseLineList.size(); ++i) {
            rootEl.appendChild(baseLineList.at(i));
        }
    }

    if (addIntRanges)  rootEl.appendChild(prangeEl);
    if (addCurveFits)  rootEl.appendChild(cFitEl);
    if (addExcel)      rootEl.appendChild(elAxExWorkbook);

    return doc;
}

bool BgmnPresetHandler::writePresetFile(const QDomDocument &doc, const QFileInfo &fiPresetFile)
{
    QString presOutFile(fiPresetFile.absoluteFilePath() + (fiPresetFile.suffix().isEmpty() ? ".pfp" : ""));
    return BgmnFileIO::writeTextFile(presOutFile, doc.toString());
}

/********************************************************************************************
 *                                    Restoring a preset                                    *
 ********************************************************************************************/

void BgmnPresetHandler::parsePresetFile(const QString &s)
{
    presetFileInfo = QFileInfo(s);
    QDomDocument doc("preset");
    doc.setContent(BgmnFileIO::readTextFile(s));

    for (int i = 0; i < allTags.size(); ++i) {
        QDomNodeList nodes = doc.elementsByTagName(allTags.at(i));

        for (int j = 0; j < nodes.size(); ++j) {
            if (nodes.at(j).isElement()) {
                elements.append(nodes.at(j).toElement());
            }
        }
    }
}

bool BgmnPresetHandler::hasBgmnProject()
{
    for (int i = 0; i < elements.size(); ++i) {
        if (elements.at(i).tagName().toLower() == "device") return true;
    }

    return false;
}

bool BgmnPresetHandler::hasBaseLine()
{
    for (int i = 0; i < elements.size(); ++i) {
        if (elements.at(i).tagName().toLower() == "baseline") return true;
    }

    return false;
}

bool BgmnPresetHandler::hasPeakIntegrals()
{
    for (int i = 0; i < elements.size(); ++i) {
        if (elements.at(i).tagName().toLower() == "peakintegral") return true;
    }

    return false;
}

bool BgmnPresetHandler::hasCurveFit()
{
    for (int i = 0; i < elements.size(); ++i) {
        if (elements.at(i).tagName().toLower() == "curvefunction") return true;
    }

    return false;
}

bool BgmnPresetHandler::hasAxExcelConfig()
{
    for (int i = 0; i < elements.size(); ++i) {
        if (elements.at(i).tagName().toLower() == "excelworkbook") return true;
    }

    return false;
}

QString BgmnPresetHandler::createProject(const QString &cfile, const QString &id, int nth)
{
    QFileInfo fiControlFile(cfile);

    QList<BgmnPresetFile> files(domElementsToFiles(elements));
    files = toAbsolutePath(files, presetFileInfo.absolutePath());

    if (!fileChecks(files, fiControlFile.absolutePath())) {
        return QString();
    }

    QList<FileType> exclude;
    exclude << TEMPLATE;
    copyFilesToDestination(files, fiControlFile.absolutePath(), exclude);

    QList<BgmnPresetFile> templateFiles(getFilesOfType(files, TEMPLATE));
    QList<BgmnPresetFile> deviceBinFiles(getFilesOfType(files, DEVICE_BINARY));

    if (templateFiles.isEmpty() || deviceBinFiles.isEmpty()) {
        qDebug() << QString("BgmnPresetHandler::createProject: No template or device file found. Exiting.");
        return QString();
    }

    BgmnSavParser sparser(BgmnFileIO::readTextFile(templateFiles.first().fileInfo.absoluteFilePath()), cfile);
    sparser.setDeviceFile(toDestinationFilePath(deviceBinFiles.first(), QString()));
    sparser.setNumberOfThreads(nth);
    sparser.setSampleId(id);

    return sparser.getContent();
}

QList<PeakRange> BgmnPresetHandler::getPeakIntegrals()
{
    return domElementsToPeakIntegrals(elements);
}

bool BgmnPresetHandler::getCurveFits(QList<CurveFitCurve> &curves, QList<CurveFitVariable> &variables, QMap<QString, QVariant> &controls)
{
    curves = domElementsToCurveFitCurves(elements);
    variables = domElementsToCurveFitVariables(elements);
    controls = domElementsToCurveFitControls(elements);

    return (curves.size() && variables.size()) ? true : false;
}

QList<AxObjectExcel> BgmnPresetHandler::getAxExcelConfig()
{
    return domElementsToAxExcel(elements);
}

QList<BgmnPresetFile> BgmnPresetHandler::getFilesOfType(const QList<BgmnPresetFile> &files, FileType t)
{
    QList<BgmnPresetFile> tFiles;

    for (int i = 0; i < files.size(); ++i) {
        if (files.at(i).type == t) {
            tFiles.append(files.at(i));
        }
    }

    return tFiles;
}

QMap<QString, QVariant> BgmnPresetHandler::baseLineParameters()
{
    return domElementsToBaseLineParams(elements);
}

/*
 * returns -1 if no base line is defined
 * returns 0 for SNIP baseline
 * returns 1 for Golotvin baseline
 */
int BgmnPresetHandler::domElementsToBaseLineMode(const QList<QDomElement> &l)
{
    for (int i = 0; i < l.size(); ++i) {
        if (l.at(i).tagName().toLower() == "baseline") {
            if (l.at(i).attribute("Mode", QString()) == "SNIP") return 0;
            if (l.at(i).attribute("Mode", QString()) == "GOLOTVIN") return 1;
        }
    }

    return -1;
}

QMap<QString, QVariant> BgmnPresetHandler::domElementsToBaseLineParams(const QList<QDomElement> &l)
{
    QMap<QString, QVariant> map;

    for (int i = 0; i < l.size(); ++i) {
        if (l.at(i).tagName().toLower() == "baseline") {
            QDomNamedNodeMap nodes = l.at(i).attributes();
            for (int j = 0; j < nodes.size(); ++j) {
                map[nodes.item(j).toAttr().name()] = QVariant(nodes.item(j).toAttr().value());
            }
            break;
        }
    }

    return map;
}

QList<BgmnPresetFile> BgmnPresetHandler::domElementsToFiles(const QList<QDomElement> &l)
{
    QList<BgmnPresetFile> files;

    for (int i = 0; i < l.size(); ++i) {
        if (!fileTags.contains(l.at(i).tagName())) continue;

        QFileInfo fi(l.at(i).text());

        BgmnPresetFile file;
        file.fileInfo = fi;
        file.md5Hash = l.at(i).attribute("MD5hash", QString());
        file.relativePath = l.at(i).attribute("RelativePath", QString());

        if ((l.at(i).tagName().toLower() == "device") && (file.fileInfo.suffix().toLower() == "ger")) file.type = DEVICE_ASCII;
        if ((l.at(i).tagName().toLower() == "device") && (file.fileInfo.suffix().toLower() == "sav")) file.type = DEVICE_ASCII;
        if ((l.at(i).tagName().toLower() == "device") && (file.fileInfo.suffix().toLower() == "geq")) file.type = DEVICE_BINARY;
        if (l.at(i).tagName().toLower() == "template")   file.type = TEMPLATE;
        if (l.at(i).tagName().toLower() == "structure")  file.type = STRUCTURE;
        if (l.at(i).tagName().toLower() == "textfile")   file.type = TEXT;
        if (l.at(i).tagName().toLower() == "binaryfile") file.type = BINARY;

        files.append(file);
    }

    return files;
}

QList<PeakRange> BgmnPresetHandler::domElementsToPeakIntegrals(const QList<QDomElement> &l)
{
    QList<PeakRange> ranges;

    for (int i = 0; i < l.size(); ++i) {
        if (l.at(i).tagName().toLower() == "peakintegral") {
            PeakRange pr;
            pr.name       = l.at(i).attribute("Name", QString());
            pr.start      = l.at(i).attribute("Start", "0.0").toDouble();
            pr.end        = l.at(i).attribute("End", "0.0").toDouble();
            pr.background = l.at(i).attribute("Background", "0").toInt();
            ranges.append(pr);
        }
    }

    return ranges;
}

QList<CurveFitCurve> BgmnPresetHandler::domElementsToCurveFitCurves(const QList<QDomElement> &l)
{
    QList<CurveFitCurve> curves;

    for (int i = 0; i < l.size(); ++i) {
        if (l.at(i).tagName().toLower() == "curvefunction") {
            CurveFitCurve cv;
            cv.type           = l.at(i).attribute("Type", 0).toInt();

            QDomNodeList pnln = l.at(i).elementsByTagName("ParameterNames");
            if (pnln.size()) cv.parameterNames = pnln.at(0).toElement().text().split(";");

            QDomNodeList pnlv = l.at(i).elementsByTagName("Variables");
            if (pnlv.size()) cv.variables = pnlv.at(0).toElement().text().split(";");

            curves.append(cv);
        }
    }

    return curves;
}

QList<CurveFitVariable> BgmnPresetHandler::domElementsToCurveFitVariables(const QList<QDomElement> &l)
{
    QList<CurveFitVariable> variables;

    for (int i = 0; i < l.size(); ++i) {
        if (l.at(i).tagName().toLower() == "curvevariable") {
            CurveFitVariable cv;
            cv.name       = l.at(i).attribute("Name", "$-");
            cv.value      = l.at(i).attribute("Value", "0.00000");
            cv.lowerLimit = l.at(i).attribute("LowerLimit", "-inf");
            cv.upperLimit = l.at(i).attribute("UpperLimit", "+inf");
            cv.checkState = l.at(i).attribute("CheckState", "1");
            variables.append(cv);
        }
    }

    return variables;
}

QMap<QString, QVariant> BgmnPresetHandler::domElementsToCurveFitControls(const QList<QDomElement> &l)
{
    QMap<QString, QVariant> controls;

    for (int i = 0; i < l.size(); ++i) {
        if (l.at(i).tagName().toLower() == "curvecontrols") {
            if (l.at(i).hasAttribute("MaxInt"))   controls["MaxInt"]   = l.at(i).attribute("MaxInt", "0").toInt();
            if (l.at(i).hasAttribute("EpsX"))     controls["EpsX"]     = l.at(i).attribute("EpsX", "0.0").toDouble();
            if (l.at(i).hasAttribute("DiffStep")) controls["DiffStep"] = l.at(i).attribute("DiffStep", "0.0001").toDouble();
        }

        if (l.at(i).tagName().toLower() == "curverange") {
            if (l.at(i).hasAttribute("RangeStart")) controls["RangeStart"] = l.at(i).attribute("RangeStart", "0.0").toDouble();
            if (l.at(i).hasAttribute("RangeEnd"))   controls["RangeEnd"]   = l.at(i).attribute("RangeEnd", "0.0").toDouble();
        }

        if (l.at(i).tagName().toLower() == "curvestepsize") {
            if (l.at(i).hasAttribute("StepSizeValue"))  controls["StepSizeValue"]  = l.at(i).attribute("StepSizeValue", "0.001").toDouble();
            if (l.at(i).hasAttribute("StepSizeCustom")) controls["StepSizeCustom"] = l.at(i).attribute("StepSizeCustom", "0") == "1" ? true : false;
        }
    }

    return controls;
}

QList<global::AxObjectExcel> BgmnPresetHandler::domElementsToAxExcel(const QList<QDomElement> &l)
{
    QList<global::AxObjectExcel> out;

    for (int i = 0; i < l.size(); ++i) {
        if (l.at(i).tagName() == "excelWorkbook") {
            QDomNodeList exCl = l.at(i).elementsByTagName("excelCell");

            for (int j = 0; j < exCl.size(); ++j) {
                QDomElement elCl = exCl.at(j).toElement();                
                global::AxObjectExcel aoEx;

                aoEx.file      = l.at(i).attribute("File", QString());
                aoEx.worksheet = elCl.attribute("Worksheet", "1").toInt();
                aoEx.row       = elCl.attribute("Row", "0").toInt();
                aoEx.col       = elCl.attribute("Column", "0").toInt();
                aoEx.phase     = elCl.attribute("Phase", QString());
                aoEx.parameter = elCl.attribute("Parameter", QString());
                aoEx.filter    = elCl.attribute("Filter", QString());
                aoEx.output    = elCl.attribute("Output", QString());

                out.append(aoEx);
            }
        }
    }

    return out;
}

QList<BgmnPresetFile> BgmnPresetHandler::toAbsolutePath(const QList<BgmnPresetFile> &files, const QString &src)
{
    QList<BgmnPresetFile> absFiles;

    for (int i = 0; i < files.size(); ++i) {
        qDebug() << QString("BgmnPresetHandler::absoluteSourcePath: Current file path is %1").arg(files.at(i).fileInfo.filePath());

        if (files.at(i).fileInfo.isAbsolute()) {
            qDebug() << QString("    Path is absolute, nothing to change");
            absFiles.append(files.at(i));
        } else {
            BgmnPresetFile file = files.at(i);

            if ((file.fileInfo.path() != ".") && (file.relativePath.isEmpty())) {
                // if the file is stored in a sub dir and relative path is not specified,
                // we must use the subdir as relative path for compatibility with previous
                // versions
                file.relativePath = file.fileInfo.path();
                qDebug() << QString("    Setting relative path to %1").arg(file.fileInfo.path());
            }

            file.fileInfo = QFileInfo(QDir(src), file.fileInfo.filePath());
            qDebug() << QString("    File path changed to %1").arg(file.fileInfo.filePath());
            absFiles.append(file);
        }
    }

    return absFiles;
}

/*
 * true: test passed, continue
 * false: test failed or aborted, abort
 */
bool BgmnPresetHandler::fileChecks(const QList<BgmnPresetFile> &files, const QString &destDir)
{
    Q_UNUSED(destDir);

    QStringList warnMissingFile(checkExisting(files));

    if (!warnMissingFile.isEmpty()) {
        showFileMissingWarning(warnMissingFile);
        return false;
    }

    QStringList warnChecksum(checkMD5sum(files));

    if (!warnChecksum.isEmpty()) {
        if (showChecksumWarning(warnChecksum)) {
            return false;
        }
    }

    return true;
}

QStringList BgmnPresetHandler::checkExisting(const QList<BgmnPresetFile> &files)
{
    QStringList missing;

    for (int i = 0; i < files.size(); ++i) {
        BgmnPresetFile file = files.at(i);

        if (file.fileInfo.exists()) {
            qDebug() << QString("BgmnPresetHandler::checkExisting: File exists:  %1").arg(file.fileInfo.absoluteFilePath());
        } else {
            missing.append(file.fileInfo.absoluteFilePath());
            qDebug() << QString("BgmnPresetHandler::checkExisting: File missing: %1").arg(file.fileInfo.absoluteFilePath());
        }
    }

    return missing;
}

QStringList BgmnPresetHandler::checkMD5sum(const QList<BgmnPresetFile> &l)
{
    QStringList md5failed;

    for (int i = 0; i < l.size(); ++i) {
        BgmnPresetFile file = l.at(i);

        QByteArray savedHash;
        savedHash.append(file.md5Hash);

        QByteArray currentHash = BgmnFileIO::checksum(file.fileInfo.absoluteFilePath(), QCryptographicHash::Md5);

        if (currentHash != QByteArray::fromHex(savedHash)) {
            md5failed.append(file.fileInfo.absoluteFilePath());
            qDebug() << QString("BgmnPresetHandler::checkMD5sum: Checksum mismatch in file %1").arg(file.fileInfo.absoluteFilePath());
        } else {
            qDebug() << QString("BgmnPresetHandler::checkMD5sum: Checksum matches of file %1").arg(file.fileInfo.absoluteFilePath());
        }
    }

    return md5failed;
}

QStringList BgmnPresetHandler::checkOverwrite(const QList<BgmnPresetFile> &l, const QString &dest)
{
    QStringList owWarning;

    for (int i = 0; i < l.size(); ++i) {
        QString destFile;
        BgmnPresetFile file = l.at(i);

        destFile = toDestinationFilePath(file, dest);

        QFileInfo fiDestFile(destFile);
        if (fiDestFile.exists()) {
            owWarning.append(fiDestFile.absoluteFilePath());
            qDebug() << QString("BgmnPresetHandler::checkOverwrite: Destination file already exists %1").arg(fiDestFile.absoluteFilePath());
        } else {
            qDebug() << QString("BgmnPresetHandler::checkOverwrite: Destination file does not exist %1").arg(fiDestFile.absoluteFilePath());
        }
    }

    return owWarning;
}

void BgmnPresetHandler::showFileMissingWarning(const QStringList &l)
{
    QMessageBox::warning(0, tr("Problem applying preset"),
                         QString(tr("The preset cannot be applied because "
                                    "the following files are missing:\n\n%1")).arg(l.join("\n")));
}

/*
 * true = abort
 * false = continue
 */
bool BgmnPresetHandler::showChecksumWarning(const QStringList &l)
{
    QString warnText = QString(tr("The following files were modified:\n\n%1\n\n"
                                 "Do you want to continue anyway?")).arg(l.join("\n"));

    return QMessageBox::warning(0, tr("Checksum warning"),
                         warnText, QMessageBox::Yes, QMessageBox::No) == QMessageBox::Yes ? false : true;
}

/*
 * true = abort
 * false = continue
 */
bool BgmnPresetHandler::showOverwriteWarning(const QStringList &l)
{
    QString warnText = QString(tr("The following files will be overwritten:\n\n%1\n\n"
                                 "Do you want to continue anyway?")).arg(l.join("\n"));

    return QMessageBox::warning(0, tr("Files exist"),
                         warnText, QMessageBox::Yes, QMessageBox::No) == QMessageBox::Yes ? false : true;
}

void BgmnPresetHandler::copyFilesToDestination(const QList<BgmnPresetFile> &files, const QString &dest, const QList<FileType> &exclude)
{
    for (int i = 0; i < files.size(); ++i) {
        if (exclude.contains(files.at(i).type)) {
            continue;
        }

        QString src(files.at(i).fileInfo.absoluteFilePath());
        QString dst(toDestinationFilePath(files.at(i), dest));

        bool b = false;

        if ((files.at(i).type == DEVICE_BINARY) || (files.at(i).type == BINARY)) {
            b = QFile::copy(src, dst);
        } else {
            b = BgmnFileIO::copyFile(src, dst);
        }

        if (b) qDebug() << QString("BgmnPresetHandler::copyFilesToDestination: File copied from %1 to %2").arg(src).arg(dst);
        else   qDebug() << QString("BgmnPresetHandler::copyFilesToDestination: Error copying file from %1 to %2").arg(src).arg(dst);
    }
}

QString BgmnPresetHandler::toDestinationFilePath(const BgmnPresetFile &f, const QString &destDir)
{
    if (f.relativePath.isEmpty()) {
        if (destDir.isEmpty()) return f.fileInfo.fileName();
        return QDir::toNativeSeparators(destDir + "/" + f.fileInfo.fileName());
    }

    if (destDir.isEmpty()) return QDir::toNativeSeparators(f.relativePath + "/" + f.fileInfo.fileName());
    return QDir::toNativeSeparators(destDir + "/" + f.relativePath + "/" + f.fileInfo.fileName());
}
