/***************************************************************************
                          bgmnprojectwidget.cpp  -  description
                             -------------------
    begin                : Mon Jan 30 11:00:00 CEST 2012
    copyright            : (C) 2012 by Nicola Doebelin
    email                : ndoebelin@gmx.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "bgmnprojectwidget.h"
#include "syntaxHighlighter/bgmnhighlighter.h"
#include "gentemplatedialog.h"
#include "internalstandarddialog.h"
#include "../quazip/JlCompress.h"

#include "../libXrdIO/parser/bgmnlstparser.h"
#include "../libXrdIO/parser/bgmnsavparser.h"
#include "../libXrdIO/parser/bgmnstrparser.h"
#include "../libXrdIO/parser/bgmnresparser.h"
#include "../libXrdIO/parser/bgmnlamparser.h"
#include "../libXrdIO/import/importhandler.h"
#include "../libXrdIO/export/exporthandler.h"
#include "../libXrdIO/bgmnfileio.h"
#include "../libXrdIO/crystal/crystalstructure.h"
#include "../libXrdIO/scan.h"
#include "../libXrdIO/scanops.h"
#include "../libXrdIO/structs.h"
#include "../libXrdIO/parser/bgmnhtmlreportgenerator.h"
#include "bgmnbackendconfig.h"
#include "bgmninstrumentselectdialog.h"
#include "presetcontentsavedialog.h"

#include <QFileDialog>
#include <QMessageBox>
#include <QApplication>
#include <QProgressDialog>
#include <QDomDocument>
#include <QDomNode>
#include <QDebug>
#include <QDirIterator>
#include <QPair>
#include <QFont>
#include <vector>
#include <math.h>

BgmnProjectWidget::BgmnProjectWidget(QWidget *parent) :
        ProjectWidget(parent)
{
    // init the BGMN process handler
    bgmnHandler = new BgmnHandler(QUuid::createUuid(), this);
    eflechHandler = new EflechHandler(QUuid::createUuid(), this);
    diaHandler = new BgmnDiaHandler(this);
    controlFileExtension = QStringLiteral("sav");

    // these files will be loaded automatically
    textFormats << QStringLiteral("sav") << QStringLiteral("SAV") << QStringLiteral("lst") << QStringLiteral("LST");

    ImportHandler iHandler;
    graphFormats = iHandler.extensions();

    // make sure the dia file is first in the list
    // because it's our preferred file format to auto-load
    graphFormats.prepend("dia");

    // file formats directly supported by the Rietveld backend
    nativeGraphFormats << "xy";

#ifdef Q_OS_WIN
    nativeGraphFormats << "val";
#endif

    // create the add/remove phase dialog
    arPhaseDialog = nullptr;

    // create strfilebatchedit dialog
    strFileEditDlg = nullptr;

    // default local parameters, if no custom ones are available
    defaultBgmnLocalParameters = global::defaultBgmnLocalGoals.split(QRegularExpression("\\n"));

    projectTypeItem->setText(1, "BGMN");
    refStrManager = BgmnRefStructureManager::getInstance();

    excelExporter = new MsoExportExcel(graphControl, this);

    QStringList resultsHeaders;
    resultsHeaders << tr("Parameter") << tr("Value") << tr("ESD");

    resultsTree = new BgmnResultsTreeWidget(this);
    resultsTree->setColumnCount(resultsHeaders.count());
    resultsTree->setHeaderLabels(resultsHeaders);
    resultsTree->setLstParser(&lparser);

    connect(bgmnHandler, SIGNAL(pollOutput()), this, SLOT(bgmnPollOutput()));
    connect(bgmnHandler, SIGNAL(allComplete(QUuid)), this, SLOT(bgmnComplete(QUuid)));
    connect(bgmnHandler, SIGNAL(aborted()), this, SLOT(processAborted()));
    connect(eflechHandler, SIGNAL(aborted()), this, SLOT(processAborted()));
    connect(this, SIGNAL(currentChanged(int)), this, SLOT(tabChanged(int)));
    connect(searchMatchWidget, SIGNAL(refStructureSelected(QString)), this, SLOT(setReferenceStructure(QString)));
    connect(searchMatchWidget, SIGNAL(refStructureReset()), this, SLOT(resetRefStructure()));
    connect(searchMatchWidget, SIGNAL(exitStatus(int)), this, SLOT(searchMatchStatus(int)));

    initSettings();
}

BgmnProjectWidget::~BgmnProjectWidget()
{
    settings->setValue("bgmnProject/resultsTree/header", resultsTree->header()->saveState());

    if (strFileEditDlg) delete strFileEditDlg;
    if (arPhaseDialog)  delete arPhaseDialog;
    delete excelExporter;
    delete bgmnHandler;
    delete eflechHandler;
    delete diaHandler;
}

/*
 * read global and BGMN specific settings
 */
void BgmnProjectWidget::initSettings()
{
    initGlobalSettings();

    resultsTree->header()->restoreState(settings->value("bgmnProject/resultsTree/header", QByteArray()).toByteArray());
    setReferenceStructureFileList();

    QMapIterator<QString, ControlFileEdit*> it(editors);

    while (it.hasNext()) {
        it.next();
        ControlFileEdit *ed = it.value();
        bool oldState = ed->blockSignals(true);
        setHighlighting(ed);
        ed->blockSignals(oldState);
    }
}

/*
 * ***********************************************************************
 * Public process control API
 * ***********************************************************************
 */
void BgmnProjectWidget::runRefinement()
{
    bgmnRunProcess();
}

void BgmnProjectWidget::runPeakDetection()
{
    teilRunProcess();
}

bool BgmnProjectWidget::isRunning()
{
    if (bgmnHandler->isRunning()) return true;
    if (eflechHandler->isRunning()) return true;
    if (searchMatchWidget->isRunning()) return true;
    return false;
}

void BgmnProjectWidget::abort()
{
    if (bgmnHandler->isRunning()) bgmnAbortProcess();
    if (eflechHandler->isRunning()) eflechAbortProcess();
    if (searchMatchWidget->isRunning()) searchMatchWidget->abort();
}

/*
 * ***********************************************************************
 * BGMN process communication
 * ***********************************************************************
 */
void BgmnProjectWidget::bgmnRunProcess()
{
    saveAll();
    controlFile = getControlFileName();

    if (!fileEditor(controlFile, false)) {
        QMessageBox::information(this, tr("Create Control File"), tr("Please create a control file using \"Add Phase (+)\" first."));
        return;
    }

    BgmnSavParser sparser = getSavParser();

    // check if the VAL[n] file exists. If not, try to create it from the loaded scan file.
    // if this fails, return, because BGMN can't run without a VAL file
    int numValFiles = convertRawDataFile();

    if (numValFiles < 1) {
        // if conversion reported an error, we will abort
        if (numValFiles < 0) {
            // if it reported an error code of -1, we will also display a message before aborting
            QMessageBox::warning(this,
                                 tr("Problem creating XY file"),
                                 tr("Profex could not convert the scan file to a format supported by BGMN.\n"
                                    "Please convert it and run the refinement with the converted file."));
        }

        return;
    }

    if (settings->value("bgmnProject/readLamFile", false).toBool()) {
        setCustomWavelengt(sparser.lamFile());
    }

    QString diaFile(sparser.diagramFile());

    if (diaFile.isEmpty()) {
        qDebug() << QString("BgmnProjectWidget::runRefinement(): No graph file specified. Aborting.");
        return;
    }

    refinedGraphFile = getAbsoluteFilePath(diaFile);
    QFileInfo fi(refinedGraphFile);
    QString tempGraphFile = QString("%1%2%3").arg(settings->getTempLocation()).arg(QDir::separator()).arg(fi.fileName());

    QFile parFile(getAbsoluteFilePath(sparser.outputFile()));
    QFile lstFile(getAbsoluteFilePath(sparser.listFile()));

    parFile.remove();
    lstFile.remove();
    deleteReport();
    clearGui();

    const Scan *cScan = graphControl->firstActiveScan();

    diaHandler->init(graphControl, tempGraphFile, cScan ? cScan->pDataAngle().count() : 0);

    setRexpDenom(sparser);

    // initialize the process handler
    if (!bgmnHandler->init()) {
        qDebug() << QString("BgmnProjectWidget::runRefinement(): BGMN executable not found: %1").arg(refinerExec);
        QMessageBox::information(this, tr("BGMN executable not found"), tr("The BGMN executable was not found.\n\nPlease check your BGMN configuration in the preferences."));
        return;
    }

    refOutput->appendPlainText(QString("Running process: %1 %2")
                               .arg(QDir::toNativeSeparators(settings->value("bgmnProject/bgmnExec", "").toString()))
                               .arg(QDir::toNativeSeparators(controlFile)));
    processTimer.start();

    QStringList args;
    args << controlFile;
    args << QString("DIAGRAMM=%1").arg(tempGraphFile);
    args << "SAVE=N";

    if (isScheduledForBatch()) {
        int npar = settings->value("config/batchParallelJobs", 1).toInt();

        if (npar > 1) {
            int nthr = sparser.numberOfThreads();
            if (nthr < 0) nthr = settings->value("bgmnProject/nThreads", 0).toInt();
            if (nthr == 0) nthr = QThread::idealThreadCount();
            nthr /= npar;
            args << QString("NTHREADS=%1").arg(nthr == 0 ? 1 : nthr);
        }
    }

    // start the process
    if (bgmnHandler->run(projectDir, args)) {
        graphControl->setViewStatus(1);
        setStatus(RUNNING);
        setCurrentIndex(0); // switch to the graph page
    } else {
        setStatus(FAILURE);
        graphControl->setViewStatus(0);
        QMessageBox::information(this, tr("Error starting BGMN"),
            QString(tr("Refinement could not be started.\nThe following command was aborted:\n\n%1 %2")).arg(refinerExec).arg(controlFile));
    }
}

void BgmnProjectWidget::bgmnAbortProcess()
{
    bgmnHandler->abort();
    setStatus(ABORTED);
}

void BgmnProjectWidget::bgmnPollOutput()
{
    bool reload = true;
    QString output(bgmnHandler->readOutput().trimmed());
    if (protocolParser.parse(output)) {
        convDisplay->setValues(protocolParser.getRwp(), protocolParser.getRexp(), protocolParser.getFirst());
        resultsTree->setRvalues(protocolParser.getRwp(), protocolParser.getRexp());
}

    if (output == ".") {
        refOutput->moveCursor(QTextCursor::End);
        refOutput->insertPlainText(output);
        refOutput->moveCursor(QTextCursor::End);
        reload = false;
    } else {
        refOutput->appendPlainText(output + " ");
    }

    refOutput->ensureCursorVisible();

    if (reload) diaHandler->reload();

    if (!currentIndex()) {
        emit currentTabChanged(static_cast<ProjectWidget *>(this), currentIndex());
    }
}

void BgmnProjectWidget::bgmnComplete(QUuid)
{
    bgmnPollOutput();
    diaHandler->stop();
    bool error = false;

    if (refOutput->toPlainText().contains("(E) ")) {
        setStatus(FAILURE);
        error = true;
    } else {
        setStatus(COMPLETED);
        error = false;
    }

    if (!lparser.reload()) {
        parseProjectFiles();
    }

    BgmnFileIO::copyFile(diaHandler->monitoredFile(), refinedGraphFile);

    if (QFile::remove(diaHandler->monitoredFile())) {
        qDebug() << QString("BgmnProjectWidget::bgmnComplete(): Temporary file deleted: %1").arg(diaHandler->monitoredFile());
    } else {
        qDebug() << QString("BgmnProjectWidget::bgmnComplete(): Could not delete temporary file: %1").arg(diaHandler->monitoredFile());
    }

    graphControl->setAngularCorrections(lparser.getEpsN(1), lparser.getEpsN(2), lparser.getEpsN(3));
    graphControl->setViewStatus(2);
    load(QString(), refinedGraphFile, "BGMN_DIA");

    if (!currentIndex()) {
        // we have to fire this signal, else the file name in the status bar will not be updated
        emit currentTabChanged(static_cast<ProjectWidget *>(this), currentIndex());
    }

    updateResults();

    if (!error && settings->value("bgmnProject/createReport", false).toBool()) {
        writeReport();
    }

    QString timeString;

    if (error) {
        timeString = QString(tr("Refinement aborted with errors"));
    } else {
        timeString = QString(tr("Refinement completed in %1 hh:mm:ss.ms").arg(elapsedTimeFormatted(processTimer.elapsed())));
    }

    refOutput->appendPlainText(timeString);

    emit completed(this);
    qDebug() << QString("BgmnProjectWidget::complete(): %1").arg(timeString);
}

/*
 * ***********************************************************************
 * SearchMatchWidget process communication
 * ***********************************************************************
 */

void BgmnProjectWidget::runSearchMatch()
{
    setupSearchMatchWidget(false);
    searchMatchWidget->run();
}

void BgmnProjectWidget::setupSearchMatchWidget(bool instr)
{
    BgmnSavParser sparser = getSavParser();
    QStringList fn = sparser.getPhases();
    QList<QPair<QString, QString> > phases;

    for (int i = 0; i < fn.size(); ++i) {
        QString fname = projectDir + "/" + fn.at(i);

        BgmnStrParser strParser(fname);

        QPair<QString, QString> data;
        data.first = fname;
        data.second = strParser.getPhaseName();
        phases.append(data);
    }

    searchMatchWidget->setPinnedPhases(phases);

    if (instr) {
        searchMatchWidget->setInstrumentFile(sparser.deviceFile());
        searchMatchWidget->setEps2(refHdispValue->value());

        if (sparser.lamFile().isEmpty()) {

        } else {
            searchMatchWidget->setLambdaFile(sparser.lamFile());
        }
    }
}

void BgmnProjectWidget::searchMatchStatus(int i)
{
    if (i == 0) setStatus(ABORTED);
    if (i == 1) setStatus(MATCHING);
    if (i == 2) setStatus(COMPLETED);
}

/*
 * ***********************************************************************
 * TEIL process communication
 * ***********************************************************************
 */
void BgmnProjectWidget::teilRunProcess()
{
    refOutput->clear();

    if (!graphModel->count()) {
        qDebug() << QString("BgmnProjectWidget::teilRunProcess(): No scan loaded. Exiting.");
        return;
    }

    QString wmin;
    QString wmax;

    bool ok;
    BgmnSavParser sparser(getSavParser(&ok));

    if (ok) {
        double wmi = sparser.getWmin();
        double wma = sparser.getWmax();

        if (wmi > 0.0) wmin = QString("WMIN=%1").arg(wmi, 0, 'f', 4);
        if (wma > 0.0) wmax = QString("WMAX=%1").arg(wma, 0, 'f', 4);
    }

    QStringList errors;
    QString euid = eflechHandler->getStrId();

    QString sdev;
    QString lam;

    if (!eflechSourceDevFile(sdev, lam)) {
        qDebug() << QString("BgmnProjectWidget::teilRunProcess(): User aborted. Exiting.");
        return;
    }

    QFileInfo devFi(sdev);
    QStringList devFilter;
    devFilter << QString("%1.geq").arg(devFi.completeBaseName());
    devFilter << QString("%1.GEQ").arg(devFi.completeBaseName());

    QList<QFileInfo> tempFiles = devFi.absoluteDir().entryInfoList(devFilter);

    // from here on we need to remove the temporary files when aborting the process

    if (!eflechHandler->gatherTemporaryFiles(tempFiles, projectDir, euid)) {
        errors << QString("Copying files to temporary location %1 failed.").arg(projectDir);
    }

    if (!eflechActiveScanToTemp(euid)) {
        errors << QString("Writing active scan to temporary location %1 failed.").arg(projectDir);
    }

    const Scan *ascan = graphControl->firstActiveScan();

    if (ascan) {
        if (!eflechHandler->init(ascan->name(), ascan->minAngle(), ascan->maxAngle())) {
            errors << QString("Error initializing eflechHandler");
        }
    } else {
        errors << QString("Could not get pointer to active scan.");
    }

    // bail out here if errors occured above
    if (errors.size()) {
        eflechHandler->removeTemporaryFiles(projectDir, euid);

        for (int i = 0; i < errors.size(); ++i) {
            qDebug() << QString("BgmnProjectWidget::teilRunProcess(): %1").arg(errors.at(i));
        }

        return;
    }

    QStringList args;
    args << euid;
    args << QString("VERZERR=%1.geq").arg(euid);
    args << QString("OUTPUTMASK=%1-$").arg(euid);
    args << QString("TITELMASK=%1-part-$").arg(euid);
    args << QString("TEST=%1").arg(eflechTestString());
    args << QString("VAL[1]=%1.xy").arg(euid);
    args << lam;
    args << QString("NTHREADS=%1").arg(QThread::idealThreadCount());
    args << QString("SAVE=Y");
    if (!wmin.isEmpty()) args << wmin;
    if (!wmax.isEmpty()) args << wmax;

    if (settings->value("eflech/overrideEpsilon", false).toBool()) {
        args << QString("EPSILON=%1").arg(settings->value("eflech/epsilon", 0.05).toDouble(), 0, 'f', 4);
    }

    disconnect(eflechHandler, SIGNAL(allComplete(QUuid)), nullptr, nullptr);
    disconnect(eflechHandler, SIGNAL(pollOutput()), nullptr, nullptr);
    connect(eflechHandler, SIGNAL(allComplete(QUuid)), this, SLOT(teilComplete()));
    connect(eflechHandler, SIGNAL(pollOutput()), this, SLOT(teilPollOutput()));

    graphControl->setViewStatus(1);
    setStatus(RUNNING);

    processTimer.start();
    eflechHandler->run(0, projectDir, args);
}

void BgmnProjectWidget::teilAbortProcess()
{
    eflechHandler->abort();
    graphControl->setViewStatus(2);
    setStatus(ABORTED);
}

void BgmnProjectWidget::teilPollOutput()
{
    refOutput->appendPlainText(eflechHandler->readOutput().trimmed());
}

void BgmnProjectWidget::teilComplete()
{
    disconnect(eflechHandler, SIGNAL(allComplete(QUuid)), nullptr, nullptr);
    disconnect(eflechHandler, SIGNAL(pollOutput()), nullptr, nullptr);
    connect(eflechHandler, SIGNAL(allComplete(QUuid)), this, SLOT(eflechComplete()));
    connect(eflechHandler, SIGNAL(pollOutput()), this, SLOT(eflechPollOutput()));
    eflechRunProcess();
}

/*
 * ***********************************************************************
 * EFLECH process communication
 * ***********************************************************************
 */
void BgmnProjectWidget::eflechRunProcess()
{
    eflechHandler->run(1, projectDir, QStringList(eflechHandler->getStrId()));
}

void BgmnProjectWidget::eflechAbortProcess()
{
    eflechHandler->abort();
    graphControl->setViewStatus(2);
    setStatus(ABORTED);
    eflechHandler->removeTemporaryFiles(projectDir, eflechHandler->getStrId());
}

void BgmnProjectWidget::eflechPollOutput()
{
    refOutput->appendPlainText(eflechHandler->readOutput().trimmed());
    eflechAppendHklData(eflechHandler, true);
}

void BgmnProjectWidget::eflechComplete()
{
    QString euid = eflechHandler->getStrId();
    disconnect(eflechHandler, SIGNAL(allComplete(QUuid)), nullptr, nullptr);
    disconnect(eflechHandler, SIGNAL(pollOutput()), nullptr, nullptr);

    eflechAppendHklData(eflechHandler, false);

    eflechHandler->removeTemporaryFiles(projectDir, euid);
    graphControl->setViewStatus(2);
    setStatus(COMPLETED);

    QString timeString = QString(tr("Peak detection completed in %1 hh:mm:ss.ms").arg(elapsedTimeFormatted(processTimer.elapsed())));
    refOutput->appendPlainText(timeString);
    qDebug() << QString("BgmnProjectWidget::eflechComplete(): %1").arg(timeString);

    emit completed(this);
}

bool BgmnProjectWidget::eflechSourceDevFile(QString &dev, QString &lam)
{
    BgmnBackendConfig bkgCfg;
    QMap<QString, QString> devFiles(bkgCfg.getAllDevFiles());

    bool ok;
    BgmnSavParser sparser(getSavParser(&ok));

    if (ok) {
        QFileInfo fiSDev(sparser.deviceFile());
        if (devFiles.contains(fiSDev.completeBaseName())) {
            dev = devFiles.value(fiSDev.completeBaseName());
            lam = QString("LAMBDA=%1").arg(sparser.lamFile());
            return true;
        }
    }

    BgmnInstrumentSelectDialog *biDlg = new BgmnInstrumentSelectDialog(this);

    if (biDlg->exec() != QDialog::Accepted) {
        return false;
    }

    QFileInfo fiDev(biDlg->getDevFile());
    dev = fiDev.absoluteFilePath();

    if (biDlg->isLamSelected()) {
        QFileInfo fiLam(biDlg->getLamFile());
        lam = QString("LAMBDA=%1").arg(fiLam.completeBaseName());
    } else {
        lam = QString("SYNCHROTRON=%1").arg(biDlg->getSynchrotronValue(), 0, 'f', 6);
    }

    return true;
}

bool BgmnProjectWidget::eflechActiveScanToTemp(const QString &uid)
{
    if (!graphControl->hasData()) return false;

    Scan *scan = graphControl->firstActiveScan();
    if (!scan) return false;

    if (!scan->hasScanData()) {
        QString m(QString(tr("Scan has no scan data. Select a scan with measured data and run the process again.")));
        refOutput->appendPlainText(m);
        return false;
    }

    QMap<QString, QVariant> flags;
    flags["fieldSeparator"] = QVariant(QString(" "));
    flags["fixBgmnZero"]    = QVariant(true);

    QString outName(projectDir + "/" + uid + ".xy");

    ExportHandler exHandler(this, false);
    if (exHandler.save("ASCII_XY", outName, *scan, flags) > 0) return true;

    return false;
}

QString BgmnProjectWidget::eflechTestString()
{
    QString s;
    if (settings->value("eflech/testN", true).toBool())  s += "N";
    if (settings->value("eflech/testD", true).toBool())  s += "D";
    if (settings->value("eflech/testP", false).toBool()) s += "+";
    if (settings->value("eflech/testM", false).toBool()) s += "-";
    if (settings->value("eflech/test2", true).toBool())  s += "2";
    if (settings->value("eflech/test3", true).toBool())  s += "3";
    if (settings->value("eflech/test4", true).toBool())  s += "4";
    if (settings->value("eflech/testU", true).toBool())  s += "U";
    return s;
}

void BgmnProjectWidget::eflechAppendHklData(EflechHandler *ehandler, bool isTemporary)
{
    bool ok = false;
    bool isLastTemporaryPeakData = graphModel->count() > 0
                                   ? graphModel->last()->auxInfo(tempAuxString, ok).toBool()
                                   : false;

    if (!(isLastTemporaryPeakData && ok)) {
        graphControl->appendHklScan(Scan(QString("%1 peak data").arg(ehandler->getSourceName()), QColor(), 1));
    }

    double l = graphModel->getWaveLength(settings->defaultWavelength());

    Scan *hklScan = graphModel->getLast();
    if (!hklScan) return;

    hklScan->setWaveLength(l);
    hklScan->setAuxInfo(tempAuxString, QVariant(isTemporary));

    QVector<QVector<double> > peakData = ehandler->parseEflechParFiles(projectDir, ehandler->getStrId());
    hklScan->pDataHkl().clear();

    for (int i = 0; i < qMin(peakData[0].size(), peakData[1].size()); ++i) {
        hklScan->pDataHkl().append(Hkl(peakData[0][i], QString(), 0.0, QString(), QColor(), peakData[1][i], 1.0));
    }

    graphControl->normalizeHklScan(hklScan);
    graphControl->updateViews();
    peakListWidget->updateData();
}

void BgmnProjectWidget::clearGui()
{
    refOutput->clear();
    chemOutput->clearAll();
    convDisplay->clear();
    peakListWidget->clearAll();
    resultsTree->resetStats();
    resultsTree->clearGlobal();
    resultsTree->clearLocal();
    protocolParser.reset();
    searchMatchWidget->clearResults();
}

/*
 * returns the absolute file path of a file with a relative path (or no path) relative
 * to projectDir
 */
QString BgmnProjectWidget::getAbsoluteFilePath(const QString &file)
{
    QFileInfo fi(file);

    if (fi.absoluteDir() == QDir(projectDir)) {
        return fi.absoluteFilePath();
    }

    QFileInfo absFileName(projectDir + "/" + fi.fileName());
    return absFileName.absoluteFilePath();
}

void BgmnProjectWidget::setRexpDenom(const BgmnSavParser &sparser)
{
    double wmin = sparser.getWmin();
    double wmax = sparser.getWmax();

    QString valFile = QDir::fromNativeSeparators(projectDir + "/" + sparser.valFile().first());
    QPair<int, double> rexpDenom;

    QVector<Scan> scans;
    ImportHandler ihandler(this);
    QString uid = ihandler.uidByFileName(valFile);
    ihandler.load(valFile, uid, scans, true);

    if (scans.isEmpty()) return;
    rexpDenom = scans.first().getRexpDenom(wmin, wmax);

    convDisplay->setRexpDenom(rexpDenom.first, rexpDenom.second);
    protocolParser.setRexpDenom(rexpDenom.first, rexpDenom.second);
}

QString BgmnProjectWidget::elapsedTimeFormatted(int mstotal)
{
    QTime t(0, 0, 0, 0);
    t = t.addMSecs(mstotal);
    return t.toString("hh:mm:ss.zzz");
}

/*
 * uses the add/remove phase dialog to append or remove structures
 * to/from the SAV file
 */
void BgmnProjectWidget::addRemovePhase()
{
    if (!arPhaseDialog) arPhaseDialog = new BgmnAddRemoveDialog(this);

    QPlainTextEdit *e = fileEditor(controlFile, false);
    QFileInfo fiCtrlFile(controlFile);
    QString controlFileContent;

    if (e) {
        controlFileContent = e->toPlainText();
    }

    // if the search-match widget has pinned phases, we preselect them.
    // if not, we check if a reference structure is active, and preselect it.
    // else, nothing is preselected.
    QStringList selectedAddFiles = searchMatchWidget->getPinnedPhases();
    bool setDeviceFileFromSearchMatch = selectedAddFiles.size() > 0 ? true : false;

    if (!selectedAddFiles.size()) {
        if (comboBoxRefStructures->currentIndex() > 0) {
            selectedAddFiles = QStringList(comboBoxRefStructures->currentData(Qt::UserRole).toString());
        }
    }

    // read various information from the current sav file
    BgmnSavParser sparser(controlFileContent, controlFile);

    // populate the dialog with phases read from the savparser
    QStringList strucFiles = sparser.getPhases();
    QStringList strucFilesPaths;

    for (int i = 0; i < strucFiles.size(); ++i) {
        strucFilesPaths.append(QString("%1/%2").arg(projectDir).arg(strucFiles.at(i)));
    }

    arPhaseDialog->initSettings();
    arPhaseDialog->readDirectories();
    arPhaseDialog->setSelectedPhases(selectedAddFiles);
    arPhaseDialog->setDeleteStrFiles(strucFilesPaths);

    QString devFile = sparser.deviceFile();

    if (devFile.isEmpty()) {
        if (setDeviceFileFromSearchMatch) {
            QFileInfo fi(searchMatchWidget->getDeviceFile());
            arPhaseDialog->setDeviceFile(true, fi.fileName());
        } else {
            arPhaseDialog->setDeviceFile(true, QString());
        }
    } else {
        arPhaseDialog->setDeviceFile(false, sparser.deviceFile());
    }

    // if we don't have a sampleID, try to read it from the previous control file content, if there is any
    if (sampleID.isEmpty()) {
        sampleID = sparser.sampleId();
    }

    // run the dialog
    if (arPhaseDialog->exec() == QDialog::Rejected) {
        qDebug() << QString("BgmnProjectWidget::addRemovePhase(): Dialog cancelled. Exiting.");
        return;
    }

    if (refStrFavoritesButton->isChecked()) {
        // favorites may have changed
        setReferenceStructureFileList();
    } else {
        resetRefStructure();
    }

    // check if there is anything to do
    bool cf = arPhaseDialog->createDefaultControlFile();
    bool af = arPhaseDialog->getCheckedAddFiles().size() ? true : false;
    bool rf = arPhaseDialog->getCheckedDeleteFiles().size() ? true : false;

    if ((!cf) && (!af) && (!rf)) {
        qDebug() << QString("BgmnProjectWidget::addRemovePhase(): Dialog accepted, but nothing to do. Exiting.");
        return;
    }

    if (!e) {
        // if no control file editor existed, create a new one
        e = fileEditor(controlFile, true);

        if (!e) {
            // if creating a new control file editor failed, exit
            QMessageBox::warning(this, tr("Create Control File"), tr("Could not create a control file."));
            qDebug() << QString("BgmnProjectWidget::addRemovePhase(): Could not create a control file. Exiting.");
            return;
        }
    }

    qApp->setOverrideCursor(Qt::WaitCursor);

    // store the previous val files for later
    QStringList vfiles = sparser.valFile();

    // if requested, create a new control file from the template file
    if (arPhaseDialog->createDefaultControlFile()) {
        QFileInfo df(arPhaseDialog->deviceFile());
        QFileInfo dfname(BgmnFileIO::gatherDevFiles(df.absoluteFilePath(), fiCtrlFile.absolutePath(), arPhaseDialog->overwriteFiles()));

        sparser.setContent(defaultControlFile(df.absoluteFilePath()));
        sparser.setSampleId(sampleID);
        sparser.setDeviceFile(dfname.fileName());
        sparser.setNumberOfThreads(settings->value("bgmnProject/nThreads", 0).toInt());

        // if a file with a measured background is specified as UNT=filename.xy, copy it, too
        QString untFile(sparser.untFile());
        if (!untFile.isEmpty()) {
            QFileInfo uFiSource(df.absolutePath() + "/" + untFile);
            QFileInfo uFiDest(projectDir + "/" + uFiSource.fileName());
            // qDebug() << QString("*** UNT file, copying from %1 to %2").arg(uFiSource.absoluteFilePath()).arg(uFiDest.absoluteFilePath());

            if (uFiSource.exists() && !uFiDest.exists()) {
                BgmnFileIO::copyFile(uFiSource.absoluteFilePath(), uFiDest.absoluteFilePath());
                // just in case, we re-write the UNT file name without any subdir,
                // because it was copied to projectDir without sub dirs
                sparser.setUntFile(uFiDest.fileName());
            }
        }
    }

    removeStrucFiles(sparser, arPhaseDialog->getCheckedDeleteFiles(), arPhaseDialog->deleteStrFiles());
    insertStrucFiles(sparser, arPhaseDialog->getCheckedAddFiles(), arPhaseDialog->overwriteFiles());

    // write back the VAL[]= files from the original text if they were not empty
    if (!vfiles.isEmpty()) {
        sparser.setValFile(vfiles);
    }

    // get the complete sav file text from the sav parser
    // and write it to the text editor
    e->setPlainText(sparser.getContent());

    autoGuessOutputFiles(false);
    setupSearchMatchWidget(cf);

    qApp->restoreOverrideCursor();
}

void BgmnProjectWidget::removeStrucFiles(BgmnSavParser &sparser, const QStringList &removeStrFiles, bool deleteFiles)
{
    bool mgeGoals = settings->value("bgmnProject/manageGoals", true).toBool();

    for (int i = 0; i < removeStrFiles.size(); ++i) {
        // remove the STRUC entry
        sparser.removeStructureFile(removeStrFiles.at(i), projectDir, mgeGoals);

        // if requested, delete the file
        // but check if there is another entry in the sav file. Don't delete it
        // if there is, because the refinement would not work anymore
        if (deleteFiles && (!sparser.hasPhase(removeStrFiles.at(i)))) {
            QFile f(QString("%1/%2").arg(projectDir).arg(removeStrFiles.at(i)));

            if (f.exists()) {
                f.remove();
            }
        }
    }
}

void BgmnProjectWidget::insertStrucFiles(BgmnSavParser &sparser, const QStringList &insertStrFiles, bool overwriteFiles)
{
    bool mgeGoals = settings->value("bgmnProject/manageGoals", true).toBool();
    QFileInfo fiCtrlFile(controlFile);
    QStringList strFiles = BgmnFileIO::gatherStrFiles(insertStrFiles, fiCtrlFile.absolutePath(), overwriteFiles);
    sparser.addStructureFiles(strFiles, mgeGoals, false);
}

void BgmnProjectWidget::load(const QString &txt, const QString &grph, const QString &uid)
{
    if (!QFile::exists(txt) && !QFile::exists(grph)) return;

    int tabToRaise = 0;
    bool autoDetectGraph = true;

    if (grph.isEmpty()) {
        tabToRaise = count();
    } else {
        loadGraph(grph, uid);
        autoDetectGraph = false;
    }

    if (txt.isEmpty()) {
        autoLoadTextFiles(grph);
    } else {
        loadTextFile(txt); // load manually because the auto-load function ignores the caller file
        autoLoadTextFiles(txt);
    }

    if (autoDetectGraph) {
        QString autoGraph = autoDetermineGraphFile();

        if (QFile::exists(autoGraph)) {
            loadGraph(autoGraph, QString());
        } else {
            QFileInfo fi(txt);
            projectBasename = fi.completeBaseName();
            projectDir = QDir::fromNativeSeparators(fi.absolutePath());
            controlFile = QString();
            sampleID = QString();

            pSelectItem->setText(0, projectBasename);
            projectFileNameItem->setText(1, fi.fileName());
            projectPathItem->setText(1, projectDir);

            projectSampleIdItem->setText(1, QString());

        }
    }

    parseProjectFiles();

    double eps1 = 0.0;
    double eps2 = 0.0;
    double eps3 = 0.0;

    getSampleDisplacements(eps1, eps2, eps3);
    sampleDisplacementsChanged(eps1, eps2, eps3);
    matchStrongestPeak();
    setupSearchMatchWidget(true);

    setCurrentIndex(tabToRaise);
    updateResults();
}

void BgmnProjectWidget::matchStrongestPeak()
{
    if (!settings->value("config/autoSelRefPhase", false).toBool()) return;
    if (!graphControl->hasData()) return;

    double maxIntPos = graphControl->first()->angleOfMaxIntensity();
    referencePosClicked(0.0, 0.0, global::Functions::twoThetaToD(maxIntPos, graphControl->getWaveLength(settings->defaultWavelength())));
}

/*
 * returns the phase quantities as a ; separated string
 */
QStringList BgmnProjectWidget::getQuantitiesCsv()
{
    if (!lparser.hasData()) {
        parseProjectFiles();
    }

    sampleID = getSampleID();

    QStringList goals = lparser.getGoalsCsv(getGlobalIncludeList(), sampleID);
    goals.append(lparser.getRValues(sampleID));
    return goals;
}

/*
 * returns refined parameters as a ; separated string
 */
QStringList BgmnProjectWidget::getResultsCsv()
{
    if (!lparser.hasData()) {
        parseProjectFiles();
    }

    sampleID = getSampleID();

    QStringList localParams(settings->value("bgmnProject/reportedLocalParameters", defaultBgmnLocalParameters).toStringList());

    return lparser.getLocalParametersCsv(localParams, sampleID);
}

QStringList BgmnProjectWidget::getChemistry(bool header)
{
    sampleID = getSampleID();
    return chemOutput->getCsvList(header, sampleID);
}

void BgmnProjectWidget::updateResults()
{
    if (!lparser.hasData()) {
        parseProjectFiles();
    }

    resultsTree->updateGoals(getAbsoluteFilePath(controlFile));
    chemOutput->setData(lparser);
    peakListWidget->updateData();
}

/*
 * check if all val files exists. If not, try to generate them from the scans currently loaded in "graph"
 */
int BgmnProjectWidget::convertRawDataFile()
{
    BgmnSavParser sparser = getSavParser();
    QStringList val = sparser.valFile();
    QMap<QString, QVariant> flags;
    flags["fieldSeparator"] = QVariant(QString(" "));
    flags["fixBgmnZero"]    = QVariant(true);

    int n = 0;
    QStringList errors;

    // compile a list of graphHeap indices with scans containing scan data (skip pure hkl scans)
    QList<int> scanIndices;
    for (int i = 0; i < graphControl->count(); ++i) {
        if (graphControl->at(i)->hasScanData()) {
            scanIndices.append(i);
        }
    }

    for (int i = 0; i < val.size(); ++i) {
        QFileInfo fiOut(QDir::fromNativeSeparators(projectDir + "/" + val.at(i)));

        if (fiOut.exists()) {
            qDebug() << QString("BgmnProjectWidget::generateValFiles(): Scan file %1 exists. No conversion needed.").arg(fiOut.fileName());
            ++n;
            continue;
        }

        if (i < scanIndices.size()) {
            Scan *scan = graphControl->getScan(scanIndices.at(i));

            ExportHandler exHandler(this, false);
            n += exHandler.save("ASCII_XY", fiOut.absoluteFilePath(), *scan, flags) > 0 ? 1 : 0;
        } else {
            errors << fiOut.absoluteFilePath();
        }
    }

    if (errors.size()) {
        qDebug() << QString("BgmnProjectWidget::convertRawDataFile(): The following files could not be converted: %1").arg(errors.join(", "));
    }

    if (n == 0) return -1;
    return n;
}

void BgmnProjectWidget::clearFileNames()
{
    QPlainTextEdit *e = fileEditor(controlFile, false);

    if (!e) {
        qDebug() << QString("BgmnProjectWidget::clearFileNames(): Could not find an control editor for %1").arg(projectBasename);
        return;
    }

    BgmnSavParser sparser(e->toPlainText(), controlFile);
    sparser.setDiagramFile(QString());
    sparser.setOutputFile(QString());
    sparser.setListFile(QString());
    sparser.setValFile(QStringList());
    e->setPlainText(sparser.getContent());
}

/*
 * creates default entries in the control file for VAL, OUTPUT, LIST files etc.
 */
void BgmnProjectWidget::autoGuessOutputFiles(bool clearFirst)
{
    if (clearFirst) clearFileNames();

    QPlainTextEdit *e = fileEditor(controlFile, false);

    if (!e) {
        qDebug() << QString("BgmnProjectWidget::autoGuessOutputFiles(): Could not find an control editor for %1").arg(projectBasename);
        return;
    }

    BgmnSavParser sparser(e->toPlainText(), controlFile);

    // hard code all output filenames to "projectBasename".xxx, except for the
    // val files (these will be handled later)
    // qDebug() << QString("BgmnProjectWidget::autoGuessOutputFiles(): Hard-coding all file names to %1.*").arg(projectBasename);
    sparser.setDiagramFile(QString("%1.dia").arg(projectBasename));
    sparser.setOutputFile(QString("%1.par").arg(projectBasename));
    sparser.setListFile(QString("%1.lst").arg(projectBasename));
    sparser.setStrucOutBaseName(projectBasename);
    sparser.setSampleId(sampleID);

    // count the number of scans that actually contain scan data (skip pure hkl scans)
    int numberOfDataScans = 0;
    for (int i = 0; i < graphControl->count(); ++i) {
        numberOfDataScans += graphControl->at(i)->hasScanData() ? 1 : 0;
    }

    if (sparser.valFile().isEmpty()) {
        bool convert = settings->value("bgmnProject/convertToXy", true).toBool();
        sparser.generateValFiles(loadedGraphFile, nativeGraphFormats, convert, numberOfDataScans);
    }

    e->setPlainText(sparser.getContent());
}

void BgmnProjectWidget::setHighlighting(ControlFileEdit *ed)
{
    BgmnHighlighter *bhl = new BgmnHighlighter(settings->getSyntaxHighlightingMode(), this);
    bhl->setDocument(ed->document());
    highlighters.insert(ed, bhl);
}

QStringList BgmnProjectWidget::contextHelp()
{
    QStringList list;
    ControlFileEdit *e = getCurrentEditor();

    if (!e) {
        qDebug() << QString("BgmnProjectWidget::contextHelp(): Current widget is not a QTextEditor.");
        return list;
    }

    QString word = e->wordUnderSelection();
    QString keyword;
    QString dummy;

    if (!e->splitWordUnderCursor(word, dummy, keyword, dummy, dummy, dummy)) {
        qDebug() << QString("BgmnProjectWidget::contextHelp(): Could not grep the keyword");
        return list;
    }

    list << editors.key(e) << keyword;
    qDebug() << QString("BgmnProjectWidget::contextHelp(): Requesting context help for %1").arg(keyword);
    return list;
}

void BgmnProjectWidget::displayReferenceStructure(const QString &s)
{
    if (s.isEmpty()) {
        graphView->setReferenceReflections(Scan());
        return;
    }

    QVector<QVector<double> > xy = getReferenceXyPattern(s);

    Scan rScan;
    rScan.pDataHkl() = getReferenceHklLines(s);
    rScan.pDataAngle()     = xy.at(0);
    rScan.pDataIntensity() = xy.at(1);
    rScan.setWaveLength(graphView->getWaveLength());
    rScan.setColor(settings->value("graph/hklLineColor", "#009900").toString());

    if (xy.at(0).size() &&  xy.at(1).size()) {
        rScan.setTypes(Scan::XY | Scan::SYNTHETIC);
    } else {
        rScan.setTypes(Scan::HKL | Scan::SYNTHETIC);
    }

    if (rScan.pDataHkl().size() || rScan.pDataAngle().size()) {
        graphView->setReferenceReflections(rScan);
    } else {
        refOutput->appendPlainText(QString("\nNo hkl data found for file %1.").arg(s));
        refOutput->appendPlainText(QString("Please index your reference structure database."));
    }
}

QVector<Hkl> BgmnProjectWidget::getReferenceHklLines(const QString &s)
{
    return refStrManager->getReferenceLines(s);
}

QVector<QVector<double> > BgmnProjectWidget::getReferenceXyPattern(const QString &s)
{
    return refStrManager->getReferencePattern(s);
}

Scan BgmnProjectWidget::getReferenceScan(const QString &s)
{
    return refStrManager->getReferenceScan(s);
}

/*
 * A double click on the graph searches the phase with the strongest peak
 * closest to the double click position and selects it from the
 * reference structure combo box
 */
void BgmnProjectWidget::referencePosClicked(double x, double y, double d)
{
    Q_UNUSED(x);
    Q_UNUSED(y);

    bool fav = refStrFavoritesButton->isChecked();
    QMap<double, PeakFile> nearestMatches;
    int strongPeaks = settings->value("bgmnProject/doubleClickNearestPeaks", 1).toInt();
    refStrManager->referencePosClicked(d, nearestMatches, fav, strongPeaks);

    if (!nearestMatches.size()) return;

    QMapIterator<double, PeakFile> it(nearestMatches);

    refOutput->appendPlainText(QString("Clicked at d=%1").arg(d));
    if (fav) refOutput->appendPlainText(QString("Searching in favorites only!"));
    refOutput->appendPlainText(QString("Best matching phases:"));

    int n = 0;
    int nmax = settings->value("bgmnProject/doubleClickNearestMatches", 5).toInt();
    QList<PeakFile> refFiles;

    while (it.hasNext()) {
        it.next();

        refOutput->appendPlainText(QString("%1: %2 (%3)").arg(n+1).arg(it.value().fileInfo.fileName()).arg(it.key(), 0, 'f', 4));
        refFiles.append(it.value());

        ++n;
        if (n >= nmax) break;
    }

    searchMatchWidget->addReferencePhases(refFiles);
    setReferenceStructure(nearestMatches.values().first().fileInfo.completeBaseName());
}

void BgmnProjectWidget::setReferenceStructure(QString s)
{
    int idx = comboBoxRefStructures->findText(s);
    comboBoxRefStructures->setCurrentIndex(idx > 0 ? idx : 0);
}

void BgmnProjectWidget::toggleRefStrFavorites(bool b)
{
    // receives a signal if this or any other project
    // toggled the ref str favorites button
    bool oldState = refStrFavoritesButton->blockSignals(true);
    refStrFavoritesButton->setChecked(b);
    refStrFavoritesButton->blockSignals(oldState);

    setReferenceStructureFileList();
}

/*
 * Populates the combo box with reference structures.
 */
void BgmnProjectWidget::setReferenceStructureFileList()
{
    resetRefStructure();

    bool oldState = comboBoxRefStructures->blockSignals(true);

    comboBoxRefStructures->clear();
    comboBoxRefStructures->addItem(QString(), QVariant());

    QStringList files;
    if (isRefStrFavorites()) files = refStrManager->getIndexedFavoriteFileNames();
    else                     files = refStrManager->getIndexedFileNames();

    // we'll temporarily store the file infos in a QMap, because it will be sorted alphabetically
    QMap<QString, QFileInfo> map;
    for (int i = 0; i < files.size(); ++i) {
        QFileInfo fi(files.at(i));
        map.insert(fi.fileName().toLower(), fi);
    }

    QMap<QString, QFileInfo>::const_iterator i = map.constBegin();

    while (i != map.constEnd()) {
        comboBoxRefStructures->addItem(i.value().completeBaseName(), QVariant(i.value().absoluteFilePath()));
        ++i;
    }

    comboBoxRefStructures->setCurrentIndex(0);
    comboBoxRefStructures->blockSignals(oldState);
    searchMatchWidget->initData();
}

/*
 * compiles and returns a stringlist with global parameters to be
 * shown in the output tables
 */
QStringList BgmnProjectWidget::getGlobalIncludeList()
{
    QStringList globalGoals = settings->value("bgmnProject/reportedGlobalGoals", global::defaultBgmnGlobalGoals.split("\n")).toStringList();

    if (globalGoals.isEmpty()) globalGoals << ".*";

    return globalGoals;
}

/*
 * compiles and returns a stringlist with local parameters to be
 * shown in the output tables
 */
QStringList BgmnProjectWidget::getLocalIncludeList()
{
    QStringList localGoals = settings->value("bgmnProject/reportedLocalParameters", global::defaultBgmnLocalGoals.split("\n")).toStringList();
    return localGoals;
}

/*
 * opens all structure files referenced in the control file in a text editor
 */
void BgmnProjectWidget::openProjectStrFiles()
{
    appendTextFile(getProjectStrFiles());
}

QStringList BgmnProjectWidget::getProjectStrFiles()
{
    BgmnSavParser sparser = getSavParser();
    QStringList rel = sparser.getPhases();
    QStringList abs;

    for (int i = 0; i < rel.size(); ++i) {
        abs.append(QString(QDir::fromNativeSeparators(projectDir + "/" + rel.at(i))));
    }

    return abs;
}

BgmnSavParser BgmnProjectWidget::getSavParser(bool *ok)
{
    BgmnSavParser sparser(QString(), controlFile);
    QPlainTextEdit *e = fileEditor(controlFile, false);

    if (e) {
        // get text from text editor, so it's up to date even if the sav file has not been
        // saved after modification. It is also more efficient than loading the file from
        // disk
        sparser.setContent(e->toPlainText());
        if (ok) *ok = true;
    } else {
        // load the sav file from the disk
        bool b = sparser.loadFile(controlFile);
        if (ok) *ok = b;
    }

    return sparser;
}

QString BgmnProjectWidget::getSampleID()
{
    bool ok;
    BgmnSavParser sparser = getSavParser(&ok);

    if (ok) return sparser.sampleId();
    return sampleID;
}

/*
 * closes all structure files referenced in the control file in a text editor
 */
void BgmnProjectWidget::closeProjectStrFiles()
{
    QList<int> tabList;

    QMap<QString, ControlFileEdit*>::const_iterator it = editors.constBegin();

    // loop over all tabs and check if it is a str file. If yes, store the tab index
    // in a list
    while (it != editors.constEnd()) {
        QFileInfo fi(it.key());

        if (fi.suffix().toLower() == "str") {
            tabList.append(editorsTabIndex(it.value()));
        }

        ++it;
    }

    // close all tabs stored in the index list.
    // Note: it is important to start at the end of a sorted list, because removing
    qSort(tabList);

    while (tabList.size()) {
        closeTab(tabList.takeLast());
    }
}

void BgmnProjectWidget::compileSumFormula(const QMap<QString, CrystalStructure> &phases)
{
//    refParams->append("");

    QMap<QString, CrystalStructure>::const_iterator it = phases.constBegin();
    while (it != phases.constEnd()) {
        CrystalStructure struc = it.value();
//        refParams->append(QString("%1: %2").arg(it.key()).arg(struc.toSumFormula()));
        ++it;
    }
}

int BgmnProjectWidget::applyPreset(const QString &s, bool overwrite)
{
    if (s.isEmpty()) {
        qDebug() << QString("BgmnProjectWidget::applyPreset(): No preset name specified, exiting.");
        return -1;
    } else {
        qDebug() << QString("BgmnProjectWidget::applyPreset(): Creating project from preset %1").arg(s);
    }

    BgmnPresetHandler phandler(this);
    phandler.parsePresetFile(s);

    if (phandler.hasBaseLine()) {
        applyPresetBaseLine(phandler);
    }

    if (phandler.hasPeakIntegrals()) {
        applyPresetIntegrals(phandler);
    }

    if (phandler.hasCurveFit()) {
        applyPresetCurveFits(phandler);
    }

    int r = 1;

    if (phandler.hasBgmnProject()) {
        r = applyPresetBgmn(phandler, overwrite);
        if (r > 0) autoGuessOutputFiles(true);
    }

    if (phandler.hasAxExcelConfig()) {
        applyPresetMsoExcelExport(phandler);
    }

    resetRefStructure();
    updateResults();
    return r;
}

int BgmnProjectWidget::applyPresetBgmn(BgmnPresetHandler &phandler, bool overwrite)
{
    int nThreads = settings->value("bgmnProject/nThreads", 0).toInt();
    QString controlFileContent(phandler.createProject(controlFile, sampleID, nThreads));

    if (controlFileContent.isEmpty()) {
        qDebug() << QString("BgmnProjectWidget::applyPresetBgmn(): Preset does not contain a refinement project");
        return 1;
    }

    QPlainTextEdit *e = fileEditor(controlFile, true);

    if (!e) {
        qDebug() << QString("BgmnProjectWidget::applyPresetBgmn(): Could not create a text editor for file %1").arg(controlFile);
        return -1;
    }

    if (!e->toPlainText().isEmpty()) {
        if (!overwrite) {
            qDebug() << QString("BgmnProjectWidget::applyPresetBgmn(): Control file %1 already exists. Overwriting not requested. Exiting.").arg(controlFile);
            return 0;
        } else {
            qDebug() << QString("BgmnProjectWidget::applyPresetBgmn(): Control file %1 already exists. Overwriting requested. Applying preset.").arg(controlFile);
        }
    }

    e->setPlainText(controlFileContent);
    return 1;
}

void BgmnProjectWidget::applyPresetIntegrals(BgmnPresetHandler &phandler)
{
    peakIntegrationWidget->setRangeData(phandler.getPeakIntegrals());
}

void BgmnProjectWidget::applyPresetBaseLine(BgmnPresetHandler &phandler)
{
    QMap<QString, QVariant> blParams = phandler.baseLineParameters();

    if (blParams.value("Mode", "SNIP").toString() == "SNIP") {
        int m = blParams.value("m", 60).toInt();
        int w = blParams.value("w", 1).toInt();
        Scan baseline = ScanOps::baseLineSNIP(*(graphModel->first()), m, w);
        graphControl->appendScan(baseline);
    }
}

void BgmnProjectWidget::applyPresetCurveFits(BgmnPresetHandler &phandler)
{
    QList<global::CurveFitCurve> cfCurves;
    QList<global::CurveFitVariable> cfVariables;
    QMap<QString, QVariant> cfControls;

    if (phandler.getCurveFits(cfCurves, cfVariables, cfControls)) {
        peakFitWidget->restorePreset(cfCurves, cfVariables, cfControls);
    }
}

void BgmnProjectWidget::applyPresetMsoExcelExport(BgmnPresetHandler &phandler)
{
    excelExporter->initStructure(phandler.getAxExcelConfig());
}

/*
 * creates and returns the content (as a QString) of a default
 * control file, either read from a *.tpl file for the provided
 * *.geq file, or by showing the GenTemplateDialog
 */
QString BgmnProjectWidget::defaultControlFile(const QString &geq)
{
    QString savFileContent;
    QFileInfo bgmnExec(settings->value("bgmnProject/bgmnExec", "").toString());

    QFileInfo df(geq);
    QFile tplSource(df.absolutePath() + "/" + df.completeBaseName() + ".tpl");

    if (!tplSource.open(QIODevice::ReadOnly | QIODevice::Text)) {
        qDebug() << QString("BgmnProjectWidget::defaultControlFile(): No template file found for device %1.tpl").arg(df.completeBaseName());
        qDebug() << QString("BgmnProjectWidget::defaultControlFile(): Showing GenTemplateDialog");

        GenTemplateDialog *gtdlg = new GenTemplateDialog(this);
        gtdlg->setBgmnDir(bgmnExec.absolutePath());

        if (gtdlg->exec()) {
            savFileContent = gtdlg->getString();
        }

        delete gtdlg;
    } else {
        QTextStream in(&tplSource);
        savFileContent = in.readAll();
    }

    return savFileContent;
}

/*
 * creates a refinement template from the current refinement
 */
void BgmnProjectWidget::createPreset()
{
    ControlFileEdit *cEditor = fileEditor(controlFile, false);
    QString controlFileContent(cEditor ? cEditor->toPlainText() : QString());
    QList<global::PeakRange> peakRanges(peakIntegrationWidget->getRangeData());
    QList<global::CurveFitCurve> curveFitCurves(peakFitWidget->getCurves());
    QList<global::CurveFitVariable> curveFitVariables(peakFitWidget->getVariables());
    QMap<QString, QVariant> curveFitControls(peakFitWidget->getControls());
    QList<global::AxObjectExcel> exclExport(excelExporter->getConfig());
    QList<global::BaseLineSnip> baseLines;

    for (int i = 0; i < graphControl->count(); ++i) {
        const Scan *blSc = graphControl->at(i);
        bool ok;

        bool isBl = blSc->auxInfo("isBaseLine", ok).toBool();

        if (isBl && ok) {
            QList<QVariant> blPars = blSc->auxInfo("baseLineParameters", ok).toList();
            if (blPars.size() >= 2) baseLines.append(global::BaseLineSnip(blPars.at(0).toInt(), blPars.at(1).toInt()));
        }
    }

    bool wrtRef       = !controlFileContent.isEmpty();
    bool wrtIntegrals = !peakRanges.isEmpty();
    bool wrtCurveFit  = !curveFitCurves.isEmpty();
    bool wrtExcelExp  = !exclExport.isEmpty();
    bool wrtBaseLines = !baseLines.isEmpty();

    PresetContentSaveDialog *pcdlg = new PresetContentSaveDialog(this);
    pcdlg->setCheckStatus(wrtRef, wrtIntegrals, wrtCurveFit, wrtExcelExp, wrtBaseLines);

    if (pcdlg->exec() == QDialog::Rejected) {
        delete pcdlg;
        return;
    }

    pcdlg->getCheckStatus(wrtRef, wrtIntegrals, wrtCurveFit, wrtExcelExp, wrtBaseLines);
    delete pcdlg;

    QString presetFile;
    QString presetName;

    if (!getPresetName(presetFile, presetName)) return;

    bool success = false;
    BgmnPresetHandler phandler(this);

    success = phandler.createPreset(cEditor ? controlFile : QString(),
                                    presetFile,
                                    wrtRef       ? controlFileContent : QString(),
                                    wrtIntegrals ? peakRanges         : QList<global::PeakRange>(),
                                    wrtCurveFit  ? curveFitCurves     : QList<global::CurveFitCurve>(),
                                    wrtCurveFit  ? curveFitVariables  : QList<global::CurveFitVariable>(),
                                    wrtCurveFit  ? curveFitControls   : QMap<QString, QVariant>(),
                                    wrtExcelExp  ? exclExport         : QList<global::AxObjectExcel>(),
                                    wrtBaseLines ? baseLines          : QList<global::BaseLineSnip>());

    // show confirmations or error messages
    if (success) {
        emit sigUpdatePresetMenu();
        qDebug() << QString("BgmnProjectWidget::createPreset(): Preset %1 created in %2").arg(presetName).arg(presetFile);
        refOutput->appendPlainText(QString("Preset %1 created in %2").arg(presetName).arg(presetFile));
        QMessageBox::information(this, tr("Refinement Preset"), QString(tr("Preset %1 created.")).arg(presetName));
    } else {
        qDebug() << QString("BgmnProjectWidget::createPreset(): Creating preset %1 in %2 failed").arg(presetName).arg(presetFile);
        refOutput->appendPlainText(QString("Creating preset %1 in %2 failed").arg(presetName).arg(presetFile));
        QMessageBox::warning(this, tr("Refinement Preset"), QString(tr("Preset %1 could not be created.\n"
                                                                       "Please read the log file for further information."))
                                                            .arg(presetName));
    }
}

bool BgmnProjectWidget::getPresetName(QString &presetFile, QString &presetName)
{
    QString pFile;
    QString pName;

    BgmnBackendConfig bkgConfig;
    QString pRepo = bkgConfig.getUserPresetRepo();

    if (pRepo.isEmpty()) {
        QMessageBox::information(this, tr("No Preset Repositore configures"), tr("Please specify a directory to store presets in\nEdit -> Preferences -> BGMN -> Respositories"));
        return false;
    }

    // endless loop, until an empty or valid file name is selected
    while (true) {
        pName = QInputDialog::getText(this, tr("Preset Name"), tr("Preset Name"));

        if (pName.isEmpty()) {
            // input dialog was aborted or closed without entering a name
            return false;
        }

        pFile = QString(pRepo + "/" + pName + "/" + pName + ".pfp");

        if (QFile::exists(pFile)) {
            QMessageBox::information(this, tr("Preset exists"),
                                     QString(tr("A preset named %1 already exists.\n"
                                                "Please enter a different name.").arg(pName)));
            // stay in the loop
        } else {
            // a valid name was provided
            presetFile = pFile;
            presetName = pName;
            return true;
        }
    }

    return false;
}

/*
 * creates a zip archive with all files referenced in the control file.
 * overwrites existing zip files. Abort before calling this function
 * if overwriting is not desired
 */
void BgmnProjectWidget::createZipArchive(const QString &zf)
{
    if (QFile::exists(zf)) QFile::remove(zf);
    qDebug() << QString("BgmnProjectWidget::createZipArchive(): Creating project backup");

    // get a list of all project files
    QFileInfoList flist = getAllProjectFilesList();

    // extract file names
    QStringList clist;

    for (int i = 0; i < flist.size(); ++i) {
        clist.append(flist.at(i).absoluteFilePath());
        qDebug() << QString("    Adding %1").arg(clist.last());
    }

    // create zip archive
    if (JlCompress::compressFiles(zf, clist)) {
        refOutput->appendPlainText(QString("Backup saved to %1").arg(QDir::toNativeSeparators(zf)));
        qDebug() << QString("    Archive created in %1").arg(QDir::toNativeSeparators(zf));
    } else {
        refOutput->appendPlainText(QString("Could not write backup file %1").arg(zf));
        qDebug() << QString("BgmnProjectWidget::createZipArchive(): Problem creating %1").arg(QDir::toNativeSeparators(zf));
    }
}

/*
 * returns a list with all files belonging to a project.
 */
QFileInfoList BgmnProjectWidget::getAllProjectFilesList()
{
    // get references to all files
    BgmnSavParser sparser(controlFile);

    QFileInfoList flist;
    checkFileExists(flist, controlFile);
    checkFileExists(flist, sparser.outputFile(),  projectDir);
    checkFileExists(flist, sparser.listFile(),    projectDir);
    checkFileExists(flist, sparser.diagramFile(), projectDir);
    checkFileExists(flist, sparser.untFile(),     projectDir);
    checkFileExists(flist, sparser.untcFile(),    projectDir);

    checkFilesExist(flist, sparser.getPhases(),   projectDir);
    checkFilesExist(flist, sparser.getStrucOut(), projectDir);
    checkFilesExist(flist, sparser.getSimpleStrucOut(), projectDir);

    checkFilesExist(flist, sparser.getFcfOut(), projectDir);
    checkFilesExist(flist, sparser.getResOut(), projectDir);
    checkFilesExist(flist, sparser.getPdbOut(), projectDir);

    QStringList vfiles = sparser.valFile();
    checkFilesExist(flist, vfiles, projectDir);

    ImportHandler ihandler;
    QStringList rdExtensions = ihandler.extensions();

    for (int i = 0; i < vfiles.size(); ++i) {
        QFileInfo fiVf(vfiles.at(i));

        // also add raw data files with the same name as the val files (e.g. x.raw belonging to x.xy)
        for (int j = 0; j < rdExtensions.size(); ++j) {
            checkFileExists(flist, fiVf.completeBaseName() + "." + rdExtensions.at(j).toLower(), projectDir);
            checkFileExists(flist, fiVf.completeBaseName() + "." + rdExtensions.at(j).toUpper(), projectDir);
        }
    }

    // add raw data files with the same name as the project basename
    for (int i = 0; i < rdExtensions.size(); ++i) {
        checkFileExists(flist, projectBasename + "." + rdExtensions.at(i).toLower(), projectDir);
        checkFileExists(flist, projectBasename + "." + rdExtensions.at(i).toUpper(), projectDir);
    }

    QFileInfo fiDf(sparser.deviceFile());
    checkFileExists(flist, fiDf.fileName(), projectDir);

    if (fiDf.suffix().toLower() != "geq") {
        checkFileExists(flist, fiDf.completeBaseName() + ".geq", projectDir);
        checkFileExists(flist, fiDf.completeBaseName() + ".GEQ", projectDir);
    }

    checkFileExists(flist, fiDf.completeBaseName() + ".sav", projectDir);
    checkFileExists(flist, fiDf.completeBaseName() + ".ger", projectDir);
    checkFileExists(flist, fiDf.completeBaseName() + ".tpl", projectDir);
    checkFileExists(flist, fiDf.completeBaseName() + ".SAV", projectDir);
    checkFileExists(flist, fiDf.completeBaseName() + ".GER", projectDir);
    checkFileExists(flist, fiDf.completeBaseName() + ".TPL", projectDir);

    // lam file (is a bit complicated, because lam file names can be lower/upper case or mixed)
    QString lamFile(sparser.lamFile());
    QStringList lamFilters;
    lamFilters << "*.lam" << "*.LAM" << "*.ano" << "*.ANO" << "*.mdr" << "*.MDR";
    QFileInfo fiBgmn(settings->value("bgmnProject/bgmnExec", QString()).toString());
    QFileInfoList fiAllLam = fiBgmn.absoluteDir().entryInfoList(lamFilters, QDir::Files);

    for (int i = 0; i < fiAllLam.size(); ++i) {
        if (fiAllLam.at(i).completeBaseName().toLower() == lamFile.toLower()) {
            flist << fiAllLam.at(i);
        }
    }

   for (int i = 0; i < flist.size(); ++i) {
        qDebug() << QString("+++ File exists: %1").arg(flist.at(i).absoluteFilePath());
    }

    return flist;
}

bool BgmnProjectWidget::checkFilesExist(QFileInfoList &fil, const QStringList &flist, const QString &dir)
{
    int n = 0;

    for (int i = 0; i < flist.size(); ++i) {
        if (checkFileExists(fil, flist.at(i), dir)) ++n;
    }

    return n > 0;
}

bool BgmnProjectWidget::checkFileExists(QFileInfoList &fil, const QString &file, const QString &dir)
{
    QString path = dir.isEmpty() ? QString() : dir + QDir::separator();

    if (QFileInfo::exists(path + file)) { // is faster than QFileInfo(path+fil).exists() according to docu
        QFileInfo fi(path + file);

        if (fi.isFile() && !fil.contains(fi)) {
            fil.append(fi);
            return true;
        }
    }

    return false;
}

/*
 * returns a list with all input files belonging to a project. This is very much dependent on
 * the type of project (BGMN, FP). Input files are structure, device, and background files.
 * Not included are: DIA, PAR, VAL, LIST, STRUCOUT, SimpleSTRUCOUT, FCFOUT, RESOUT, and PDBOUT
 */
QFileInfoList BgmnProjectWidget::getSharedInputProjectFilesList()
{
    // get references to all files
    BgmnSavParser sparser(controlFile);

    QFileInfoList flist;
    flist.append(QFileInfo(projectDir + "/" + sparser.untFile()));

    // add structures
    QStringList strfiles = sparser.getPhases();
    for (int i = 0; i < strfiles.size(); ++i) {
        flist.append(QFileInfo(projectDir + "/" + strfiles.at(i)));
    }

    // add device files
    QFileInfo fiDf(projectDir + "/" + sparser.deviceFile());
    flist.append(fiDf);
    flist.append(QFileInfo(projectDir + "/" + fiDf.completeBaseName() + ".ger"));
    flist.append(QFileInfo(projectDir + "/" + fiDf.completeBaseName() + ".geq"));

    // test which files exist, skip the ones that don't exist
    QFileInfoList clist;

    for (int i = 0; i < flist.size(); ++i) {
        if (flist.at(i).isFile() && flist.at(i).exists()) {
            clist.append(flist.at(i));
        }
    }

    return clist;
}

/*
 * copes the input files shared with other projects to the current directory.
 */
bool BgmnProjectWidget::gatherSharedInputFiles(const QString &sourceDir, const QFileInfoList &sourceFiles, bool forceOverwrite)
{
    QDir sdir(sourceDir);
    QFileInfoList destFiles;

    // create a list with destination file names
    for (int i = 0; i < sourceFiles.size(); ++i) {
        QString sourceRelPath = sdir.relativeFilePath(sourceFiles.at(i).absoluteFilePath());
        QFileInfo destFi(projectDir + "/" + sourceRelPath);

        // check if source and dest are identical
        if (destFi.absolutePath() != sourceFiles.at(i).absolutePath()) {
            destFiles.append(destFi);
        }
    }

    // check if any of the dest files exists, but ignore *.geq, *.ger, *.sav
    QStringList ex;
    for (int i = 0; i < destFiles.size(); ++i) {
        if (destFiles.at(i).suffix().toLower() == "geq") continue;
        if (destFiles.at(i).suffix().toLower() == "ger") continue;
        if (destFiles.at(i).suffix().toLower() == "sav") continue;

        if (destFiles.at(i).exists()) ex.append(destFiles.at(i).absoluteFilePath());
    }

    // ask for permission to overwrite files
    if (ex.size()) {
        if (!forceOverwrite) {
            if (QMessageBox::question(this,
                                      tr("Overwriting files"),
                                      QString(tr("The following files exists and may be used by other projects.\nDo you want to overwrite them?\n\n%1")).arg(ex.join("\n")))
                    == QMessageBox::No) return false;
        }
    }

    // delete existing files
    for (int i = 0; i < ex.size(); ++i) {
        QFile::remove(ex.at(i));
    }

    // copy the files
    for (int i = 0; i < qMin(sourceFiles.size(), destFiles.size()); ++i) {
        qDebug() << QString("Copying %1 to %2").arg(sourceFiles.at(i).absoluteFilePath()).arg(destFiles.at(i).absoluteFilePath());
        QFile fi(sourceFiles.at(i).absoluteFilePath());
        fi.copy(destFiles.at(i).absoluteFilePath());
    }

    return true;
}

/*
 * returns the EPS values (sample height displacement) from the list file.
 */
void BgmnProjectWidget::getSampleDisplacements(double &e1, double &e2, double &e3)
{
    if (!lparser.hasData()) {
        parseProjectFiles();
    }

    e1 = lparser.getEpsN(1);
    e2 = lparser.getEpsN(2);
    e3 = lparser.getEpsN(3);
}

/*
 * Saves all structures found in the list file as cif files.
 * File names are assigned automatically:
 *
 * projectbasename-phasename.cif
 */
void BgmnProjectWidget::saveCifFiles(bool complete, const QMap<QString, QVariant> &auxDataGlobal)
{
    if (!lparser.hasData()) {
        parseProjectFiles();
    }

    QStringList phases = lparser.getPhaseNames();

    if (phases.isEmpty()) {
        qDebug() << QString("BgmnProjectWidget::saveCifFiles(): No phases read from file %1, exiting").arg(lparser.getFileName());
        return;
    }

    QFileInfo bname(controlFile);
    QStringList fnames;

    QMap<QString, QVariant> auxData(auxDataGlobal);

    if (graphModel->hasData()) {
        QList<QVariant> wls;

        double wl1 = graphModel->first()->waveLength();
        double wl2 = graphModel->first()->waveLength2();
        double wl3 = graphModel->first()->waveLength3();

        if (!qFuzzyIsNull(wl1)) wls.append(QVariant(wl1));
        if (!qFuzzyIsNull(wl2)) wls.append(QVariant(wl2));
        if (!qFuzzyIsNull(wl3)) wls.append(QVariant(wl3));

        if (wls.size()) auxData["_diffrn_radiation_wavelength"] = QVariant(wls);
    }

    for (int i = 0; i < phases.size(); ++i) {
        CrystalStructure structure = lparser.getCrystalStructure(phases.at(i));
        QString fn(bname.absolutePath() + "/" + bname.completeBaseName() + "-" + phases.at(i) + ".cif");

        qDebug() << QString("BgmnProjectWidget::saveCifFiles(): Saving phase %1 to file %2").arg(phases.at(i)).arg(fn);

        QFile f(fn);

        if (!f.open(QIODevice::WriteOnly | QIODevice::Text)) {
            qDebug() << QString("BgmnProjectWidget::saveCifFiles(): Could not open file for writing %1, continuing").arg(fn);
            continue;
        }

        QTextStream out(&f);
        out << structure.toCif(complete, controlFile, auxData);
        f.close();

        fnames.append(fn);
    }

    refOutput->appendPlainText(QString("Exported CIF files:\n%1\n").arg(fnames.join("\n")));
}

QList<phaseData> BgmnProjectWidget::getCifData(const QMap<QString, QVariant> &auxDataGlobal)
{
    QList<phaseData> list;

    if (!lparser.hasData()) {
        parseProjectFiles();
    }

    QStringList phases = lparser.getPhaseNames();

    if (phases.isEmpty()) {
        qDebug() << QString("BgmnProjectWidget::getCifData(): No phases read from file %1, exiting").arg(lparser.getFileName());
        return list;
    }

    QMap<QString, QVariant> auxData(auxDataGlobal);

    if (graphModel->hasData()) {
        QList<QVariant> wls;
        double wl[3] = {0.0, 0.0, 0.0};
        graphModel->getScanWaveLength(wl[0], wl[1], wl[2], settings->defaultWavelength());

        if (!qFuzzyIsNull(wl[0])) wls.append(QVariant(wl[0]));
        if (!qFuzzyIsNull(wl[1])) wls.append(QVariant(wl[1]));
        if (!qFuzzyIsNull(wl[2])) wls.append(QVariant(wl[2]));

        if (wls.size()) auxData["_diffrn_radiation_wavelength"] = QVariant(wls);
    }

    for (int i = 0; i < phases.size(); ++i) {
        CrystalStructure structure = lparser.getCrystalStructure(phases.at(i));

        phaseData d;
        d.project = projectBasename;
        d.phase = phases.at(i);
        d.data = QVariant(structure.toCif(false, controlFile, auxData));
        list.append(d);
    }

    return list;
}

/*
 * Saves all structures found in the list file as cif files.
 * File names are assigned automatically:
 *
 * projectbasename-phasename.cell
 */
void BgmnProjectWidget::saveCastepCellFiles()
{
    bool ok;
    BgmnSavParser sparser(controlFile, &ok);

    if (!ok) {
        qDebug() << QString("BgmnProjectWidget::saveCastepCellFiles(): Could not read control file %1, exiting").arg(controlFile);
        return;
    }

    QStringList resFileNames = sparser.getResOut();
    QStringList resFiles;

    // store the res files that really exist in resFiles
    for (int i = 0; i < resFileNames.size(); ++i) {
        if (QFileInfo::exists(projectDir + "/" + resFileNames.at(i))) {
            resFiles.append(projectDir + "/" + resFileNames.at(i));
        }
    }

    if (!resFiles.size()) {
        qDebug() << QString("BgmnProjectWidget::saveCastepCellFiles(): None of the referenced res files exists");
        refOutput->appendPlainText(QString(tr("CELL file export: No RES files available")));
        refOutput->appendPlainText(QString(tr("  CELL files are converted from RES files")));
        refOutput->appendPlainText(QString(tr("  Please add \"RESOUT[n]=mystructure.res\" to your control file")));
        refOutput->appendPlainText(QString(tr("  and re-run the refinement.")));
        return;
    }

    QFileInfo bname(controlFile);
    refOutput->appendPlainText(QString(tr("Exported CELL files:")));

    for (int r = 0; r < resFiles.size(); ++r) {
        BgmnResParser rparser(resFiles.at(r));
        CrystalStructure structure = rparser.structure();

        QString fn(bname.absolutePath() + "/" + bname.baseName() + "-" + structure.name() + ".cell");

        qDebug() << QString("BgmnProjectWidget::saveCastepCellFiles(): Saving phase %1 to file %2")
                    .arg(structure.name()).arg(fn);

        QFile f(fn);

        if (!f.open(QIODevice::WriteOnly | QIODevice::Text)) {
            qDebug() << QString("BgmnProjectWidget::saveCastepCellFiles(): Could not open file for reading %1, continuing").arg(fn);
            continue;
        }

        QTextStream out(&f);
        out << structure.toCastepCell();
        f.close();

        refOutput->appendPlainText(fn);
    }
}

void BgmnProjectWidget::parseProjectFiles()
{
    QFileInfo fi(controlFile);

    if (!fi.exists()) {
        qDebug() << QString("BgmnProjectWidget::parseProjectFiles(): Control file does not exist: %1").arg(controlFile);
        return;
    }

    BgmnSavParser sparser(fi.absoluteFilePath());

    if (!lparser.load(fi.absolutePath() + "/" + sparser.listFile())) {
        qDebug() << QString("BgmnProjectWidget::parseProjectFiles(): Loading list file failed: %1").arg(lparser.getFileName());
    }

    if (settings->value("bgmnProject/readLamFile", false).toBool()) {
        setCustomWavelengt(sparser.lamFile());
    }
}

void BgmnProjectWidget::setInternalStandard()
{
    // get the text editor for the controlFile
    QPlainTextEdit *e = fileEditor(controlFile, false);

    if (!e) {
        qDebug() << QString("BgmnProjectWidget::setInternalStandard(): No control file available");
        return;
    }

    // create a savparser to access and modify the control file content
    BgmnSavParser sparser(e->toPlainText(), controlFile);

    // populate the dialog with phases read from the savparser
    QStringList strucFiles = sparser.getPhases();
    QStringList phaseNames;

    for (int i = 0; i < strucFiles.size(); ++i) {
        QString absStrucFile = QString("%1/%2").arg(projectDir).arg(strucFiles.at(i));
        BgmnStrParser strParser(absStrucFile);
        phaseNames.append(strParser.getQuantGoal());
    }

    if (phaseNames.isEmpty()) {
        QMessageBox::information(this, tr("No phases found"), tr("Please add at least one phase to the project."));
        return;
    }

    InternalStandardDialog *idlg = new InternalStandardDialog(this);
    idlg->setPhases(phaseNames);
    idlg->setCurrentPhase(sparser.internalStandard());
    idlg->setCurrentQuantity(sparser.internalStandardQuantity());

    if (idlg->exec() == QDialog::Accepted) {
        sparser.setInternalStandard(idlg->getPhase(), idlg->getQuantity());
        e->setPlainText(sparser.getContent());
    }

    delete idlg;
}

void BgmnProjectWidget::unsetInternalStandard()
{
    // get the text editor for the controlFile
    QPlainTextEdit *e = fileEditor(controlFile, false);

    if (!e) {
        qDebug() << QString("BgmnProjectWidget::unsetInternalStandard(): No control file available");
        return;
    }

    // create a savparser to access and modify the control file content
    BgmnSavParser sparser(e->toPlainText(), controlFile);
    sparser.unsetInternalStandard();
    e->setPlainText(sparser.getContent());
}

void BgmnProjectWidget::generateReport()
{
    QString fileName = writeReport();
    refOutput->appendPlainText("Report saved to " + fileName);
    QDesktopServices::openUrl(QUrl("File:" + fileName, QUrl::TolerantMode));
}

QString BgmnProjectWidget::writeReport()
{
    QString fileName(projectDir + "/" + projectBasename + ".html");

    BgmnHtmlReportGenerator reportGen(lparser);
    reportGen.setDocumentStructure(settings->value("bgmnProject/report/documentStructure", BgmnFileIO::readTextFile(":/resources/report-structure.xml")).toString());

    BgmnSavParser sparser(controlFile);

    QMap<QString, QVariant> projectData;
    projectData.insert("rawFileName", loadedGraphFile);
    projectData.insert("workingDir", workingDir());
    projectData.insert("sampleId", QVariant(sampleID));
    projectData.insert("geqFile", sparser.deviceFile());
    projectData.insert("wavelength", graphView->getWaveLength());
    projectData.insert("wavelengthFile", sparser.lamFile());
    projectData.insert("globalGoalsIncludes", QVariant(getGlobalIncludeList()));
    projectData.insert("localGoalsIncludes", QVariant(getLocalIncludeList()));
    projectData.insert("quantGoals100percent", QVariant(settings->value("bgmnProject/quantGoals100percent", QVariant(false))));
    projectData.insert("parFileName", getAbsoluteFilePath(sparser.outputFile()));
    projectData.insert("chemistryData", chemOutput->getHtmlTable());

    projectData.insert("diffpatternSvg", graphView->getSvg(reportGen.patternAspectRatio()));

    QFont reportFont;
    reportFont.setFamily("Helvetica");
    reportFont.setPointSize(10);

    reportGen.setProjectData(projectData);

    reportGen.generateDocument();
    BgmnFileIO::writeTextFile(fileName, reportGen.getContents());
    return fileName;
}

void BgmnProjectWidget::deleteReport()
{
    QString fileName(projectDir + "/" + projectBasename + ".html");
    if (QFile::exists(fileName)) QFile::remove(fileName);
}

void BgmnProjectWidget::resetMarginColor()
{
    if (!bgmnHandler->isRunning()) {
        graphView->setIdle();
    }
}

void BgmnProjectWidget::setCustomWavelengt(const QString &lambda)
{
    if (!settings->value("bgmnProject/readLamFile", false).toBool()) return;

    // we must determine the lam file name case insensitively (often the case in the control
    // file doesn't match the actual file name's case). That's why the following routine
    // is a bit complicated.
    QFileInfo fibgmn(settings->value("bgmnProject/bgmnExec", QString()).toString());
    QFileInfo filambda(fibgmn.absolutePath() + "/" + lambda + ".lam");
    QDir ldir(filambda.absolutePath());

    QStringList filters;
    filters << "*.lam" << "*.LAM";

    QStringList files = ldir.entryList(filters, QDir::Files | QDir::Readable).filter(filambda.fileName(), Qt::CaseInsensitive);

    if (files.size()) {
        QString foundLamFile(filambda.absolutePath() + "/" + files.first());
        qDebug() << QString("BgmnProjectWidget::setCustomWavelengt(): Parsing lam file %1").arg(foundLamFile);
        BgmnLamParser lamparser(foundLamFile);
        // graph->setCustomWaveLengths(lamparser.getWavelengths());
        graphControl->setCustomWaveLength(lamparser.getWavelengths());
    } else {
        qDebug() << QString("BgmnProjectWidget::setCustomWavelengt(): Lam file %1 does not exist").arg(filambda.absoluteFilePath());
    }
}

QString BgmnProjectWidget::getResultsFile()
{
    if (!lparser.hasData()) {
        parseProjectFiles();
    }

    return lparser.getFileName();
}

/*
 * returns the file name (and absolute path) of the device SAV file.
 *
 * - If the device SAV file is found in the project dir, its full path and
 *   name will be returned.
 *
 * - If it is not found in the project dir (only GEQ and GER must be present),
 *   it will search for the file in the device file repository. If found, the
 *   return value will depend on the state of the flag "copy":
 *
 *   - copy = true:  The SAV file will be copied from the device repo to the
 *                   project dir, and the destination file path and name
 *                   will be returned.
 *   - copy = false: The SAV file will not be copied, the source file path
 *                   and name will be returned.
 *
 * - If the SAV file is neither found in the project dir, nor in the device repo,
 *   an empty string is returned.
 */
QString BgmnProjectWidget::getInstrumentConfigFile(bool copy)
{
    BgmnSavParser sparser(controlFile);
    QFileInfo fiGeq(sparser.deviceFile());

    QStringList filters;
    filters << fiGeq.completeBaseName() + ".sav" << fiGeq.completeBaseName() + ".SAV";

    QDir pdir(projectDir);
    QFileInfoList savList = pdir.entryInfoList(filters, QDir::Files);

    if (savList.size()) {
        return savList.first().absoluteFilePath();
    }

    QStringList devRepos = settings->value("bgmnProject/deviceDatabase", QStringList()).toStringList();

    for (int i = 0; i < devRepos.size(); ++i)  {
        QDir repo(devRepos.at(i));
        savList = repo.entryInfoList(filters, QDir::Files);

        if (!savList.size()) continue;

        QFileInfo fiSrc = savList.first();
        QFileInfo fiDst = QFileInfo(projectDir + "/" + savList.first().fileName());

        if (copy) {
            if (!BgmnFileIO::copyFile(fiSrc.absoluteFilePath(), fiDst.absoluteFilePath())) {
                return QString();
            }

            return fiDst.absoluteFilePath();
        }
        return fiSrc.absoluteFilePath();
    }

    return QString();
}

void BgmnProjectWidget::editProjectStrFiles()
{
    if (!strFileEditDlg) {
        strFileEditDlg = new StrFileBatchEditDialog(this);
        connect(strFileEditDlg, SIGNAL(accepted()), this, SLOT(applyEditProjectStrFiles()));
        connect(strFileEditDlg, SIGNAL(sigApply()), this, SLOT(applyEditProjectStrFiles()));
    }

    strFileEditDlg->reset();


    QFileInfo cfi(currentFileName());
    QString absCurFileName;

    if (cfi.suffix().toLower() == "str") {
        absCurFileName = cfi.absoluteFilePath();
    }

    strFileEditDlg->setCurrentFile(absCurFileName);
    strFileEditDlg->setOpenFileList(getOpenFileNames("str"));
    strFileEditDlg->setStrFileList(getProjectStrFiles());

    strFileEditDlg->show();
}

void BgmnProjectWidget::applyEditProjectStrFiles()
{
    // read values from dialog
    QMap<QString, QVariant> values = strFileEditDlg->getParameters();
    QStringList modFiles = values.value("files").toStringList();

    // save all open str files
    for (int i = 0; i < modFiles.size(); ++i) {
        saveEditorFile(modFiles.at(i));
    }

    // apply changes to files on disk
    BgmnStrParser strParser;

    for (int i = 0; i < modFiles.size(); ++i) {
        strParser.loadFile(modFiles.at(i));

        QMapIterator<QString, QVariant> it(values);

        while (it.hasNext()) {
            it.next();
            strParser.setRefinementState(it.key(), it.value().toInt());
        }

        strParser.writeToFile(modFiles.at(i));
        // reload all open str files
        loadTextFile(modFiles.at(i));
    }
}

/*
 * sets the current scan as reference scan in the currently shown reference structure
 */
void BgmnProjectWidget::setReferenceXY()
{
    if (!graphModel->hasData()) return;

    int n = scanList->currentIndex();

    if (n < 0) {
        qDebug() << QString("ProjectWidget::setReferenceXY(): No selected scan found, index was %1").arg(n);
        return;
    }

    Scan *scan = graphModel->getScan(n);
    QString rFile = comboBoxRefStructures->currentData(Qt::UserRole).toString();
    if (rFile.isEmpty()) {
        refOutput->appendPlainText("Select a reference structure first.");
        return;
    }

    if (QMessageBox::question(this,
                              tr("Overwrite reference pattern"),
                              tr("Do you really want to overwrite the current reference structure's pattern?"))
            == QMessageBox::Yes) {
        refStrManager->setXyData(rFile, *scan);
    }
}

void BgmnProjectWidget::addAmorphousPeak()
{
    QPlainTextEdit *e = fileEditor(controlFile, false);

    if (!e) return;

    bool ok;
    double tt = settings->value("bgmnProject/amorphPeak2t", 30.0).toDouble();

    tt = QInputDialog::getDouble(this, tr("Amorphous Peak"), QString(tr("Center position (°2%1)").arg(global::theta)), tt, 1.0, 180.0, 2, &ok);
    if (!ok) return;

    settings->setValue("bgmnProject/amorphPeak2t", tt);

    double wl = graphView->getWaveLength();
    double d = 0.1 * wl / (2.0 * qSin(qDegreesToRadians(tt / 2.0)));
    double dmin = qMax(0.05 * wl / (2.0 * qSin(qDegreesToRadians(tt / 2.0))),
                       0.1 * wl / (2.0 * qSin(qDegreesToRadians(180.0 / 2.0))));
    double dmax = 2.0 * d;

    QString str("PHASE=amorphous\n");
    str += QString("SpacegroupNo=195 GeneralCondition=eq(1,sqr(h)+sqr(k)+sqr(l))\n");
    str += QString("PARAM=B1=0.1_0^1 PARAM=B2=0_0^0.1 PARAM=A=%1_%2^%3").arg(d, 0, 'f', 6).arg(dmin, 0, 'f', 6).arg(dmax, 0, 'f', 6);

    QFileInfo fi(workingDir() + "/" + "amorphous.str");

    BgmnFileIO::writeTextFile(fi.absoluteFilePath(), str);
    BgmnSavParser sparser = getSavParser();
    sparser.addAmorphous(fi.fileName());
    e->setPlainText(sparser.getContent());
}

void BgmnProjectWidget::exportToExcel()
{
    bool ok;
    BgmnSavParser sparser = getSavParser(&ok);
    if (!ok) return;
    excelExporter->exportData(controlFile, projectDir + "/" + sparser.listFile());
}

void BgmnProjectWidget::editExcelExport()
{
    bool ok;
    BgmnSavParser sparser = getSavParser(&ok);
    if (!ok) return;
    excelExporter->showEditor(controlFile, projectDir + "/" + sparser.listFile());
}

/* EOF */

