/***************************************************************************
                          bgmndiahandlernoarch.cpp  -  description
                             -------------------
    begin                : Fri Mar 25 10:22:00 CEST 2016
    copyright            : (C) 2016 by Nicola Doebelin
    email                : ndoebelin@gmx.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


/*
 * The NOARCH dia file poller works on all platforms. It was used up to profex
 * version 3.8, and tries to determine the ideal delay time to read the dia
 * file from disk. With fast update rates, however, it may skip many updates.
 */

#include "bgmndiahandler.h"
#include <QDateTime>
#include <QDebug>

BgmnDiaHandler::BgmnDiaHandler(QObject *parent) : QObject(parent)
{
    settings = SettingsManager::getInstance();
    scanControl = nullptr;
    delayTimer.setSingleShot(true);

    connect(&delayTimer, SIGNAL(timeout()), this, SLOT(reloadDelayed()));

    delayPerLine = settings->value("bgmnProject/diaReadDelayPerLine", 0.001).toDouble();
    increaseSkipTolerance = 1;         // max number of skipped reloads before the delay interval is increased
    reduceSkipTolerance = 2;           // max number of skipped reloads before the delay interval is reduced
    increaseSkips = 0;                 // counter of skipped reloads
    reduceSkips = 0;                   // counter of excessively long reloads
    maxPartialLinesRead = 0;           // number of lines read if reading was incomplete
    delayIncrement = 5;                // increment by which delayInterval will grow if graph files are incomplete
    delayDecrement = 1;                // decrement by which delayInterval will shrink if the timer doesn't have time to fire between calls
    hadCompleteRead = false;           // will be set to true the first time a dia file was read completely
}

void BgmnDiaHandler::init(GraphDataController *g, const QString &s, int nLines)
{
    if (QFile::exists(s)) {
        QFile::remove(s);
    }

    diaFile = s;
    scanControl = g;
    delayInterval = int(nLines * delayPerLine + 0.5);
    hadCompleteRead = false;
    activeTimers = 0;
    increaseSkips = 0;
    reduceSkips = 0;
    qDebug() << QString("BgmnDiaHandler::init(): Monitoring file %1 with delay interval %2 ms").arg(s).arg(delayInterval);
    qDebug() << QString("BgmnDiaHandler::init(): delayPerLine = %1").arg(delayPerLine);
}

void BgmnDiaHandler::reload()
{
    reload(diaFile);
}

void BgmnDiaHandler::reload(const QString &s)
{
    diaFile = s;

    if (!scanControl || diaFile.isEmpty()) {
        return;
    }

    activeTimers++;

    if (delayTimer.isActive()) {
        if (reduceSkips < reduceSkipTolerance) {
            reduceSkips++;
            qDebug() << QString("BgmnDiaHandler::reload(): Active timer detected. Skipping %1 of %2").arg(reduceSkips).arg(reduceSkipTolerance);
        } else {
            delayInterval *= 0.75;
            reduceSkips = 0;
            qDebug() << QString("BgmnDiaHandler::reload(): Active timer detected. Reducing delayInterval to %1").arg(delayInterval);
        }

        activeTimers--;
        return;
    }

    reduceSkips = 0;
    delayTimer.start(delayInterval);
}

void BgmnDiaHandler::stop()
{
    if (delayTimer.isActive()) activeTimers = 0;
    delayTimer.stop();
    settings->setValue("bgmnProject/diaReadDelayPerLine", delayPerLine);
    qDebug() << QString("BgmnDiaHandler::stop(): delayPerLine = %1 ms").arg(delayPerLine);
}

void BgmnDiaHandler::reloadDelayed()
{
    activeTimers--;
    if (!QFile::exists(diaFile)) return;

    int loaded = -1;
    int linesRead, linesTotal;

    if (scanControl->fileName() == diaFile) {
        loaded = scanControl->reloadDiaFile(true, false, linesRead, linesTotal);
    } else {
        loaded = scanControl->loadDiaFile(diaFile, true, scanControl->getSampleId(), false, linesRead, linesTotal);
    }

    if (loaded > 0) {
        scanControl->updateViews();
        hadCompleteRead = true;
        increaseSkips = 0;
        maxPartialLinesRead = 0;
        delayPerLine = double(delayInterval) / double(linesTotal);
        // qDebug() << QString("delayInterval = %1, delayPerLine = %2").arg(delayInterval).arg(delayPerLine);
    } else {
        incrementDelay(linesRead, linesTotal);
    }
}

void BgmnDiaHandler::incrementDelay(int linesRead, int linesTotal)
{
    if (!hadCompleteRead) {
        // the dia file has never been read completely. So we try to estimate the right delay time
        // without accepting skips

        if ((delayInterval == 0) || (linesRead == 0)) {
            delayInterval += delayIncrement;
        } else {
            double tPerLine = double(delayInterval) / double(linesRead);
            delayInterval = int(tPerLine * linesTotal + 0.5) + increaseSkips * delayIncrement;
            increaseSkips++;
        }
    } else {
        // the dia file has been read completely before. So we don't want to change the delay time
        // too aggressively. We also accept a certain number of missed reads before changing the delay
        // so as not to over-react in case of single delayed flushes.

        if (increaseSkips < increaseSkipTolerance) {
            maxPartialLinesRead = qMax(maxPartialLinesRead, linesRead);
            increaseSkips++;
            qDebug() << QString("BgmnDiaHandler::incrementDelay(): Read %1 of %2 lines. Skipping missed read %3 of %4")
                        .arg(linesRead).arg(linesTotal).arg(increaseSkips).arg(increaseSkipTolerance);
            return;
        }

        if ((delayInterval == 0) || (maxPartialLinesRead == 0)) {
            delayInterval += delayIncrement;
        } else {
            double tPerLine = double(delayInterval) / double(maxPartialLinesRead);
            delayInterval = int(tPerLine * linesTotal + 0.5) + delayIncrement; // add one increment, just to have some additional buffer time
        }

        increaseSkips = 0;
        maxPartialLinesRead = 0;
    }

    qDebug() << QString("BgmnDiaHandler::incrementDelay(): Read %1 of %2 lines. Delay interval adjusted to %3 ms")
                .arg(linesRead).arg(linesTotal).arg(delayInterval);
}
