/***************************************************************************
                          graphdatamodel.cpp  -  description
                             -------------------
    begin                : Tue Feb 11 18:00:00 CEST 2020
    copyright            : (C) 2020 by Nicola Doebelin
    email                : ndoebelin@gmx.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "graphdatamodel.h"
#include "../libXrdIO/import/importhandler.h"
#include "../libXrdIO/export/exporthandler.h"
#include "../libXrdIO/scanops.h"
#include "../libXrdIO/import/bgmndiaimport.h"
#include "../libXrdIO/functions.h"
#include <QFileInfo>
#include <QDebug>
#include <limits>

GraphDataModel::GraphDataModel()
{
    graphHeap = new QVector<Scan>;
    graphBuffer = new QVector<Scan>;
    strucRefls = new QVector<Hkl>;
    integralRanges = new QVector<QPointF>;
    waveLengthTable = new QVector<QVector<double> >();

    QList<global::CharWaveLength> wls = global::Functions::getAllWavelengths();
    for (int i = 0; i < wls.size(); ++i) {
        QVector<double> v;
        v.append(10.0 * wls.at(i).ka1); // convert from nm to A
        v.append(10.0 * wls.at(i).ka2);
        v.append(10.0 * wls.at(i).kb);
        waveLengthTable->append(v);
    }

    globalXrangeMin = 0.0;
    globalXrangeMax = 0.0;
    globalYrangeMin = 0.0;
    globalYrangeMax = 0.0;

    eps1 = 0.0;
    eps2 = 0.0;
    eps3 = 0.0;

    stacked = false;
    nFiles = 0;
}

GraphDataModel::~GraphDataModel()
{
    delete graphHeap;
    delete graphBuffer;
    delete strucRefls;
    delete integralRanges;
    delete waveLengthTable;
}

int GraphDataModel::count() const
{
    return graphHeap->size();
}

bool GraphDataModel::hasData()
{
    return graphHeap->size() > 0;
}

void GraphDataModel::keepTemporary()
{
    for (int i = 0; i < graphHeap->size(); ++i) {
        Scan::ScanTypes flags = graphHeap->at(i).scanTypes();

        if (flags.testFlag(Scan::TEMPORARY)) {
            (*graphHeap)[i].setTypes(Scan::XY | Scan::SYNTHETIC);
        }
    }
}

void GraphDataModel::keepTemporary(const QList<QUuid> &uids)
{
    for (int i = 0; i < uids.size(); ++i) {
        Scan *scan = getScan(uids.at(i));
        if (scan) scan->setTypes(Scan::XY | Scan::SYNTHETIC);
    }
}

void GraphDataModel::clearTemporary()
{
    for (int i = graphHeap->size() - 1; i >= 0; --i) {
        Scan::ScanTypes flags = graphHeap->at(i).scanTypes();

        if (flags.testFlag(Scan::TEMPORARY)) {
            graphHeap->remove(i);
        }
    }
}

void GraphDataModel::removeScans(const QList<QUuid> &uids)
{
    for (int i = 0; i < uids.size(); ++i) {
        removeScan(uids.at(i));
    }
}

double GraphDataModel::getWaveLength(double fallback)
{
    double ka1;
    double ka2;
    double kb;

    getScanWaveLength(ka1, ka2, kb, fallback);
    return ka1;
}

void GraphDataModel::getScanWaveLength(double &ka1, double &ka2, double &kb, double fallback)
{
    if (graphHeap->size()) {
        if (!qFuzzyIsNull(graphHeap->first().waveLength())) {
            getFixedWaveLength(ka1, ka2, kb, graphHeap->first().waveLength());
            return;
        }
    }

    getFixedWaveLength(ka1, ka2, kb, fallback);
}

void GraphDataModel::getFixedWaveLength(double &ka1, double &ka2, double &kb, double fixwl)
{
    for (int i = 0; i < waveLengthTable->size(); ++i) {
        if (qAbs(waveLengthTable->at(i).at(0) - fixwl) < 0.002) {
            ka1 = waveLengthTable->at(i).at(0);
            ka2 = waveLengthTable->at(i).at(1);
            kb  = waveLengthTable->at(i).at(2);
            return;
        }
    }

    ka1 = fixwl;
    ka2 = fixwl;
    kb  = fixwl;
}

int GraphDataModel::loadScanFile(const QString &s, const QString &uid, bool min, const QString &sid)
{
    QFileInfo fi(s);
    QString fext = fi.suffix().toLower();

    *graphBuffer = *graphHeap;
    metricsBuffer = scanMetrics;

    graphHeap->clear();
    scanMetrics.clear();

    int n = addScanFile(fi, uid, min);

    if (n <= 0) {
        // if reloading failed (probably due to timing issues), we roll back the
        // buffered version of the file to avoid a blank screen
        *graphHeap = *graphBuffer;
        scanMetrics = metricsBuffer;
    }

    if (n > 0) {
        if ((fext == "prf") || (fext == "dia")) {
            sampleId = sid;
        } else {
            sampleId = graphHeap->first().name();
        }
    }

    return n;
}

int GraphDataModel::loadDiaFile(const QString &s, bool min, const QString &sid, int &ll, int &tl)
{
    QFileInfo fi(s);
    QString fext = fi.suffix().toLower();

    *graphBuffer = *graphHeap;
    metricsBuffer = scanMetrics;

    graphHeap->clear();
    scanMetrics.clear();

    int n = addDiaFile(fi, "BGMN_DIA", min, ll, tl);

    if (n <= 0) {
        // if reloading failed (probably due to timing issues), we roll back the
        // buffered version of the file to avoid a blank screen
        *graphHeap = *graphBuffer;
        scanMetrics = metricsBuffer;
    }

    if (n > 0) {
        if ((fext == "prf") || (fext == "dia")) {
            sampleId = sid;
        } else {
            sampleId = graphHeap->first().name();
        }
    }

    return n;
}

bool GraphDataModel::saveScanFile(const QString &f, const QString &uid, int n, const QMap<QString, QVariant> &flags)
{
    if (n >= graphHeap->size()) {
        return false;
    }

    ExportHandler exhandler;

    if (n < 0) {
        return exhandler.save(uid, f, *graphHeap, flags);
    }

    return exhandler.save(uid, f, (*graphHeap)[n], flags);
}

int GraphDataModel::addScanFile(const QString &s, const QString &uid, bool min)
{
    return addScanFile(QFileInfo(s), uid, min);
}

int GraphDataModel::addScanFile(const QFileInfo &fi, const QString &uid, bool min)
{
    QString fuid = uid;
    ImportHandler iHandler;

    if (fuid.isEmpty()) {
        fuid = iHandler.uidByFileName(fi.absoluteFilePath());
    }

    completeFileName = fi.absoluteFilePath();
    fileUid = fuid;

    // some import filters clear the graphHeap. Here we don't want this, so we
    // first load the file to a temporary graphHeap and then append it to
    // the main graphHeap.
    QVector<Scan> *tGraphHeap = new QVector<Scan>;

    int n = iHandler.load(completeFileName, fileUid, *tGraphHeap, min);
    if (n < 1) return n;

    graphHeap->append(*tGraphHeap);
    delete tGraphHeap;

    if (!min) {
        iHandler.getReflections(completeFileName, *graphHeap);
    }

    for (int i = graphHeap->size() - n; i < graphHeap->size(); ++i) {
        QVector<double> smetr(4, 0.0);
        ScanOps::scanMetrics(graphHeap->at(i), smetr[0], smetr[1], smetr[2], smetr[3]);
        scanMetrics[graphHeap->at(i).uid()] = smetr;
    }

    countFiles();
    updateMetrics();

    return n;
}

int GraphDataModel::addDiaFile(const QString &s, const QString &uid, bool min, int &ll, int &tl)
{
    return addDiaFile(QFileInfo(s), uid, min, ll, tl);
}

int GraphDataModel::addDiaFile(const QFileInfo &fi, const QString &uid, bool min, int &ll, int &tl)
{
    fileUid = uid;
    BgmnDiaImport diaImport;
    ImportHandler iHandler;

    completeFileName = fi.absoluteFilePath();

    int n = diaImport.loadAndCheck(completeFileName, *graphHeap, ll, tl, min);
    if (n < 1) return n;

    if (!min) {
        iHandler.getReflections(completeFileName, *graphHeap);
    }

    for (int i = graphHeap->size() - n; i < graphHeap->size(); ++i) {
        QVector<double> smetr(4, 0.0);
        ScanOps::scanMetrics(graphHeap->at(i), smetr[0], smetr[1], smetr[2], smetr[3]);
        scanMetrics[graphHeap->at(i).uid()] = smetr;
    }

    countFiles();
    updateMetrics();

    return n;
}
int GraphDataModel::reloadScanFile(bool min)
{
    QList<bool> vis;
    QList<double> yoff;

    for (int i = 0; i < graphHeap->size(); ++i) {
        vis.append(graphHeap->at(i).isVisible());
        yoff.append(graphHeap->at(i).yOffset());
    }

    *graphBuffer = *graphHeap;
    metricsBuffer = scanMetrics;

    graphHeap->clear();
    scanMetrics.clear();
    nFiles = 0;

    int n = addScanFile(QFileInfo(completeFileName), fileUid, min);

    if (n <= 0) {
        // if reloading failed (probably due to timing issues), we roll back the
        // buffered version of the file to avoid a blank screen
        *graphHeap = *graphBuffer;
        scanMetrics = metricsBuffer;
    }

    for (int j = 0; j < graphHeap->size();  ++j) {
        if (j < vis.size())  {
            (*graphHeap)[j].setVisible(vis.at(j));
        }

        if (j < yoff.size()) {
            if ((*graphHeap)[j].yOffset() >= 0.0) {
                // skip negative scans, otherwise the difference curve
                // would not move closer to the base line as the fit
                // gets better
                (*graphHeap)[j].setYoffset(yoff.at(j));
            }
        }
    }

    countFiles();
    return n;
}


int GraphDataModel::reloadDiaFile(bool min, int &ll, int &tl)
{
    QList<bool> vis;
    QList<double> yoff;

    for (int i = 0; i < graphHeap->size(); ++i) {
        vis.append(graphHeap->at(i).isVisible());
        yoff.append(graphHeap->at(i).yOffset());
    }

    *graphBuffer = *graphHeap;
    metricsBuffer = scanMetrics;

    graphHeap->clear();
    scanMetrics.clear();
    nFiles = 0;

    int n = addDiaFile(QFileInfo(completeFileName), fileUid, min, ll, tl);

    if (n <= 0) {
        // if reloading failed (probably due to timing issues), we roll back the
        // buffered version of the file to avoid a blank screen
        *graphHeap = *graphBuffer;
        scanMetrics = metricsBuffer;
    }

    for (int j = 0; j < graphHeap->size();  ++j) {
        if (j < vis.size())  {
            (*graphHeap)[j].setVisible(vis.at(j));
        }

        if (j < yoff.size()) {
            if ((*graphHeap)[j].yOffset() >= 0.0) {
                // skip negative scans, otherwise the difference curve
                // would not move closer to the base line as the fit
                // gets better
                (*graphHeap)[j].setYoffset(yoff.at(j));
            }
        }
    }

    countFiles();
    return n;
}

void GraphDataModel::appendScan(const Scan &s)
{
    graphHeap->append(s);
    countFiles();
}

void GraphDataModel::appendHklScan(const Scan &s)
{
    Scan hklScan(s);
    hklScan.setHklXunit("dnm");
    normalizeHklScan(&hklScan);
    graphHeap->append(hklScan);
}

void GraphDataModel::replaceScan(int n, const Scan &s)
{
    if ((n >= graphHeap->size()) || (n < 0)) {
        appendScan(s);
    } else {
        graphHeap->replace(n, s);
        (*graphHeap)[n].setUid(s.uid());
        updateMetrics();
    }
}

void GraphDataModel::replaceHklScan(int n, const Scan &s)
{
    Scan hklScan(s);
    hklScan.setHklXunit("dnm");
    normalizeHklScan(&hklScan);
    replaceScan(n, hklScan);
}

void GraphDataModel::removeScan(int n)
{
    if ((n >= graphHeap->size()) || (n < 0)) return;
    graphHeap->remove(n);
}

void GraphDataModel::removeScan(const QUuid &uid)
{
    for (int i = graphHeap->size() - 1; i >= 0; --i) {
        if (graphHeap->at(i).uid() == uid) {
            graphHeap->remove(i);
            return;
        }
    }
}

void GraphDataModel::updateMetrics()
{
    globalXrangeMin = std::numeric_limits<double>::max();
    globalXrangeMax = std::numeric_limits<double>::min();
    globalYrangeMin = std::numeric_limits<double>::max();
    globalYrangeMax = std::numeric_limits<double>::min();

    QMapIterator<QUuid, QVector<double> > it(scanMetrics);

    while (it.hasNext()) {
        it.next();
        globalXrangeMin = qMin(globalXrangeMin, it.value().at(0));
        globalXrangeMax = qMax(globalXrangeMax, it.value().at(1));
        globalYrangeMin = qMin(globalYrangeMin, it.value().at(2));
        globalYrangeMax = qMax(globalYrangeMax, it.value().at(3));
    }
}

void GraphDataModel::normalizeHklScan(Scan *scan)
{
    double _ymaxHkl = 0.0;

    for (int i = 0; i < scan->pDataHkl().size(); ++i) {
        _ymaxHkl = qMax(_ymaxHkl, scan->pDataHkl().at(i).intensity());
    }

    double _ymaxScan = qFuzzyIsNull(globalYrangeMax) ? 100.0 : globalYrangeMax;

    for (int i = 0; i < scan->pDataHkl().size(); ++i) {
        double _d = scan->pDataHkl()[i].intensity();
        scan->pDataHkl()[i].setIntensity(_ymaxScan * _d / _ymaxHkl);
    }
}

Scan * GraphDataModel::getScan(int i)
{
    if ((i >= 0) && (i < graphHeap->size())) {
        return &((*graphHeap)[i]);
    }

    return nullptr;
}

Scan * GraphDataModel::getScan(const QUuid &u)
{
    for (int i = 0; i < graphHeap->size(); ++i) {
        if (graphHeap->at(i).uid() == u) {
            return &((*graphHeap)[i]);
        }
    }

    return nullptr;
}

Scan * GraphDataModel::getFirst()
{
    if (graphHeap->size()) {
        return &((*graphHeap)[0]);
    }

    return nullptr;
}

Scan * GraphDataModel::getLast()
{
    if (graphHeap->size()) {
        return &((*graphHeap)[graphHeap->size() - 1]);
    }

    return nullptr;
}

const Scan * GraphDataModel::first()
{
    if (graphHeap->size()) {
        return &(graphHeap->first());
    }

    return nullptr;
}

const Scan * GraphDataModel::last()
{
    if (graphHeap->size()) {
        return &(graphHeap->last());
    }

    return nullptr;
}

const Scan * GraphDataModel::at(int i)
{
    if ((i >= 0) && (i < graphHeap->size())) {
        return &(graphHeap->at(i));
    }

    return nullptr;
}

const QVector<const Scan *> GraphDataModel::xyScans()
{
    QVector<const Scan *> vec;

    for (int i = 0; i < graphHeap->size(); ++i) {
        if (graphHeap->at(i).hasScanData()) {
            vec.append(&(graphHeap->at(i)));
        }
    }

    return vec;
}

const QVector<const Scan *> GraphDataModel::hklScans()
{
    QVector<const Scan *> vec;

    for (int i = 0; i < graphHeap->size(); ++i) {
        if (graphHeap->at(i).hasHklData()) {
            vec.append(&(graphHeap->at(i)));
        }
    }

    return vec;
}

const QVector<const Scan *> GraphDataModel::allScans()
{
    QVector<const Scan *> vec;

    for (int i = 0; i < graphHeap->size(); ++i) {
        vec.append(&(graphHeap->at(i)));
    }

    return vec;
}

int GraphDataModel::indexOf(const Scan &s) const
{
    if (!graphHeap->size()) return -1;

    for (int i = 0; i < graphHeap->size(); ++i) {
        if (graphHeap->at(i).uid() == s.uid()) {
            return i;
        }
    }

    return -1;
}

int GraphDataModel::indexOf(const QUuid &u) const
{
    if (!graphHeap->size()) return -1;

    for (int i = 0; i < graphHeap->size(); ++i) {
        if (graphHeap->at(i).uid() == u) {
            return i;
        }
    }

    return -1;
}

void GraphDataModel::clear()
{
    graphHeap->clear();
    graphBuffer->clear();
    strucRefls->clear();
    integralRanges->clear();
    integralRangeNames.clear();
    completeFileName = QString();
    fileUid = QString();
    sampleId = QString();
}

QUuid GraphDataModel::scanUid(int i) const
{
    if ((i >= 0) && (i < graphHeap->size())) {
        return graphHeap->at(i).uid();
    }

    return QUuid();
}

void GraphDataModel::setAngularCorrections(double e1, double e2, double e3)
{
    eps1 = e1;
    eps2 = e2;
    eps3 = e3;
}

double GraphDataModel::getAngularCorrEPS1() const
{
    return eps1;
}

double GraphDataModel::getAngularCorrEPS2() const
{
    return eps2;
}

double GraphDataModel::getAngularCorrEPS3() const
{
    return eps3;
}

bool GraphDataModel::sortScanUp(int n)
{
    if ((n >= graphHeap->size()) || (n < 1)) {
        return false;
    }

    Scan scan = graphHeap->takeAt(n);
    graphHeap->insert(n - 1, scan);

    return true;
}

bool GraphDataModel::sortScanDown(int n)
{
    if ((n >= graphHeap->size() - 1) || (n < 0)) {
        return false;
    }

    Scan scan = graphHeap->takeAt(n);
    graphHeap->insert(n + 1, scan);

    return true;
}


void GraphDataModel::yOffsetUp()
{
    double initial = 2.0 * getYmax() / 3.0;
    int n = 1;

    for (int i = 1; i < graphHeap->size(); ++i) {
        Scan *scan = &(*graphHeap)[i];
        if (!scan) continue;

        if (scan->yOffset() == 0.0) {
            scan->setYoffset(initial * double(n));
            ++n;
        }

        if (scan->yOffset() > 0.0) {
            scan->setYoffset(scan->yOffset() * 1.5);
        }
    }

    stacked = true;
}

void GraphDataModel::yOffsetDown()
{
    if (stacked) {
        for (int i = 1; i < graphHeap->size(); ++i) {
            Scan *scan = &(*graphHeap)[i];
            if (!scan) continue;

            if (scan->yOffset() > 0.0) {
                scan->setYoffset(scan->yOffset() / 1.5);
            }
        }
    } else {
        // if not stacked, this function will displace each scan to
        // the maximum intensity of the previous scan. Scans with
        // negative yOffset are ignored
        int p = 0;

        for (int i = 1; i < graphHeap->size(); ++i) {
            Scan *prevScan = &(*graphHeap)[p];
            Scan *scan = &(*graphHeap)[i];

            if (!scan || !prevScan) continue;

            if (scan->yOffset() >= 0.0) {
                double pMx = prevScan->maxIntensity();
                if (pMx < 0.0) pMx = 0.0;

                double d = pMx * prevScan->scaleFactor() + prevScan->yOffset();

                if (d > 0.0) scan->setYoffset(d);
                p = i;
            }
        }
        stacked = true;
    }
}

void GraphDataModel::xOffsetRight()
{
    double initial = (getXmax() - getXmin()) / 100.0;
    int n = 1;

    for (int i = 1; i < graphHeap->size(); ++i) {
        Scan *scan = &(*graphHeap)[i];
        if (!scan) continue;

        if (scan->xOffset() == 0.0) {
            scan->setXoffset(initial * double(n));
            ++n;
        }

        if (scan->xOffset() > 0.0) {
            scan->setXoffset(scan->xOffset() * 1.5);
        }

        if (scan->xOffset() < 0.0) {
            scan->setXoffset(scan->xOffset() / 1.5);
        }
    }

    stacked = true;
}

void GraphDataModel::xOffsetLeft()
{
    double initial = -(getXmax() - getXmin()) / 100.0;
    int n = 1;

    for (int i = 1; i < graphHeap->size(); ++i) {
        Scan *scan = &(*graphHeap)[i];
        if (!scan) continue;

        if (scan->xOffset() == 0.0) {
            scan->setXoffset(initial * double(n));
            ++n;
        }

        if (scan->xOffset() < 0.0) {
            scan->setXoffset(scan->xOffset() * 1.5);
        }

        if (scan->xOffset() > 0.0) {
            scan->setXoffset(scan->xOffset() / 1.5);
        }
    }

    stacked = true;
}

void GraphDataModel::resetOffset()
{
    for (int i = 0; i < graphHeap->size(); ++i) {
        Scan *scan = &(*graphHeap)[i];
        if (!scan) continue;

        scan->setXoffset(0.0);

        if (scan->yOffset() > 0.0) {
            // don't touch negative scans
            scan->setYoffset(0.0);
        }
    }

    stacked = false;
}

double GraphDataModel::getXmin() const
{
    double xmin = std::numeric_limits<double>::max();

    for (int i = 0; i < graphHeap->size(); ++i) {
        xmin = qMin(xmin, graphHeap->at(i).minAngle());
    }

    return xmin;
}

double GraphDataModel::getXmax() const
{
    double xmax = std::numeric_limits<double>::min();

    for (int i = 0; i < graphHeap->size(); ++i) {
        xmax = qMax(xmax, graphHeap->at(i).maxAngle());
    }

    return xmax;
}

double GraphDataModel::getYmin() const
{
    double xmin = std::numeric_limits<double>::max();

    for (int i = 0; i < graphHeap->size(); ++i) {
        double xm = graphHeap->at(i).minIntensity();
        if (xm >= 0.0) xmin = qMin(xmin, xm);
    }

    return xmin < 0.0 ? 0.0 : xmin;
}

double GraphDataModel::getYmax() const
{
    double xmax = std::numeric_limits<double>::min();

    for (int i = 0; i < graphHeap->size(); ++i) {
        double xm = graphHeap->at(i).maxIntensity();
        if (xm >= 0.0) xmax = qMax(xmax, xm);
    }

    return xmax < 0.0 ? 0.0 : xmax;
}

void GraphDataModel::countFiles()
{
    QSet<QString> fn;

    for (int i = 0; i < graphHeap->size(); ++i) {
        fn.insert(graphHeap->at(i).sourceFileName());
    }

    nFiles = fn.size();
}

QVector<Scan *> GraphDataModel::activeScans()
{
    QVector<Scan *> vec;

    for (int i = 0; i < graphHeap->size(); ++i) {
        Scan *sc = &(*graphHeap)[i];
        if (sc->isActive()) vec.append(sc);
    }

    return vec;
}

void GraphDataModel::setActiveScans(const QList<int> &l)
{
    for (int i = 0; i < graphHeap->size(); ++i) {
        Scan *sc = &(*graphHeap)[i];
        if (sc) sc->setActive(l.contains(i));
    }
}

void GraphDataModel::setActiveScans(const QList<QUuid> &l)
{
    for (int i = 0; i < graphHeap->size(); ++i) {
        Scan *sc = &(*graphHeap)[i];
        if (sc) sc->setActive(l.contains(sc->uid()));
    }
}

QList<int> GraphDataModel::activeScanIndices() const
{
    QList<int> l;

    for (int i = 0; i < graphHeap->size(); ++i) {
        if (graphHeap->at(i).isActive()) l.append(i);
    }

    return l;
}

void GraphDataModel::clearActiveScans() const
{
    for (int i = 0; i < graphHeap->size(); ++i) {
        Scan *sc = &(*graphHeap)[i];
        if (sc) sc->setActive(false);
    }
}
