/***************************************************************************
                          graphwindow.cpp  -  description
                             -------------------
    begin                : Die Jun 17 2003
    copyright            : (C) 2003 by Nicola Doebelin
    email                : ndoebelin@gmx.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "graphwindow.h"

#include "../../../libXrdIO/import/importhandler.h"
#include "../../../libXrdIO/import/bgmndiaimport.h"
#include "../../../libXrdIO/structs.h"
#include "../../../libXrdIO/scanops.h"

#include <QApplication>
#include <QClipboard>
#include <QEvent>
#include <QHelpEvent>
#include <QToolTip>
#include <QPrinter>
#include <QTime>
#include <QSvgGenerator>
#include <QPolygon>
#include <QDesktopWidget>
#include <QToolTip>
#include <QStaticText>
#include <QDebug>

#include <limits>

#ifndef M_PI
    #define M_PI acos(-1.0)
#endif

GraphWindow::GraphWindow(GraphDataController *c, QWidget *parent)
    : AbstractGraphView(c, parent)
{
    QSizePolicy sp = sizePolicy();
    sp.setControlType(QSizePolicy::Frame);
    sp.setVerticalPolicy(QSizePolicy::Expanding);
    sp.setHorizontalPolicy(QSizePolicy::Expanding);
    setSizePolicy(sp);

    setMouseTracking(true);
    recalcPlot = true;
    darkTheme = false;

    ihandler = new ImportHandler();
    rubberBand = new QRubberBand(QRubberBand::Rectangle, this);
    legend = new GraphWindowLegend(c, this);
    connect(legend, SIGNAL(sigRequestRedraw()), this, SLOT(forceUpdate()));

    contextMenuXaxis    = new QMenu(this);
    contextMenuYaxis    = new QMenu(this);
    contextMenuDataArea = new QMenu(this);

    contextMenuXaxis->addAction(QString(tr("2%1 Scale")).arg(global::theta), this, SLOT(setXaxis2theta()));
    contextMenuXaxis->addAction(tr("d Spacing Scale"), this, SLOT(setXaxisD()));
    contextMenuXaxis->addAction(tr("Reset Zoom"), this, SLOT(resetZoomX()));

    contextMenuYaxis->addAction(tr("Linear Scale"), this, SLOT(setYaxisLinear()));
    contextMenuYaxis->addAction(tr("Sqrt Scale"), this, SLOT(setYaxisSqrt()));
    contextMenuYaxis->addAction(tr("Log10 Scale"), this, SLOT(setYaxisLog10()));
    contextMenuYaxis->addSeparator();
    contextMenuYaxis->addAction(tr("Zoom to Baseline"), this, SLOT(zoomFromZeroY()));
    contextMenuYaxis->addAction(tr("Zoom to Minimum Intensity"), this, SLOT(zoomFromMinY()));
    contextMenuYaxis->addAction(tr("Reset Zoom"), this, SLOT(resetZoomY()));

    pBuffer = QPixmap(width(), height());

    crossHair = false;
    noiseCursor = false;
    strucScaling = false;
    inspector = false;
    drawRubberBand = false;
    drawLivePeakIntegral = false;
    dragging = false;
    zoomed = false;
    specLines = false;
    useBackgroundColors = false;
    hasActiveIntegrationRange = false;
    firstDoubleClickPoint = QPoint();
    peakPreviewMode = PPMNONE;
    rangeSelectMode = false;
    fillActive = true;
    hasOverrideWaveLength = false;

    maxi = 0.0;
    mini = 0.0;
    points = 0;
    maxang = 0.0;
    minang = 0.0;
    difOffset = 0.0;

    waveLength[0] = 0.0;
    waveLength[1] = 0.0;
    waveLength[2] = 0.0;

    // sample height displacement, as used by BGMN's EPS2 parameter
    // delta = sample displacement in mm
    // iRadius = goniometer radius in mm
    //
    // EPS2 = 2delta / iRadius
    eps1 = 0.0;
    eps2 = 0.0;
    eps3 = 0.0;

    pScale = 1.0;
    scanMaxIntensity = 0.0;
    prevStrucScaleFactor = 1.0;
    prevScanScaleFactor = 1.0;
    integrationRangeStart = -1.0;
    integrationRangeEnd = -1.0;

    completeFileName = QString();
    fileName = QString();
    fileUid = QString();

    colorTable.append(QColor(0, 0, 0, 255));
    colorTable.append(QColor(255, 0, 0, 255));
    colorTable.append(QColor(0, 0, 255, 255));
    colorTable.append(QColor(0, 0, 0, 255));

    updateCoordinateTimer.setSingleShot(true);

    xScaling = XSCALETWOTHETA;
    yScaling = YSCALELIN;
    antiAliasing = false;
    showGridMinor = false;
    showGridMajor = false;

    initSettings();
}

// Destructor
GraphWindow::~GraphWindow()
{
    delete ihandler;

    delete contextMenuXaxis;
    delete contextMenuYaxis;
    delete contextMenuDataArea;
    delete legend;
}

void GraphWindow::initSettings()
{
//    tungstenLines.insert(QString("WL%1%2").arg(global::beta).arg(QChar(0x2084)), 1.30162);  // WLb4
//    tungstenLines.insert(QString("WL%1%2").arg(global::beta).arg(QChar(0x2083)), 1.26269);  // WLb3
    tungstenLines.insert(QString("WL%1%2").arg(global::beta).arg(global::subOne), 1.281809); // WLb1
//    tungstenLines.insert(QString("WL%1%2").arg(global::alpha).arg(QChar(0x2082)), 1.48743);  // WLa2
    tungstenLines.insert(QString("WL%1%2").arg(global::alpha).arg(global::subOne), 1.47639);  // WLa1
//    tungstenLines.insert(QString("WL%1%2").arg(global::beta).arg(QChar(0x2086)), 1.28989);  // WLb6
//    tungstenLines.insert(QString("WL%1%2%3").arg(global::beta).arg(QChar(0x2081)).arg(QChar(0x2085)), 1.24631);  // WLb15
//    tungstenLines.insert(QString("WL%1%2").arg(global::beta).arg(QChar(0x2082)), 1.24460);  // WLb2
//    tungstenLines.insert(QString("WL%1%2").arg(global::beta).arg(QChar(0x2087)), 1.22400);  // WLb7
//    tungstenLines.insert(QString("WL%1%2").arg(global::beta).arg(QChar(0x2085)), 1.21545);  // WLb5
//    tungstenLines.insert(QString("WL%1%2%3").arg(global::beta).arg(QChar(0x2081)).arg(QChar(0x2080)), 1.21218);  // WLb10
//    tungstenLines.insert(QString("WL%1%2").arg(global::beta).arg(QChar(0x2089)), 1.20479);  // WLb9

    oFontTitle.fromString(settings->value("graph/fontTitle", font().toString()).toString());
    oFontAxis.fromString(settings->value("graph/fontAxis", font().toString()).toString());
    oFontTicks.fromString(settings->value("graph/fontTicks", font().toString()).toString());
    oFontLegend.fromString(settings->value("graph/fontLegend", font().toString()).toString());
    lineWidth = settings->value("graph/lineWidth", 1).toInt();
    symbolSize = settings->value("graph/crossSize", 1).toInt();
    printingLineWidth = settings->value("graph/printingLineWidth", 3).toInt();
    legend->setDrawn(settings->value("graph/drawLegend", true).toBool());
    drawRefls = settings->value("graph/drawHkl", true).toBool();
    antiAliasing = settings->value("graph/antiAliasing", false).toBool();
    fillActive = settings->value("graph/fillActiveScan", true).toBool();

    int xUnit = settings->value("graph/xAxisUnit", 0).toInt();
    int yScale = settings->value("graph/yAxisScale", 0).toInt();
    xScaling = static_cast<Scale>(xUnit + 4); // convert to enum Scale
    yScaling = static_cast<Scale>(yScale);    // convert to enum Scale

    if (qFuzzyIsNull(waveLength[0])) {
        scanControl->getFixedWaveLength(waveLength[0], waveLength[1], waveLength[2], settings->defaultWavelength());
    }
    useCountsPerSecond = settings->value("graph/countsPerSecond", false).toBool();
    showGridMinor = settings->value("graph/showMinorGridLines", false).toBool();
    showGridMajor = settings->value("graph/showMajorGridLines", false).toBool();

    tickDensityX = settings->value("graph/tickMarkDensityX", 10).toInt();
    tickDensityY = settings->value("graph/tickMarkDensityY", 5).toInt();

    fontTitle = oFontTitle;
    fontAxis = oFontAxis;
    fontTicks = oFontTicks;
    fontLegend = oFontLegend;

    QList<QVariant> cl = settings->value("graph/colorTable", QList<QVariant>()).toList();
    QList<QVariant> sl = settings->value("graph/styleTable", QList<QVariant>()).toList();
    int hklStl = settings->value("config/hklSymbol", 0).toInt();

    for (int i = 0; i < qMax(cl.size(), sl.size()); ++i) {
        QColor col = i >= cl.size() ? settings->getRandomColor(i) : QColor(cl.at(i).toString());
        int    stl = i >= sl.size() ? 0 : sl.at(i).toInt();
        scanStyles.append(ScanStyle(col, stl, hklStl));
    }

    useBackgroundColors = settings->value("graph/useBackgroundColors", true).toBool();
    darkTheme = settings->isDarkMode();

    if (darkTheme) bgColor = QGuiApplication::palette().color(QPalette::Base);
    else           bgColor = Qt::white;

    printingBgColor = Qt::white;
    activeColor = darkTheme ? bgColor : QColor(settings->value("graph/activeColor", "#ff9999").toString());
    completedColor = darkTheme ? bgColor : QColor(settings->value("graph/completeColor", "#99ff99").toString());
    peakSelectColor = darkTheme ? bgColor : QColor(settings->value("graph/peakSelectColor", "#c8c8ff").toString());
    idleColor = bgColor;
    marginColor = idleColor;
    axisColor = QColor(settings->value("graph/axisColor", "#000000").toString());
    integralRangeColor = QColor(settings->value("graph/integralRangeColor", QString("#cccccc")).toString());
    integralCenterLineColor = QColor(settings->value("graph/integralCenterLineColor", QString("#999999")).toString());

    if (!settings->value("bgmnProject/readLamFile", false).toBool()) {
         scanControl->clearCustomWaveLength();
    }

    legend->initSettings();
    changeGraphStyles();
}

void GraphWindow::updateView()
{
    eps1 = scanControl->getAngularCorrEPS1();
    eps2 = scanControl->getAngularCorrEPS2();
    eps3 = scanControl->getAngularCorrEPS3();
    setupNewScans();
    forceUpdate();
}

void GraphWindow::resetView()
{
    hasOverrideWaveLength = false;
}

void GraphWindow::resizeEvent(QResizeEvent *)
{
      pBuffer = QPixmap(width(), height());
      recalcPlot = true;
}

void GraphWindow::paintEvent(QPaintEvent *e)
{
    if (pBuffer.width() == 0 || pBuffer.height() == 0) {
        return;
    }

    QPainter p;

    if (recalcPlot) {
      p.begin(&pBuffer);
        calcMargins(p, true);
        initPlot(pBuffer.rect());

        p.setRenderHint(QPainter::Antialiasing, antiAliasing);
        p.setRenderHint(QPainter::TextAntialiasing, true);
        p.setRenderHint(QPainter::SmoothPixmapTransform, true);

        if (useBackgroundColors) {
            QRegion widgetRegion(this->rect());
            QRegion plotRegion(plot);
            p.setClipRegion(widgetRegion.subtracted(plotRegion));
            p.fillRect(this->rect(), marginColor);
            p.setClipping(false);
            p.fillRect(plot.x(), plot.y(), plot.width(), plot.height(), bgColor);
        } else {
            p.fillRect(this->rect(), bgColor);
        }

        drawWindow(p);
        if (hasActiveIntegrationRange) drawMouseBackgroundRegion(p, getXcoord(integrationRangeStart), getXcoord(integrationRangeEnd));
        drawHighlightedRegions(p);
        drawPlot(p, false);
        drawAnchorPoints(p);
        drawHklTicks(p);
        drawHklScans(p);
        drawHklReferenceLines(p);
        drawTicks(p);
        drawLabels(p);
        drawLegend(p);
      p.end();
    }

    p.begin(this);
        p.drawPixmap(e->rect(), pBuffer, e->rect());

        p.setClipRect(plot);

        // draw temporary overlays
        if (crossHair)                  drawCrossHairCursor(p);
        if (noiseCursor)                drawNoiseCursor(p);
        if (specLines)                  drawSpecLineCursor(p);
        if (inspector)                  drawInspector(p);
        if (peakPreviewMode != PPMNONE) drawPeakPreviews(p);
        if (rangeSelectMode)            drawRange(p);
    p.end();

    recalcPlot = false;
}

void GraphWindow::drawCrossHairCursor(QPainter &p)
{
    QPen pen(darkTheme ? Qt::white : Qt::black);
    pen.setStyle(Qt::DashLine);
    pen.setWidth(lineWidth);
    p.setPen(pen);

    QPoint pt = mapFromGlobal(cursor().pos());
    p.drawLine(plot.x(), pt.y(), plot.width() + plot.x() + 1, pt.y());
    p.drawLine(pt.x(), plot.y(), pt.x(), plot.height() + plot.y() + 1);
}

void GraphWindow::drawNoiseCursor(QPainter &p)
{
    if (!scanControl->count()) return;
    QFontMetrics fml(fontLegend);
    QPen pen(darkTheme ? Qt::white : Qt::black);
    pen.setStyle(Qt::SolidLine);
    pen.setWidth(lineWidth);
    p.setPen(pen);

    int px = mapFromGlobal(cursor().pos()).x();
    int py = mapFromGlobal(cursor().pos()).y();
    double yc = getRealY(py);
    bool dCurve = false;

    if (yc < 0.0) {
        // on difference curve use the intensity of the active scan
        Scan *sc = scanControl->firstActiveScan();
        double xc = getRealX(px);
        int angIdx = sc->indexOfAngle(xc + angularCorrection(xc) - sc->xOffset(), 0);
        yc = sc->intensity(angIdx);
        dCurve = true;
    } else {
        if (yScaling == YSCALELOG10) yc = getInvLog10(yc);
        if (yScaling == YSCALESQRT)  yc = getInvSqrt(yc);
    }

    if (yc > 0.0) {
        int endWidth = fml.width("m");
        int midWidth = endWidth / 2;

        int yTop = 0;
        int yCnt = 0;
        int yBot = 0;
        int yDif = 0;

        if (yScaling == YSCALELIN) {
            yTop = getYcoord(yc + sqrt(yc));
            yCnt = getYcoord(yc);
            yBot = getYcoord(yc - sqrt(yc));
            yDif = getYcoord(sqrt(yc));
        }

        if (yScaling == YSCALESQRT) {
            yTop = getYcoord(getSqrt(yc + sqrt(yc)));
            yCnt = getYcoord(getSqrt(yc));
            yBot = getYcoord(getSqrt(yc - sqrt(yc)));
            yDif = getYcoord(getSqrt(sqrt(yc)));
        }

        int pyt = py + yTop - yCnt;
        int pyc = py;
        int pyb = py - yCnt + yBot;
        int pyd = yDif - getYcoord(0.0);

        if (dCurve) {
            // on the difference curve
            p.drawLine(px,          pyc+pyd, px,          pyc-pyd); // vertical
            p.drawLine(px-endWidth, pyc-pyd, px+endWidth, pyc-pyd); // upper end
            p.drawLine(px-midWidth, pyc,     px+midWidth, pyc);     // center
            p.drawLine(px-endWidth, pyc+pyd, px+endWidth, pyc+pyd); // lower end
        } else {
            // in the positive area
            p.drawLine(px,          pyt, px,          pyb); // vertical
            p.drawLine(px-endWidth, pyt, px+endWidth, pyt); // upper end
            p.drawLine(px-midWidth, pyc, px+midWidth, pyc); // center
            p.drawLine(px-endWidth, pyb, px+endWidth, pyb); // lower end
        }
    }
}

void GraphWindow::drawSpecLineCursor(QPainter &p)
{
    QFontMetrics fml(fontLegend);
    QPen pen;
    pen.setStyle(Qt::DashLine);
    pen.setWidth(lineWidth);
    p.setFont(fontLegend);

    double twotheta = getRealX(mapFromGlobal(cursor().pos()).x());
    double d = global::Functions::twoThetaToD(twotheta, waveLength[0]);

    QMap<QString, double> wlengths;

    if (!scanControl->getCustomWaveLength()->isEmpty()) {
        pen.setColor(darkTheme ? Qt::white : Qt::black);
        p.setPen(pen);

        wlengths = *(scanControl->getCustomWaveLength());
    } else if (waveLength[0] > 0.0) {
        pen.setColor(darkTheme ? global::Functions::colorToDarkMode(Qt::blue) : Qt::blue);
        p.setPen(pen);

        wlengths[QString("K%1%2").arg(global::alpha).arg(global::subOne)] = waveLength[0];
        wlengths[QString("K%1%2").arg(global::alpha).arg(global::subTwo)] = waveLength[1];
        wlengths[QString("K%1").arg(global::beta)] = waveLength[2];
    }

    if (!wlengths.isEmpty()) {
        p.setClipRect(plot.x() + lineWidth, plot.y() + lineWidth, plot.width()-lineWidth, plot.height()-lineWidth);

        QMap<QString, double>::ConstIterator itw = wlengths.constBegin();

        while (itw != wlengths.constEnd()) {
            int ttip = getXcoord(global::Functions::dToTwoTheta(d, itw.value())) + plot.x();

            p.drawLine(ttip, plot.y() + 3 * fml.ascent(), ttip, plot.height() + plot.y() + 1);
            p.drawText(ttip - fml.width(itw.key()) / 2, plot.y() + 2 * fml.ascent(), itw.key());

            pen.setColor(darkTheme ? Qt::white : Qt::black);
            p.setPen(pen);

            ++itw;
        }

        if (settings->value("graph/showTungstenSpecLines", false).toBool()) {
            pen.setColor(darkTheme ? global::Functions::colorToDarkMode(Qt::gray) : Qt::gray);
            p.setPen(pen);

            QMap<QString, double>::ConstIterator itt = tungstenLines.constBegin();

            while (itt != tungstenLines.constEnd()) {
                int wline  = getXcoord(global::Functions::dToTwoTheta(d, itt.value())) + plot.x();
                p.drawLine(wline, plot.y() + 3 * fml.ascent(), wline, plot.height() + plot.y() + 1);
                p.drawText(wline - fml.width(itt.key()) / 2, plot.y() + 2 * fml.ascent(), itt.key());
                ++itt;
            }
        }
    }
}

void GraphWindow::drawInspector(QPainter &p)
{
    QPoint pt(mapFromGlobal(cursor().pos()));

    if (!plot.contains(pt)) return;

    QFontMetrics fms(fontLegend);
    int size = fms.width("n");

    QPen pen(darkTheme ? Qt::white : Qt::black);
    pen.setStyle(Qt::SolidLine);
    pen.setWidth(lineWidth);
    p.setPen(pen);

    int n = scanControl->hasActiveScan() ? 0 : scanControl->activeScanIncides().first();
    double rc = getRealX(pt.x());

    int angIdx = scanControl->at(n)->indexOfAngle(rc + angularCorrection(rc) - scanControl->at(n)->xOffset(), 0);

    if (angIdx < 0) return;

    double tps = 1.0;

    if (useCountsPerSecond) {
        tps = scanControl->at(n)->timePerStep() > 0.0 ? scanControl->at(n)->timePerStep() : 1.0;
    }

    double xc  = scanControl->at(n)->angle(angIdx);
    double yc  = scanControl->at(n)->intensity(angIdx) * scanControl->at(n)->scaleFactor() / tps;

    QString s_deg = QString("%1%2%3").arg(global::degree).arg("2").arg(global::theta);
    QString s_unt = useCountsPerSecond ? "cps" : "counts";
    QString str   = QString("%1 %2 / %3 %4")
            .arg(xc, 0, 'f', 4)
            .arg(s_deg)
            .arg(yc, 0, 'f', 2)
            .arg(s_unt);

    int vpos = pt.y() - size;
    if (vpos < plot.top() + fms.height()) vpos = pt.y() + size + fms.ascent();

    int hpos = pt.x() + size;
    if (hpos > plot.right() - fms.width(str) - size) hpos = pt.x() - size - fms.width(str);

    p.drawText(hpos, vpos, str);

    double yo = scanControl->at(n)->yOffset();

    if (yScaling == YSCALESQRT)  {
        yo = getSqrt(yo);
        yc = getSqrt(yc);
    }

    if (yScaling == YSCALELOG10) {
        yo = getLog10(yo);
        yc = getLog10(yc);
    }

    int sx = getXcoord(xc - angularCorrection(xc) + scanControl->at(n)->xOffset()) + plot.x();
    int sy = getYcoord(yc + yo) + plot.y();

    p.drawLine(sx - size, sy - size, sx + size, sy + size);
    p.drawLine(sx - size, sy + size, sx + size, sy - size);
    p.drawLine(sx,        sy + size, sx,        sy - size);
    p.drawLine(sx - size, sy,        sx + size, sy);
}

void GraphWindow::drawPeakPreviews(QPainter &p)
{
    QPoint pt(mapFromGlobal(cursor().pos()));

    if (firstDoubleClickPoint.isNull()) {
        if (peakPreviewMode == PPMLIN)   emit sigCursorMessage(tr("Double-click to set first point"));
        if (peakPreviewMode == PPMPEAK)  emit sigCursorMessage(tr("Double-click to set tip of peak"));
    } else {
        if (peakPreviewMode == PPMLIN)   emit sigCursorMessage(tr("Double-click to set second point"));
        if (peakPreviewMode == PPMPEAK)  emit sigCursorMessage(tr("Double-click to set base of peak"));
    }

    if (firstDoubleClickPoint.isNull()) return;

    QPen pen(darkTheme ? Qt::white : Qt::black);
    pen.setStyle(Qt::DashLine);
    pen.setWidth(lineWidth);

    p.save();
    p.setCompositionMode(QPainter::CompositionMode_Multiply);
    p.setPen(pen);

    if (peakPreviewMode == PPMLIN) {
        p.drawLine(firstDoubleClickPoint.x(), firstDoubleClickPoint.y(), pt.x(), pt.y());
    }

    if (peakPreviewMode == PPMPEAK) {
        int dx = pt.x() - firstDoubleClickPoint.x();
        int dy = pt.y() - firstDoubleClickPoint.y();
        p.drawLine(firstDoubleClickPoint.x(),
                   firstDoubleClickPoint.y(),
                   firstDoubleClickPoint.x() - dx,
                   firstDoubleClickPoint.y() + dy);
        p.drawLine(firstDoubleClickPoint.x(),
                   firstDoubleClickPoint.y(),
                   firstDoubleClickPoint.x() + dx,
                   firstDoubleClickPoint.y() + dy);
        p.drawLine(firstDoubleClickPoint.x() - int(0.5*dx),
                   firstDoubleClickPoint.y() + int(0.5*dy),
                   firstDoubleClickPoint.x() + int(0.5*dx),
                   firstDoubleClickPoint.y() + int(0.5*dy));
    }

    p.restore();
}

void GraphWindow::drawRange(QPainter &p)
{
    if (firstDoubleClickPoint.isNull()) {
        emit sigCursorMessage(tr("Double-click to set start of range"));
    } else {
        emit sigCursorMessage(tr("Double-click to set end of range"));
    }

    if (firstDoubleClickPoint.isNull()) return;

    QPoint ptA(firstDoubleClickPoint.x(), plot.top());
    QPoint ptB(mapFromGlobal(cursor().pos()).x(), plot.bottom());

    QColor rgCol(darkTheme ? global::Functions::colorToDarkMode(integralRangeColor) : integralRangeColor);

    p.save();

    p.setCompositionMode(QPainter::CompositionMode_Multiply);
    p.fillRect(QRect(ptA, ptB), rgCol);

    p.restore();
}

/*
 * calculates the margins as a function of font sizes.
 * three constants are required:
 * margin_o: outer margin between border and axis labels / title, will be
 *           set to 0 for exporting bitmaps
 * margin_m: space between axis lables and tick lables
 * margin_i: inner space between lables and axes
 */
void GraphWindow::calcMargins(const QPainter &p, bool border)
{
    // painter seems to be inactive
    if (!p.isActive()) return;

    QFontMetrics fmTitle(fontTitle);
    QFontMetrics fmAxis(fontAxis);
    QFontMetrics fmTicks(fontTicks);

    // add an outer margin (e.g. on screen, but not on printing or rendering)
    if (border) {
        margin_o = fmAxis.height() / 2;
    } else {
        margin_o = 0;
    }

    margin_m = fmAxis.height() / 2;
    margin_i = fmTicks.height() / 2;

    QString stra = QString("%1").arg(maxi, 0, 'f', 0);
    int spc = margin_o + margin_i;
    int vTickLabelLength = qMax(fmTicks.width(stra), fmTicks.width("000000"));
    int spr = int(fmTicks.width("MMM.MM") / 2.0);

    margin_l = fmAxis.height() + vTickLabelLength + margin_m + spc;
    margin_r = spr + margin_o;
    margin_t = fmTitle.height() + spc;
    margin_b = fmAxis.height() + fmTicks.height() + margin_m + spc;
}

/*
 * calculates the size of the plot area
 */
void GraphWindow::initPlot(const QRect &rc)
{
    canvas = rc;
    plot.setRect(margin_l, margin_t, rc.width() - (margin_r + margin_l), rc.height() - (margin_t + margin_b));
    topMarginRect.setRect(plot.left(), 0, plot.width(), plot.top());
    bottomMarginRect.setRect(plot.left(), plot.bottom(), plot.width(), margin_b);
    leftMarginRect.setRect(0, plot.top(), margin_l, plot.height());
    rightMarginRect.setRect(plot.right(), plot.top(), margin_r, plot.height());
}

/*
 * draws axis labels and file name
 */
void GraphWindow::drawWindow(QPainter &p, bool replaceChars)
{
    if (!p.isActive()) return;

    QPen pen;
    QString xa;
    QString ya;

    if (xScaling == XSCALED) {
        if (replaceChars) {
            xa = tr("d Spacing [Angstrom]");
        } else {
            xa = tr("d Spacing [") + QString(global::angstrom) + "]";
        }
    } else {
        if (replaceChars) {
            xa = tr("Diffraction Angle [degrees 2theta]");
        } else {
            xa = tr("Diffraction Angle [") + QString(global::degree) + "2" + QString(global::theta) + "]";
        }
    }

    if (useCountsPerSecond && scanControl->first()->timePerStep() > 0.0) {
        ya = QString(tr("Intensity [cps]"));
    } else {
        ya = QString(tr("Intensity [counts]"));
    }

    pen.setWidth(lineWidth);
    pen.setColor(darkTheme ? global::Functions::colorToDarkMode(axisColor) : axisColor);

    QFontMetrics fmt(fontTitle);
    QFontMetrics fma(fontAxis);

    p.setPen(pen);
    p.setFont(fontTitle);
    p.drawText(margin_o, fmt.ascent() + margin_o, scanControl->getSampleId());

    QString displayName = settings->value("graph/drawCompleteFileName", false).toBool()
            ? QDir::toNativeSeparators(scanControl->fileInfo().absoluteFilePath())
            : scanControl->fileInfo().fileName();

    p.drawText(p.device()->width() - margin_o - fmt.horizontalAdvance(displayName), fmt.ascent() + margin_o, displayName);

    // x-axis
    p.setFont(fontAxis);
    p.drawText(plot.center().x() - fma.horizontalAdvance(xa) / 2, canvas.height() - margin_o - (fma.descent() + 2), xa);

    // y-axis
    int dx = fma.ascent() + margin_o;
    int dy = plot.center().y() + fma.width(ya) / 2;

    p.save();
    p.translate(dx, dy);
    p.rotate(-90);

    p.drawText(0, 0, ya);

    p.restore();
}

void GraphWindow::drawHighlightedRegions(QPainter &p)
{
    if (!p.isActive()) return;
    if (!integralRanges.count()) return;

    QColor clCol(darkTheme ? global::Functions::colorToDarkMode(integralCenterLineColor) : integralCenterLineColor);
    QColor rgCol(darkTheme ? global::Functions::colorToDarkMode(integralRangeColor) : integralRangeColor);

    QPen pen(clCol);
    pen.setWidth(lineWidth);

    QFontMetrics fm(font());
    p.save();
    p.translate(plot.x(), plot.y());
    p.setClipRect(lineWidth, lineWidth, plot.width() - lineWidth, plot.height() - lineWidth);
    p.setPen(pen);
    p.setCompositionMode(QPainter::CompositionMode_Multiply);

    for (int i = 0; i < integralRanges.size() - 1; i += 2) {
        double _sx = getXcoord(integralRanges.at(i).x());
        double _ex = getXcoord(integralRanges.at(i+1).x());
        double _cx = (_sx + _ex) / 2.0;

        p.fillRect(QRectF(_sx, 0.0, _ex - _sx, double(plot.height())), rgCol);
        p.drawLine(QPointF(_cx, 0.0), QPointF(_cx, double(plot.height())));
        p.drawText(QPointF(_sx + fm.width("M"), 2 * fm.ascent()), i / 2 < integralRangeNames.size() ? integralRangeNames.at(i/2) : QString("Range %1").arg(i/2));
    }

    p.restore();
}

void GraphWindow::setIntegralRanges(const QVector<QPointF> &l, const QStringList &n)
{
    integralRanges = l;
    integralRangeNames = n;
}

void GraphWindow::drawMouseBackgroundRegion(QPainter &p, int _startX, int _endX)
{
    if (!p.isActive()) return;

    p.save();
    p.translate(plot.x(), plot.y());
    p.fillRect(_startX, 0, _endX - _startX, plot.height(), darkTheme ? QColor(75, 75, 75) : QColor(180, 180, 180));
    p.restore();
}

/*
 * draws hkl tick marks of scans that contain both xy and hkl data.
 * hkl lines of pure hkl scans (no xy data) will be drawn elsewhere.
 */
void GraphWindow::drawHklTicks(QPainter &p)
{
    tt.clear();

    // painting of hkl reflections set to off: exit
    if (!drawRefls) return;

    // painter not active: exit
    if (!p.isActive()) return;

    // no reflection available: exit
    if (!scanControl->count()) return;

    p.save();

    QFontMetrics fm(fontLegend);
    int markerSize = fm.ascent()/2;

    QRect r;

    // vertical length of the lines, use font size as maximum length,
    // but allow shorter lines
    int len = qMin(fm.ascent(), int(plot.height()/40.0));

    // spacing between tops of the lines
    int vspacing = int(double(len) + 0.5 * double(len));

    int top = int(pScale * 6);
    int bot = top + len;

    QPen pen;
    pen.setWidth(lineWidth);
    p.translate(plot.x(), plot.y());
    p.setClipRect(lineWidth, lineWidth, plot.width() - lineWidth, plot.height() - lineWidth);

    int vpos = 0;

    // loop over all scans and draw their hkl lines
    for (int n = 0; n < scanControl->count(); ++n) {
        const Scan *scan = scanControl->at(n);

        // skip scans that don't contain hkl data or scan data
        if (scan->dataAngle().isEmpty() || scan->dataIntensity().isEmpty() || scan->dataHkl().isEmpty()) {
            continue;
        }

        // skip invisible scans
        if (!scan->isVisible()) {
            continue;
        }

        const QVector<Hkl> &vec = scan->dataHkl();

        if (n < colorTable.size()) {
            pen.setColor(colorTable.at(n));
        } else {
            pen.setColor(scan->color());
        }

        if (darkTheme) pen.setColor(global::Functions::colorToDarkMode(pen.color()));
        p.setPen(pen);

        bool convertToTT = scan->getHklXunit() == "dnm";

        // draw line for reflections
        for (int i = 0; i < vec.size(); ++i) {
            Hkl hkl = vec.at(i);
            int x = getXcoord(convertToTT ? global::Functions::dToTwoTheta(hkl.position(), 0.1*waveLength[0]) : hkl.position());
            int of = vpos * vspacing + top;

            // ignore x-positions outside the displayed range
            if ((x < 0) || x > plot.width()) {
                continue;
            }

            if (hkl.status() == 2) {
                // hkl lines can be highlighted individually. trying to
                // keep the code as fast as possible for all non-highlighted ones
                pen.setWidth(2 * lineWidth);
                p.setPen(pen);
                p.drawLine(x, top + of, x, bot + of);

                p.drawLine(x - markerSize/2, top + of - markerSize, x, top + of);
                p.drawLine(x + markerSize/2, top + of - markerSize, x, top + of);
                p.drawLine(x - markerSize/2, bot + of + markerSize, x, bot + of);
                p.drawLine(x + markerSize/2, bot + of + markerSize, x, bot + of);

                pen.setWidth(lineWidth);
                p.setPen(pen);
            } else {
                p.drawLine(x, top + of, x, bot + of);
            }

            r.setLeft(x - 2 + plot.x());
            r.setRight(x + 2 + plot.x());
            r.setTop(top + plot.y() + of);
            r.setBottom(bot + plot.y() + of);

            QString str;

            if (hkl.phase().isEmpty()) {
                    str = QString("(%1) Texture=%2")
                            .arg(hkl.hkl())
                            .arg(hkl.texture(), 0, 'f', 4);
            } else {
                    str = QString("%1 (%2) Texture=%3")
                            .arg(hkl.phase())
                            .arg(hkl.hkl())
                            .arg(hkl.texture(), 0, 'f', 4);
            }

            if (tt.contains(r.left())) {
                // if a ttip at position r.left() already exists, just append the new string to
                // the existing ttip's string
                if (!tt[r.left()].str.contains(str)) {
                    tt[r.left()].str.append(str);
                }
            } else {
                // if no ttip exists at position r.left(), add a new one
                tTip _ttip;

                _ttip.pos = r;
                _ttip.str = QStringList(str);

                tt[r.left()] = _ttip;
            }
        }

        ++vpos;
    }

    p.restore();
}

/*
 * draw scans that only contain hkl lines
 *
 * if a scan contains scan data AND hkl data, the hkl tick lines are drawn at the top of the scan (not here)
 * if it only contains hkl data, the hkl lines are drawn from the base line with correct (normalized) intensities
 * (here).
 */
void GraphWindow::drawHklScans(QPainter &p)
{
    if (!p.isActive()) return;

    double blMax = 0.0;
    double blOff = 0.0;

    const Scan *bkgrScan = getBackgroundMetrics(bkgrScanNumber, blMax, blOff);

    for (int i = 0; i < scanControl->count(); ++i) {
        const Scan *scan = scanControl->at(i);
        if (!scan->isVisible()) continue; // skip scans set to invisible

        if (scan->dataIntensity().size()) {
            continue; // skip scans with raw data
        }

        if (!scan->dataHkl().size()) {
            continue; // skip scans without hkl data
        }

        int lw = scan->isActive() ? 2 * lineWidth : lineWidth;

        drawHklScanLines(p, scan, bkgrScan, lw);
    }
}

/*
 * draws reference hkl reflections
 *
 * note: reference reflections are stored in d-positions, not 2theta
 */
void GraphWindow::drawHklReferenceLines(QPainter &p)
{
    if (!p.isActive()) return;
    if (!refStructure.hasHklData()) return;

    double blMax = 0.0;
    double blOff = 0.0;

    const Scan *bkgrScan = getBackgroundMetrics(bkgrScanNumber, blMax, blOff);
    QColor col(settings->value("graph/hklLineColor", "#009900").toString());

    refStructure.setColor(darkTheme ? global::Functions::colorToDarkMode(col) : col);
    refStructure.setHklXunit("dnm");
    refStructure.setYoffset(blOff);
    refStructure.setHklSymbol(settings->value("config/hklSymbol", 0).toInt());

    drawHklScanLines(p, &refStructure, bkgrScan, lineWidth);
}

void GraphWindow::drawHklScanLines(QPainter &p, const Scan *scan, const Scan *bkgrScan, int lwidth)
{
    if (!scan) return;
    int x = 0;
    int y0 = 0;
    int y1 = 0;

    p.save();

    QFontMetrics fm(fontLegend);
    int markerSize = fm.ascent()/2;

    QPen pen;
    pen.setColor(darkTheme ? global::Functions::colorToDarkMode(scan->color()) : scan->color());
    pen.setWidth(lwidth);
    p.setPen(pen);
    p.translate(plot.x(), plot.y());
    p.setClipRect(QRect(lwidth, lwidth, plot.width() - lwidth, plot.height() - lwidth));

    // draw line for reflections
    for ( int i = 0; i < scan->dataHkl().size(); ++i) {
        Hkl hkl = scan->dataHkl().at(i);

        // skip lines set to invisible
        if (hkl.status() == 0) {
            continue;
        }

        double pos = hkl.position();
        if (scan->getHklXunit() == "dnm") pos = global::Functions::dToTwoTheta(10.0 * pos, waveLength[0]);

        x = getXcoord(pos + scan->xOffset());

        // ignore x-positions outside the displayed range
        if ((x < 0) || (x > plot.width())) {
            continue;
        }

        double blLin = bkgrScan ? bkgrScan->intensity(pos) : 0.0;
        double baseLine = 0.0;

        // relative intensity of the hkl line
        double yoffset = scan->yOffset();
        double hklIntens = scan->scaleFactor() * hkl.intensity();

        if (yoffset > -1.0) {
            if (yScaling == YSCALELIN) {
                baseLine    = blLin;
                yoffset     = scan->yOffset();
                hklIntens   = hklIntens + blLin;
            }

            if (yScaling == YSCALESQRT) {
                baseLine    = getSqrt(blLin);
                yoffset     = getSqrt(scan->yOffset());
                hklIntens   = getSqrt(hklIntens + blLin);
            }

            if (yScaling == YSCALELOG10) {
                baseLine    = getLog10(blLin);
                yoffset     = getLog10(scan->yOffset());
                hklIntens   = getLog10(hklIntens + blLin);
            }

            y0 = getYcoord(yoffset + baseLine);
            y1 = getYcoord(yoffset + hklIntens);
        } else {
            y0 = getYcoord(yoffset);
            y1 = qFuzzyIsNull(scanMaxIntensity) ? getYcoord(yoffset - hklIntens)
                                                : getYcoord(yoffset - yoffset * hklIntens / scanMaxIntensity);
        }

        if (hkl.status() == 2) {
            // use bold lines for hkl lines with status = 2
            pen.setWidth(lwidth == 2 * lineWidth ? lwidth : 2 * lwidth);
            p.setPen(pen);
            p.drawLine(x - markerSize/2, y0 + markerSize, x, y0);
            p.drawLine(x + markerSize/2, y0 + markerSize, x, y0);
            p.drawLine(x - markerSize/2, y1 - markerSize, x, y1);
            p.drawLine(x + markerSize/2, y1 - markerSize, x, y1);
        }

        p.drawLine(x, y0, x, y1);
        int dx =  symbolSize * lineWidth + 1;
        if (scan->hklSymbol() == 1) p.drawLine(x - dx, y1, x + dx, y1);
        if (scan->hklSymbol() == 2) p.drawEllipse(x - dx, y1 - dx, 4, 4);

        if (hkl.status() == 2) {
            // reset bold lines
            pen.setWidth(lwidth);
            p.setPen(pen);
        }

        QString str;

        if (hkl.phase().isEmpty()) {
            str = QString("Ref: (%1)").arg(hkl.hkl());
        } else {
            str = QString("Ref: %1 (%2)").arg(hkl.phase()).arg(hkl.hkl());
        }

        // create the tool tip
        tTip _ttip;

        _ttip.pos = QRect(x - 2 + plot.x(), plot.y() + y1, 4, y0 - y1);
        _ttip.str = QStringList(str);

        tt[_ttip.pos.left()] = _ttip;
    }

    p.restore();
}

const Scan * GraphWindow::getBackgroundMetrics(int sc, double &ymax, double &yoff)
{
    int bkgrMode = settings->value("graph/hklOnBackground", 1).toInt();
    const Scan *scan = nullptr;
    double cps = 1.0;

    if ((sc >= 0) && (sc < scanControl->count()) && (bkgrMode > 0)) {
        if (bkgrMode == 1) {
            scan = scanControl->at(bkgrScanNumber);

            if (useCountsPerSecond) {
                cps = scan->timePerStep() <= 0.0 ? 1.0 : scan->timePerStep();
            }

            double yMx = scan->maxIntensity(minang, maxang);
            if (yMx < 0.0) yMx = 0.0;

            if (yScaling == YSCALELIN) {
                yoff = scan->yOffset() / cps;
                ymax = yMx / cps;
            }

            if (yScaling == YSCALESQRT) {
                yoff = getSqrt(scan->yOffset()) / cps;
                ymax = getSqrt(yMx) / cps;
            }

            if (yScaling == YSCALELOG10) {
                yoff = getLog10(scan->yOffset()) / cps;
                ymax = getLog10(yMx) / cps;
            }
        } else if (bkgrMode == 2) {
            yoff = difOffset;
        }
    }

    return scan;
}

/*
 * draws the plot lines
 */
void GraphWindow::drawPlot(QPainter &p, bool clipManually)
{
    if (!scanControl->count()) return;
    if (!p.isActive()) return;

    int aLineWidth = settings->value("graph/useActiveLineWidth", true).toBool()
            ? settings->value("graph/activeLineWidth", 2).toInt()
            : lineWidth;

    QPainter::CompositionMode fillCompMode = static_cast<QPainter::CompositionMode>(settings->value("graph/compositionMode", darkTheme ? 12 : 13).toInt());
    QPolygon bgScreenPolygon = getBackgroundScreenPolygon(clipManually);

    // reset the offset value for the difference curve
    difOffset = 0.0;

    QPen pen;
    pen.setWidth(lineWidth);
    pen.setColor(darkTheme ? Qt::white : Qt::black);
    pen.setCapStyle(Qt::FlatCap);
    pen.setJoinStyle(Qt::RoundJoin);

    QBrush brush(QColor::fromRgbF(1.0, 1.0, 1.0, 1.0));

    p.setPen(pen);
    p.setBrush(brush);
    p.save();
    p.translate(plot.x(), plot.y());

    if (qFuzzyIsNull(waveLength[0])) {
        waveLength[0] = settings->defaultWavelength();
    }

    // now clip the drawing rect, so as not to print over the axes
    p.setClipRect(0, 0, plot.width(), plot.height());

    // loop over all scans
    for (int ix = 0; ix <= scanControl->count(); ++ix) {
        const Scan *scan = nullptr;

        if (ix < scanControl->count()) {
            scan = scanControl->at(ix);
        } else {
            if (settings->value("graph/showRefStructureScan", false).toBool()) {
                scan = &refStructure;
            }
        }

        if (!scan) break;

        // skip invisible scans
        if (!scan->isVisible()) continue;

        // skip scans of size 0
        if (!scan->size()) continue;

        pen.setColor(darkTheme ? global::Functions::colorToDarkMode(scan->color()) : scan->color());

        bool isBg = scan->scanTypes().testFlag(Scan::BACKGROUND);
        bool isDiff = scan->scanTypes().testFlag(Scan::DIFF);

        // don't convert the background scan to screen again if we already have done it
        QPolygon polygon = (isBg && bgScreenPolygon.size() > 1) ? bgScreenPolygon : getScanScreenPolygon(scan, clipManually);
        QPolygon fillPoly;

        // only draw the scan if the polygon contains at least 2 points
        if (polygon.size() > 2) {
            // check if this is the active scan. If yes, use double line width
            int cLineWidth = scan->isActive() ? aLineWidth : lineWidth;

            // close the polygon if the scan is active and filling is requested
            if (scan->isActive() && fillActive && !isBg && !isDiff) {
                fillPoly = appendBackgroundPolygon(polygon, bgScreenPolygon);
                brush.setColor(settings->toFillColor(pen.color()));
            }

            // draw a solid line
            if (scanStyles.at(ix).pointSymbol <= 0) {
                pen.setWidth(cLineWidth);
                pen.setStyle(Qt::SolidLine);
                p.setPen(pen);
                p.drawPolyline(polygon);
            }

            // or draw points only
            if (scanStyles.at(ix).pointSymbol == 1) {
                pen.setWidth(cLineWidth * symbolSize);
                pen.setStyle(Qt::SolidLine);
                p.setPen(pen);
                p.drawPoints(polygon);
            }

            // or draw crosses
            if (scanStyles.at(ix).pointSymbol == 2) {
                pen.setWidth(cLineWidth);
                pen.setStyle(Qt::SolidLine);
                p.setPen(pen);
                double sz = symbolSize * lineWidth + 1.0;
                for (int pp = 0; pp < polygon.size(); ++pp) {
                    p.drawLine(polygon.at(pp).x() - sz, polygon.at(pp).y() - sz,
                               polygon.at(pp).x() + sz, polygon.at(pp).y() + sz);
                    p.drawLine(polygon.at(pp).x() - sz, polygon.at(pp).y() + sz,
                               polygon.at(pp).x() + sz, polygon.at(pp).y() - sz);
                }
            }

            if (scan->isActive() && fillActive) {
                p.setCompositionMode(fillCompMode);
                p.setBrush(brush);
                QPen fillPen(QColor(0, 0, 0, 0));
                fillPen.setWidth(0);
                p.setPen(fillPen);
                p.drawPolygon(fillPoly);
                p.setCompositionMode(QPainter::CompositionMode_SourceOver);
                p.setBrush(QBrush());
            }

            // draw a dash line
            if (scanStyles.at(ix).pointSymbol >= 3) {
                pen.setWidth(cLineWidth);
                pen.setStyle(Qt::DashLine);
                p.setPen(pen);
                p.drawPolyline(polygon);
            }

            if (ix == scanControl->count()) break;

            // if there are active integral ranges, draw the linear background
            if (integralRanges.size()) {
                // the y-coordinates of the integral range points have no other meaning than indicating whether or not
                // the background line should be drawn: y > 0 = draw, y < 0 = don't draw
                if (integralRanges.first().y() > 0.0) {
                    for (int i = 0; i < integralRanges.size() - 1; i += 2) {
                        QPointF pt_s(integralRanges.at(i));
                        QPointF pt_e(integralRanges.at(i+1));

                        double intens_s = scan->intensity(scan->indexOfAngle(pt_s.x() + angularCorrection(pt_s.x()), 1));
                        double intens_e = scan->intensity(scan->indexOfAngle(pt_e.x() + angularCorrection(pt_e.x()), 2));

                        int x_s = getXcoord(pt_s.x());
                        int x_e = getXcoord(pt_e.x());
                        int y_s = getYcoord(intens_s);
                        int y_e = getYcoord(intens_e);

                        p.drawLine(x_s, y_s, x_e, y_e);
                    }
                }
            }
        }

        difOffset = qMin(scan->yOffset(), difOffset);
    }

    // draw a horizontal line at yOffset if yOffset is < 0. This is only used to draw a line
    // at the center of the difference curve
    if (difOffset < 0.0) {
        QPen hlinePen(darkTheme ? Qt::white : Qt::black, lineWidth, Qt::DashLine);
        p.setPen(hlinePen);
        p.drawLine(0, getYcoord(difOffset), plot.width(), getYcoord(difOffset));
    }

    // draw axes
    pen.setColor(darkTheme ? Qt::white : Qt::black);
    pen.setWidth(lineWidth);
    p.setPen(pen);
    p.setBrush(QBrush());
    // increase the clipped rect
    p.setClipRect(-lineWidth, -lineWidth, plot.width() + 2*lineWidth, plot.height() + 2*lineWidth);
    p.drawRect(0, 0, plot.width() , plot.height());

    p.restore();
}

QPolygon GraphWindow::getScanScreenPolygon(const Scan *scan, bool clipManually)
{
    if (!scan) return QPolygon();
    QPolygon polygon;

    double yoff = scan->yOffset();
    double tps = 1.0;

    if (useCountsPerSecond) {
        tps = scan->timePerStep() > 0.0 ? scan->timePerStep() : 1.0;
    }

    if (yScaling == YSCALELOG10) yoff = getLog10(yoff);
    if (yScaling == YSCALESQRT)  yoff = getSqrt(yoff);

    int firstIdx = 0;
    int lastIdx = scan->size() - 1;

    double leftAng = getRealX(plot.left());
    double rightAng = getRealX(plot.right());

    // loop from the left to determine the start of the visible range
    for (int i = 0; i < scan->size(); ++i) {
        if (scan->angle(i) - angularCorrection(scan->angle(i)) + scan->xOffset() >= leftAng) {
            break;
        }

        firstIdx = i;
    }

    // loop from the right to determine the end of the visible range
    for (int i = scan->size() - 1; i > firstIdx; --i) {
        if (scan->angle(i) - angularCorrection(scan->angle(i)) + scan->xOffset() <= rightAng) {
            break;
        }

        lastIdx = i;
    }

    // loop over the visible range and calculate screen coordinates
    for (int i = firstIdx; i <= lastIdx; ++i) {
        // calculate intensity
        double scint = scan->intensity(i) * scan->scaleFactor() / tps;

        if (yScaling == YSCALESQRT)  scint = getSqrt(scint);
        if (yScaling == YSCALELOG10) scint = getLog10(scint);

        // calculate screen coordinates
        int x = getXcoord(scan->angle(i) + scan->xOffset() - angularCorrection(scan->angle(i)));
        int y = getYcoord(scint + yoff);

        if (clipManually) {
            if (y < 0)         y = 0;
            else if (y > plot.height()) y = plot.height();
        }

        polygon << QPoint(x, y);
    }

    if (polygon.size() >= 2) {
        if (polygon.at(0).x() < 0) {
            polygon[0] = interpolate(polygon.at(0), polygon.at(1), 0);
        }

        if (polygon.last().x() > plot.width()) {
            polygon[polygon.size() - 1] = interpolate(polygon.at(polygon.size() - 2), polygon.last(), plot.width());
        }
    }

    return polygon;
}

QPolygon GraphWindow::getBackgroundScreenPolygon(bool clipManually)
{
    if (fillActive && (scanControl->hasActiveScan())) {
        for (int i = 0; i < scanControl->count(); ++i) {
            const Scan *scan = scanControl->at(i);
            if (!scan) continue;
            if (scan->scanTypes().testFlag(Scan::BACKGROUND)) return getScanScreenPolygon(scan, clipManually);
        }
    }

    return QPolygon();
}

QPolygon GraphWindow::appendBackgroundPolygon(const QPolygon &sc, const QPolygon &bg)
{
    QPolygon poly = sc;

    if (bg.size() < 2) {
        int fx = sc.first().x();
        int lx = sc.last().x();
        int fy = getYcoord(0.0);

        poly.prepend(QPoint(fx, fy));
        poly.append(QPoint(lx, fy));
    } else {
        for (int i = bg.size() - 1; i >= 0; --i) {
            poly.append(bg.at(i));
        }
    }

    return poly;
}

void GraphWindow::drawAnchorPoints(QPainter &p)
{
    if (anchorPoints.isEmpty()) return;
    if (!scanControl->count()) return;
    if (!p.isActive()) return;

    int dia = int(pScale * 5);

    const Scan *scan = scanControl->last();

    QPen pen;
    pen.setWidth(lineWidth);
    pen.setColor(darkTheme ? global::Functions::colorToDarkMode(scan->color()) : scan->color());
    pen.setCapStyle(Qt::FlatCap);
    pen.setJoinStyle(Qt::RoundJoin);

    p.save();
    p.setPen(pen);
    p.translate(plot.x(), plot.y());

    QBrush brush(scan->color(), Qt::SolidPattern);

    // now clip the drawing rect, so as not to print over the axes
    p.setClipRect(0, 0, plot.width() + lineWidth, plot.height() + lineWidth);

    double yoff = scan->yOffset();
    if (yScaling == YSCALELOG10) yoff = getLog10(yoff);
    if (yScaling == YSCALESQRT)  yoff = getSqrt(yoff);

    double tps = 1.0;

    if (useCountsPerSecond) {
        tps = scan->timePerStep() > 0.0 ? scan->timePerStep() : 1.0;
    }

    int firstIdx = 0;
    int lastIdx = anchorPoints.size() - 1;

    double leftAng = getRealX(plot.left());
    double rightAng = getRealX(plot.right());

    // loop from the left to determine the start of the visible range
    for (int i = 0; i < anchorPoints.size(); ++i) {
        if (anchorPoints.at(i).angle >= leftAng) {
            break;
        }

        firstIdx = i;
    }

    // loop from the right to determine the end of the visible range
    for (int i = anchorPoints.size() - 1; i > firstIdx; --i) {
        if (anchorPoints.at(i).angle <= rightAng) {
            break;
        }

        lastIdx = i;
    }

    // loop over the visible range and calculate screen coordinates
    for (int i = firstIdx; i <= lastIdx; ++i) {
        // calculate intensity
        double ancint = anchorPoints.at(i).intensity * scan->scaleFactor() / tps;

        if (yScaling == YSCALESQRT)  ancint = getSqrt(ancint);
        if (yScaling == YSCALELOG10) ancint = getLog10(ancint);

        // calculate screen coordinates
        int x = getXcoord(anchorPoints.at(i).angle + scan->xOffset());
        int y = getYcoord(ancint + yoff);

        if (anchorPoints.at(i).highlight) p.setBrush(brush);
        else                              p.setBrush(Qt::NoBrush);

        p.drawEllipse(QPoint(x, y), dia, dia);
    }

    p.restore();
}

/*
 * draws the legend
 */
void GraphWindow::drawLegend(QPainter &p)
{
    legend->redraw(p, fontLegend, plot);
}

/*
 * returns a tooltip at position pt
 */
QString GraphWindow::tip(const QPoint &pt)
{
    QMapIterator<int, tTip> it(tt);

    while (it.hasNext()) {
        it.next();
        if (it.value().pos.contains(pt)) {
            return it.value().str.join("\n");
        }
    }

    return QString();
}

QString GraphWindow::getXaxisUnit()
{
    if (scanControl->hasData()) return scanControl->first()->xAxisLabel();

    QString xa = "Diffraction Angle [" + QString(global::degree) + "2" + QString(global::theta) + "]";
    return xa;
}

/*
 * this function accepts an angle and returns it
 * in *widget* coordinates
 */
int GraphWindow::getXcoord(double i)
{
    return int(plot.width() * (i - minang) / (maxang - minang));
}

/*
 * this function accepts an intensity and returns it
 * in *widget* coordinates
 */
int GraphWindow::getYcoord(double i)
{
    return int(plot.height() - plot.height() * (i - startintens ) / (endintens - startintens));
}

/*
 * conversion from *widget* X coordinate to angle
 */
double GraphWindow::getRealX(int i)
{
    return minang + (maxang - minang) * (i - margin_l) / plot.width();
}

/*
 * conversion from *widget* Y coordinate to intensity
 */
double GraphWindow::getRealY(int i)
{
    return startintens + double(plot.height() - i + margin_t) * (endintens - startintens) / plot.height();
}

/*
 * shows the tooltip if placed on a reflection
 */
bool GraphWindow::event(QEvent *e)
{
    if (e->type() == QEvent::ToolTip) {
        return showToolTip(e);
    }

    e->ignore();
    return QWidget::event(e);
}

bool GraphWindow::showToolTip(QEvent *e)
{
    QHelpEvent *helpEvent = static_cast<QHelpEvent *>(e);
    QString s = tip(helpEvent->pos());

    if (!s.isNull()) {
        QToolTip::showText(helpEvent->globalPos(), s);
        e->accept();
    } else {
        QToolTip::hideText();
        e->ignore();
    }

    return true;
}

void GraphWindow::mousePressEvent(QMouseEvent *e)
{
    clickedPosition = e->pos();

    if (e->button() == Qt::LeftButton) {
        mouseButtonLeft(e);
    }

    if (e->button() == Qt::RightButton) {
        mouseButtonRight(e);
    }

    if (e->button() == Qt::MiddleButton) {
        mouseButtonMiddle(e);
    }

    e->ignore();
}

void GraphWindow::mouseReleaseEvent(QMouseEvent *e)
{
    clickedPosition = QPoint();

    if (drawRubberBand) {
        mouseButtonStopRubberBand(e);
    }

    if (strucScaling) {
        mouseButtonStopStruScaling(e);
    }

    if (dragging) {
        mouseButtonStopDragging(e);
    }

    if (drawLivePeakIntegral) {
        mouseButtonStopPeakIntegral(e);
    }

    e->ignore();
}

void GraphWindow::mouseMoveEvent(QMouseEvent *e)
{
    // no scans: do nothing
    if (!scanControl->count()) {
        e->ignore();
        return;
    }

    broadcastCoordinates(e->pos());

    if (drawRubberBand) {
        mouseMoveRubberBand(e);
    }

    if (strucScaling) {
        mouseMoveStrucScaling(e);
    }

    if (dragging) {
        mouseMoveDragging(e);
    }

    if (crossHair || noiseCursor || specLines || inspector || rangeSelectMode || (peakPreviewMode != PPMNONE)) {
        // special cursors only need a buffered update
        update();
    }

    if (drawLivePeakIntegral) {
        integrationRangeEnd = getRealX(mapFromGlobal(QCursor::pos()).x());
        forceUpdate();
    }

    e->accept();
}

void GraphWindow::wheelEvent(QWheelEvent *e)
{
    if (!plot.contains(e->pos())) return;

    if (e->modifiers().testFlag(Qt::NoModifier)) {
        zoomStep(e->pos(), e->delta(), true, false);
        e->accept();
        return;
    }

    if (e->modifiers().testFlag(Qt::ControlModifier)) {
        zoomStep(e->pos(), e->delta(), false, true);
        e->accept();
        return;
    }

    e->ignore();
}

void GraphWindow::mouseDoubleClickEvent(QMouseEvent *e)
{
    if (e->button() == Qt::LeftButton) {
        if (peakPreviewMode != PPMNONE) {
            if (firstDoubleClickPoint.isNull()) {
                firstDoubleClickPoint = e->pos();
            } else {
                double ptD  = 0.0; // unused
                double ptAx = 0.0;
                double ptAy = 0.0;
                double ptBx = 0.0;
                double ptBy = 0.0;
                calcCoordinates(firstDoubleClickPoint, ptAx, ptAy, ptD);
                calcCoordinates(e->pos(),       ptBx, ptBy, ptD);
                emit sigPeakPreviewPoints(QPointF(ptAx, ptAy), QPointF(ptBx, ptBy), peakPreviewMode);
                firstDoubleClickPoint = QPoint();
                emit sigCursorMessage(QString());
                e->accept();
                update();
            }
        } else if (rangeSelectMode) {
            if (firstDoubleClickPoint.isNull()) {
                firstDoubleClickPoint = e->pos();
            } else {
                double ptD  = 0.0; // unused
                double ptAx = 0.0;
                double ptAy = 0.0;
                double ptBx = 0.0;
                double ptBy = 0.0;
                calcCoordinates(firstDoubleClickPoint, ptAx, ptAy, ptD);
                calcCoordinates(e->pos(),       ptBx, ptBy, ptD);
                emit sigRangePoints(QPointF(ptAx, ptAy), QPointF(ptBx, ptBy));
                firstDoubleClickPoint = QPoint();
                emit sigCursorMessage(QString());
                e->accept();
                update();
            }
        } else {
            double x = 0.0;
            double y = 0.0;
            double d = 0.0;

            calcCoordinates(e->pos(), x, y, d);

            // We emit two different signals, depending on whether Ctrl is pressed or not.
            // These can be used for different actions in the parent object.
            if (e->modifiers() == Qt::NoModifier) emit sigDoubleClickA(x, y, d);
            if (e->modifiers() == Qt::ControlModifier) emit sigDoubleClickB(x, y, d);
            if (e->modifiers() == Qt::ShiftModifier) {
                if (integralRanges.size()) emit sigClearIntegrals();
            }

            e->accept();
            return;
        }
    }

    e->ignore();
}

void GraphWindow::mouseButtonLeft(QMouseEvent *e)
{
    if (e->modifiers().testFlag(Qt::AltModifier)) {
        mouseButtonStartStrucScaling(e);
    }

    if (e->modifiers().testFlag(Qt::NoModifier)) {
        mouseButtonStartRubberBand(e);
    }

    if (e->modifiers().testFlag(Qt::ControlModifier)) {
        mouseButtonStartDragging(e);
    }

    if (e->modifiers().testFlag(Qt::ShiftModifier)) {
        mouseButtonStartPeakIntegral(e);
    }
}

void GraphWindow::mouseButtonMiddle(QMouseEvent *e)
{
    if (e->modifiers().testFlag(Qt::NoModifier)) {
        mouseButtonStartStrucScaling(e);
    }
}

void GraphWindow::mouseButtonRight(QMouseEvent *e)
{
    if (e->modifiers().testFlag(Qt::NoModifier)) {
        if (plot.contains(e->pos())) {
            if (legend->boundingRect().contains(e->pos())) {
                legend->showContextMenu(e);
            } else {
                rightClickPlot(e);
            }
            e->accept();
        } else if (leftMarginRect.contains(e->pos())) {
            rightClickLeftMargin(e);
            e->accept();
        } else if (bottomMarginRect.contains(e->pos())) {
            rightClickBottomMargin(e);
            e->accept();
        }
    }
}

void GraphWindow::rightClickPlot(QMouseEvent *)
{
    bool fullUpdate = false;

    if (refStructure.hasHklData()) {
        if (!qFuzzyCompare(1.0, refStructure.scaleFactor())) {
            refStructure.setScaleFactor(1.0);
            fullUpdate = true;
        }
    }

    QVector<Scan *> aScans = scanControl->activeScans();

    for (int i = 0; i < aScans.size(); ++i) {
        if (!qFuzzyCompare(1.0, aScans.at(i)->scaleFactor())) {
            aScans[i]->setScaleFactor(1.0);
            fullUpdate = true;
        }
    }

    if (fullUpdate) {
        scanControl->updateViews();
    } else {
        resetZoom();
        forceUpdate();
    }
}

void GraphWindow::rightClickLeftMargin(QMouseEvent *e)
{
    contextMenuYaxis->exec(e->globalPos());
}

void GraphWindow::rightClickBottomMargin(QMouseEvent *e)
{
    contextMenuXaxis->exec(e->globalPos());
}

void GraphWindow::mouseButtonStartStrucScaling(QMouseEvent *e)
{
    prevStrucScaleFactor = refStructure.scaleFactor();

    if (scanControl->hasActiveScan()) {
        prevScanScaleFactor = scanControl->firstActiveScan()->scaleFactor();
    }

    strucScaling = true;
    e->accept();
}

void GraphWindow::mouseButtonStartDragging(QMouseEvent *e)
{
    point_a = e->pos();
    dragging = true;
    e->accept();
}

void GraphWindow::mouseButtonStartRubberBand(QMouseEvent *e)
{
    point_a = e->pos();
    rubberBand->setGeometry(QRect(e->pos(), QSize()));
    rubberBand->show();
    drawRubberBand = true;
    e->accept();
}

void GraphWindow::mouseButtonStartPeakIntegral(QMouseEvent *e)
{
    integrationRangeStart = getRealX(mapFromGlobal(QCursor::pos()).x());
    integrationRangeEnd   = integrationRangeStart;
    drawLivePeakIntegral = true;
    hasActiveIntegrationRange = true;
    e->accept();
    update();
}

void GraphWindow::mouseButtonStopStruScaling(QMouseEvent *e)
{
    strucScaling = false;
    e->accept();
}

void GraphWindow::mouseButtonStopDragging(QMouseEvent *e)
{
    dragging = false;
    e->accept();
}

void GraphWindow::mouseButtonStopRubberBand(QMouseEvent *e)
{
    drawRubberBand = false;
    rubberBand->hide();
    zoom(point_a, e->pos());
    e->accept();
}

void GraphWindow::mouseButtonStopPeakIntegral(QMouseEvent *e)
{
    int cpos = mapFromGlobal(QCursor::pos()).x();
    integrationRangeEnd = getRealX(cpos);
    hasActiveIntegrationRange = false;
    drawLivePeakIntegral = false;
    e->accept();

    // check for a certain threshold, ignore very small ranges
    if (qAbs(cpos - margin_l - getXcoord(integrationRangeStart)) > 5) {
        integrateScan();
    }

    forceUpdate();
}

void GraphWindow::mouseMoveStrucScaling(QMouseEvent *e)
{
    if (clickedPosition.isNull()) return;

    int y0 = clickedPosition.y();
    int zeroLine = getYcoord(0.0);
    int diffLine = getYcoord(difOffset);

    int baseline = margin_t + plot.height();

    if (y0 < diffLine)      baseline = diffLine;
    else if (y0 < zeroLine) baseline = zeroLine;

    double delta = double(baseline - e->y() + margin_t) / double(baseline - y0 + margin_t);

    QVector<Scan*> aScans = scanControl->activeScans();

    if (aScans.size()) {
        for (int i = 0; i < aScans.size(); ++i) {
            double sfac = prevScanScaleFactor * delta;
            aScans[i]->setScaleFactor(sfac);
        }

        scanControl->updateViews();
    } else if (refStructure.hasHklData()) {
        double sfac = prevStrucScaleFactor * delta;
        refStructure.setScaleFactor(sfac);
        scanControl->updateViews();
    }
}

void GraphWindow::mouseMoveDragging(QMouseEvent *e)
{
    int dx = point_a.x() - e->pos().x();
    int dy = point_a.y() - e->pos().y();

    double startX = getRealX(margin_l + dx);
    double endX   = getRealX(margin_l + plot.width() + dx);
    double startY = getRealY(margin_t + plot.height() + dy);
    double endY   = getRealY(margin_t + dy);

    if ((startX < minangAbs) || (endX > maxangAbs)) {
        startX = minang;
        endX   = maxang;
    }

    if ((startY < mini) || (endY > maxi)) {
        startY = startintens;
        endY   = endintens;
    }

    setZoomRange(startX, endX, startY, endY, true);
    point_a = e->pos();
}

void GraphWindow::mouseMoveRubberBand(QMouseEvent *e)
{
    rubberBand->setGeometry(QRect(point_a, e->pos()).normalized());
}

void GraphWindow::broadcastCoordinates(const QPoint &p)
{
    if (updateCoordinateTimer.isActive()) return;

    double x = 0.0;
    double y = 0.0;
    double d = 0.0;

    calcCoordinates(p, x, y, d);
    emit sigCoordinates(x, y, d);

    // block coordinate updates for 50 ms to reduce CPU load
    // note: intervals > 50 ms will be noticably choppy
    updateCoordinateTimer.start(40);
}

/*
 * handles key press events
 */
void GraphWindow::keyPressEvent( QKeyEvent *e )
{
    if (e->isAutoRepeat()) {
        e->ignore();
        return;
    }

    if (e->key() == Qt::Key_H) {
        QToolTip::showText(cursor().pos(), global::helpTextMouseActions, this);
        e->accept();
        return;
    }

    int doUpdate = 0;

    if (e->key() == Qt::Key_C) {
        crossHair = !crossHair;
        doUpdate++;
    }

    if (e->key() == Qt::Key_N) {
        noiseCursor = !noiseCursor;
        doUpdate++;
    }

    if (e->key() == Qt::Key_S) {
        specLines = !specLines;
        doUpdate++;
    }

    if (e->key() == Qt::Key_I) {
        inspector = !inspector;
        doUpdate++;
    }

    if (e->key() == Qt::Key_Escape) {
        if ((peakPreviewMode != PPMNONE) || (rangeSelectMode)) {
            firstDoubleClickPoint = QPoint();
            doUpdate++;
        }
    }

    if (doUpdate > 0) {
        if (noiseCursor || specLines || crossHair || inspector) {
            setCursor(Qt::CrossCursor);
        } else {
            setCursor(Qt::ArrowCursor);
        }

        e->accept();
        update();
    } else {
        e->ignore();
    }
}

/*
 * handles key release events
 */
void GraphWindow::keyReleaseEvent( QKeyEvent *e )
{
    if (dragging) {
        dragging = false;
        e->accept();
        update();
        return;
    }

    e->ignore();
}

void GraphWindow::copy()
{
    copyPixmapToClipboard();
}

/*
 * calculates the real coordinates of widget point p.
 * returns x (in x-axis units), y (in y-axis units), and d (in A)
 */
void GraphWindow::calcCoordinates(const QPoint &p, double &x, double &y, double &d)
{
    // calculate the coordinates and emit a signal
    if (plot.contains(p)) {
        x = getRealX(p.x());
        y = getRealY(p.y());

        // avoid division by 0
        if (!qFuzzyIsNull(x)) {
            d = 10.0 * global::Functions::twoThetaToD(x, waveLength[0] / 10.0);
        }

        // for the difference curve, add the vertical offset from the base line
        if ((y < 0.0) && (difOffset < 0.0)) {
            y -= difOffset;
        }

        if (yScaling == YSCALESQRT)  y = getInvSqrt(y);
        if (yScaling == YSCALELOG10) y = getInvLog10(y);
    }
}

/*
 * zoom region between pa and pb (in widget coordinates)
 */
void GraphWindow::zoom(QPoint pa, QPoint pb)
{
    // request a minimum difference, otherwise do nothing
    int limit = 5;

    setFocus();

    if ((abs(pa.x() - pb.x()) <= limit)
        || (abs(pa.y() - pb.y()) <= limit)) {
        return;
    }

    setZoomRange(getRealX(pa.x()), getRealX(pb.x()), getRealY(pa.y()), getRealY(pb.y()), true);
}

/*
 * zoom a certain step into the graph
 *
 * p = mouse position, or center of plot
 * i = number of zoom steps, can be > 0 or < 0
 *
 * horizontal: set to true to zoom in horizontal direction
 * vertical:   set to true to zoom in vertical direction
 *
 * both can be true at the same time
 */
void GraphWindow::zoomStep(QPoint p, int i, bool horizontal, bool vertical)
{
    double base = 1.0;  // fraction which is multiplied, increase to increase the zoom step

    int x_a = plot.left();
    int x_b = plot.right() + 1;
    int y_a = plot.top();
    int y_b = plot.bottom() + 1;

    if (horizontal) {
        x_a += int(i * base * (double(p.x() - plot.left()) / double(plot.width())));
        x_b -= int(i * base * (double(plot.width() - p.x() + plot.left()) / double(plot.width())));
    }

    if (vertical) {
        y_a += int(i * base * (double(p.y() - plot.left()) / double(plot.height())));
        y_b -= int(i * base * (double(plot.height() - p.y() + plot.top()) / double(plot.height())));
    }

    QPoint pa(x_a, y_a);
    QPoint pb(x_b, y_b);
    zoom(pa, pb);
}

void GraphWindow::resetZoom()
{
    calcLimits();

    maxang = maxangAbs;
    minang = minangAbs;
    endintens = maxi;
    startintens = mini;

    zoomed = false;
}

void GraphWindow::resetZoomX()
{
    calcLimits();

    maxang = maxangAbs;
    minang = minangAbs;

    if (qFuzzyCompare(endintens, maxi) && qFuzzyCompare(startintens, mini)) {
        zoomed = false;
    } else {
        zoomed = true;
    }

    forceUpdate();
}

void GraphWindow::resetZoomY()
{
    calcLimits();

    startintens = mini;
    endintens = maxi;

    if (qFuzzyCompare(maxang, maxangAbs) && qFuzzyCompare(minang, minangAbs)) {
        zoomed = false;
    } else {
        zoomed = true;
    }

    forceUpdate();
}

/*
 * calculate max and min angle and intensity of all scans
 */
void GraphWindow::calcLimits()
{
    if (!scanControl->count()) {
        mini = 0.0;
        maxi = 0.0;
        minangAbs = 0.0;
        maxangAbs = 0.0;
        endpoint = 0;
        return;
    }

    mini = 0.0;
    maxi = 0.0;
    minangAbs = std::numeric_limits<double>::max();
    maxangAbs = 0.0;
    endpoint = 0;

    for (int i = 0; i < scanControl->count(); i++) {
        const Scan *scan = scanControl->at(i);
        if (!scan) continue;

        double _xmin = std::numeric_limits<double>::max();
        double _xmax = std::numeric_limits<double>::min();
        double _ymin = 0.0;
        double _ymax = 0.0;

        if (scan->dataIntensity().size()) {
            ScanOps::scanMetrics(*scan, _xmin, _xmax, _ymin, _ymax);
        } else {
            if (scan->dataHkl().size()) {
                for (int j = 0; j < scan->dataHkl().size(); ++j) {
                    _ymax = qMax(_ymax, scan->dataHkl().at(j).intensity());
                }
            } else {
                // skip scans that neither contain raw data nor hkl data
                continue;
            }
        }

        double cps = 1.0;
        if (useCountsPerSecond) {
            cps = scan->timePerStep() < 0.0 ? 1.0 : scan->timePerStep();
        }

        minangAbs = qMin(minangAbs, _xmin + scan->xOffset());
        maxangAbs = qMax(maxangAbs, _xmax + scan->xOffset());
        endpoint = uint(qMax(int(endpoint), scan->size()));

        double scMinInt = _ymin / cps;
        double scMaxInt = _ymax / cps;
        scanMaxIntensity = qMax(scanMaxIntensity, scMaxInt);

        double yoff = scan->yOffset();
        double scint_i = scMinInt;
        double scint_a = scMaxInt;

        if (yScaling == YSCALESQRT) {
            yoff = getSqrt(scan->yOffset());
            scint_i = getSqrt(scMinInt);
            scint_a = getSqrt(scMaxInt);
        }

        if (yScaling == YSCALELOG10) {
            yoff = getLog10(scan->yOffset());
            scint_i = getLog10(scMinInt);
            scint_a = getLog10(scMaxInt);
        }

        mini = qMin(mini, scint_i + yoff);
        maxi = qMax(maxi, scint_a + yoff);
    }

    // give a slight margin above the highest and lowest point
    maxi *= 1.05;
    mini *= 1.05;
}

/*
 * draw minor tick marks. Major marks are drawn in ::drawLabels()
 */
void GraphWindow::drawTicks(QPainter &p)
{
    if (!p.isActive()) return;

    // check if valid axes are defined
    if (startintens >= endintens) return;
    if (minang >= maxang) return;

    int len = int(pScale * 5);
    int pos = 0;

    QPen pen;
    pen.setColor(darkTheme ? global::Functions::colorToDarkMode(axisColor) : axisColor);
    pen.setWidth(lineWidth);
    p.setPen(pen);
    p.setClipRect(plot);

    QVector<int> vPos;
    QVector<int> hPos;
    QVector<double> tVec;

    // x-axis
    if (xScaling == XSCALED)  {
        tVec = global::Functions::scaleAxisD(minang, maxang, waveLength[0], 1, tickDensityX);

        for (int i = tVec.size() - 1; i >= 0; --i) {
            pos = getXcoord(global::Functions::dToTwoTheta(tVec.at(i), waveLength[0])) + plot.left();
            hPos.append(pos);

            p.drawLine(pos, plot.top(), pos, plot.top() + len);
            p.drawLine(pos, plot.bottom(), pos, plot.bottom() - len);
        }
    } else /* XSCALETWOTHETA */ {
        tVec = global::Functions::scaleAxis1(minang, maxang, 6*tickDensityX);

        for (int i = 0; i < tVec.size(); ++i) {
            pos = getXcoord(tVec.at(i)) + plot.left();
            hPos.append(pos);

            p.drawLine(pos, plot.top(), pos, plot.top() + len);
            p.drawLine(pos, plot.bottom(), pos, plot.bottom() - len);
        }
    }

    // y-axis
    if (startintens <= 0) {
        if      (yScaling == YSCALESQRT)   tVec = global::Functions::scaleAxis1(0.0, getInvSqrt(endintens), 6*tickDensityY);
        else if (yScaling == YSCALELOG10)  tVec = global::Functions::scaleAxisLog(1.0, getInvLog10(endintens), 6*tickDensityY);
        else              /* YSCALELIN */  tVec = global::Functions::scaleAxis1(0.0, endintens, 6*tickDensityY);

        // zero line
        pos = getYcoord(0.0) + plot.top();
        p.drawLine(plot.left(), pos, plot.right(), pos);
    } else {
        if      (yScaling == YSCALESQRT)   tVec = global::Functions::scaleAxis1(getInvSqrt(startintens), getInvSqrt(endintens), 6*tickDensityY);
        else if (yScaling == YSCALELOG10)  tVec = global::Functions::scaleAxisLog(getInvLog10(startintens), getInvLog10(endintens), 6*tickDensityY);
        else              /* YSCALELIN */  tVec = global::Functions::scaleAxis1(startintens, endintens, 6*tickDensityY);
    }

    for (int i = 0; i < tVec.size(); ++i) {
        if      (yScaling == YSCALESQRT)   pos = getYcoord(getSqrt(tVec.at(i))) + plot.top();
        else if (yScaling == YSCALELOG10)  pos = getYcoord(getLog10(tVec.at(i))) + plot.top();
        else              /* YSCALELIN */  pos = getYcoord(tVec.at(i)) + plot.top();
        vPos.append(pos);

        p.drawLine(plot.left(),  pos, plot.left() + len, pos);
        p.drawLine(plot.right(), pos, plot.right() - len, pos);
    }


    if (showGridMinor) {
        pen.setColor(darkTheme ? global::Functions::colorToDarkMode(Qt::lightGray) : Qt::lightGray);
        pen.setStyle(Qt::DotLine);
        p.setPen(pen);

        for (int i = 0; i < hPos.size(); ++i) {
            p.drawLine(hPos.at(i), plot.top(), hPos.at(i), plot.bottom());
        }

        for (int i = 0; i < vPos.size(); ++i) {
            p.drawLine(plot.left(), vPos.at(i), plot.right(), vPos.at(i));
        }
    }
}

/*
 * draws tickmark lables
 */
void GraphWindow::drawLabels(QPainter &p)
{
    if (!p.isActive()) return;

    // check if valid axes are defined
    if (startintens >= endintens) return;

    if (minang >= maxang) return;

    int len = int(pScale * 10);  //length of major ticks
    int sp = margin_i;           //distance of labels to axes
    int pos = 0;

    QFontMetrics fm(fontTicks);
    p.setFont(fontTicks);
    p.setClipRect(canvas);

    QPen pen;
    pen.setColor(darkTheme ? global::Functions::colorToDarkMode(axisColor) : axisColor);
    pen.setWidth(lineWidth);
    pen.setCapStyle(Qt::FlatCap);
    p.setPen(pen);

    QVector<int> vPos;
    QVector<int> hPos;
    QVector<double> tVec;

    // x-axis:
    if (xScaling == XSCALED) {
        tVec = global::Functions::scaleAxisD(minang, maxang, waveLength[0], 0, tickDensityX);

        for (int i = tVec.size() - 1; i >= 0; --i) {
            pos = getXcoord(global::Functions::dToTwoTheta(tVec.at(i), waveLength[0])) + plot.left();
            hPos.append(pos);

            QString s = QString("%1").arg(tVec.at(i), 0, 'f', 2);

            if ((pos >= plot.left()) && (pos <= plot.right() + 1)) {
                p.drawText(pos - int(fm.width(s) / 2.0), plot.bottom() + fm.height() + sp, s);
                p.drawLine(pos, plot.top(), pos, plot.top() + len);
                p.drawLine(pos, plot.bottom(), pos, plot.bottom() - len);
            }
        }
    } else {
        tVec = global::Functions::scaleAxis1(minang, maxang, tickDensityX);

        for (int i = 0; i < tVec.size(); ++i) {
            pos = getXcoord(tVec.at(i)) + plot.left();
            hPos.append(pos);

            QString s = QString("%1").arg(tVec.at(i), 0, 'f', 2);

            if ((pos >= plot.left()) && (pos <= plot.right() + 1)) {
                p.drawText(pos - int(fm.width(s) / 2.0), plot.bottom() + fm.height() + sp, s);
                p.drawLine(pos, plot.top(), pos, plot.top() + len);
                p.drawLine(pos, plot.bottom(), pos, plot.bottom() - len);
            }
        }
    }

    // y-axis:
    if (startintens <= 0) {
        if (yScaling == YSCALELIN)    tVec = global::Functions::scaleAxis1(0.0, endintens, tickDensityY);
        if (yScaling == YSCALESQRT)   tVec = global::Functions::scaleAxis1(0.0, getInvSqrt(endintens), tickDensityY);
        if (yScaling == YSCALELOG10)  tVec = global::Functions::scaleAxisLog(1.0, getInvLog10(endintens), tickDensityY);
    } else {
        if (yScaling == YSCALELIN)    tVec = global::Functions::scaleAxis1(startintens, endintens, tickDensityY);
        if (yScaling == YSCALESQRT)   tVec = global::Functions::scaleAxis1(getInvSqrt(startintens), getInvSqrt(endintens), tickDensityY);
        if (yScaling == YSCALELOG10)  tVec = global::Functions::scaleAxisLog(getInvLog10(startintens), getInvLog10(endintens), tickDensityY);
    }

    for (int i = 0; i < tVec.size(); ++i) {
        if (yScaling == YSCALELIN)    pos = getYcoord(tVec.at(i)) + plot.top();
        if (yScaling == YSCALESQRT)   pos = getYcoord(getSqrt(tVec.at(i))) + plot.top();
        if (yScaling == YSCALELOG10)  pos = getYcoord(getLog10(tVec.at(i))) + plot.top();
        vPos.append(pos);

        QString s = QString("%1").arg(tVec.at(i), 0, 'f', 0);

        p.drawText(plot.left() - (fm.width(s) + sp), pos + int(fm.ascent() / 2.0), s);
        p.drawLine(plot.left(), pos, plot.left() + len, pos);
        p.drawLine(plot.right(), pos, plot.right() - len, pos);
    }

    if (showGridMajor) {
        pen.setColor(darkTheme ? global::Functions::colorToDarkMode(Qt::gray) : Qt::gray);
        pen.setStyle(Qt::DashLine);
        p.setPen(pen);

        for (int i = 0; i < hPos.size(); ++i) {
            p.drawLine(hPos.at(i), plot.top(), hPos.at(i), plot.bottom());
        }

        for (int i = 0; i < vPos.size(); ++i) {
            p.drawLine(plot.left(), vPos.at(i), plot.right(), vPos.at(i));
        }
    }
}

/*
 * toggles drawing of the legend
 */
void GraphWindow::slotSetDrawLegend(bool a)
{
    legend->setDrawn(a);
    forceUpdate();
}

/*
 * toggles drawing of hkl reflections
 */
void GraphWindow::slotSetDrawRefls(bool a)
{
    drawRefls = a;
    forceUpdate();
}

/*
 * moves visible section to the left
 * used by tool buttons or key press events
 */
void GraphWindow::moveLeft( double fraction )
{
    double old_minang = minang;
    double old_maxang = maxang;

    // determine shifted range on x-axis
    minang = old_minang - ((old_maxang - old_minang) * (plot.width() / fraction)) / plot.width();
    maxang = old_maxang - ((old_maxang - old_minang) * (plot.width() / fraction)) / plot.width();

    // respect limits, points outside the range are not accessible
    if (minang <= 0) {
        minang = 0;
        maxang = old_maxang - old_minang;
    }

    update();
}

/*
 * moves visible section to the right
 * used by tool buttons or key press events
 */
void GraphWindow::moveRight( double fraction )
{
    double old_minang = minang;
    double old_maxang = maxang;

    // determine shifted range on x-axis
    minang = old_minang + ((old_maxang - old_minang) * (plot.width() / fraction)) / plot.width();
    maxang = old_maxang + ((old_maxang - old_minang) * (plot.width() / fraction)) / plot.width();

    // respect limits, points outside the range are not accessible
    if (maxang >= maxangAbs) {
        maxang = maxangAbs;
        minang = old_minang + (maxangAbs - old_maxang);
    }

    update();
}

/*
 * moves visible section up
 * used by tool buttons or key press events
 */
void GraphWindow::moveUp( double fraction )
{
    double old_starti = startintens;
    double old_endi = endintens;

    startintens = old_starti + (old_endi - old_starti) * (plot.height() / fraction) / plot.height();
    endintens = old_endi + (old_endi - old_starti) * (plot.height() / fraction) / plot.height();

    // respect limits, points outside the range are not accessible
    if (endintens >= maxi) endintens = maxi;

    update();
}

/*
 * moves visible section down
 * used by tool buttons or key press events
 */
void GraphWindow::moveDown( double fraction )
{
    double old_starti = startintens;
    double old_endi = endintens;

    startintens = old_starti - (old_endi - old_starti) * (plot.height() / fraction) / plot.height();
    endintens = old_endi - (old_endi - old_starti) * (plot.height() / fraction) / plot.height();

    update();
}

/*
 * returns the minimum and maximum angle/intensity of the graph
 */
void GraphWindow::getMaxRange(double &a_min, double &a_max, double &i_min, double &i_max)
{
    a_min = minangAbs;
    a_max = maxangAbs;
    i_min = mini;
    i_max = maxi;
}

/*
 * writes the currently active zoom range in angle/intensity units
 */
void GraphWindow::getZoomRange(double &a_min, double &a_max, double &i_min, double &i_max)
{
    a_min = minang;
    a_max = maxang;
    i_min = startintens;
    i_max = endintens;
}

/*
 * sets the zoom range in angle/intensity units
 */
void GraphWindow::setZoomRange(double a_min, double a_max, double i_min, double i_max, bool checkLimits)
{
    if (checkLimits) {
        minang = qMax(minangAbs, qMin(a_min, a_max));
        maxang = qMin(maxangAbs, qMax(a_min, a_max));
        startintens = qMax(mini, qMin(i_min, i_max));
        endintens = qMin(maxi, qMax(i_min, i_max));
    } else {
        minang = qMin(a_min, a_max);
        maxang = qMax(a_min, a_max);
        startintens = qMin(i_min, i_max);
        endintens = qMax(i_min, i_max);
    }

    //check if we zoomed all the way out
    if (qFuzzyCompare(minang, minangAbs) && qFuzzyCompare(maxang, maxangAbs) && qFuzzyCompare(startintens, mini) && qFuzzyCompare(endintens, maxi)) {
        zoomed = false;
    } else {
        zoomed = true;
    }

    forceUpdate();
}

/*
 * forces a complete redraw of the plot
 * use this instead of update() if zoom ranges,
 * number of scans, visibilities etc. changed.
 *
 * update() only uses the buffered plot to redraw
 * the widget, e.g. after drawing the crosshair
 */
void GraphWindow::forceUpdate()
{
    recalcPlot = true;
    update();
}

bool GraphWindow::setStatus(int i)
{
    switch (i) {
    case 0:
        marginColor = idleColor;
        break;
    case 1:
        marginColor = activeColor;
        break;
    case 2:
        marginColor = completedColor;
        break;
    }

    return true;
}

/*
 * changes the state of the graph to active
 */
void GraphWindow::setActive()
{
    marginColor = activeColor;
    recalcPlot = true;
    update();
}

/*
 * changes the state of the graph to idle
 */
void GraphWindow::setIdle()
{
    marginColor = idleColor;
    recalcPlot = true;
    update();
}

/*
 * changes the state of the graph to complete
 */
void GraphWindow::setComplete()
{
    marginColor = completedColor;
    recalcPlot = true;
    update();
}

void GraphWindow::setPeakPreviewMode(PeakPreviewMode p)
{
    peakPreviewMode = p;

    if (p != PPMNONE) {
        marginColor = peakSelectColor;
        setCursor(Qt::CrossCursor);
    } else {
        marginColor = idleColor;
        setCursor(Qt::ArrowCursor);
        emit sigCursorMessage(QString());
    }

    recalcPlot = true;
    update();
}

void GraphWindow::setRangeSelectMode(bool b)
{
    rangeSelectMode = b;

    if (b) {
        marginColor = peakSelectColor;
        setCursor(Qt::CrossCursor);
    } else {
        marginColor = idleColor;
        setCursor(Qt::ArrowCursor);
        emit sigCursorMessage(QString());
    }

    recalcPlot = true;
    update();
}

void GraphWindow::setReferenceReflections(const Scan &sc)
{
    tt.clear();
    refStructure = sc;

    const Scan *scanIobs = scanControl->first();
    const Scan *scanIbkg = scanControl->at(bkgrScanNumber);

    if (!scanIobs) {
        forceUpdate();
        return;
    }

    double _refYHklMax = 0.0;

    // find maximum hkl intensity
    for (int i = 0; i < refStructure.dataHkl().size(); ++i) {
        _refYHklMax = qMax(_refYHklMax, refStructure.dataHkl().at(i).intensity());
    }

    // get the maximum intensity of the displayed scans
    int n = scanIobs->indexOfMaxIntensity();
    double _yTarget = scanIobs->at(n);

    if (scanIbkg && settings->value("graph/hklOnBackground", true).toBool()) {
        _yTarget -= scanIbkg->at(n);
    }

    // normalize the hkl intensities
    for (int i = 0; i < refStructure.dataHkl().size(); ++i) {
        refStructure.pDataHkl()[i].setIntensity(refStructure.dataHkl().at(i).intensity() * _yTarget / _refYHklMax);
    }

    // exit here if only hkl lines are available in refStructure
    if (!settings->value("graph/showRefStructureScan", false).toBool()) {
        forceUpdate();
        return;
    }

    if (!refStructure.hasScanData()) {
        forceUpdate();
        return;
    }

    // normalize the xy reference pattern
    double _refYXyMax  = refStructure.maxIntensity();

    if (_refYXyMax < 0.0) return;

    for (int i = 0; i < refStructure.dataIntensity().size(); ++i) {
        refStructure.pDataIntensity()[i] = _yTarget * refStructure.dataIntensity().at(i) / _refYXyMax;
        if (bkgrScanNumber >= 0) {
            refStructure.pDataIntensity()[i] += scanIbkg->intensity(refStructure.dataAngle().at(i));
        }
    }

    forceUpdate();
}

void GraphWindow::normalizeHkl(Scan *scan)
{
    double _ymaxHkl = 0.0;
    for (int i = 0; i < scan->dataHkl().size(); ++i) {
        _ymaxHkl = qMax(_ymaxHkl, scan->dataHkl().at(i).intensity());
    }

    double _ymaxScan = qFuzzyIsNull(scanMaxIntensity) ? 100.0 : scanMaxIntensity;

    for (int i = 0; i < scan->dataHkl().size(); ++i) {
        double _d = scan->dataHkl().at(i).intensity();
        scan->pDataHkl()[i].setIntensity(_ymaxScan * _d / _ymaxHkl);
    }
}

/*
 * call this function after changing the content of scanHeap
 * (e.g. after loading, adding, reloading scans).
 * It applies the correct styles, wavelength, loads hkl reflections etc.
 */
void GraphWindow::setupNewScans()
{
    if (!hasOverrideWaveLength) {
        scanControl->getScanWaveLength(waveLength[0], waveLength[1], waveLength[2], settings->defaultWavelength());
    }

    changeGraphStyles();
    setBackgroundIndex();
    legend->updateView();

    if (zoomed) calcLimits();
    else        resetZoom();
}

void GraphWindow::setBackgroundIndex()
{
    bkgrScanNumber = -1;

    for (int i = 0; i < scanControl->count(); ++i) {
        if (scanControl->at(i)->name().toLower().trimmed() == QString("background")) {
            bkgrScanNumber = i;
            return;
        }
    }
}

/*
 * calculates the y coordinate at position x by interpolating between a and b
 */
QPoint GraphWindow::interpolate(const QPoint &a, const QPoint &b, int x)
{
    if (a.x() == b.x()) return a;

    int dy = int(0.5 + a.y() + (b.y() - a.y()) * (double(x) - a.x()) / (b.x() - a.x()));

    return QPoint(x, dy);
}

void GraphWindow::changeGraphStyles()
{
    bool writeSettings = false;

    // write the point style values and color to the newly loaded scans
    for (int i = 0; i < scanControl->count(); ++i) {
        Scan *scan = scanControl->getScan(i);
        if (!scan) continue;

        if (i >= scanStyles.size()) {
            ScanStyle s;
            s.color = settings->getRandomColor(i);
            s.pointSymbol = 0;
            s.hklSymbol = settings->value("config/hklSymbol", 0).toInt();
            scanStyles.append(s);
            writeSettings = true;
        }

        if (scan->scanTypes().testFlag(Scan::REFINED)) {
            switch (i) {
                case 0: scan->setColor(QColor(settings->value("graph/iobsColor", "#000000").toString()));
                        break;
                case 1: scan->setColor(QColor(settings->value("graph/icalcColor", "#ff0000").toString()));
                        break;
                case 2: scan->setColor(QColor(settings->value("graph/idiffColor", "#bbbbbb").toString()));
                        break;
                case 3: scan->setColor(QColor(settings->value("graph/ibkgrColor", "#0000ff").toString()));
                        break;
                default: scan->setColor(scanStyles.at(i).color);
            }
        } else {
            scan->setColor(scanStyles.at(i).color);
        }
    }

    // if new colors were added to the table (because the table was too short)
    // we will store the longer list in the settings, so next time the additional
    // colors are not random anymore and can be modified in the preferences dialog.
    if (writeSettings) {
        QList<QVariant> cl;
        QList<QVariant> sl;

        for (int i = 0; i < scanStyles.size(); ++i) {
            cl.append(scanStyles.at(i).color.name());
            sl.append(scanStyles.at(i).pointSymbol);
        }

        settings->setValue("graph/colorTable", cl);
        settings->setValue("graph/styleTable", sl);
    }
}

void GraphWindow::print(QPrinter &printer, QPainter &painter)
{
    // scale factor used to scale the output for printing (will be 1.0 for screen painting)
    // we pretend the widget has the size of the desktop, to make the output independent
    // of the widget size. Else by resizing the profex window the printed font size would change.
    QDesktopWidget desktop;
    double pFontSize = settings->value("graph/printingFontSize", 1.25).toDouble();
    pScale = pFontSize * qMin(double(printer.pageRect().width()) / double(desktop.screenGeometry().width()),
                        double(printer.pageRect().height()) / double(desktop.screenGeometry().height()));

    QFontInfo fiTitle(oFontTitle);
    QFontInfo fiAxis(oFontAxis);
    QFontInfo fiTicks(oFontTicks);
    QFontInfo fiLegend(oFontLegend);
    bool oDarkTheme = darkTheme;
    darkTheme = false;

    fontTitle.setPixelSize(int(pScale * double(fiTitle.pixelSize())));
    fontAxis.setPixelSize(int(pScale * double(fiAxis.pixelSize())));
    fontTicks.setPixelSize(int(pScale * double(fiTicks.pixelSize())));
    fontLegend.setPixelSize(int(pScale * double(fiLegend.pixelSize())));

    int w = printer.pageRect().width();
    int h = printer.pageRect().height();

    // temporarily switch the line width to the printing line width
    int screenLineWidth = lineWidth;
    lineWidth = printingLineWidth;

    calcMargins(painter, false);

    painter.setRenderHints(QPainter::Antialiasing | QPainter::TextAntialiasing | QPainter::SmoothPixmapTransform, antiAliasing);
    painter.fillRect(0, 0, w, h, idleColor);
    initPlot(printer.pageRect());
    drawWindow(painter);
    drawHighlightedRegions(painter);
    drawPlot(painter, false);
    drawHklTicks(painter);
    drawHklScans(painter);
    drawTicks(painter);
    drawLabels(painter);
    drawLegend(painter);

    lineWidth = screenLineWidth;
    pScale = 1.0;

    fontTitle = oFontTitle;
    fontAxis = oFontAxis;
    fontTicks = oFontTicks;
    fontLegend = oFontLegend;
    darkTheme = oDarkTheme;

    resize(width(), height());
    update();
}

/*
 * renders the graph to a bitmap
 */
void GraphWindow::renderBitmap(const QString &s, int w, int h)
{
    QPixmap pix = renderBitmap(w, h, settings->value("graph/rasterBackgroundFilled", false).toBool());

    if (pix.isNull()) {
        return;
    }

    pix.save(s, "PNG");
}

QPixmap GraphWindow::renderBitmap(int w, int h, bool fillbg)
{
    if ((w <= 0) || (h <= 0)) {
        return QPixmap();
    }

    int screenLineWidth = lineWidth;
    QPixmap pix(w, h);

    if (fillbg) {
        pix.fill(bgColor);
    } else {
        pix.fill(Qt::transparent);
    }

    // leave the reference hard-coded to obtain resolution-independent font sizes
    pScale = qMin(double(pix.width()) / 1152.0, double(pix.height()) / 768.0);

    QFontInfo fiTitle(oFontTitle);
    QFontInfo fiAxis(oFontAxis);
    QFontInfo fiTicks(oFontTicks);
    QFontInfo fiLegend(oFontLegend);

    fontTitle.setPixelSize(int(pScale * fiTitle.pixelSize()));
    fontAxis.setPixelSize(int(pScale * fiAxis.pixelSize()));
    fontTicks.setPixelSize(int(pScale * fiTicks.pixelSize()));
    fontLegend.setPixelSize(int(pScale * fiLegend.pixelSize()));
    bool oDarkTheme = darkTheme;
    darkTheme = false;

    QPainter painter;
    painter.begin(&pix);
        calcMargins(painter, false);

        painter.setRenderHints(QPainter::Antialiasing | QPainter::TextAntialiasing | QPainter::SmoothPixmapTransform, true);

        initPlot(pix.rect());
        drawWindow(painter);
        drawHighlightedRegions(painter);
        drawPlot(painter, false);
        drawAnchorPoints(painter);
        drawHklTicks(painter);
        drawHklScans(painter);
        drawHklReferenceLines(painter);
        drawTicks(painter);
        drawLabels(painter);
        drawLegend(painter);
    painter.end();

    lineWidth = screenLineWidth;
    pScale = 1.0;

    fontTitle = oFontTitle;
    fontAxis = oFontAxis;
    fontTicks = oFontTicks;
    fontLegend = oFontLegend;
    darkTheme = oDarkTheme;

    return pix;
}

void GraphWindow::renderSvg(QBuffer *buf, double a)
{
    QSvgGenerator svg;
    svg.setOutputDevice(buf);
    svg.setTitle(scanControl->getSampleId());

    // using an aspect ratio of sqrt(2) : 1
    double w = 27.0 * double(svg.resolution()) / 2.54;
    double h = w / (qFuzzyIsNull(a) ? sqrt(2.0) : a);
    QRectF pageRect(0.0, 0.0, w, h);

    svg.setSize(QSize(int(w), int(h)));
    svg.setViewBox(pageRect);

    pScale = settings->value("graph/svgExportFontSize", 1.0).toDouble(); // smaller values reduce font size

    QFontInfo fiTitle(oFontTitle);
    QFontInfo fiAxis(oFontAxis);
    QFontInfo fiTicks(oFontTicks);
    QFontInfo fiLegend(oFontLegend);
    bool oDarkTheme = darkTheme;
    darkTheme = false;

    fontTitle.setPixelSize(int(pScale * fiTitle.pixelSize()));
    fontAxis.setPixelSize(int(pScale * fiAxis.pixelSize()));
    fontTicks.setPixelSize(int(pScale * fiTicks.pixelSize()));
    fontLegend.setPixelSize(int(pScale * fiLegend.pixelSize()));

    QPainter painter;
    painter.begin(&svg);
        calcMargins(painter, false);

        painter.setRenderHints(QPainter::Antialiasing | QPainter::TextAntialiasing | QPainter::SmoothPixmapTransform, true);
        initPlot(pageRect.toRect());
        drawWindow(painter, true);
        drawHighlightedRegions(painter);
        drawPlot(painter, true);
        drawAnchorPoints(painter);
        drawHklTicks(painter);
        drawHklScans(painter);
        drawHklReferenceLines(painter);
        drawTicks(painter);
        drawLabels(painter);
        drawLegend(painter);
    painter.end();

    pScale = 1.0;

    fontTitle = oFontTitle;
    fontAxis = oFontAxis;
    fontTicks = oFontTicks;
    fontLegend = oFontLegend;
    darkTheme = oDarkTheme;
}

/*
 * renders the graph to a SVG file
 */
void GraphWindow::saveSvg(const QString &s)
{
    QBuffer buf;
    buf.open(QBuffer::ReadWrite);
    renderSvg(&buf);

    QFile f(s);

    if (f.open(QIODevice::WriteOnly | QIODevice::Truncate))     {
        QTextStream str(&f);
        buf.seek(0);
        str << buf.readAll();
        f.close();
    }

    buf.close();
}

/*
 * returns an SVG file as a string
 */
QByteArray GraphWindow::getSvg(double a)
{
    QBuffer buf;
    buf.open(QBuffer::ReadWrite);
    renderSvg(&buf, a);

    buf.seek(0);
    QByteArray data = buf.readAll();

    buf.close();
    return data;
}

void GraphWindow::setXaxis2theta()
{
    if (xScaling == XSCALETWOTHETA) return;

    double amin, amax, imin, imax;
    getZoomRange(amin, amax, imin, imax);

    xScaling = XSCALETWOTHETA;

    calcLimits();

    if (zoomed) {
        setZoomRange(amin, amax, imin, imax, true);
    } else {
        resetZoom();
    }

    forceUpdate();
}

void GraphWindow::setXaxisD()
{
    if (xScaling == XSCALED) return;

    double amin, amax, imin, imax;
    getZoomRange(amin, amax, imin, imax);

    xScaling = XSCALED;

    calcLimits();

    if (zoomed) {
        setZoomRange(amin, amax, imin, imax, true);
    } else {
        resetZoom();
    }

    forceUpdate();
}

void GraphWindow::setYaxisLinear()
{
    double amin, amax, imin, imax;
    getZoomRange(amin, amax, imin, imax);

    if (yScaling == YSCALESQRT) {
        imin = getInvSqrt(imin);
        imax = getInvSqrt(imax);
    }

    if (yScaling == YSCALELOG10) {
        imin = getInvLog10(imin);
        imax = getInvLog10(imax);
    }

    yScaling = YSCALELIN;

    calcLimits();

    if (zoomed) {
        setZoomRange(amin, amax, imin, imax, true);
    } else {
        resetZoom();
    }

    forceUpdate();
}

void GraphWindow::setYaxisSqrt()
{
    double amin, amax, imin, imax;
    getZoomRange(amin, amax, imin, imax);

    if (yScaling == YSCALELOG10) {
        imin = getInvLog10(imin);
        imax = getInvLog10(imax);
    }

    yScaling = YSCALESQRT;

    calcLimits();

    if (zoomed) {
        setZoomRange(amin, amax, getSqrt(imin), getSqrt(imax), true);
    } else {
        resetZoom();
    }

    forceUpdate();
}

void GraphWindow::setYaxisLog10()
{
    double amin, amax, imin, imax;
    getZoomRange(amin, amax, imin, imax);

    if (yScaling == YSCALESQRT) {
        imin = getInvSqrt(imin);
        imax = getInvSqrt(imax);
    }

    yScaling = YSCALELOG10;

    calcLimits();

    if (zoomed) {
        setZoomRange(amin, amax, getLog10(imin), getLog10(imax), true);
    } else {
        resetZoom();
    }

    forceUpdate();
}

void GraphWindow::zoomFromZeroY()
{
    double amin, amax, imin, imax;
    getZoomRange(amin, amax, imin, imax);

    calcLimits();
    setZoomRange(amin, amax, 0.0, imax, true);

    forceUpdate();
}

void GraphWindow::zoomFromMinY()
{
    if (!scanControl->count()) return;

    double amin, amax, imin, imax;
    getZoomRange(amin, amax, imin, imax);

    double currentScanMinInt = scanControl->first()->minIntensity(amin, amax);
    imin = currentScanMinInt < 0.0 ? std::numeric_limits<double>::max() : currentScanMinInt;

    for (int i = 1; i < scanControl->count(); ++i) {
        currentScanMinInt = scanControl->at(i)->minIntensity(amin, amax);
        if (currentScanMinInt >= 0.0) imin = qMin(currentScanMinInt, imin);
    }

    calcLimits();

    if (yScaling == YSCALELIN) setZoomRange(amin, amax, imin, imax, true);
    if (yScaling == YSCALESQRT)   setZoomRange(amin, amax, getSqrt(imin), imax, true);
    if (yScaling == YSCALELOG10)  setZoomRange(amin, amax, getLog10(imin), imax, true);

    forceUpdate();
}

void GraphWindow::integrateScan()
{
    emit sigIntegration(integrationRangeStart, integrationRangeEnd);
}

void GraphWindow::copyPixmapToClipboard()
{
    QColor oldMarginColor = marginColor;
    QColor oldBgColor = bgColor;
    bool oldDarkTheme = darkTheme;
    darkTheme = false;

    marginColor = Qt::white;
    bgColor = Qt::white;
    forceUpdate();

    int w = settings->value("graph/rasterResolutionWidth", 1536).toInt();
    int h = settings->value("graph/rasterResolutionHeight", 1024).toInt();

    QPixmap px = renderBitmap(w, h, true);
    QClipboard *clipBoard = qApp->clipboard();
    clipBoard->setPixmap(px);

    qDebug() << QString("GraphWindow::doClipboardVisualAction(): Pixmap of size %1 x %2 copied to clipboard").arg(w).arg(h);

    // flicker the background color to grey to give visual feedback
    marginColor = Qt::lightGray;
    bgColor = Qt::lightGray;

    forceUpdate();
    qApp->processEvents();

    marginColor = oldMarginColor;
    bgColor = oldBgColor;
    darkTheme = oldDarkTheme;

    QTimer::singleShot(200, this, SLOT(forceUpdate()));
}

void GraphWindow::overrideWaveLength(double ka1, bool characteristic)
{
    hasOverrideWaveLength = true;

    // ka1 in Angstrom
    if (characteristic) {
        scanControl->getFixedWaveLength(waveLength[0], waveLength[1], waveLength[2], ka1);
    } else {
        waveLength[0] = ka1;
        waveLength[1] = ka1;
        waveLength[2] = ka1;
    }

    forceUpdate();
}

/** EOF **/


