/***************************************************************************
                          graphwindow.h  -  description
                             -------------------
    begin                : Die Jun 17 2003
    copyright            : (C) 2003 by Nicola Doebelin
    email                : ndoebelin@gmx.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef GRAPHWINDOW_H
#define GRAPHWINDOW_H

#include <QtCore>
#include <QMenu>
#include <QWidget>
#include <QRubberBand>
#include <QPainter>
#include <QPrinter>
#include <QMap>
#include <QVector>
#include <QLabel>
#include <math.h>

#include "abstractgraphview.h"
#include "graphwindowlegend.h"

#include "../../../libXrdIO/scan.h"
#include "../../../libXrdIO/hkl.h"
#include "../../../libXrdIO/import/importhandler.h"
#include "../../../libXrdIO/structs.h"
#include "../../../libXrdIO/functions.h"

/**
  *@author Nicola Doebelin
  */

struct tTip {
    QRect pos;
    QStringList str;
};

struct ScanStyle {
    QColor color;
    int pointSymbol;
    int hklSymbol;

    ScanStyle() : color(), pointSymbol(), hklSymbol() {}
    ScanStyle(QColor c, int ps, int hs) : color(c), pointSymbol(ps), hklSymbol(hs) {}
};

enum Scale {YSCALELIN, YSCALESQRT, YSCALELOG10, YSCALELOGIT, XSCALETWOTHETA, XSCALED};
enum PeakPreviewMode {PPMNONE, PPMLIN, PPMPEAK};

class GraphWindow : public AbstractGraphView  {
   Q_OBJECT

public:
  GraphWindow(GraphDataController *c, QWidget *parent=nullptr);
  ~GraphWindow();

  void initSettings();
  bool setStatus(int);

  void changeGraphStyles();

  void moveLeft( double );
  void moveRight( double );
  void moveUp( double );
  void moveDown( double );
  inline double getTmax()            {return maxangAbs;}
  inline double getTmin()            {return 0;}
  inline double getImax()            {return maxi;}
  inline double getImin()            {return mini;}
  inline double getWaveLength()      {return waveLength[0];} //in Angstrom
  void getMaxRange(double &, double &, double &, double &);
  void getZoomRange(double &, double &, double &, double &);
  void setZoomRange(double, double, double, double, bool checkLimits = true);
  inline double getScanMaxIntensity() {return scanMaxIntensity;}
  void print(QPrinter &, QPainter &);
  void setIntegralRanges(const QVector<QPointF> &, const QStringList &);
  void setAnchorPoints(const QVector<global::AnchorPoint> &a) {anchorPoints = a;}

  void setReferenceReflections(const Scan &);

  QString tip(const QPoint &);
  QString getXaxisUnit();
  inline int getNumberOfScans() {return graphHeap.size();} /** deprecated **/
  inline int size() {return graphHeap.size();}             /** deprecated **/
  inline int count() {return graphHeap.size();}            /** deprecated **/
  inline bool isZoomed() {return zoomed;}
  void setDrawPhaseScans(bool b);
  inline Scale getXScaling() const {return xScaling;}
  inline Scale getYScaling() const {return yScaling;}

  void renderBitmap(const QString &s, int, int);
  QPixmap renderBitmap(int, int, bool fillbg = false);

  void saveSvg(const QString &);
  QByteArray getSvg(double a); // h = w / a

  void normalizeHkl(Scan *);

  void yOffsetUp();
  void yOffsetDown();
  void xOffsetRight();
  void xOffsetLeft();
  void xyOffsetReset();
  void resetZoom();
  inline double angularCorrection(double tt) {return global::Functions::angularCorrection(tt, eps1, eps2, eps3);}

  void setPeakPreviewMode(PeakPreviewMode);
  void setRangeSelectMode(bool);
  void overrideWaveLength(double, bool characteristic = true);

private:
  bool drawRubberBand, zoomed, antiAliasing,
       crossHair, strucScaling, noiseCursor, inspector,
       dragging, specLines, drawLivePeakIntegral,
       rangeSelectMode, fillActive;

  ImportHandler *ihandler;
  QColor bgColor, activeColor, completedColor, idleColor, peakSelectColor, printingBgColor;
  QColor marginColor;
  QColor axisColor;
  QColor integralRangeColor;
  QColor integralCenterLineColor;
  QList<QColor> colorTable;
  QList<uint> styleTable;
  QList<ScanStyle> scanStyles;
  QPixmap pBuffer;
  int margin_t, margin_b, margin_l, margin_r, margin_o, margin_m, margin_i;
  double difOffset;
  unsigned int points;
  double minang, maxang, minangAbs, maxangAbs;
  QPoint point_a;
  QRect plot, canvas, legendRect;
  QString completeFileName, fileName;
  unsigned int startpoint, endpoint;
  double startintens, endintens, maxi, mini;
  double pScale;
  bool drawLeg, drawRefls;
  bool recalcPlot;
  bool useBackgroundColors;
  bool hasOverrideWaveLength;
  QRubberBand *rubberBand;
  int lineWidth;
  int symbolSize;
  int printingLineWidth;
  Scale xScaling, yScaling;
  double waveLength[3];
  QFont fontTitle, fontAxis, fontTicks, fontLegend;
  QFont oFontTitle, oFontAxis, oFontTicks, oFontLegend;
  QMap<int, tTip> tt;
  QVector<Hkl> strucRefls;
  QVector<Scan> graphHeap;
  Scan refStructure;
  QVector<global::AnchorPoint> anchorPoints;
  QMap<QString, QColor> hklColorList;
  double scanMaxIntensity;
  int bkgrScanNumber;
  double prevStrucScaleFactor;
  double prevScanScaleFactor;
  double integrationRangeStart, integrationRangeEnd;
  bool hasActiveIntegrationRange;
  QString geometry;
  QTimer updateCoordinateTimer;
  // bool drawPhaseScans;
  QString fileUid;
  bool useCountsPerSecond;
  bool showGridMajor, showGridMinor;
  QMap<QString, double> tungstenLines;
  int tickDensityX, tickDensityY;
  double eps1, eps2, eps3;
  QVector<QPointF> integralRanges;
  QStringList integralRangeNames;
  QPoint clickedPosition;
  QPoint firstDoubleClickPoint;
  PeakPreviewMode peakPreviewMode;
  bool darkTheme;
  QRect leftMarginRect;
  QRect topMarginRect;
  QRect rightMarginRect;
  QRect bottomMarginRect;
  QMenu *contextMenuYaxis;
  QMenu *contextMenuXaxis;
  QMenu *contextMenuDataArea;

  GraphWindowLegend *legend;

  bool event(QEvent *);
  void paintEvent(QPaintEvent *);
  void resizeEvent(QResizeEvent *);
  void mousePressEvent(QMouseEvent *);
  void mouseReleaseEvent(QMouseEvent *);
  void mouseMoveEvent(QMouseEvent *);
  void wheelEvent(QWheelEvent *);
  void mouseDoubleClickEvent(QMouseEvent *);
  void keyPressEvent(QKeyEvent *);
  void keyReleaseEvent(QKeyEvent *);

  void mouseButtonLeft(QMouseEvent *);
  void mouseButtonMiddle(QMouseEvent *);
  void mouseButtonRight(QMouseEvent *);

  void mouseButtonStartStrucScaling(QMouseEvent *);
  void mouseButtonStartRubberBand(QMouseEvent *);
  void mouseButtonStartDragging(QMouseEvent *);
  void mouseButtonStartPeakIntegral(QMouseEvent *);

  void mouseButtonStopStruScaling(QMouseEvent *);
  void mouseButtonStopRubberBand(QMouseEvent *);
  void mouseButtonStopDragging(QMouseEvent *);
  void mouseButtonStopPeakIntegral(QMouseEvent *);

  void mouseMoveStrucScaling(QMouseEvent *);
  void mouseMoveRubberBand(QMouseEvent *);
  void mouseMoveDragging(QMouseEvent *);

  void calcLimits();
  void renderSvg(QBuffer *, double a = 0.0);

  int getXcoord(double);
  int getYcoord(double);
  double getRealX(int);
  double getRealY(int);

  void zoom(QPoint, QPoint);
  void zoomStep(QPoint, int, bool, bool);

  void initPlot(const QRect &);
  void calcMargins(const QPainter &, bool border);
  const Scan *getBackgroundMetrics(int, double &, double &);

  void drawWindow(QPainter &, bool replaceChars = false);
  void drawTicks(QPainter &);
  void drawLabels(QPainter &);
  void drawLegend(QPainter &);
  void drawHighlightedRegions(QPainter &);
  void drawHklTicks(QPainter &);
  void drawHklScans(QPainter &);
  void drawHklReferenceLines(QPainter &);
  void drawHklScanLines(QPainter &, const Scan *, const Scan *, int);
  void drawPlot(QPainter &, bool clipManually);
  void drawCrossHairCursor(QPainter &);
  void drawNoiseCursor(QPainter &);
  void drawSpecLineCursor(QPainter &);
  void drawInspector(QPainter &);
  void drawMouseBackgroundRegion(QPainter &, int, int);
  void drawAnchorPoints(QPainter &);
  void drawPeakPreviews(QPainter &);
  void drawRange(QPainter &);

  QPolygon getScanScreenPolygon(const Scan *, bool clipManually);
  QPolygon getBackgroundScreenPolygon(bool clipManually);
  QPolygon appendBackgroundPolygon(const QPolygon &, const QPolygon &);
  void broadcastCoordinates(const QPoint &);
  bool showToolTip(QEvent *);
  void setBackgroundIndex();
  void setupNewScans();

  void regionSetBegin(int);
  void regionSetEnd(int);
  QPoint interpolate(const QPoint &, const QPoint &, int);
  void copyPixmapToClipboard();

  inline double getLog10(double d) {return d < 1.0 ? d < -1.0 ? -log10(-d) : 0.0 : log10(d);}
  inline double getSqrt(double d)  {return d < 0.0 ? -sqrt(-d) : sqrt(d);}

  inline double getInvLog10(double d) {return d < 0.0 ? -pow(10.0, -d) : pow(10.0, d);}
  inline double getInvSqrt(double d)  {return d < 0.0 ? -d * d : d * d;}

  void integrateScan();

  void rightClickPlot(QMouseEvent *);
  void rightClickLeftMargin(QMouseEvent *);
  void rightClickBottomMargin(QMouseEvent *);
  void rightClickLegend(QMouseEvent *);

public slots:
  void resetView();
  void updateView();
  void setIdle();
  void setActive();
  void setComplete();
  void slotSetDrawLegend(bool);
  void slotSetDrawRefls(bool);
  void forceUpdate();
  void copy();

private slots:
  void calcCoordinates(const QPoint &, double &x, double &y, double &d);
  void setXaxis2theta();
  void setXaxisD();
  void setYaxisLinear();
  void setYaxisSqrt();
  void setYaxisLog10();
  void resetZoomX();
  void resetZoomY();
  void zoomFromZeroY();
  void zoomFromMinY();

signals:
  void sigCoordinates( double, double, double );
  void sigDoubleClickA(double, double, double);
  void sigDoubleClickB(double, double, double);
  void sigIntegration(double, double);
  void sigClearIntegrals();
  void sigPeakPreviewPoints(QPointF, QPointF, PeakPreviewMode);
  void sigRangePoints(QPointF, QPointF);
  void sigCursorMessage(QString);
};

#endif
