/***************************************************************************
                          graphwindowlegend.h  -  description
                             -------------------
    begin                : Mon Feb 01 2021
    copyright            : (C) 2021 by Nicola Doebelin
    email                : ndoebelin@gmx.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "graphwindowlegend.h"
#include <QPainter>
#include <QGuiApplication>
#include <QMouseEvent>

GraphWindowLegend::GraphWindowLegend(GraphDataController *c, QWidget *parent)
    : AbstractGraphView(c, parent)
{
    standardColors.append(QColor(0, 0, 0, 255));
    standardColors.append(QColor(255, 0, 0, 255));
    standardColors.append(QColor(0, 0, 255, 255));
    standardColors.append(QColor(0, 0, 0, 255));

    contextMenuLegend   = new QMenu(this);
    actLegendShow = new QAction(tr("Show legend"));
    actLegendFrame = new QAction(tr("Draw frame"));
    actLegendFill  = new QAction(tr("Fill background"));

    connect(actLegendShow, SIGNAL(triggered(bool)), this, SLOT(toggleLegendVisibility()));
    connect(actLegendFrame, SIGNAL(triggered(bool)), this, SLOT(toggleLegendFrame()));
    connect(actLegendFill, SIGNAL(triggered(bool)), this, SLOT(toggleLegendFill()));

    actLegendShow->setCheckable(true);
    actLegendFrame->setCheckable(true);
    actLegendFill->setCheckable(true);

    contextMenuLegend->addAction(actLegendShow);
    contextMenuLegend->addAction(actLegendFrame);
    contextMenuLegend->addAction(actLegendFill);
}

void GraphWindowLegend::initSettings()
{
    lineWidth = settings->value("graph/lineWidth", 1).toInt();
    boundingBoxIsDrawn = settings->value("graph/drawLegendBoundingBox", false).toBool();
    backgroundIsFilled = settings->value("graph/fillLegendBox", true).toBool();
    isDrawn = settings->value("graph/drawLegend", true).toBool();

    if (settings->isDarkMode()) bgColor = QGuiApplication::palette().color(QPalette::Base);
    else                        bgColor = Qt::white;
}

void GraphWindowLegend::updateView()
{
    colors.clear();
    labels.clear();
    lineWidths.clear();

    bool addFn = scanControl->countFiles() > 1;

    const Scan *scan = nullptr;

    for (int i = 0; i < scanControl->count(); ++i) {
        scan = scanControl->at(i);

        if (!scan)              continue;
        if (!scan->isVisible()) continue;

        QString str = scan->name();

        if (addFn) {
            QFileInfo fi(scan->sourceFileName());
            str += QString(" (%1)").arg(fi.fileName());
        }

        labels.append(str);

        if (scan->color().isValid()) {
            colors.append(settings->isDarkMode() ? global::Functions::colorToDarkMode(scan->color()) : scan->color());
        } else {
            if (i < colors.size()) {
                colors.append(settings->isDarkMode() ? global::Functions::colorToDarkMode(standardColors.at(i)) : standardColors.at(i));
            } else {
                colors.append(settings->isDarkMode() ? Qt::white : Qt::black);
            }
        }

        lineWidths.append(scan->lineWidth() < lineWidth ? lineWidth : scan->lineWidth());
    }
}

void GraphWindowLegend::redraw(QPainter &p, const QFont &f, const QRect &plt)
{
    bRect = QRect();
    if (!isDrawn)      return;
    if (!p.isActive()) return;
    if (plt.isNull())  return;

    p.save();
    QFontMetrics fm(f);
    updateBoundingRect(fm, plt);
    drawBackground(p);
    drawContent(p, f, fm);
    p.restore();
}

void GraphWindowLegend::updateBoundingRect(const QFontMetrics &fm, const QRect &plt)
{
    lTextWidth = 0;

    for (int i = 0; i < labels.size(); ++i) {
        lTextWidth = qMax(fm.boundingRect(labels.at(i)).width(), lTextWidth);
    }

    lLineSpacing = fm.lineSpacing();
    lTopBoundingBox = plt.top() + fm.lineSpacing() + fm.ascent();
    lRightBoundingBox = lTextWidth + fm.boundingRect("MM").width();
    lLineLength = fm.boundingRect("MMMM").width();
    lHorizontalSpacing = fm.boundingRect("M").width();

    lLeftBoundingBox = plt.right() - lRightBoundingBox - 2*lHorizontalSpacing - lLineLength;
    lLeftText = plt.right() - lRightBoundingBox - lHorizontalSpacing;
    lTopText = lTopBoundingBox;

    bRect = QRect(lLeftBoundingBox - lHorizontalSpacing,
                  lTopBoundingBox - int(fm.height() / 2.0),
                  lTextWidth + 2 * lHorizontalSpacing + lLeftText - lLeftBoundingBox,
                  labels.size() * lLineSpacing + int(fm.descent()));
}

void GraphWindowLegend::drawBackground(QPainter &p)
{
    if (!backgroundIsFilled && !boundingBoxIsDrawn) return;

    QPen pen(QColor(0, 0, 0, boundingBoxIsDrawn ? 255 : 0));
    QBrush brush;

    if (backgroundIsFilled) brush = QBrush(settings->isDarkMode() ? bgColor : Qt::white);
    if (boundingBoxIsDrawn) pen = QPen(settings->isDarkMode() ? Qt::white : Qt::black);

    p.setPen(pen);
    p.setBrush(brush);
    p.drawRect(bRect);
}

void GraphWindowLegend::drawContent(QPainter &p, const QFont &f, const QFontMetrics &fm)
{
    int v = lTopText;
    p.setFont(f);

    for (int i = 0; i < qMin(labels.size(), qMin(colors.size(), lineWidths.size())); i++) {
        QPen pen(colors.at(i));
        pen.setWidth(lineWidths.at(i));

        p.setPen(pen);
        p.drawLine(lLeftBoundingBox, v, lLeftBoundingBox + lLineLength, v);
        p.drawText(lLeftText, v + int(fm.ascent() / 2.0), labels.at(i));
        v += lLineSpacing;
    }
}

void GraphWindowLegend::setDrawn(bool b)
{
    isDrawn = b;
    settings->setValue("graph/drawLegend", isDrawn);
}

void GraphWindowLegend::toggleIsDrawn()
{
    setDrawn(!isDrawn);
}

void GraphWindowLegend::showContextMenu(QMouseEvent *e)
{
    actLegendShow->setChecked(isDrawn);
    actLegendFrame->setChecked(settings->value("graph/drawLegendBoundingBox", false).toBool());
    actLegendFill->setChecked(settings->value("graph/fillLegendBox", true).toBool());

    contextMenuLegend->exec(e->globalPos());
}

void GraphWindowLegend::toggleLegendVisibility()
{
    toggleIsDrawn();
    emit sigRequestRedraw();
}

void GraphWindowLegend::toggleLegendFrame()
{
    boundingBoxIsDrawn = !boundingBoxIsDrawn;
    settings->setValue("graph/drawLegendBoundingBox", boundingBoxIsDrawn);
    emit sigRequestRedraw();
}

void GraphWindowLegend::toggleLegendFill()
{
    backgroundIsFilled = !backgroundIsFilled;
    settings->setValue("graph/fillLegendBox", backgroundIsFilled);
    emit sigRequestRedraw();
}

