/***************************************************************************
                          peakfitwidget.cpp  -  description
                             -------------------
    begin                : Tue Nov 16 22:00:00 CEST 2019
    copyright            : (C) 2019 by Nicola Doebelin
    email                : ndoebelin@gmx.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "peakfitwidget.h"
#include "ui_peakfitwidget.h"

#include <QComboBox>
#include <QSpinBox>
#include <QShowEvent>
#include <QCloseEvent>
#include <QTableWidget>
#include <QMessageBox>
#include <QInputDialog>
#include <QToolButton>
#include <QList>
#include <limits.h>
#include <QDebug>

PeakFitWidget::PeakFitWidget(GraphDataController *c, GraphWindow *v, QWidget *parent) :
    AbstractGraphView(c, parent),
    ui(new Ui::PeakFitWidget)
{
    graphView = v;

    ui->setupUi(this);

    wasUsed = false;
    curveNumber = 1;

    ui->treeWidgetFunctions->setColumnCount(3);
    ui->treeWidgetFunctions->setHeaderLabels(QStringList() << tr("Parameter") << tr("Variable") << tr("Value"));

    ui->treeWidgetVariables->setColumnCount(4);
    ui->treeWidgetVariables->setHeaderLabels(QStringList() << tr("Variable") << tr("Value") << tr("Lower boundary") << tr("Upper boundary"));

    initSettings();

    connect(ui->toolButtonRemovePeak, SIGNAL(clicked()), this, SLOT(removeCurve()));
    connect(ui->toolButtonPinTemporary, SIGNAL(clicked()), this, SLOT(keepTemporary()));
    connect(ui->toolButtonFit, SIGNAL(clicked()), this, SLOT(fit()));
    connect(ui->toolButtonClearAll, SIGNAL(clicked()), this, SLOT(clearAll()));
    connect(ui->pushButtonAddPeak, SIGNAL(toggled(bool)), this, SLOT(toggleAddPeakMode(bool)));
    connect(ui->pushButtonSelectRange, SIGNAL(toggled(bool)), this, SLOT(toggleSelectRangeMode(bool)));
    connect(ui->comboBoxPeakFunction, SIGNAL(currentIndexChanged(int)), this, SLOT(togglePeakFunction(int)));
    connect(ui->checkBoxStepSize, SIGNAL(toggled(bool)), ui->doubleSpinBoxStepSize, SLOT(setEnabled(bool)));
    connect(ui->treeWidgetFunctions, SIGNAL(itemDoubleClicked(QTreeWidgetItem*,int)), this, SLOT(functionDoubleClicked(QTreeWidgetItem*,int)));
    connect(ui->treeWidgetVariables, SIGNAL(itemDoubleClicked(QTreeWidgetItem*,int)), this, SLOT(variableDoubleClicked(QTreeWidgetItem*,int)));
    connect(ui->treeWidgetFunctions, SIGNAL(itemClicked(QTreeWidgetItem*,int)), this, SLOT(activeCurveChanged(QTreeWidgetItem*,int)));
    connect(ui->toolButtonResetControl, SIGNAL(clicked()), this, SLOT(undoControls()));
    connect(ui->toolButtonApplyToAll, SIGNAL(clicked()), this, SIGNAL(applyToAll()));
    connect(ui->checkBoxControlsAuto, SIGNAL(toggled(bool)), ui->controlsContainer, SLOT(setDisabled(bool)));

    if (graphView) {
        connect(graphView, SIGNAL(sigPeakPreviewPoints(QPointF, QPointF, PeakPreviewMode)), this, SLOT(peakPreviewPoints(QPointF, QPointF, PeakPreviewMode)));
        connect(graphView, SIGNAL(sigRangePoints(QPointF,QPointF)), this, SLOT(rangePoints(QPointF,QPointF)));
        connect(graphView, SIGNAL(sigCursorMessage(QString)), this, SLOT(cursorMessage(QString)));
    }
}

PeakFitWidget::~PeakFitWidget()
{
    if (wasUsed) saveSettings();
    delete ui;
}

void PeakFitWidget::initSettings()
{
    ui->treeWidgetFunctions->header()->restoreState(settings->value("peakFitting/functionsHeader", QByteArray()).toByteArray());
    ui->treeWidgetVariables->header()->restoreState(settings->value("peakFitting/variablesHeader", QByteArray()).toByteArray());

    ui->comboBoxPeakFunction->addItem(LinearCurve::descriptionStatic(), int(PPMLIN));
    ui->comboBoxPeakFunction->addItem(GaussianCurve::descriptionStatic(), int(PPMPEAK));
    ui->comboBoxPeakFunction->addItem(LorentzianCurve::descriptionStatic(), int(PPMPEAK));
    ui->comboBoxPeakFunction->addItem(PseudoVoigtCurve::descriptionStatic(), int(PPMPEAK));
    ui->comboBoxPeakFunction->addItem(PearsonCurve::descriptionStatic(), int(PPMPEAK));

    ui->comboBoxPeakFunction->addItem(GaussianSplitCurve::descriptionStatic(), int(PPMPEAK));
    ui->comboBoxPeakFunction->addItem(LorentzianSplitCurve::descriptionStatic(), int(PPMPEAK));
    ui->comboBoxPeakFunction->addItem(PseudoVoigtSplitCurve::descriptionStatic(), int(PPMPEAK));
    ui->comboBoxPeakFunction->addItem(PearsonSplitCurve::descriptionStatic(), int(PPMPEAK));

    ui->comboBoxPeakFunction->addItem(QuadraticCurve::descriptionStatic(), int(PPMLIN));
    ui->comboBoxPeakFunction->addItem(CubicCurve::descriptionStatic(), int(PPMLIN));
    ui->comboBoxPeakFunction->addItem(Polynom4Curve::descriptionStatic(), int(PPMLIN));

    QList<QVariant> range = settings->value("peakFitting/range", QList<QVariant>()).toList();

    if (range.size() > 1) {
        ui->doubleSpinBoxRangeStart->setValue(range.at(0).toDouble());
        ui->doubleSpinBoxRangeEnd->setValue(range.at(1).toDouble());
    }

    ui->checkBoxStepSize->setChecked(settings->value("peakFitting/stepSizeCustom", false).toBool());
    ui->doubleSpinBoxStepSize->setValue(settings->value("peakFitting/stepSizeValue", 0.001).toDouble());
    ui->doubleSpinBoxStepSize->setEnabled(ui->checkBoxStepSize->isChecked());

    ui->spinBoxMaxIts->setValue(settings->value("peakFitting/controls/maxIts", 200).toInt());
    ui->doubleSpinBoxEpsX->setValue(settings->value("peakFitting/controls/epsX", 0.000001).toDouble());
    ui->doubleSpinBoxDiffStep->setValue(settings->value("peakFitting/controls/diffStep", 0.0001).toDouble());

    ui->checkBoxControlsAuto->setChecked(settings->value("peakFitting/controls/auto", false).toBool());
    ui->controlsContainer->setDisabled(ui->checkBoxControlsAuto->isChecked());
}

void PeakFitWidget::saveSettings()
{
    settings->setValue("peakFitting/functionsHeader", ui->treeWidgetFunctions->header()->saveState());
    settings->setValue("peakFitting/variablesHeader", ui->treeWidgetVariables->header()->saveState());

    settings->setValue("peakFitting/range", QList<QVariant>() << ui->doubleSpinBoxRangeStart->value() << ui->doubleSpinBoxRangeEnd->value());
    settings->setValue("peakFitting/stepSizeCustom", ui->checkBoxStepSize->isChecked());
    settings->setValue("peakFitting/setpSizeValue", ui->doubleSpinBoxStepSize->value());

    settings->setValue("peakFitting/controls/maxIts", ui->spinBoxMaxIts->value());
    settings->setValue("peakFitting/controls/epsX", ui->doubleSpinBoxEpsX->value());
    settings->setValue("peakFitting/controls/diffStep", ui->doubleSpinBoxDiffStep->value());

    settings->setValue("peakFitting/controls/auto", ui->checkBoxControlsAuto->isChecked());
}

void PeakFitWidget::updateView()
{
    if (!isShown) return;
}

void PeakFitWidget::toggleSelectRangeMode(bool b)
{
    if (!graphView) {
        bool oldState = ui->pushButtonSelectRange->blockSignals(true);
        ui->pushButtonSelectRange->setChecked(false);
        ui->pushButtonSelectRange->blockSignals(oldState);
        return;
    }

    if (b && ui->pushButtonAddPeak->isChecked()) {
        bool oldState = ui->pushButtonAddPeak->blockSignals(true);
        ui->pushButtonAddPeak->setChecked(false);
        graphView->setPeakPreviewMode(PPMNONE);
        ui->pushButtonAddPeak->blockSignals(oldState);
    }

    graphView->setRangeSelectMode(b);
}

void PeakFitWidget::toggleAddPeakMode(bool b)
{
    if (!graphView) {
        bool oldState = ui->pushButtonAddPeak->blockSignals(true);
        ui->pushButtonAddPeak->setChecked(false);
        ui->pushButtonAddPeak->blockSignals(oldState);
        return;
    }

    if (b && ui->pushButtonSelectRange->isChecked()) {
        bool oldState = ui->pushButtonSelectRange->blockSignals(true);
        ui->pushButtonSelectRange->setChecked(false);
        graphView->setRangeSelectMode(false);
        ui->pushButtonSelectRange->blockSignals(oldState);
    }

    if (b) {
        graphView->setPeakPreviewMode(PeakPreviewMode(ui->comboBoxPeakFunction->currentData(Qt::UserRole).toInt()));
    } else {
        graphView->setPeakPreviewMode(PPMNONE);
    }
}

void PeakFitWidget::togglePeakFunction(int)
{
    if (!graphView) return;

    if (ui->pushButtonAddPeak->isChecked()) {
        graphView->setPeakPreviewMode(PeakPreviewMode(ui->comboBoxPeakFunction->currentData(Qt::UserRole).toInt()));
    }
}

const Scan * PeakFitWidget::getActiveDataScan()
{
    QVector<Scan*> actScans = scanControl->activeScans();

    for (int i = 0; i < actScans.size(); ++i) {
        Scan::ScanTypes flags = actScans.at(i)->scanTypes();

        if (!flags.testFlag(Scan::TEMPORARY)) {
            return actScans.at(i);
        }
    }

    return scanControl->getScan(0);
}

bool PeakFitWidget::initSolver()
{
    double rStart = ui->doubleSpinBoxRangeStart->value();
    double rEnd   = ui->doubleSpinBoxRangeEnd->value();

    const Scan *curScan = getActiveDataScan();

    if (!curScan) {
        qDebug() << QString("PeakFitWidget::initSolver(): No active scan found");
        return false;
    }

    bool liveUpd = ui->checkBoxLiveUpdate->isChecked();
    int itsPerStep = ui->checkBoxControlsAuto->isChecked() ? 0 : ui->spinBoxMaxIts->value();

    if (liveUpd) itsPerStep = 1;

    double epsX = ui->checkBoxControlsAuto->isChecked() ? 0.0 : ui->doubleSpinBoxEpsX->value();
    double diffStep = ui->checkBoxControlsAuto->isChecked() ? 0.0001 : ui->doubleSpinBoxDiffStep->value();

    cfitManager.setControls(itsPerStep, epsX, diffStep);
    cfitManager.setData(curScan->mid(rStart, rEnd));

    QVector<double> val;
    QVector<double> loLim;
    QVector<double> upLim;

    readInitialValues(val, loLim, upLim);

    cfitManager.setCurves(curveList, val, loLim, upLim);
    return true;
}

void PeakFitWidget::fit()
{
    wasUsed = true;
    if (!ui->treeWidgetFunctions->topLevelItemCount()) return;

    qApp->setOverrideCursor(Qt::WaitCursor);

    bool oldState = ui->pushButtonAddPeak->blockSignals(true);
    ui->pushButtonAddPeak->setChecked(false);
    graphView->setPeakPreviewMode(PPMNONE);
    ui->pushButtonAddPeak->blockSignals(oldState);

    bool liveUpd = ui->checkBoxLiveUpdate->isChecked();
    int maxIts = ui->checkBoxControlsAuto->isChecked() ? 100 : ui->spinBoxMaxIts->value();
    // todo: check for convergence if auto is checked (for now: hard coded to 100 cycles)

    if (!liveUpd) maxIts = 1;

    QMap<QString, QVariant> report;
    emit protocol("Running curve fit...");

    for (int i = 0; i < maxIts; ++i) {
        if (!initSolver()) {
            emit protocol("    No active scan found. Exiting.");
            clearTemporary();
            qApp->restoreOverrideCursor();
            return;
        }

        QVector<double> fitValues = cfitManager.fit();
        report = cfitManager.getReport();

        emit protocol(QString("    %1: R2 = %2").arg(i).arg(report.value("R2", 0.0).toDouble()));

        if (fitValues.isEmpty()) {
            emit protocol(QString("    Convergence error\n  %1 - %2")
                          .arg(report.value("ExitCode").toInt())
                          .arg(report.value("ExitString").toString()));
            clearTemporary();
            qApp->restoreOverrideCursor();
            return;
        }

        writeFittedValues(fitValues);
        updateCurves();
        qApp->processEvents();
    }

    emit protocol(QString("    %1 - %2")
                  .arg(report.value("ExitCode").toInt())
                  .arg(report.value("ExitString").toString()));
    qApp->restoreOverrideCursor();
}

void PeakFitWidget::updateCurves()
{
    updateCurveValues();
    clearTemporary();

    if (!initSolver()) return;

    QList<Scan> curveFits = cfitManager.calculateCurve(ui->checkBoxStepSize->isChecked() ? ui->doubleSpinBoxStepSize->value() : -1.0, true);
    QMap<QUuid, QString> areas;

    QMap<int, QString> curveNames;
    for (int i = 0; i < ui->treeWidgetFunctions->topLevelItemCount(); ++i) {
        curveNames[i] = ui->treeWidgetFunctions->topLevelItem(i)->text(0);
    }

    bool ok;
    for (int i = 0; i < curveFits.size(); ++i) {
        Scan *cv = &(curveFits[i]);
        cv->setActive(false);
        if (curveNames.contains(i)) cv->setName(curveNames.value(i));
        scanControl->appendScan(*cv);
        currentScans.append(scanControl->last()->uid());

        QUuid cUid;
        double a = cfitManager.getArea(i, cUid, ok);
        if (ok) areas[cUid] = QString("%1").arg(a, 0, 'f', 6);
    }

    updateAreas(areas);
    scanControl->updateViews();
}

void PeakFitWidget::updateCurveValues()
{
    for (int t = 0; t < ui->treeWidgetFunctions->topLevelItemCount(); ++t) {
        QTreeWidgetItem *tIt = ui->treeWidgetFunctions->topLevelItem(t);
        if (!tIt) continue;

        updateCurveValues(tIt);
    }
}

void PeakFitWidget::updateCurveValues(const QTreeWidgetItem *tIt)
{
    QMap<QString, global::CurveFitVariable> vars = getVariablesByName();

    for (int c = 0; c < tIt->childCount(); ++c) {
        QTreeWidgetItem *cIt = tIt->child(c);
        if (!cIt) continue;

        QString val;

        if (vars.contains(cIt->text(1))) {
            val = vars.value(cIt->text(1)).value;
        }

        cIt->setText(2, val);
    }
}

void PeakFitWidget::updateAreas(const QMap<QUuid, QString> &m)
{
    for (int i = 0; i < ui->treeWidgetFunctions->topLevelItemCount(); ++i) {
        QTreeWidgetItem *tlIt = ui->treeWidgetFunctions->topLevelItem(i);
        int c = tlIt->childCount();

        if (c < 2) continue;

        QUuid tlUid = tlIt->data(0, Qt::UserRole).toUuid();
        QTreeWidgetItem *cIt = tlIt->child(c - 1);

        // if the child item has its UserRole set to "false", it is the Area item
        if (!cIt->data(0, Qt::UserRole).toBool() && m.contains(tlUid)) {
            cIt->setText(2, m.value(tlUid));
        }
    }
}

QPointF PeakFitWidget::uncorrectedPoint(const QPointF &pt)
{
    double eps1 = scanControl->getAngularCorrEPS1();
    double eps2 = scanControl->getAngularCorrEPS2();
    double eps3 = scanControl->getAngularCorrEPS3();

    return QPointF(pt.x() + global::Functions::angularCorrection(pt.x(), eps1, eps2, eps3), pt.y());
}

void PeakFitWidget::rangePoints(QPointF pA, QPointF pB)
{
    bool oldState = ui->pushButtonSelectRange->blockSignals(true);
    ui->pushButtonSelectRange->setChecked(false);
    ui->pushButtonSelectRange->blockSignals(oldState);

    graphView->setRangeSelectMode(false);

    ui->doubleSpinBoxRangeStart->setValue(uncorrectedPoint(pA).x());
    ui->doubleSpinBoxRangeEnd->setValue(uncorrectedPoint(pB).x());
}

void PeakFitWidget::peakPreviewPoints(QPointF pA, QPointF pB, PeakPreviewMode mode)
{
    Q_UNUSED(mode);

    if (qFuzzyCompare(pA.x(), pB.x())) return;

    QPointF pAc = uncorrectedPoint(pA);
    QPointF pBc = uncorrectedPoint(pB);

    double slope = (pBc.y() - pAc.y()) / (pBc.x() - pAc.x());
    double xoff = pAc.y() - pAc.x() * slope;

    double mu     = pAc.x();
    double sigma  = qAbs(pAc.x() - pBc.x()) * 0.5;
    double intens = qAbs(pAc.y() - pBc.y());

    if (ui->comboBoxPeakFunction->currentText() == LinearCurve::descriptionStatic())
        addCurve(new LinearCurve(),           QList<double>() << xoff << slope);

    if (ui->comboBoxPeakFunction->currentText() == GaussianCurve::descriptionStatic())
        addCurve(new GaussianCurve(),         QList<double>() << mu << sigma << intens);

    if (ui->comboBoxPeakFunction->currentText() == LorentzianCurve::descriptionStatic())
        addCurve(new LorentzianCurve(),       QList<double>() << mu << sigma << intens);

    if (ui->comboBoxPeakFunction->currentText() == PseudoVoigtCurve::descriptionStatic())
        addCurve(new PseudoVoigtCurve(),      QList<double>() << mu << sigma << intens << 0.5);

    if (ui->comboBoxPeakFunction->currentText() == PearsonCurve::descriptionStatic())
        addCurve(new PearsonCurve(),          QList<double>() << mu << sigma << intens << 2.0);

    if (ui->comboBoxPeakFunction->currentText() == GaussianSplitCurve::descriptionStatic())
        addCurve(new GaussianSplitCurve(),    QList<double>() << mu << sigma << sigma << intens);

    if (ui->comboBoxPeakFunction->currentText() == LorentzianSplitCurve::descriptionStatic())
        addCurve(new LorentzianSplitCurve(),  QList<double>() << mu << sigma << sigma << intens);

    if (ui->comboBoxPeakFunction->currentText() == PseudoVoigtSplitCurve::descriptionStatic())
        addCurve(new PseudoVoigtSplitCurve(), QList<double>() << mu << sigma << sigma << intens << 0.5 << 0.5);

    if (ui->comboBoxPeakFunction->currentText() == PearsonSplitCurve::descriptionStatic())
        addCurve(new PearsonSplitCurve(),     QList<double>() << mu << sigma << sigma << intens << 2.0 << 2.0);

    if (ui->comboBoxPeakFunction->currentText() == QuadraticCurve::descriptionStatic())
        addCurve(new QuadraticCurve(),        QList<double>() << xoff << slope << 0.0);

    if (ui->comboBoxPeakFunction->currentText() == CubicCurve::descriptionStatic())
        addCurve(new CubicCurve(),            QList<double>() << xoff << slope << 0.0 << 0.0);

    if (ui->comboBoxPeakFunction->currentText() == Polynom4Curve::descriptionStatic())
        addCurve(new Polynom4Curve(),         QList<double>() << xoff << slope << 0.0 << 0.0 << 0.0);

    updateCurves();
}

void PeakFitWidget::readInitialValues(QVector<double> &val, QVector<double> &loLim, QVector<double> &upLim)
{
    QMap<QString, int>  idxMap;

    for (int i = 0; i < ui->treeWidgetVariables->topLevelItemCount(); ++i) {
        QTreeWidgetItem *it = ui->treeWidgetVariables->topLevelItem(i);

        bool doFit = (it->checkState(0) == Qt::Checked);
        double startVal = it->text(1).toDouble();

        idxMap[it->text(0)] = i;
        val << startVal;

        if (doFit) {
            if (it->text(2) == "-inf") loLim << std::numeric_limits<double>::lowest();
            else                       loLim << it->text(2).toDouble();

            if (it->text(3) == "+inf") upLim << std::numeric_limits<double>::max();
            else                       upLim << it->text(3).toDouble();
        } else {
            loLim << startVal;
            upLim << startVal;
        }
    }

    for (int i = 0; i < ui->treeWidgetFunctions->topLevelItemCount(); ++i) {
        QTreeWidgetItem *it = ui->treeWidgetFunctions->topLevelItem(i);
        GenericCurve *cv = nullptr;

        for (int j = 0; j < curveList.size(); ++j) {
            if (curveList.at(j)->uid() == it->data(0, Qt::UserRole).toUuid()) {
                cv = curveList[j];
                break;
            }
        }

        if (!cv) continue;

        QVector<int> cidx;

        for (int j = 0; j < it->childCount(); ++j) {
            cidx << idxMap.value(it->child(j)->text(1));
        }

        cv->setIndices(cidx);
    }
}

void PeakFitWidget::writeFittedValues(const QVector<double> &val)
{
    for (int i = 0; i < qMin(val.size(), ui->treeWidgetVariables->topLevelItemCount()); ++i) {
        QTreeWidgetItem *it = ui->treeWidgetVariables->topLevelItem(i);
        it->setText(1, QString("%1").arg(val.at(i), 0, 'f', 6));
    }
}

void PeakFitWidget::removeCurve()
{
    if (!ui->treeWidgetFunctions->currentItem()) return;

    QUuid u = ui->treeWidgetFunctions->currentItem()->data(0, Qt::UserRole).toUuid();

    for (int i = ui->treeWidgetFunctions->topLevelItemCount(); i > 0; --i) {
        if (ui->treeWidgetFunctions->topLevelItem(i-1)->data(0, Qt::UserRole).toUuid() == u) {
            QTreeWidgetItem *it = ui->treeWidgetFunctions->takeTopLevelItem(i-1);
            delete it;
        }
    }

    for (int i = ui->treeWidgetVariables->topLevelItemCount(); i > 0; --i) {
        if (ui->treeWidgetVariables->topLevelItem(i-1)->data(0, Qt::UserRole).toUuid() == u) {
            QTreeWidgetItem *it = ui->treeWidgetVariables->takeTopLevelItem(i-1);
            delete it;
        }
    }

    for (int i = 0; i < curveList.size(); ++i) {
        if (curveList.at(i)->uid() == u) {
            GenericCurve *g = curveList.takeAt(i);
            delete g;
            break;
        }
    }

    updateCurves();
}

void PeakFitWidget::append()
{
    keepTemporary();
}

void PeakFitWidget::keepTemporary()
{
    scanControl->keepTemporary(currentScans);
    currentScans.clear();
    scanControl->updateViews();
}

void PeakFitWidget::clearTemporary()
{
    scanControl->removeScans(currentScans);
    currentScans.clear();
    scanControl->updateViews();
}

void PeakFitWidget::addCurve(GenericCurve *curve, const QList<double> &values)
{
    const Scan *curScan = getActiveDataScan();

    if (!curScan) {
        if (curve) delete curve;
        return;
    }

    double minang = ui->doubleSpinBoxRangeStart->value();
    double maxang = ui->doubleSpinBoxRangeEnd->value();
    double minss = curScan->stepSize();

    curveList.append(curve);

    int m = 1;

    QString name = QString(tr("#%1 %2")).arg(curveNumber).arg(curve->description());
    QTreeWidgetItem *curveIt = new QTreeWidgetItem(QStringList() << name);
    curveIt->setData(0, Qt::UserRole, curve->uid());
    curveIt->setData(1, Qt::UserRole, curve->type());

    QList<double> defaultValues = curve->defaultValues();
    QStringList lowerLimits = curve->lowerLimits(minang, maxang, minss);
    QStringList upperLimits = curve->upperLimits(minang, maxang, minss);
    QStringList valNames = curve->parameterNames();

    for (int i = 0; i < curve->nParameters(); ++i) {
        QString val = QString::number(i < values.count() ? values.at(i) : defaultValues.at(i), 'f', 6);
        QTreeWidgetItem *varIt = new QTreeWidgetItem(QStringList() << QString("$%1-%2").arg(curveNumber).arg(m)
                                                                    << val
                                                                    << lowerLimits.at(i)
                                                                    << upperLimits.at(i));

        varIt->setTextAlignment(1, Qt::AlignRight);
        varIt->setTextAlignment(2, Qt::AlignRight);
        varIt->setTextAlignment(3, Qt::AlignRight);

        varIt->setData(0, Qt::UserRole, curve->uid());
        varIt->setData(1, Qt::UserRole, QString("$%1-%2").arg(curveNumber).arg(m));
        varIt->setCheckState(0, Qt::Checked);

        QTreeWidgetItem *paramIt = new QTreeWidgetItem(QStringList() << valNames.at(i) << varIt->text(0));
        paramIt->setData(0, Qt::UserRole, true); // mark the item as editable by double click
        paramIt->setTextAlignment(2, Qt::AlignRight);
        curveIt->addChild(paramIt);

        ui->treeWidgetVariables->addTopLevelItem(varIt);
        ++m;
    }

    if (curve->hasArea()) appendAreaItem(curveIt);
    ui->treeWidgetFunctions->addTopLevelItem(curveIt);
    curveIt->setExpanded(true);
    ++curveNumber;
}

void PeakFitWidget::restoreCurve(const global::CurveFitCurve &cv, int type, int n)
{
    GenericCurve *curve = cfitHandler.createCurve(type);

    if (!curve) return;
    curveList.append(curve);

    QString name = QString(tr("#%1 %2")).arg(n).arg(curve->description());
    QTreeWidgetItem *curveIt = new QTreeWidgetItem(QStringList() << name);
    curveIt->setData(0, Qt::UserRole, curve->uid());
    curveIt->setData(1, Qt::UserRole, curve->type());

    for (int i = 0; i < qMin(cv.parameterNames.size(), cv.variables.size()); ++i) {
        QTreeWidgetItem *paramIt = new QTreeWidgetItem(QStringList() << cv.parameterNames.at(i) << cv.variables.at(i));
        paramIt->setData(0, Qt::UserRole, true); // mark the item as editable by double click
        paramIt->setTextAlignment(2, Qt::AlignRight);
        curveIt->addChild(paramIt);
    }

    if (curve->hasArea()) appendAreaItem(curveIt);
    ui->treeWidgetFunctions->addTopLevelItem(curveIt);
    curveIt->setExpanded(true);
}

void PeakFitWidget::appendAreaItem(QTreeWidgetItem *tl)
{
    if (!tl) return;

    QTreeWidgetItem *aIt = new QTreeWidgetItem;
    aIt->setTextAlignment(2, Qt::AlignRight);
    aIt->setData(0, Qt::UserRole, false); // mark the item as not editable by double click
    aIt->setText(0, tr("Area"));

    tl->addChild(aIt);
}

void PeakFitWidget::functionDoubleClicked(QTreeWidgetItem *it, int)
{
    if (!it) return;
    if (!it->parent()) return;
    if (!it->data(0, Qt::UserRole).toBool()) return; // marked as non-editable

    QStringList items;
    int c = 0;

    for (int i = 0; i < ui->treeWidgetVariables->topLevelItemCount(); ++i) {
        QString t = ui->treeWidgetVariables->topLevelItem(i)->text(0);
        items.append(t);
        if (t == it->text(1)) c = i;
    }

    bool ok;
    QString v = QInputDialog::getItem(this, tr("Select variable"), tr("Variable"), items, c, false, &ok);

    if (!ok) return;
    it->setText(1, v);
    updateCurves();
}

void PeakFitWidget::variableDoubleClicked(QTreeWidgetItem *it, int col)
{
    if (col == 0) return;

    QString txt = it->text(col);
    bool ok;
    txt = QInputDialog::getText(this, tr("Enter value"), tr("Value"), QLineEdit::Normal, txt, &ok);

    if (!ok) return;
    it->setText(col, txt);
    updateCurves();
}

void PeakFitWidget::clearAll()
{
    if (QMessageBox::question(this, tr("Clear all curves"), tr("Do you want to delete all curves and variables?")) == QMessageBox::No) return;

    bool oldState = ui->treeWidgetFunctions->blockSignals(true);
    ui->treeWidgetFunctions->clear();
    ui->treeWidgetFunctions->blockSignals(oldState);

    oldState = ui->treeWidgetVariables->blockSignals(true);
    ui->treeWidgetVariables->clear();
    ui->treeWidgetVariables->blockSignals(oldState);

    curveList.clear();
    clearTemporary();
    curveNumber = 1;
}

void PeakFitWidget::undoControls()
{
    ui->spinBoxMaxIts->setValue(0);
    ui->doubleSpinBoxEpsX->setValue(0.0);
    ui->doubleSpinBoxDiffStep->setValue(0.0001);
}

QList<global::CurveFitCurve> PeakFitWidget::getCurves()
{
    QList<global::CurveFitCurve> l;

    for (int i = 0; i < ui->treeWidgetFunctions->topLevelItemCount(); ++i) {
        QTreeWidgetItem *tlIt = ui->treeWidgetFunctions->topLevelItem(i);

        global::CurveFitCurve cv;
        cv.type = tlIt->data(1, Qt::UserRole).toInt();

        QStringList paramNames;
        QStringList paramVariables;
        QStringList paramValues;

        for (int j = 0; j < tlIt->childCount(); ++j) {
            paramNames.append(tlIt->child(j)->text(0));
            paramVariables.append(tlIt->child(j)->text(1));
            paramValues.append(tlIt->child(j)->text(2));
        }

        cv.parameterNames = paramNames;
        cv.variables      = paramVariables;
        cv.values         = paramValues;

        l.append(cv);
    }

    return l;
}

QMap<QString, global::CurveFitVariable> PeakFitWidget::getVariablesByName()
{
    QMap<QString, global::CurveFitVariable> m;

    for (int i = 0; i < ui->treeWidgetVariables->topLevelItemCount(); ++i) {
        QTreeWidgetItem *tlIt = ui->treeWidgetVariables->topLevelItem(i);

        global::CurveFitVariable cv;

        cv.name       = tlIt->text(0);
        cv.value      = tlIt->text(1);
        cv.lowerLimit = tlIt->text(2);
        cv.upperLimit = tlIt->text(3);
        cv.checkState = tlIt->checkState(0) == Qt::Checked ? "1" : "0";

        m[cv.name] = cv;
    }

    return m;
}

QList<global::CurveFitVariable> PeakFitWidget::getVariables()
{
    // not using ::getVariablesByName because we want to keep
    // the original sorting
    QList<global::CurveFitVariable> l;

    for (int i = 0; i < ui->treeWidgetVariables->topLevelItemCount(); ++i) {
        QTreeWidgetItem *tlIt = ui->treeWidgetVariables->topLevelItem(i);

        global::CurveFitVariable cv;

        cv.name       = tlIt->text(0);
        cv.value      = tlIt->text(1);
        cv.lowerLimit = tlIt->text(2);
        cv.upperLimit = tlIt->text(3);
        cv.checkState = tlIt->checkState(0) == Qt::Checked ? "1" : "0";

        l.append(cv);
    }

    return l;
}

QMap<QString, QVariant> PeakFitWidget::getControls()
{
    QMap<QString, QVariant> m;

    m["RangeStart"]     = ui->doubleSpinBoxRangeStart->value();
    m["RangeEnd"]       = ui->doubleSpinBoxRangeEnd->value();
    m["MaxInt"]         = ui->spinBoxMaxIts->value();
    m["EpsX"]           = ui->doubleSpinBoxEpsX->value();
    m["DiffStep"]       = ui->doubleSpinBoxDiffStep->value();
    m["StepSizeValue"]  = ui->doubleSpinBoxStepSize->value();
    m["StepSizeCustom"] = ui->checkBoxStepSize->isChecked() ? "1" : "0";

    return m;
}

void PeakFitWidget::restorePreset(const QList<global::CurveFitCurve> &lc, const QList<global::CurveFitVariable> &lv, const QMap<QString, QVariant> &mc)
{
    // restore variables first, then curves
    setControls(mc);
    setVariables(lv);
    setCurves(lc);
}

void PeakFitWidget::setCurves(const QList<global::CurveFitCurve> &l)
{
    ui->treeWidgetFunctions->clear();
    curveList.clear();

    for (int i = 0; i < l.size(); ++i) {
        restoreCurve(l.at(i), l.at(i).type, i + 1);
    }

    curveNumber = l.size() + 1;
    updateCurves();
}

void PeakFitWidget::setVariables(const QList<global::CurveFitVariable> &l)
{
    ui->treeWidgetVariables->clear();

    for (int i = 0; i < l.size(); ++i) {
        QTreeWidgetItem *it = new QTreeWidgetItem;
        it->setTextAlignment(1, Qt::AlignRight);
        it->setTextAlignment(2, Qt::AlignRight);
        it->setTextAlignment(3, Qt::AlignRight);

        it->setText(0, l.at(i).name);
        it->setText(1, l.at(i).value);
        it->setText(2, l.at(i).lowerLimit);
        it->setText(3, l.at(i).upperLimit);
        it->setCheckState(0, l.at(i).checkState == "1" ? Qt::Checked : Qt::Unchecked);

        ui->treeWidgetVariables->addTopLevelItem(it);
    }
}

void PeakFitWidget::setControls(const QMap<QString, QVariant> &m)
{
    ui->doubleSpinBoxRangeStart->setValue(m.value("RangeStart", 0.0).toDouble());
    ui->doubleSpinBoxRangeEnd->setValue(m.value("RangeEnd", 0.0).toDouble());

    ui->checkBoxStepSize->setChecked(m.value("StepSizeCustom", "0") == "1");
    ui->doubleSpinBoxStepSize->setEnabled(ui->checkBoxStepSize->isChecked());
    ui->doubleSpinBoxStepSize->setValue(m.value("StepSizeValue", 0.001).toDouble());

    ui->spinBoxMaxIts->setValue(m.value("MaxIts", 0).toInt());
    ui->doubleSpinBoxEpsX->setValue(m.value("EpsX", 0.0).toDouble());
    ui->doubleSpinBoxDiffStep->setValue(m.value("DiffStep", 0.0001).toDouble());
}

QString PeakFitWidget::getReport(const QString &scanFile)
{
    if (!curveList.size()) return QString();

    QMap<QString, QVariant> cfControls = getControls();
    QList<global::CurveFitCurve> cfCurves = getCurves();

    QString scanName;

    const Scan *curScan = scanControl->firstActiveScan();
    if (curScan) scanName = curScan->name();

    QStringList out;
    out << "File;" + scanFile;
    out << "Scan;" + scanName;
    out << QString("Range;%1 - %2").arg(cfControls.value("RangeStart").toDouble()).arg(cfControls.value("RangeEnd").toString());
    out << QString("Step size;%1").arg(cfControls.value("StepSizeValue").toDouble());
    out << QString();

    for (int i = 0; i < cfCurves.size(); ++i) {
        out << QString("#%1 %2").arg(i+1).arg(cfitHandler.description(cfCurves.at(i).type));
        out << ";Equation;" + cfitHandler.equationText(cfCurves.at(i).type);
        out << ";Parameters";

        for (int j = 0; j < cfitHandler.parameterText(cfCurves.at(i).type).size(); ++j) {
            out << ";;" + cfitHandler.parameterText(cfCurves.at(i).type).at(j);
        }

        out << ";Results";
        int jmx = qMin(cfCurves.at(i).parameterNames.size(), cfCurves.at(i).values.size());

        for (int j = 0; j < jmx; ++j) {
            QString par = cfCurves.at(i).parameterNames.at(j);
            QString val = cfCurves.at(i).values.at(j);
            out << ";;" + par + ";" + val;
        }

        out << QString();
    }

    out << QString();

    QList<Scan> curveFits = cfitManager.calculateCurve(ui->checkBoxStepSize->isChecked() ? ui->doubleSpinBoxStepSize->value() : -1.0, true);
    QString scanHeader("Angle;");

    for (int i = 0; i < curveFits.size(); ++i) {
        scanHeader += curveFits.at(i).name() + ";";
    }

    out << scanHeader;

    for (int l = 0; l < curveFits.at(0).pDataAngle().size(); ++l) {
        QString line = QString("%1;").arg(curveFits.at(0).pDataAngle().at(l), 0, 'f', 6);

        for (int i = 0; i < curveFits.size(); ++i) {
            line += QString("%1;").arg(curveFits.at(i).pDataIntensity().at(l), 0, 'f', 6);
        }

        out << line;
    }

    return out.join("\n");
}

void PeakFitWidget::cursorMessage(QString s)
{
    ui->labelCursorMessage->setText(s);
}

void PeakFitWidget::activeCurveChanged(QTreeWidgetItem *it, int)
{
    if (!it) return;
    if (it->parent()) it = it->parent();

    int n = ui->treeWidgetFunctions->indexOfTopLevelItem(it);

    if (n < 0) return;
    if (n >= currentScans.size()) return;

    Scan *sc = scanControl->getScan(currentScans.at(n));
    if (!sc) return;

    scanControl->clearActiveScans();
    sc->setActive(true);
    scanControl->updateViews(this);
}
