/***************************************************************************
                          projectsearchmatchwidget.cpp  -  description
                             -------------------
    begin                : Tue Jan 22 21:00:00 CEST 2019
    copyright            : (C) 2019 by Nicola Doebelin
    email                : ndoebelin@gmx.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "peaklistwidget.h"
#include "../../libXrdIO/bgmnfileio.h"
#include "../../libXrdIO/structs.h"
#include "../../libXrdIO/scanops.h"
#include "../../libXrdIO/parser/bgmnsampledata.h"
#include "../../libXrdIO/functions.h"
#include "ui_peaklistwidget.h"

#include <QtMath>
#include <QFileDialog>
#include <QGroupBox>
#include <QItemSelectionModel>
#include <QMenu>
#include <algorithm>

PeakListWidget::PeakListWidget(GraphDataController *c, QWidget *parent) :
    AbstractGraphView(c, parent),
    ui(new Ui::PeakListWidget)
{
    ui->setupUi(this);
    refStrManager = BgmnRefStructureManager::getInstance();
    saveSettingsRequested = false;

    allPeaksId = QUuid::createUuid();
    allPeaksModel = new HklItemModel(allPeaksId, QString(tr("All")), this);

    twoThetaMin = 0.0;
    twoThetaMax = 180.0;
    waveLength = 0.0;
    prevNumberOfScans = -1;

    hklHeaderLabels << QString("Phase")
                    << QString("h")
                    << QString("k")
                    << QString("l")
                    << QString("Angle (%1%2%3)").arg(global::degree).arg(2).arg(global::theta)
                    << QString("d (nm)")
                    << QString("Intensity")
                    << QString("Texture")
                    << QString("B1")
                    << QString("b2%1").arg(global::superTwo);

    columns["phase"]     = 0;
    columns["h"]         = 1;
    columns["k"]         = 2;
    columns["l"]         = 3;
    columns["2theta"]    = 4;
    columns["dnm"]       = 5;
    columns["intensity"] = 6;
    columns["texture"]   = 7;
    columns["b1"]        = 8;
    columns["b2"]        = 9;

    allPeaksModel->setSortRole(Qt::EditRole);
    allPeaksModel->setHorizontalHeaderLabels(hklHeaderLabels);

    contextMenuHeader = new QMenu(this);

    for (int i = 0; i < hklHeaderLabels.size(); ++i) {
        QAction *act = new QAction(hklHeaderLabels.at(i), contextMenuHeader);
        act->setCheckable(true);
        act->setData(i);
        contextMenuHeader->addAction(act);
    }

    ui->tableViewHklData->horizontalHeader()->setContextMenuPolicy(Qt::CustomContextMenu);
    connect(ui->tableViewHklData->horizontalHeader(), SIGNAL(sectionClicked(int)), ui->tableViewHklData, SLOT(sortByColumn(int)));
    connect(ui->tableViewHklData->horizontalHeader(), SIGNAL(sectionResized(int,int,int)), this, SLOT(headerResized()));
    connect(ui->tableViewHklData->horizontalHeader(), SIGNAL(customContextMenuRequested(QPoint)), this, SLOT(customContextMenuRequested(QPoint)));
    connect(contextMenuHeader, SIGNAL(triggered(QAction*)), this, SLOT(toggleColumnHidden(QAction*)));

    initSettings();
}

PeakListWidget::~PeakListWidget()
{
    saveSettings();

    while (hklModelStack.count()) {
        HklItemModel *m = hklModelStack.take(hklModelStack.firstKey());
        if (m) delete m;
    }

    delete ui;
}

void PeakListWidget::initSettings()
{
    ui->tableViewHklData->horizontalHeader()->restoreState(settings->value("peakListWidget/hklHeader", QByteArray()).toByteArray());
    ui->comboBoxHklData->addItem(tr("All"), QVariant(allPeaksId));
    setCurrentModel(allPeaksId);

    QList<QVariant> hiddenCols;
    hiddenCols.append(QVariant(columns.value("b1"))); // hide the b1 column by default
    hiddenCols.append(QVariant(columns.value("k2"))); // hide the k2 column by default
    hiddenCols = settings->value("peakListWidget/hiddenColumns", hiddenCols).toList();

    for (int i = 0; i < ui->tableViewHklData->horizontalHeader()->count(); ++i) {
        bool b = hiddenCols.contains(QVariant(i));
        ui->tableViewHklData->setColumnHidden(i, b);
        contextMenuHeader->actions().at(i)->setChecked(!b);
    }
}

void PeakListWidget::saveSettings()
{
    if (!saveSettingsRequested) return;

    settings->setValue("peakListWidget/hklHeader", ui->tableViewHklData->horizontalHeader()->saveState());

    QList<QVariant> hiddenCols;
    for (int i = 1; i < ui->tableViewHklData->horizontalHeader()->count(); ++i) {
        if (ui->tableViewHklData->isColumnHidden(i)) hiddenCols.append(QVariant(i));
    }

    settings->setValue("peakListWidget/hiddenColumns", hiddenCols);
}

void PeakListWidget::customContextMenuRequested(const QPoint &p)
{
    contextMenuHeader->popup(ui->tableViewHklData->viewport()->mapToGlobal(p));
}

void PeakListWidget::updateView()
{
    if (!isShown) return;

    /*
     * We only update the view if the number of scans has changed,
     * because updating this widget's data can be a massive resource hog.
     * To force updates even if the number of scans has not changed (e.g.
     * when a refinement completes), call ::updateData() manually.
     */
    if (scanControl->count() != prevNumberOfScans) {
        prevNumberOfScans = scanControl->count();
        updateData();
    }
}

void PeakListWidget::updateData()
{
    if (!isShown) return;
    updateList();
}

void PeakListWidget::clearAll()
{
    bool bcb = ui->comboBoxHklData->blockSignals(true);
    bool btv = ui->tableViewHklData->blockSignals(true);

    allPeaksModel->clear();
    allPeaksModel->setHorizontalHeaderLabels(hklHeaderLabels);

    ui->comboBoxHklData->clear();
    ui->tableViewHklData->setModel(allPeaksModel);
    ui->tableViewHklData->horizontalHeader()->setSortIndicatorShown(true);

    while (hklModelStack.count()) {
        HklItemModel *m = hklModelStack.take(hklModelStack.firstKey());
        if (m) delete m;
    }

    initSettings();

    ui->comboBoxHklData->blockSignals(bcb);
    ui->tableViewHklData->blockSignals(btv);
}

void PeakListWidget::updateList()
{
    int n = ui->comboBoxHklData->currentIndex();

    clearAll();
    setBoundariesTwoTheta(scanControl->getXmin(),
                          scanControl->getXmax(),
                          scanControl->getWaveLength(settings->defaultWavelength()));

    for (int i = 0; i < scanControl->count(); ++i) {
        if (scanControl->getScan(i)->hasHklData()) {
            addHklData(scanControl->getScan(i));
        }
    }

    if (n >= ui->comboBoxHklData->count()) n = ui->comboBoxHklData->count() - 1;
    if (n < 0) n = 0;

    bool oldState = ui->comboBoxHklData->blockSignals(true);
    ui->comboBoxHklData->setCurrentIndex(n);
    showHklList(n);
    ui->comboBoxHklData->blockSignals(oldState);
}

void PeakListWidget::setBoundariesTwoTheta(double tmin, double tmax, double wl)
{
    waveLength = wl;
    twoThetaMin = tmin;
    twoThetaMax = tmax;
}

void PeakListWidget::addHklData(const Scan *scan)
{
    clearHklSelection();
    if (!scan->hasHklData()) return;

    bool oldState = ui->comboBoxHklData->blockSignals(true);

    // remove and delete the previous data
    if (hklModelStack.contains(scan->uid())) {
        HklItemModel *m = hklModelStack.take(scan->uid());
        if (m) delete m;
    }

    // add a new model to the model stack
    hklModelStack[scan->uid()] = dataToModel(scan, true);

    // also append the peaks to the allPeaks model
    appendDataToModel(allPeaksModel, scan, true);

    int idx = ui->comboBoxHklData->findData(QVariant(scan->uid()), Qt::UserRole);

    if (idx < 0) {
        ui->comboBoxHklData->addItem(scan->name(), QVariant(scan->uid()));
    } else {
        ui->comboBoxHklData->removeItem(idx);
        ui->comboBoxHklData->insertItem(idx, scan->name(), QVariant(scan->uid()));
    }

    ui->comboBoxHklData->setCurrentIndex(ui->comboBoxHklData->findData(QVariant(scan->uid()), Qt::UserRole));
    setCurrentModel(scan->uid());
    ui->comboBoxHklData->blockSignals(oldState);
}

void PeakListWidget::setCurrentModel(const QUuid &uid)
{
    HklItemModel *m;

    if (uid == allPeaksId) {
        m = allPeaksModel;
    } else {
        m = hklModelStack.value(uid, nullptr);
    }

    if (!m) return;

    ui->tableViewHklData->setModel(m);
    ui->tableViewHklData->horizontalHeader()->setSortIndicatorShown(true);

    connect(ui->tableViewHklData->selectionModel(),
            SIGNAL(selectionChanged(QItemSelection, QItemSelection)),
            this,
            SLOT(selectHkl(QItemSelection, QItemSelection)));
}

/*
 * creates a new model for scan and adds scan's hkl data
 */
HklItemModel * PeakListWidget::dataToModel(const Scan *scan, bool normalize)
{
    HklItemModel *model = new HklItemModel(scan->uid(), scan->name(), this);
    model->setSortRole(Qt::EditRole);
    model->setHorizontalHeaderLabels(hklHeaderLabels);

    if (scan->hasHklData()) {
        appendDataToModel(model, scan, normalize);
    }

    return model;
}

/*
 * appends scan's hkl data to an existing model
 */
void PeakListWidget::appendDataToModel(HklItemModel *model, const Scan *scan, bool normalize)
{
    if (!model) return;
    removeDataFromModel(model, scan->uid());

    double maxHklIntens = normalize ? getMaxHklIntensity(scan) : 1.0;
    double maxScanIntens = normalize ? scanControl->getYmax() : 1.0;

    for (int i = 0; i < scan->dataHkl().size(); ++i) {
        appendHklToModel(model, scan->dataHkl().at(i), scan, maxScanIntens / maxHklIntens);
    }

    model->sort(4, Qt::AscendingOrder);
}

/*
 * appends a single hkl object to an existing model
 */
void PeakListWidget::appendHklToModel(HklItemModel *model, const Hkl &hkl, const Scan *scan, double normFactor)
{
    double hklIntens = hkl.intensity();
    double b1 = hkl.b1();
    double b2 = hkl.k2();
    double pos = hkl.position();
    double d  = scan->getHklXunit() == "dnm" ? pos : global::Functions::twoThetaToD(pos, waveLength) / 10.0;
    double tt = scan->getHklXunit() == "dnm" ? global::Functions::dToTwoTheta(10.0 * pos, waveLength) : pos;
    double tex = hkl.texture();
    QList<int> hklList = hklToInt(hkl.hkl());

    hklIntens *= normFactor;

    QString tTwoTheta(QString("%1").arg(tt, 0, 'f', 4));
    QString tD(QString("%1").arg(d, 0, 'f', 4));
    QString tInt(QString("%1").arg(hklIntens, 0, 'f', 4));
    QString tTex(QString("%1").arg(tex, 0, 'f', 4));
    QString tB1(QString("%1").arg(b1, 0, 'f', 7));
    QString tB2(QString("%1").arg(b2, 0, 'f', 9));

    QStandardItem *itPhase    = new QStandardItem(scan->name());
    QStandardItem *itH        = new QStandardItem(QString("%1").arg(hklList.at(0)));
    QStandardItem *itK        = new QStandardItem(QString("%1").arg(hklList.at(1)));
    QStandardItem *itL        = new QStandardItem(QString("%1").arg(hklList.at(2)));
    QStandardItem *iTwoTheta  = new QStandardItem(tTwoTheta);
    QStandardItem *iD         = new QStandardItem(tD);
    QStandardItem *iInt       = new QStandardItem(tInt);
    QStandardItem *iTex       = new QStandardItem(tTex);
    QStandardItem *iB1        = new QStandardItem(tB1);
    QStandardItem *iB2        = new QStandardItem(tB2);

    // edit role data
    itPhase->setData(scan->name(), Qt::EditRole);
    itH->setData(hklList.at(0), Qt::EditRole);
    itK->setData(hklList.at(1), Qt::EditRole);
    itL->setData(hklList.at(2), Qt::EditRole);
    iTwoTheta->setData(tt, Qt::EditRole);
    iD->setData(d, Qt::EditRole);
    iInt->setData(hklIntens, Qt::EditRole);
    iTex->setData(tex, Qt::EditRole);
    iB1->setData(b1, Qt::EditRole);
    iB2->setData(b2, Qt::EditRole);

    // user role data
    itPhase->setData(QVariant(scan->uid()), Qt::UserRole);
    // itH->setData(hklList.at(0), Qt::UserRole);
    // itK->setData(hklList.at(1), Qt::UserRole);
    // itL->setData(hklList.at(2), Qt::UserRole);
    iTwoTheta->setData(QVariant(hkl.uid()), Qt::UserRole);
    // iD->setData(d, Qt::UserRole);
    // iInt->setData(hklIntens, Qt::UserRole);
    // iTex->setData(tex, Qt::UserRole);
    // iB1->setData(b1, Qt::UserRole);
    // iB2->setData(b2, Qt::UserRole);

    itPhase->setEditable(false);
    itH->setEditable(false);
    itK->setEditable(false);
    itL->setEditable(false);
    iTwoTheta->setEditable(false);
    iD->setEditable(false);
    iInt->setEditable(false);
    iTex->setEditable(false);
    iB1->setEditable(false);
    iB2->setEditable(false);

    QList<QStandardItem *> row {nullptr, nullptr, nullptr, nullptr, nullptr, nullptr, nullptr, nullptr, nullptr, nullptr};
    row[columns.value("phase")]     = itPhase;
    row[columns.value("h")]         = itH;
    row[columns.value("k")]         = itK;
    row[columns.value("l")]         = itL;
    row[columns.value("2theta")]    = iTwoTheta;
    row[columns.value("dnm")]       = iD;
    row[columns.value("intensity")] = iInt;
    row[columns.value("texture")]   = iTex;
    row[columns.value("b1")]        = iB1;
    row[columns.value("b2")]        = iB2;

    model->appendRow(row);
}

/*
 * removes all hkl lines of scan uid from the model
 */
void PeakListWidget::removeDataFromModel(HklItemModel *model, const QUuid &uid)
{
    if (!model) return;

    for (int i = model->rowCount() - 1; i >= 0; --i) {
        if (model->item(i, columns.value("phase"))->data(Qt::UserRole).toUuid() == uid) {
            model->removeRow(i);
        }
    }
}

/*
 * removes one hkl with uid from the model
 */
void PeakListWidget::removeHklFromModel(HklItemModel *model, const QUuid &uidHkl)
{
    if (!model) return;

    for (int i = model->rowCount() - 1; i >= 0; --i) {
        if (model->item(i, columns.value("2theta"))->data(Qt::UserRole).toUuid() == uidHkl) {
            model->removeRow(i);
            return;
        }
    }
}

void PeakListWidget::saveHklList()
{
    QString fDia = scanControl->fileName();

    QString csvFilter("CSV File (*.csv *.CSV)");
    QString qualxDFilter("d values (*.dif *.DIF)");
    QString qualxTFilter("2theta values (*.dif *.DIF)");

    QStringList filters;
    filters << csvFilter << qualxDFilter << qualxTFilter;

    QString selectedFilter;

    QString f = QFileDialog::getSaveFileName(this, tr("Save hkl list to CSV"), fDia, filters.join(";;"), &selectedFilter);
    if (f.isEmpty()) return;

    if (selectedFilter == csvFilter) saveToCsv(f);
    if (selectedFilter == qualxDFilter) saveToQualxD(f);
    if (selectedFilter == qualxTFilter) saveToQualxT(f);
}

void PeakListWidget::saveToCsv(const QString &f)
{
    QString sep(settings->value("config/asciiFieldSeparator", " ").toString());
    QStringList out(hklHeaderLabels.join(sep));

    HklItemModel *m = static_cast<HklItemModel*>(ui->tableViewHklData->model());

    for (int r = 0; r < m->rowCount(); ++r) {
        QStringList line;
        for (int c = 0; c < m->columnCount(); ++c) {
            line.append(QString("%1").arg(m->item(r, c)->data(Qt::EditRole).toDouble(), 0, 'f', 9));
        }

        out.append(line.join(sep));
    }

    BgmnFileIO::writeTextFile(f, out.join("\n"));
}

void PeakListWidget::saveToQualxD(const QString &f)
{
    QStringList out;
    HklItemModel *m = static_cast<HklItemModel*>(ui->tableViewHklData->model());

    for (int r = 0; r < m->rowCount(); ++r) {
        QStringList line;
        line.append(QString("%1").arg(10.0 * m->item(r, columns.value("dnm"))->data(Qt::EditRole).toDouble(), 0, 'f', 6));
        line.append(QString("%1").arg(m->item(r, columns.value("intensity"))->data(Qt::EditRole).toDouble(), 0, 'f', 6));

        out.append(line.join(" "));
    }

    BgmnFileIO::writeTextFile(f, out.join("\n"));
}

void PeakListWidget::saveToQualxT(const QString &f)
{
    QStringList out;
    HklItemModel *m = static_cast<HklItemModel*>(ui->tableViewHklData->model());

    for (int r = 0; r < m->rowCount(); ++r) {
        QStringList line;
        line.append(QString("%1").arg(m->item(r, columns.value("2theta"))->data(Qt::EditRole).toDouble(), 0, 'f', 6));
        line.append(QString("%1").arg(m->item(r, columns.value("intensity"))->data(Qt::EditRole).toDouble(), 0, 'f', 6));

        out.append(line.join(" "));
    }

    BgmnFileIO::writeTextFile(f, out.join("\n"));
}

void PeakListWidget::showHklList(int i)
{
    ui->tableViewHklData->clearSelection();

    QUuid uid = ui->comboBoxHklData->itemData(i, Qt::UserRole).toUuid();
    if (uid.isNull()) return;

    setCurrentModel(uid);
}

void PeakListWidget::clearHklSelection()
{
    ui->tableViewHklData->clearSelection();
}

double PeakListWidget::getMaxHklIntensity(const Scan *scan)
{
    if (!scan->hasHklData()) return 0.0;

    double d = scan->dataHkl().first().intensity();

    for (int i = 1; i < scan->dataHkl().size(); ++i) {
        d = qMax(d, scan->dataHkl().at(i).intensity());
    }

    return d;
}

void PeakListWidget::resultSelected(QTreeWidgetItem *it)
{
    if (!it) return;
    emit resultsSelectionChanged(it->data(0, Qt::UserRole).toString());
}

void PeakListWidget::selectHkl(const QItemSelection &current, const QItemSelection &previous)
{
    QModelIndexList prevIdx = previous.indexes();
    QModelIndexList currIdx = current.indexes();

    HklItemModel *m = static_cast<HklItemModel*>(ui->tableViewHklData->model());

    int colScanUid = columns.value("phase");
    int colHklUid  = columns.value("2theta");

    for (int i = 0; i < prevIdx.size(); ++i) {
        QUuid scanUid = m->item(prevIdx.at(i).row(), colScanUid)->data(Qt::UserRole).toUuid();
        QUuid hklUid = m->item(prevIdx.at(i).row(), colHklUid)->data(Qt::UserRole).toUuid();

        Scan *scan = scanControl->getScan(scanUid);
        if (!scan) continue;

        int idx = indexOfHkl(scan, hklUid);

        if (idx >= 0) {
            scan->pDataHkl()[idx].setStatus(1);
        }
    }

    for (int i = 0; i < currIdx.size(); ++i) {
        QUuid scanUid = m->item(currIdx.at(i).row(), colScanUid)->data(Qt::UserRole).toUuid();
        QUuid hklUid = m->item(currIdx.at(i).row(), colHklUid)->data(Qt::UserRole).toUuid();

        Scan *scan = scanControl->getScan(scanUid);
        if (!scan) continue;

        int idx = indexOfHkl(scan, hklUid);

        if (idx >= 0) {
            scan->pDataHkl()[idx].setStatus(2);
        }
    }

    scanControl->updateViews(this);
}

void PeakListWidget::removeCurrentHkl()
{
    QItemSelectionModel *sm = ui->tableViewHklData->selectionModel();
    if (!sm) return;

    QModelIndex idx = sm->currentIndex();
    if (!idx.isValid()) return;

    HklItemModel *m = static_cast<HklItemModel*>(ui->tableViewHklData->model());

    QUuid uidScan = m->item(idx.row(), columns.value("phase"))->data(Qt::UserRole).toUuid();
    QUuid uidHkl  = m->item(idx.row(), columns.value("2theta"))->data(Qt::UserRole).toUuid();

    qDebug() << QString("removing from %1 and %2").arg(m->uid().toString()).arg(allPeaksModel->uid().toString());
    removeHklFromModel(hklModelStack.value(uidScan, nullptr), uidHkl);
    removeHklFromModel(allPeaksModel, uidHkl);

    Scan *scan = scanControl->getScan(uidScan);
    if (!scan) return;

    int hklIndex = indexOfHkl(scan, uidHkl);
    if (hklIndex < 0) return;

    scan->pDataHkl().remove(hklIndex);
    scanControl->updateViews();
}

/*
 * always returns a list of size 3
 */
QList<int> PeakListWidget::hklToInt(const QString &strHkl)
{
    QStringList l = strHkl.split(QRegularExpression("\\s+"));
    QList<int> intHkl;

    for (int i = 0; i < 3; ++i) {
        if (i >= l.size()) intHkl.append(0);
        else               intHkl.append(l.at(i).toInt());
    }

    return intHkl;
}

int PeakListWidget::indexOfHkl(const Scan *scan, const QUuid &uid)
{
    if (!scan) return -1;

    for (int i = 0; i < scan->dataHkl().size(); ++i) {
        if (uid == scan->dataHkl().at(i).uid()) return i;
    }

    return -1;
}

void PeakListWidget::toggleColumnHidden(QAction *a)
{
    int c = a->data().toInt();
    ui->tableViewHklData->setColumnHidden(c, !a->isChecked());
    saveSettingsRequested = true;
}

void PeakListWidget::headerResized()
{
    saveSettingsRequested = true;
}

QUuid PeakListWidget::currentScanUid()
{
    return ui->comboBoxHklData->currentData(Qt::UserRole).toUuid();
}

HklItemModel::HklItemModel(const QUuid &u, const QString &n, QObject *parent)
    :  QStandardItemModel(parent), _uid(u), _name(n)
{
    setSortRole(Qt::UserRole + 1);
}

