/***************************************************************************
                          scanlistwidget.cpp  -  description
                             -------------------
    begin                : Thu May 12 15:27:00 CEST 2016
    copyright            : (C) 2016 by Nicola Doebelin
    email                : ndoebelin@gmx.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "scanlistwidget.h"
#include "../libXrdIO/export/asciihklexport.h"
#include "../libXrdIO/export/asciixyexport.h"
#include "../libXrdIO/export/exporthandler.h"
#include "../libXrdIO/functions.h"
#include <QString>
#include <QStringList>
#include <QHeaderView>
#include <QClipboard>
#include <QGuiApplication>
#include <QFileDialog>
#include <QHBoxLayout>
#include <QObject>
#include <QDebug>
#include <QMenu>

ScanListWidget::ScanListWidget(GraphDataController *c, QWidget *parent)
    : AbstractGraphView(c, parent)
{
    treeWidget = new QTreeWidget(this);
    treeWidget->setContextMenuPolicy(Qt::CustomContextMenu);

    QHBoxLayout *layout = new QHBoxLayout;
    layout->setContentsMargins(0, 0, 0, 0);
    layout->addWidget(treeWidget);
    this->setLayout(layout);

    slContextMenu = new QMenu(this);
    slContextMenu->addAction(QIcon::fromTheme("profex-save"), tr("Export Scan..."), this, SLOT(exportSingleScan()));
    slContextMenu->addAction(QIcon::fromTheme("profex-delete"), tr("Remove Scan"), this, SLOT(removeScan()));
    slContextMenu->addAction(QIcon::fromTheme("profex-document-copy"), tr("Copy Scan Data"), this, SLOT(copySingleScan()));
    slContextMenu->addAction(QIcon::fromTheme("profex-document-copy"), tr("Copy HKL Data"), this, SLOT(copyHklData()));
    slContextMenu->addSeparator();
    slContextMenu->addAction(QIcon::fromTheme("profex-save"), tr("Export Scan with angular corrections..."), this, SLOT(exportSingleScanWithAngularCorrections()));
    slContextMenu->addAction(QIcon::fromTheme("profex-document-copy"), tr("Copy Scan Data with angular corrections"), this, SLOT(copySingleScanWithAngularCorrections()));
    slContextMenu->addSeparator();
    slContextMenu->addAction(QIcon::fromTheme("profex-arrow-up"), tr("Move Scan Up"), this, SLOT(sortScanUp()));
    slContextMenu->addAction(QIcon::fromTheme("profex-arrow-down"), tr("Move Scan Down"), this, SLOT(sortScanDown()));
    slContextMenu->addSeparator();
    slContextMenu->addAction(tr("Add synthetic noise"), this, SLOT(addNoise()));

    initSettings();
    initConnections();
}

ScanListWidget::~ScanListWidget()
{
    saveSettings();
}

void ScanListWidget::initSettings()
{
    QStringList slHeaders;
    slHeaders << QString() << tr("Scan") << tr("Scaling") << tr("Vertical Offset") << tr("Horizontal Offset");

    treeWidget->setColumnCount(slHeaders.size());
    treeWidget->setRootIsDecorated(false);
    treeWidget->setSortingEnabled(false);
    treeWidget->setHeaderLabels(slHeaders);
    treeWidget->header()->setSectionResizeMode(QHeaderView::Interactive);
    headerStateRestored = treeWidget->header()->restoreState(settings->value("config/scanListState", QByteArray()).toByteArray());
}

void ScanListWidget::saveSettings()
{
    settings->setValue("config/scanListState", treeWidget->header()->saveState());
}

void ScanListWidget::initConnections()
{
    connect(treeWidget->header(), SIGNAL(sectionResized(int,int,int)), this, SLOT(saveSettings()));
    connect(treeWidget, SIGNAL(customContextMenuRequested(QPoint)), this, SLOT(showContextMenu(QPoint)));
    connect(treeWidget, SIGNAL(itemChanged(QTreeWidgetItem*,int)), this, SLOT(scanParameterChanged(QTreeWidgetItem*,int)));
    connect(treeWidget, SIGNAL(itemClicked(QTreeWidgetItem*,int)), this, SLOT(activeScanChanged(QTreeWidgetItem*,int)));
}

void ScanListWidget::adjustHeaderSize()
{
    if (!headerStateRestored) {
        treeWidget->header()->resizeSections(QHeaderView::ResizeToContents);
    }
}

int ScanListWidget::currentIndex()
{
    return treeWidget->indexOfTopLevelItem(treeWidget->currentItem());
}

void ScanListWidget::updateView()
{
    if (!isShown) return;
    bool oldState = treeWidget->blockSignals(true);

    QMap<QUuid, Qt::CheckState> checkStates = adjustTree();
    generateTreeItems(checkStates, settings->value("graph/phaseVisibility", true).toBool());
    treeWidget->blockSignals(oldState);
}

QMap<QUuid, Qt::CheckState> ScanListWidget::adjustTree()
{
    QMap<QUuid, Qt::CheckState> checkStates;

    while (treeWidget->topLevelItemCount()) {
        QTreeWidgetItem *itm = treeWidget->takeTopLevelItem(0);
        checkStates[itm->data(0, Qt::UserRole).toUuid()] = itm->checkState(0);
        delete itm;
    }

    return checkStates;
}

void ScanListWidget::generateTreeItems(const QMap<QUuid, Qt::CheckState> &cs, bool cdefault)
{
    bool addFn = scanControl->countFiles() > 1;

    for (int i = 0; i < scanControl->count(); ++i) {
        Scan *scan = scanControl->getScan(i);
        if (!scan) continue;

        Qt::CheckState dcs = Qt::Checked; // default check state
        bool dvs = true;                  // default visibility

        if (scan->scanTypes().testFlag(Scan::PHASE)) {
            dcs = cdefault ? Qt::Checked : Qt::Unchecked;
            dvs = cdefault ? true : false;
        }

        QTreeWidgetItem *it = new QTreeWidgetItem;
        it->setFlags(it->flags() | Qt::ItemIsEditable);

        if (scan->isActive()) {
            it->setText(0, scanActiveText);
        }

        QFileInfo fi(scan->sourceFileName());
        QString name = addFn ? QString("%1 (%2)").arg(scan->name()).arg(fi.fileName()) : scan->name();

        it->setData(0, Qt::UserRole, QVariant(scan->uid()));
        it->setText(1, name);
        it->setText(2, QString::number(scan->scaleFactor(), 'f', 2));
        it->setText(3, QString::number(scan->yOffset(), 'f', 2));
        it->setText(4, QString::number(scan->xOffset(), 'f', 4));

        if (cs.contains(scan->uid())) {
            it->setCheckState(0, cs.value(scan->uid()));
            scan->setVisible(cs.value(scan->uid()) == Qt::Checked);
        } else {
            it->setCheckState(0, dcs);
            scan->setVisible(dvs);
        }

        if (scan->color().isValid()) {
            if (settings->isDarkMode()) {
                it->setForeground(1, QBrush(global::Functions::colorToDarkMode(scan->color())));
            } else {
                it->setForeground(1, QBrush(scan->color()));
            }
        }

        treeWidget->addTopLevelItem(it);
    }

    this->adjustHeaderSize();
}

void ScanListWidget::showContextMenu(const QPoint &p)
{
    slContextMenu->popup(treeWidget->viewport()->mapToGlobal(p));
}

void ScanListWidget::exportSingleScan()
{
    int n = currentIndex();

    if (n < 0) {
        qDebug() << QString("ScanListWidget::exportSingleScan(): No selected scan found, index was %1").arg(n);
        return;
    }

    Scan *scan = scanControl->getScan(n);

    if (!scan) {
        qDebug() << QString("ScanListWidget::exportSingleScanWithAngularCorrections(): No valid scan retrieved from controller, index was %1").arg(n);
        return;
    }

    exportScan(*scan);
}

void ScanListWidget::exportSingleScanWithAngularCorrections()
{
    double eps1 = scanControl->getAngularCorrEPS1();
    double eps2 = scanControl->getAngularCorrEPS2();
    double eps3 = scanControl->getAngularCorrEPS3();

    int n = currentIndex();

    if (n < 0) {
        qDebug() << QString("ScanListWidget::exportSingleScanWithAngularCorrections(): No selected scan found, index was %1").arg(n);
        return;
    }

    Scan *scan = scanControl->getScan(n);

    if (!scan) {
        qDebug() << QString("ScanListWidget::exportSingleScanWithAngularCorrections(): No valid scan retrieved from controller, index was %1").arg(n);
        return;
    }

    Scan corScan = scan->clone();
    for (int i = 0; i < corScan.pDataAngle().size(); ++i) {
        double tt = corScan.pDataAngle().at(i);
        corScan.pDataAngle()[i] = tt - global::Functions::angularCorrection(tt, eps1, eps2, eps3) + corScan.xOffset();
    }

    exportScan(corScan);
}

void ScanListWidget::exportScan(Scan &scan)
{
    QMap<QString, QVariant> flags;
    flags["fieldSeparator"] = settings->value("config/asciiFieldSeparator", QVariant(QString(" ")));
    flags["fixBgmnZero"]    = QVariant(false);

    ExportHandler exHandler(this, true);
    QMap<QString, QString> formats = exHandler.uidsByFilter();
    QStringList keys = static_cast<QStringList>(formats.keys());

    QString selFilter = settings->value("graph/saveAsFilter", QString()).toString();

    QString f = QDir::fromNativeSeparators(QString("%1/%2-%3.%4")
                                           .arg(scanControl->fileInfo().absolutePath())
                                           .arg(scanControl->fileInfo().baseName())
                                           .arg(scan.name())
                                           .arg(exHandler.extensionByFilter(selFilter)));

    f = QFileDialog::getSaveFileName(this, tr("Save File As"), f, keys.join(";;"), &selFilter);

    if (f.isEmpty()) {
        return;
    }

    QFileInfo fout(f);

    if (fout.suffix().isEmpty()) {
        fout = QFileInfo(fout.absoluteFilePath() + QString(".%1").arg(exHandler.extensionByFilter(selFilter)));
    }

    QString uid = formats.value(selFilter);
    settings->setValue("graph/saveAsFilter", selFilter);

    if(exHandler.save(uid, fout.absoluteFilePath(), scan, flags) > 0) {
        qDebug() << QString("ProjectWidget::exportScan(): Scan file saved to %1 using format id = %2").arg(fout.absoluteFilePath()).arg(uid);
    } else {
        qDebug() << QString("ProjectWidget::exportScan(): No scan saved to file %1 using format id = %2)").arg(fout.absoluteFilePath()).arg(uid);
    }
}

void ScanListWidget::removeScan()
{
    int n = currentIndex();
    if (n < 0) return;
    scanControl->removeScan(n, true);
}

void ScanListWidget::copySingleScan()
{
    int n = currentIndex();

    if (n < 0) {
        qDebug() << QString("ScanListWidget::copySingleScan(): No selected scan found, index was %1").arg(n);
        return;
    }

    Scan *scan = scanControl->getScan(n);
    if (!scan) return;

    copyScan(*scan);
}

void ScanListWidget::copySingleScanWithAngularCorrections()
{
    double eps1 = scanControl->getAngularCorrEPS1();
    double eps2 = scanControl->getAngularCorrEPS2();
    double eps3 = scanControl->getAngularCorrEPS3();

    int n = currentIndex();

    if (n < 0) {
        qDebug() << QString("ScanListWidget::copySingleScan(): No selected scan found, index was %1").arg(n);
        return;
    }

    const Scan *scan = scanControl->at(n);
    if (!scan) return;

    Scan corScan = scan->clone();
    for (int i = 0; i < corScan.pDataAngle().size(); ++i) {
        double tt = corScan.pDataAngle().at(i);
        corScan.pDataAngle()[i] = tt - global::Functions::angularCorrection(tt, eps1, eps2, eps3) + corScan.xOffset();
    }

    copyScan(corScan);
}

void ScanListWidget::copyScan(Scan &scan)
{
    AsciiXyExport xyExport;

    QString data = xyExport.getData(scan, settings->value("config/asciiFieldSeparator", " ").toString(), false);

    QClipboard *clipBoard = qApp->clipboard();
    clipBoard->setText(data);
}

void ScanListWidget::copyHklData()
{
    int n = currentIndex();

    if (n < 0) {
        qDebug() << QString("ScanListWidget::copyHklData(): No selected scan found, index was %1").arg(n);
        return;
    }

    Scan *scan = scanControl->getScan(n);
    if (!scan) return;

    AsciiHklExport hklExport;
    QString data = hklExport.getData(*scan, settings->value("config/asciiFieldSeparator", " ").toString());

    QClipboard *clipBoard = qApp->clipboard();
    clipBoard->setText(data);
}

void ScanListWidget::sortScanUp()
{
    scanControl->sortScanUp(currentIndex(), true);
}

void ScanListWidget::sortScanDown()
{
    scanControl->sortScanDown(currentIndex(), true);
}

void ScanListWidget::activeScanChanged(QTreeWidgetItem *it, int c)
{
    if (c == 0) return; // ignore clicks on the check box in column 0
    if (!it) return;

    bool ctrlKey = QGuiApplication::keyboardModifiers().testFlag(Qt::ControlModifier);
    bool oldState = treeWidget->blockSignals(true);

    int n = treeWidget->indexOfTopLevelItem(it);
    bool b = scanControl->getScan(n)->isActive();
    scanControl->getScan(n)->setActive(!b);
    it->setText(0, b ? QString() : scanActiveText);

    // if ctrl is pressed, we leave the state of all other scans and only toggle the clicked one
    // if ctrl is not pressed, we set all other scans inactive
    if (!ctrlKey) {
        for (int i = 0; i < treeWidget->topLevelItemCount(); ++i) {
            if (i != n) {
                scanControl->getScan(i)->setActive(false);
                treeWidget->topLevelItem(i)->setText(0, QString());
            }
        }
    }

    scanControl->updateViews(this);
    treeWidget->blockSignals(oldState);
}

void ScanListWidget::addNoise()
{
    int n = currentIndex();

    if (n < 0) {
        qDebug() << QString("ScanListWidget::copyHklData(): No selected scan found, index was %1").arg(n);
        return;
    }

    Scan *scan = scanControl->getScan(n);
    if (!scan) return;

    scan->addNoise();
    scanControl->updateViews(this);
}

void ScanListWidget::scanParameterChanged(QTreeWidgetItem *it, int c)
{
    if (!it) return;
    Scan *scan = scanControl->getScan(treeWidget->indexOfTopLevelItem(it));
    if (!scan) return;

    if (c == 0) changeScanVisibility(scan, it->checkState(0));
    if (c == 1) scan->setName(it->text(1));
    if (c == 2) changeScanScaling(scan, it->text(2));
    if (c == 3) changeScanYOffset(scan, it->text(3));
    if (c == 4) changeScanXOffset(scan, it->text(4));

    scanControl->updateViews(this);
}

void ScanListWidget::changeScanVisibility(Scan *scan, int visible)
{
    if (visible == Qt::Unchecked) {
        scan->setVisible(false);
    } else {
        scan->setVisible(true);
    }
}

void ScanListWidget::changeScanScaling(Scan *scan, const QString &s)
{
    bool ok;
    double d = s.toDouble(&ok);

    if (ok) scan->setScaleFactor(d);
}

void ScanListWidget::changeScanYOffset(Scan *scan, const QString &s)
{
    bool ok;
    double d = s.toDouble(&ok);

    if (ok) scan->setYoffset(d);
}

void ScanListWidget::changeScanXOffset(Scan *scan, const QString &s)
{
    bool ok;
    double d = s.toDouble(&ok);

    if (ok) scan->setXoffset(d);
}

void ScanListWidget::toggleSubScans(bool b)
{
    for (int i = 0; i < treeWidget->topLevelItemCount(); ++i) {
        Scan *scan = scanControl->getScan(i);
        if (!scan) break;

        if (scan->scanTypes().testFlag(Scan::PHASE)) {
            scan->setVisible(b);
            treeWidget->topLevelItem(i)->setCheckState(0, b ? Qt::Checked : Qt::Unchecked);
        }
    }

    scanControl->updateViews();
}
