/***************************************************************************
                          searchmatchwidget.cpp  -  description
                             -------------------
    begin                : Mon May 20 18:00:00 CEST 2019
    copyright            : (C) 2019 by Nicola Doebelin
    email                : ndoebelin@gmx.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "ui_searchmatchwidget.h"
#include "searchmatchwidget.h"
#include "strfilefilter.h"
#include "periodictablewidget.h"
#include "../libXrdIO/bgmnfileio.h"
#include "../libXrdIO/parser/bgmnlstparser.h"
#include "../libXrdIO/parser/bgmnsavparser.h"
#include "../libXrdIO/parser/bgmnstrparser.h"
#include "../libXrdIO/import/importhandler.h"
#include "../libXrdIO/export/exporthandler.h"

SearchMatchWidget::SearchMatchWidget(GraphDataController *c, QWidget *parent) :
    AbstractGraphView(c, parent),
    ui(new Ui::SearchMatchWidget)
{
    ui->setupUi(this);

    refStrManager = BgmnRefStructureManager::getInstance();
    wasUsed = false;

    refOutput = nullptr;
    convDisplay = nullptr;

    defaultEps2 = 0.0;

    QStringList sheader;
    sheader << "Repository" << "Number of phases";
    ui->treeWidgetDirectories->setColumnCount(sheader.size());
    ui->treeWidgetDirectories->setHeaderLabels(sheader);

    QStringList theader;
    theader << "File" << "Phase" << "Fraction" << "FoM";
    ui->treeWidgetResults->setColumnCount(theader.size());
    ui->treeWidgetResults->setHeaderLabels(theader);

    QStringList pheader;
    pheader << "File" << "Phase" << "Fraction" << "Source";
    ui->treeWidgetPinned->setColumnCount(pheader.size());
    ui->treeWidgetPinned->setHeaderLabels(pheader);

    ui->comboBoxCrystallinity->addItem(tr("Infinite"), 0.0);
    ui->comboBoxCrystallinity->addItem(tr("High"), 0.0025);
    ui->comboBoxCrystallinity->addItem(tr("Medium"), 0.01);
    ui->comboBoxCrystallinity->addItem(tr("Low"), 0.025);
    ui->comboBoxCrystallinity->addItem(tr("Nano"), 0.1);

    ui->comboBoxUnitCellRefinement->addItem(tr("Fix"), 0.0);
    ui->comboBoxUnitCellRefinement->addItem(tr("Strict"), 0.001);
    ui->comboBoxUnitCellRefinement->addItem(tr("Medium"), 0.005);
    ui->comboBoxUnitCellRefinement->addItem(tr("Loose"), 0.01);

    nativeFormats << "xy" << "val";

    ui->toolButtonSave->setEnabled(false);
    ui->toolButtonPin->setEnabled(false);
    ui->toolButtonUnpin->setEnabled(false);

    ui->progressBar->setEnabled(false);
    ui->progressBar->reset();
    ui->progressBar->hide();

    buttonToggleAllOptional     = new PeriodicTableToggleAllButton(0, this);
    buttonToggleAllMandatory    = new PeriodicTableToggleAllButton(1, this);
    buttonToggleAllOne          = new PeriodicTableToggleAllButton(2, this);
    buttonToggleAllDiscard      = new PeriodicTableToggleAllButton(3, this);
    buttonToggleAllDiscOptional = new PeriodicTableToggleAllButton(4, this);
    buttonToggleAllDiscOptional->setText(QChar(0x25CF));

    statistics.n = 0;
    statistics.sum = 0.0;
    statistics.mean = 0.0;
    statistics.stdev = 0.0;

    bgmnHandler = new BgmnHandler(QUuid::createUuid());
    diaHandler = new BgmnDiaHandler(this);
    connect(bgmnHandler, SIGNAL(pollOutput()), this, SLOT(pollOutput()));
    connect(bgmnHandler, SIGNAL(allComplete(QUuid)), this, SLOT(complete(QUuid)));
    connect(ui->treeWidgetResults, SIGNAL(currentItemChanged(QTreeWidgetItem*,QTreeWidgetItem*)), this, SLOT(itemChanged(QTreeWidgetItem*,QTreeWidgetItem*)));
    connect(ui->treeWidgetResults, SIGNAL(currentItemChanged(QTreeWidgetItem*,QTreeWidgetItem*)), this, SLOT(showQuantities(QTreeWidgetItem*,QTreeWidgetItem*)));
    connect(ui->treeWidgetResults, SIGNAL(itemClicked(QTreeWidgetItem*,int)), this, SLOT(itemClicked(QTreeWidgetItem*,int)));
    connect(ui->treeWidgetResults, SIGNAL(itemClicked(QTreeWidgetItem*,int)), this, SLOT(showQuantities(QTreeWidgetItem*,int)));
    connect(ui->treeWidgetPinned, SIGNAL(currentItemChanged(QTreeWidgetItem*,QTreeWidgetItem*)), this, SLOT(itemChanged(QTreeWidgetItem*,QTreeWidgetItem*)));
    connect(ui->treeWidgetPinned, SIGNAL(itemClicked(QTreeWidgetItem*,int)), this, SLOT(itemClicked(QTreeWidgetItem*,int)));

    ui->treeWidgetResults->header()->setContextMenuPolicy(Qt::CustomContextMenu);
    ui->treeWidgetPinned->header()->setContextMenuPolicy(Qt::CustomContextMenu);

    initGui();
    initSettings();
}

SearchMatchWidget::~SearchMatchWidget()
{
    if (wasUsed) {
        // every open project creates its instance of SearchMatchWidget
        // which saves its current settings when destroyed. Since we
        // don't want unused instances to overwrite the settings of the
        // actually used instances, we only save the settings if
        // run() was called.
        saveSettings();
    }

    delete bgmnHandler;
    delete diaHandler;
    delete ui;
}

void SearchMatchWidget::updateView()
{
    if (!isShown) return;
}

void SearchMatchWidget::customMenuRequestedResults(QPoint p)
{
    contextMenuHeaderResults->popup(ui->treeWidgetResults->header()->viewport()->mapToGlobal(p));
}

void SearchMatchWidget::customMenuRequestedPinned(QPoint p)
{
    contextMenuHeaderPinned->popup(ui->treeWidgetPinned->header()->viewport()->mapToGlobal(p));
}

void SearchMatchWidget::initGui()
{
    contextMenuHeaderResults = new QMenu();
    contextMenuHeaderPinned = new QMenu();

    connect(contextMenuHeaderResults, SIGNAL(triggered(QAction*)), this, SLOT(toggleColumnHiddenResults(QAction*)));
    connect(contextMenuHeaderPinned, SIGNAL(triggered(QAction*)), this, SLOT(toggleColumnHiddenPinned(QAction*)));
    connect(ui->treeWidgetResults->header(), SIGNAL(customContextMenuRequested(QPoint)), this, SLOT(customMenuRequestedResults(QPoint)));
    connect(ui->treeWidgetPinned->header(), SIGNAL(customContextMenuRequested(QPoint)), this, SLOT(customMenuRequestedPinned(QPoint)));

    for (int i = 1; i < ui->treeWidgetResults->header()->count() - 1; ++i) {
        QAction *act = new QAction(ui->treeWidgetResults->headerItem()->text(i), contextMenuHeaderResults);
        act->setCheckable(true);
        act->setData(i);
        contextMenuHeaderResults->addAction(act);
    }

    for (int i = 1; i < ui->treeWidgetPinned->header()->count(); ++i) {
        QAction *act = new QAction(ui->treeWidgetPinned->headerItem()->text(i), contextMenuHeaderPinned);
        act->setCheckable(true);
        act->setData(i);
        contextMenuHeaderPinned->addAction(act);
    }

    connect(buttonToggleAllOptional, SIGNAL(clicked(bool)), ui->periodicTable, SLOT(allOptional()));
    connect(buttonToggleAllMandatory, SIGNAL(clicked(bool)), ui->periodicTable, SLOT(allMandatory()));
    connect(buttonToggleAllOne, SIGNAL(clicked(bool)), ui->periodicTable, SLOT(allOne()));
    connect(buttonToggleAllDiscard, SIGNAL(clicked(bool)), ui->periodicTable, SLOT(allDisabled()));
    connect(buttonToggleAllDiscOptional, SIGNAL(clicked(bool)), ui->periodicTable, SLOT(disableOptionals()));

    QGridLayout *periodicTableToggleLayout = new QGridLayout;

    periodicTableToggleLayout->addWidget(buttonToggleAllOptional,     0, 0);
    periodicTableToggleLayout->addWidget(buttonToggleAllMandatory,    1, 0);
    periodicTableToggleLayout->addWidget(buttonToggleAllOne,          0, 2);
    periodicTableToggleLayout->addWidget(buttonToggleAllDiscard,      1, 2);
    periodicTableToggleLayout->addWidget(buttonToggleAllDiscOptional, 0, 4);

    periodicTableToggleLayout->addWidget(new QLabel(tr("Optional")),         0, 1);
    periodicTableToggleLayout->addWidget(new QLabel(tr("Mandatory")),        1, 1);
    periodicTableToggleLayout->addWidget(new QLabel(tr("At least one")),     0, 3);
    periodicTableToggleLayout->addWidget(new QLabel(tr("Discarded")),        1, 3);
    periodicTableToggleLayout->addWidget(new QLabel(tr("Discard optional")), 0, 5);

    ui->groupBoxToggleAllRestrictions->setLayout(periodicTableToggleLayout);
}

void SearchMatchWidget::initSettings()
{
    ui->radioButtonFav->setChecked(settings->value("searchMatch/databaseFavsOnly", true).toBool());
    ui->radioButtonDir->setChecked(!ui->radioButtonFav->isChecked());
    ui->treeWidgetDirectories->setEnabled(ui->radioButtonDir->isChecked());

    ui->spinBoxItMax->setValue(settings->value("searchMatch/itMax", 10).toInt());
    ui->checkBoxAniso->setChecked(settings->value("searchMatch/allowAniso", false).toBool());
    ui->doubleSpinBoxWmax->setValue(settings->value("searchMatch/wmax", 60.0).toDouble());
    ui->checkBoxSetWmax->setChecked(settings->value("searchMatch/setWmax", true).toBool());
    ui->doubleSpinBoxWmin->setValue(settings->value("searchMatch/wmin", 5.0).toDouble());
    ui->checkBoxSetWmin->setChecked(settings->value("searchMatch/setWmin", true).toBool());
    ui->checkBoxEps2->setChecked(settings->value("searchMatch/eps2", false).toBool());
    ui->checkBoxRU->setChecked(settings->value("searchMatch/setRU", false).toBool());
    ui->spinBoxRU->setEnabled(ui->checkBoxRU->isChecked());
    ui->spinBoxRU->setValue(settings->value("searchMatch/valueRU", 10).toInt());

    ui->splitter->restoreState(settings->value("searchMatch/splitter", QByteArray()).toByteArray());
    ui->treeWidgetDirectories->header()->restoreState(settings->value("searchMatch/drHeader", QByteArray()).toByteArray());
    ui->treeWidgetResults->header()->restoreState(settings->value("searchMatch/twHeader", QByteArray()).toByteArray());
    ui->treeWidgetPinned->header()->restoreState(settings->value("searchMatch/pwHeader", QByteArray()).toByteArray());

    ui->comboBoxCrystallinity->setCurrentIndex(settings->value("searchMatch/crystallinity", 1).toInt());
    ui->comboBoxUnitCellRefinement->setCurrentIndex(settings->value("searchMatch/unitCellRefinement", 1).toInt());

    QList<QVariant> hiddenColsResults;
    hiddenColsResults = settings->value("searchMatch/hiddenColumnsResults", hiddenColsResults).toList();

    int n = qMin(ui->treeWidgetResults->header()->count() - 1, contextMenuHeaderResults->actions().count() + 1);

    for (int i = 1; i < n; ++i) {
        bool b = hiddenColsResults.contains(QVariant(i));
        ui->treeWidgetResults->setColumnHidden(i, b);
        contextMenuHeaderResults->actions().at(i - 1)->setChecked(!b);
    }

    QList<QVariant> hiddenColsPinned;
    hiddenColsPinned = settings->value("searchMatch/hiddenColumnsPinned", hiddenColsPinned).toList();

    n = qMin(ui->treeWidgetPinned->header()->count(), contextMenuHeaderPinned->actions().count() + 1);

    for (int i = 1; i < n; ++i) {
        bool b = hiddenColsPinned.contains(QVariant(i));
        ui->treeWidgetPinned->setColumnHidden(i, b);
        contextMenuHeaderPinned->actions().at(i - 1)->setChecked(!b);
    }
}

void SearchMatchWidget::initData()
{
    int errors = 0;
    if (!initInstruments()) errors++;
    if (!initLam())         errors++;
    if (!initDbDirs())      errors++;

    hasValidBgmnConfig = errors == 0;
}

bool SearchMatchWidget::initInstruments()
{
    QStringList devDir = settings->value("bgmnProject/deviceDatabase", QStringList()).toStringList();
    if (devDir.isEmpty()) return false;

    QStringList devExt;
    devExt << "geq" << "GEQ";

    for (int d = 0; d < devDir.size(); ++d) {
        QMap<QString,QString> devFiles = BgmnFileIO::getFileList(devDir.at(d), devExt);
        QMap<QString,QString>::const_iterator i = devFiles.constBegin();

        while (i != devFiles.constEnd()) {
            QFileInfo fi(i.value());
            ui->comboBoxInstrument->addItem(fi.baseName(), i.value());
            ++i;
        }
    }

    int c = ui->comboBoxInstrument->findText(settings->value("searchMatch/defaultInstrument",
                                                             settings->value("bgmnProject/defaultInstrument", "").toString()).toString());
    if (c >= 0) ui->comboBoxInstrument->setCurrentIndex(c);
    return true;
}

bool SearchMatchWidget::initLam()
{
    QFileInfo fiBgmnExec(settings->value("bgmnProject/bgmnExec", "").toString());
    if (!fiBgmnExec.exists()) return false;
    QString lamDir = fiBgmnExec.absolutePath();

    QStringList lamExt;
    lamExt << "lam" << "LAM";

    QMap<QString,QString> lamFiles = BgmnFileIO::getFileList(lamDir, lamExt);
    QMap<QString,QString>::const_iterator i = lamFiles.constBegin();

    while (i != lamFiles.constEnd()) {
        QFileInfo fi(i.value());
        ui->comboBoxLambda->addItem(fi.baseName(), i.value());
        ++i;
    }

    int c = ui->comboBoxLambda->findText(settings->value("searchMatch/lamFile", "cu").toString());
    if (c >= 0) ui->comboBoxLambda->setCurrentIndex(c);
    ui->doubleSpinBoxSynchrotron->setValue(settings->value("searchMatch/synchrotron", 0.05).toDouble());
    return true;
}

bool SearchMatchWidget::initDbDirs()
{
    QStringList repositories = settings->value("bgmnProject/structureDatabase", QStringList()).toStringList();
    QStringList checkedDirs = settings->value("searchMatch/databaseCheckedDirs", QStringList()).toStringList();
    if (repositories.isEmpty()) return false;

    ui->labelNumPhasesFavs->setText(QString("%1").arg(refStrManager->countFavPhases()));
    ui->labelNumPhasesAll->setText(QString("%1").arg(refStrManager->countAllPhases()));

    QMap<QString, QTreeWidgetItem *> items = refStrManager->createDirTree(ui->treeWidgetDirectories,
                                                                          repositories,
                                                                          false,
                                                                          Qt::ItemFlags(Qt::ItemIsEnabled
                                                                                        | Qt::ItemIsAutoTristate
                                                                                        | Qt::ItemIsUserCheckable));

    QMapIterator<QString, QTreeWidgetItem*> it(items);

    while (it.hasNext()) {
        it.next();
        it.value()->setText(1, QString("%1").arg(refStrManager->countPhasesInDirectory(it.value()->data(0, Qt::UserRole).toString())));

        if (checkedDirs.contains(it.value()->data(0, Qt::UserRole).toString())) {
            it.value()->setCheckState(0, Qt::Checked);
        } else {
            it.value()->setCheckState(0, Qt::Unchecked);
        }
    }

    return true;
}

void SearchMatchWidget::saveSettings()
{    
    settings->setValue("searchMatch/databaseFavsOnly", ui->radioButtonFav->isChecked());

    settings->setValue("searchMatch/itMax", ui->spinBoxItMax->value());
    settings->setValue("searchMatch/lamFile", ui->comboBoxLambda->currentText());
    settings->setValue("searchMatch/synchrotron", ui->doubleSpinBoxSynchrotron->value());
    settings->setValue("searchMatch/defaultInstrument", ui->comboBoxInstrument->currentText());
    settings->setValue("searchMatch/allowAniso", ui->checkBoxAniso->isChecked());
    settings->setValue("searchMatch/wmax", ui->doubleSpinBoxWmax->value());
    settings->setValue("searchMatch/setWmax", ui->checkBoxSetWmax->isChecked());
    settings->setValue("searchMatch/wmin", ui->doubleSpinBoxWmin->value());
    settings->setValue("searchMatch/setWmin", ui->checkBoxSetWmin->isChecked());
    settings->setValue("searchMatch/eps2", ui->checkBoxEps2->isChecked());

    settings->setValue("searchMatch/setRU", ui->checkBoxRU->isChecked());
    settings->setValue("searchMatch/valueRU", ui->spinBoxRU->value());

    settings->setValue("searchMatch/splitter", ui->splitter->saveState());
    settings->setValue("searchMatch/drHeader", ui->treeWidgetDirectories->header()->saveState());
    settings->setValue("searchMatch/twHeader", ui->treeWidgetResults->header()->saveState());
    settings->setValue("searchMatch/pwHeader", ui->treeWidgetPinned->header()->saveState());

    settings->setValue("searchMatch/crystallinity", ui->comboBoxCrystallinity->currentIndex());
    settings->setValue("searchMatch/unitCellRefinement", ui->comboBoxUnitCellRefinement->currentIndex());

    QStringList databaseCheckedDirs;

    QTreeWidgetItemIterator it(ui->treeWidgetDirectories);
    while (*it) {
        if ((*it)->checkState(0) == Qt::Checked) {
            databaseCheckedDirs.append((*it)->data(0, Qt::UserRole).toString());
        }
        ++it;
    }

    settings->setValue("searchMatch/databaseCheckedDirs", databaseCheckedDirs);

    QList<QVariant> hiddenColsResults;
    for (int i = 1; i < ui->treeWidgetResults->header()->count(); ++i) {
        if (ui->treeWidgetResults->isColumnHidden(i)) hiddenColsResults.append(QVariant(i));
    }

    settings->setValue("searchMatch/hiddenColumnsResults", hiddenColsResults);

    QList<QVariant> hiddenColsPinned;
    for (int i = 1; i < ui->treeWidgetPinned->header()->count(); ++i) {
        if (ui->treeWidgetPinned->isColumnHidden(i)) hiddenColsPinned.append(QVariant(i));
    }

    settings->setValue("searchMatch/hiddenColumnsPinned", hiddenColsPinned);
}

void SearchMatchWidget::setWidgets(QPlainTextEdit *o, ConvergenceDisplay *c)
{
    refOutput = o;
    convDisplay = c;
}

/*
 * files in l must be given with absolute file path
 */
void SearchMatchWidget::setPinnedPhases(const QList<QPair<QString, QString> > &l)
{
    for (int i = 0; i < l.size(); ++i) {
        QFileInfo fi(l.at(i).first);
        bool skip = false;

        // check if the phase is already pinned
        QList<QTreeWidgetItem*> pPhase = ui->treeWidgetPinned->findItems(fi.fileName(), Qt::MatchExactly, FILENAME);

        if (pPhase.size()) {
            for (int j = 0; j < pPhase.size(); ++j) {
                if (pPhase.at(j)->data(FILENAME, Qt::UserRole).toString() == fi.absoluteFilePath()) {
                    skip = true;
                }
            }
        }

        if (skip) continue;

        QTreeWidgetItem *it = new QTreeWidgetItem(ui->treeWidgetPinned);
        it->setText(FILENAME,    fi.fileName());
        it->setText(PHASE,       l.at(i).second);
        it->setText(GEWICHT,     QString());
        it->setText(FOMSOURCE,   "project");
        it->setData(FILENAME,    Qt::UserRole, fi.absoluteFilePath());
        it->setData(FOMSOURCE,   Qt::UserRole, false);
        it->setToolTip(FILENAME, fi.absoluteFilePath());
        ui->treeWidgetPinned->addTopLevelItem(it);
    }
}

void SearchMatchWidget::setInstrumentFile(const QString &s)
{
    QFileInfo fi(s);
    int c = ui->comboBoxInstrument->findText(fi.baseName());
    if (c >= 0) ui->comboBoxInstrument->setCurrentIndex(c);
}

void SearchMatchWidget::setLambdaFile(const QString &s)
{
    ui->radioButtonLambda->setChecked(true);
    ui->comboBoxLambda->setCurrentText(s);
}

void SearchMatchWidget::setSynchrotron(double d)
{
    ui->radioButtonSynchrotron->setChecked(true);
    ui->doubleSpinBoxSynchrotron->setValue(d);
}

void SearchMatchWidget::pin()
{
    QList< QTreeWidgetItem*> selected = ui->treeWidgetResults->selectedItems();

    for (int i = 0; i < selected.size(); ++i) {
        QTreeWidgetItem *tit = selected.at(i);

        QTreeWidgetItem *lit = new QTreeWidgetItem(ui->treeWidgetPinned);
        lit->setText(FILENAME,   tit->text(FILENAME));
        lit->setText(PHASE,      tit->text(PHASE));
        lit->setText(GEWICHT,    QString());
        lit->setText(FOMSOURCE,  "matched");
        lit->setData(FILENAME,   Qt::UserRole, tit->data(FILENAME, Qt::UserRole));
        lit->setData(PHASE,      Qt::UserRole, tit->data(PHASE, Qt::UserRole));
        lit->setData(GEWICHT,    Qt::UserRole, 0.0);
        lit->setData(FOMSOURCE,  Qt::UserRole, true);
        lit->setToolTip(FILENAME, tit->toolTip(FILENAME));
        ui->treeWidgetPinned->addTopLevelItem(lit);

        delete tit;
    }

    ui->treeWidgetResults->clearSelection();
}

void SearchMatchWidget::unpin()
{
    qDeleteAll(ui->treeWidgetPinned->selectedItems());
}

void SearchMatchWidget::run()
{
    if (!hasValidBgmnConfig) {
        qDebug() << QString("SearchMatchWidget::run(): BGMN backend configuration is not valid. Exiting.");
        return;
    }

    wasUsed = true;
    aborted = false;

    statistics.n = 0;
    statistics.sum = 0.0;
    statistics.mean = 0.0;
    statistics.stdev = 0.0;

    if (bgmnHandler->isRunning()) {
        abort();
        return;
    }

    if (!scanControl) {
        QMessageBox::warning(this, tr("Search-Match"), tr("Cannot access internal scan data.\nSearch-match aborted."));
        qDebug() << QString("SearchMatchWidget::run(): Cannot access scanControl. Exiting.");
        return;
    }

    if (!scanControl->count()) {
        QMessageBox::warning(this, tr("Search-Match"), tr("No scans found.\nSearch-match aborted."));
        qDebug() << QString("SearchMatchWidget::run(): No scans found. Exiting.");
        return;
    }

    if (!convertRawFile()) {
        QMessageBox::warning(this, tr("Search-Match"), tr("Could not convert raw data file \nto native format.\nSearch-match aborted."));
        qDebug() << QString("SearchMatchWidget::run(): Could not convert raw file to native format.");
        return;
    }

    pinnedPhases.clear();
    convDisplay->clear();
    fomList.clear();

    for (int i = 0; i < ui->treeWidgetPinned->topLevelItemCount(); ++i) {
        QTreeWidgetItem *it = ui->treeWidgetPinned->topLevelItem(i);
        it->setText(GEWICHT, QString());
        pinnedPhases.append(it->data(FILENAME, Qt::UserRole).toString());
        it->setData(GEWICHT,  Qt::UserRole, QList<QVariant>());
    }

    ui->toolButtonSave->setEnabled(false);
    ui->treeWidgetResults->clear();
    ui->tabWidget->setCurrentIndex(3);

    diaFile = QFileInfo(valFile.absolutePath() + "/" + valFile.baseName() + ".dia");
    lstFile = QFileInfo(valFile.absolutePath() + "/" + valFile.baseName() + ".lst");
    parFile = QFileInfo(valFile.absolutePath() +"/" + valFile.baseName() + ".par");
    geqFile = QFileInfo(ui->comboBoxInstrument->currentData(Qt::UserRole).toString());
    wDir = valFile.absolutePath();

    strFiles = getStrFiles();

    if (strFiles.isEmpty()) {
        if (ui->radioButtonFav->isChecked()) {
            qDebug() << QString("SearchMatchWidget::run(): No structure files found in favorites. Exiting.");
            QMessageBox::information(this, tr("Search-Match"),
                                     tr("No favorite phases found.\nPlease define "
                                        "favorites or\ndisable favorites-only search."));
            return;
        }

        qDebug() << QString("SearchMatchWidget::run(): No structure files selected. Exiting.");
        QMessageBox::information(this, tr("Search-Match"),
                                 tr("No phases retrieved from database.\n\n"
                                    "Please check the following:\n"
                                    "- At least one structure repository is checked on the \"Phases\" page\n"
                                    "- At least one phase fulfills the restrictions on the \"Restrictions\" page"));
        return;
    }

    if (pinnedPhases.size()) {
        double thAxis  = settings->value("searchMatch/compareAxis", 0.02).toDouble();
        double thAngle = settings->value("searchMatch/compareAngle", 0.1).toDouble();

        strFiles = StrFileFilter::eliminatePinnedFiles(strFiles, pinnedPhases);

        if (strFiles.isEmpty()) {
            qDebug() << QString("SearchMatchWidget::run(): No structure files left after filtering pinned ones. Exiting.");
            QMessageBox::information(this, tr("Search-Match"), tr("All structure files in the search directories are already pinned."));
            return;
        }

        strFiles = StrFileFilter::eliminateDuplicates(strFiles, pinnedPhases, thAxis, thAngle);

        if (strFiles.isEmpty()) {
            qDebug() << QString("SearchMatchWidget::run(): No structure files left after filtering duplicates of pinned ones. Exiting.");
            QMessageBox::information(this, tr("Search-Match"), tr("No structure files left after filtering pinned ones and duplicates of pinned phases."));
            return;
        }
    }

    if (!geqFile.exists()) {
        qDebug() << QString("SearchMatchWidget::run(): Instrument file %1 not found. Exiting.").arg(geqFile.absoluteFilePath());
        return;
    }

    if (!bgmnHandler->init()) {
        qDebug() << QString("SearchMatchWidget::run(): Initializing BgmnHandler failed. Exiting.");
        return;
    }

    QString geqSrc = geqFile.absoluteFilePath();
    QString geqDst = valFile.absolutePath() + "/" + geqFile.fileName();
    QString savSrc = geqFile.absolutePath() + "/" + geqFile.baseName() + ".sav";
    QString savDst = valFile.absolutePath() + "/" + geqFile.baseName() + ".sav";
    QString gerSrc = geqFile.absolutePath() + "/" + geqFile.baseName() + ".ger";
    QString gerDst = valFile.absolutePath() + "/" + geqFile.baseName() + ".ger";
    QString tplSrc = geqFile.absolutePath() + "/" + geqFile.baseName() + ".tpl";
    QString tplDst = valFile.absolutePath() + "/" + geqFile.baseName() + ".tpl";
    QString bkgSrc;
    QString bkgDst;

    BgmnSavParser sparser(tplSrc);

    if (!sparser.untFile().isEmpty()) {
        untFile = QFileInfo(geqFile.absolutePath() + "/" + sparser.untFile());

        bkgSrc = geqFile.absolutePath() + "/" + untFile.fileName();
        bkgDst = valFile.absolutePath() + "/" + untFile.fileName();
    } else {
        untFile = QFileInfo();
    }

    QFile::copy(geqSrc, geqDst);          // binary
    BgmnFileIO::copyFile(savSrc, savDst); // text
    BgmnFileIO::copyFile(gerSrc, gerDst); // text
    BgmnFileIO::copyFile(tplSrc, tplDst); // text

    if (!bkgSrc.isEmpty()) {
        BgmnFileIO::copyFile(bkgSrc, bkgDst); // text
    }

    double wmin = ui->checkBoxSetWmin->isChecked() ? ui->doubleSpinBoxWmin->value() : -1.0;
    double wmax = ui->checkBoxSetWmax->isChecked() ? ui->doubleSpinBoxWmax->value() : -1.0;

    const Scan *cScan = scanControl->firstActiveScan();
    diaHandler->init(scanControl, diaFile.absoluteFilePath(), cScan ? cScan->mid(wmin, wmax).pDataAngle().count() : 0);

    QPair<int, double> rexpDenom = scanControl->getFirst()->getRexpDenom(wmin, wmax);
    convDisplay->setRexpDenom(rexpDenom.first, rexpDenom.second);
    protocolParser.setRexpDenom(rexpDenom.first, rexpDenom.second);

    strFile = QFileInfo(strFiles.takeFirst().file());
    ui->progressBar->setEnabled(true);
    ui->progressBar->setMaximum(strFiles.size());
    ui->progressBar->show();
    emit exitStatus(1);
    startProcess();
}

QList<HklPhaseData> SearchMatchWidget::getStrFiles()
{
    QStringList restrAll  = ui->periodicTable->getAll();
    QStringList restrOne  = ui->periodicTable->getOne();
    QStringList restrNone = ui->periodicTable->getNone();

    qDebug() << QString("SearchMatchWidget::getStrFiles(): Restrictions \"All\":  %1").arg(restrAll.join(";"));
    qDebug() << QString("SearchMatchWidget::getStrFiles(): Restrictions \"One\":  %1").arg(restrOne.join(";"));
    qDebug() << QString("SearchMatchWidget::getStrFiles(): Restrictions \"None\": %1").arg(restrNone.join(";"));

    QList<HklPhaseData> phases;

    if (ui->radioButtonFav->isChecked()) {
        phases = refStrManager->getFavPhases(restrAll, restrOne, restrNone);
        qDebug() << QString("SearchMatchWidget::getStrFiles(): %1 favorite phases match the restrictions.").arg(phases.size());
        return phases;
    }

    QTreeWidgetItemIterator it(ui->treeWidgetDirectories);

    while (*it) {
        if ((*it)->checkState(0) == Qt::Checked) {
            QString dir((*it)->data(0, Qt::UserRole).toString());
            QList<HklPhaseData> p = refStrManager->getSubDirPhases(dir, restrAll, restrOne, restrNone);
            qDebug() << QString("SearchMatchWidget::getStrFiles(): %1 phases match the restrictions in %2").arg(p.size()).arg(dir);
            phases.append(p);
        }

        ++it;
    }

    return phases;
}


bool SearchMatchWidget::convertRawFile()
{
    rawFile = scanControl->fileInfo();

    QMap<QString, QVariant> flags;
    flags["fieldSeparator"] = QVariant(QString(" "));
    flags["fixBgmnZero"]    = QVariant(true);

    valFile = QFileInfo(rawFile.absolutePath() + "/" + rawFile.baseName() + ".xy");

    Scan *scan = scanControl->firstActiveScan();

    if (!scan) {
        qDebug() << QString("SearchMatchWidget::convertRawFile(): No Scan available");
        return false;
    }

    ExportHandler exHandler(this, false);

    if (exHandler.save("ASCII_XY", valFile.absoluteFilePath(), *scan, flags) > 0) {
        return true;
    }

    return false;
}

void SearchMatchWidget::startProcess()
{
    // create a random string for the local str file name, because there might already be a local
    // copy with the original name from a different refinement project. We don't want to overwrite it.

    QUuid uid = QUuid::createUuid();

#if QT_VERSION >= QT_VERSION_CHECK(5, 11, 0)
    localStrFile = QFileInfo(valFile.absolutePath() + "/" + uid.toString(QUuid::WithoutBraces) + "-" + strFile.fileName());
#else
    localStrFile = QFileInfo(valFile.absolutePath() + "/" + uid.toString() + "-" + strFile.fileName());
#endif

    BgmnFileIO::copyFile(strFile.absoluteFilePath(), localStrFile.absoluteFilePath());

    QStringList args;

    args << QString("VERZERR=%1").arg(geqFile.fileName());
    args << QString("LAMBDA=%1").arg(ui->comboBoxLambda->currentText());
    args << QString("VAL[1]=%1").arg(valFile.fileName());

    for (int i = 0; i < pinnedPhases.size(); ++i) {
        args << QString("STRUC[%1]=%2").arg(i+1).arg(pinnedPhases.at(i));
    }

    args << QString("STRUC[%1]=%2").arg(pinnedPhases.size() + 1).arg(localStrFile.absoluteFilePath());
    args << QString("ITMAX=%1").arg(ui->spinBoxItMax->value());

    if (!ui->checkBoxAniso->isChecked()) {
        args << QString("ONLYISO=Y");
    }

    int n = settings->value("bgmnProject/nThreads", QThread::idealThreadCount()).toInt();
    if (n == 0) n = QThread::idealThreadCount();

    args << QString("NTHREADS=%1").arg(n);

    if (ui->checkBoxSetWmin->isChecked()) {
        args << QString("WMIN=%1").arg(ui->doubleSpinBoxWmin->value());
    }

    if (ui->checkBoxSetWmax->isChecked()) {
        args << QString("WMAX=%1").arg(ui->doubleSpinBoxWmax->value());
    }

    if (ui->checkBoxRU->isChecked()) {
        args << QString("RU=%1").arg(ui->spinBoxRU->value());
    }

    changeParameterState(localStrFile.absoluteFilePath());

    if (untFile.isFile()) {
        args << QString("UNT=%1").arg(untFile.fileName());
    }

    args << QString("DIAGRAMM=%1").arg(diaFile.fileName());
    args << QString("LIST=%1").arg(lstFile.fileName());
    args << QString("OUTPUT=%1").arg(parFile.fileName());
    args << QString("PROTOKOLL=Y");
    args << QString("SAVE=N");

    if (ui->checkBoxEps2->isChecked()) {
        args << QString("PARAM[1]=EPS2=%1_-0.1^0.1").arg(defaultEps2, 0, 'f', 6);
    } else {
        args << QString("EPS2=%1").arg(defaultEps2, 0, 'f', 6);
    }

    scanControl->setViewStatus(1);
    bgmnHandler->run(valFile.absolutePath(), args);
}

void SearchMatchWidget::pollOutput()
{
    QString output(bgmnHandler->readOutput().trimmed());
    if (protocolParser.parse(output)) {
        convDisplay->setValues(protocolParser.getRwp(), protocolParser.getRexp(), protocolParser.getFirst());
    }

    if (refOutput) {
        if (output == ".") {
            refOutput->moveCursor(QTextCursor::End);
            refOutput->insertPlainText(output);
            refOutput->moveCursor(QTextCursor::End);
        } else {
            refOutput->appendPlainText(output + " ");
        }

        refOutput->ensureCursorVisible();
    }

    diaHandler->reload(diaFile.absoluteFilePath());
}

void SearchMatchWidget::complete(QUuid)
{
    pollOutput();
    diaHandler->stop();

    ui->progressBar->setValue(ui->progressBar->maximum() - strFiles.size() + 1);

    parseLst();

    if (strFiles.size()) {
        // QFile::copy(lstFile.absoluteFilePath(), lstFile.absolutePath() + QDir::separator() + strFile.baseName() + ".lst");
        QFile::remove(localStrFile.absoluteFilePath());
        strFile = QFileInfo(strFiles.takeFirst().file());
        const Scan *cScan = scanControl->firstActiveScan();
        diaHandler->init(scanControl, diaFile.absoluteFilePath(), cScan ? cScan->pDataAngle().count() : 0);
        startProcess();
    } else {
        scanControl->setViewStatus(2);
        emit exitStatus(aborted ? 0 : 2);

        ImportHandler ihandler;
        QString uid = ihandler.uidByFileName(rawFile.absoluteFilePath());

        scanControl->loadScanFile(rawFile.absoluteFilePath(), uid, false, scanControl->getSampleId(), true);
        QFile::remove(diaFile.absoluteFilePath());
        QFile::remove(parFile.absoluteFilePath());
        QFile::remove(lstFile.absoluteFilePath());
        QFile::remove(localStrFile.absoluteFilePath());

        ui->toolButtonSave->setEnabled(true);
        ui->progressBar->reset();
        ui->progressBar->setEnabled(false);
        ui->progressBar->hide();
    }
}

void SearchMatchWidget::abort()
{
    aborted = true;
    strFiles.clear();
    bgmnHandler->abort();
}

bool SearchMatchWidget::isRunning()
{
    return bgmnHandler->isRunning();
}

void SearchMatchWidget::parseLst()
{
    bool ok;
    BgmnLstParser lparser(lstFile.absoluteFilePath(), ok);
    if (!ok) return;

    QStringList phases = lparser.getPhaseNames();
    QList<QMap<QString, QVariant> > params;

    for (int i = 0; i < phases.size(); ++i) {
        params.append(lparser.getAllLocalParameters(phases.at(i), true));
    }

    double rwp = lparser.getRwp();
    // double rexp = lparser.getRexp();
    // double chi2 = lparser.getChi2();

    PhaseStats stats = calculateFoM(rwp, params.last());
    fomList.append(stats.fom);

    QStringList text;
    text << strFile.fileName()
         << phases.last()
         << QString()
         << QString("%1").arg(stats.fom, 0, 'f', 6, '0');

    QTreeWidgetItem *it = new QTreeWidgetItem(ui->treeWidgetResults, text);
    it->setData(FILENAME,  Qt::UserRole, strFile.absoluteFilePath());
    it->setData(PHASE,     Qt::UserRole, phases.last());
    it->setData(GEWICHT,   Qt::UserRole, stats.gewicht);
    it->setData(FOMSOURCE, Qt::UserRole, stats.fom);
    it->setToolTip(FILENAME, strFile.absoluteFilePath());
    ui->treeWidgetResults->addTopLevelItem(it);
    ui->treeWidgetResults->sortByColumn(3, Qt::DescendingOrder);

    updateQuantities(params, it);

    setItemColors();
}

void SearchMatchWidget::addReferencePhases(const QList<PeakFile> &fmap)
{
    clearResults();

    for (int i = 0; i < fmap.size(); ++i) {
        QStringList text;
        text << fmap.at(i).fileInfo.fileName()
             << fmap.at(i).phase
             << QString()
             << QString("%1").arg(1.0 - fmap.at(i).offset, 0, 'f', 6, '0');

        QTreeWidgetItem *it = new QTreeWidgetItem(ui->treeWidgetResults, text);
        it->setData(FILENAME,  Qt::UserRole, fmap.at(i).fileInfo.absoluteFilePath());
        it->setData(PHASE,     Qt::UserRole, fmap.at(i).phase);
        it->setData(GEWICHT,   Qt::UserRole, 1.0);
        it->setData(FOMSOURCE, Qt::UserRole, 1.0 - fmap.at(i).offset);
        it->setToolTip(FILENAME, fmap.at(i).fileInfo.absoluteFilePath());
        ui->treeWidgetResults->addTopLevelItem(it);
        ui->treeWidgetResults->sortByColumn(3, Qt::DescendingOrder);
    }

    wasUsed = true;
    setItemColors();
}

PhaseStats SearchMatchWidget::calculateFoM(double rwp, QMap<QString, QVariant> localGoals)
{
    double gw = getPhaseValue("GEWICHT", localGoals);
    double b1 = getPhaseValue("B1", localGoals);
    double k2 = getPhaseValue("k2", localGoals);

    // get the starting values of a, b, c from the str file
    double az, bz, cz;
    cellZero(az, bz, cz);

    double ar = getPhaseValue("A", localGoals);
    double br = getPhaseValue("B", localGoals);
    double cr = getPhaseValue("C", localGoals);

    if (qFuzzyIsNull(ar)) ar = az;
    if (qFuzzyIsNull(br)) br = bz;
    if (qFuzzyIsNull(cr)) cr = cz;

    double tb1 = settings->value("seachMatch/weightB1", 0.02).toDouble();
    double tk2 = settings->value("seachMatch/weightK2", 0.00001).toDouble();

    double a = settings->value("seachMatch/weightA", 1.0).toDouble();
    double b = settings->value("seachMatch/weightB", 1.0).toDouble();
    double c = b1 > tb1 ? settings->value("seachMatch/weightC", 1.0).toDouble() : 0.0;
    double d = k2 > tk2 ? settings->value("seachMatch/weightD", 1.0).toDouble() : 0.0;

    // multiplied by 100 to increase the weight by default
    double aterm = az <= 0.0 ? 0.0 : 100.0 * a * ((qAbs(ar - az) / az)
                                        + (qAbs(br - bz) / bz)
                                        + (qAbs(cr - cz) / cz));

    double bterm = b * gw;
    double delta = 1.0 + (qFuzzyIsNull(b1) ? 0.0 : c/b1) + d*k2;

    double fom = ((1.0 / (rwp + aterm)) + bterm) / delta;

    PhaseStats stats;

    if (localGoals.contains("Phase")) stats.name = localGoals.value("Phase", QString()).toStringList().first();
    else stats.name = QString();

    stats.fom = fom;
    stats.gewicht = gw;
    stats.azero = az;
    stats.bzero = bz;
    stats.czero = cz;
    stats.arefined = ar;
    stats.brefined = br;
    stats.crefined = cr;

    return stats;
}

double SearchMatchWidget::getPhaseValue(const QString &p, const QMap<QString, QVariant> &m)
{
    bool ok;
    QStringList l = m.value(p).toStringList();
    if (!l.size()) return 0.0;

    double d = l.first().toDouble(&ok);
    if (!ok) return 0.0;
    return d;
}

/*
 * extracts the unit cell parameters from a str file
 */
void SearchMatchWidget::cellZero(double &a, double &b, double &c)
{
    QString content = BgmnFileIO::readTextFile(strFile.absoluteFilePath());

    QRegularExpression rxA("(?:PARAM=)?A=(\\d+\\.\\d+)(?:_\\d+\\.\\d+)?(?:\\^\\d+\\.\\d+)?");
    QRegularExpression rxB("(?:PARAM=)?B=(\\d+\\.\\d+)(?:_\\d+\\.\\d+)?(?:\\^\\d+\\.\\d+)?");
    QRegularExpression rxC("(?:PARAM=)?C=(\\d+\\.\\d+)(?:_\\d+\\.\\d+)?(?:\\^\\d+\\.\\d+)?");

    QRegularExpressionMatch rmA = rxA.match(content);
    QRegularExpressionMatch rmB = rxB.match(content);
    QRegularExpressionMatch rmC = rxC.match(content);

    if (rmA.hasMatch()) a = rmA.captured(1).toDouble();
    else a = -1.0;
    if (rmB.hasMatch()) b = rmB.captured(1).toDouble();
    else b = a;
    if (rmC.hasMatch()) c = rmC.captured(1).toDouble();
    else c = a;
}

void SearchMatchWidget::setItemColors()
{
    updateMeanAndSd();

    for (int i = 0; i < ui->treeWidgetResults->topLevelItemCount(); ++i) {
        QTreeWidgetItem *it = ui->treeWidgetResults->topLevelItem(i);
        double fom = it->data(FOMSOURCE, Qt::UserRole).toDouble();

        if (statistics.stdev > 0.0) {
            QColor c = calcFoMColor((fom - statistics.mean) / statistics.stdev);
            it->setBackgroundColor(FOMSOURCE, c);
        }
    }

}

void SearchMatchWidget::updateMeanAndSd()
{
    statistics.stdev = 0.0;

    if (fomList.size() == 0) {
        return;
    }

    statistics.n = fomList.size();
    statistics.sum += fomList.last();
    statistics.mean = statistics.sum / double(statistics.n);

    if (fomList.size() == 1) {
        return;
    }

    // calculate mean and std dev of all FoM
    double sd = 0.0;

    for (int i = 0; i < fomList.size(); ++i) {
        sd += pow(fomList.at(i) - statistics.mean, 2.0);
    }

    statistics.stdev = qSqrt(sd / double(statistics.n));
}

/*
 * d = (fom - mean) / sd
 */
QColor SearchMatchWidget::calcFoMColor(double d)
{
    double g = 512.0 * d / 3.0;
    double r = 512.0 - g;

    if (r > 255.0) r = 255.0;
    if (r < 0.0)   r = 0.0;

    if (g > 255.0) g = 255.0;
    if (g < 0.0)   g = 0.0;

    return QColor(int(r), int(g), 0);
}

void SearchMatchWidget::itemClicked(QTreeWidgetItem *i, int)
{
    itemChanged(i, Q_NULLPTR);
}

void SearchMatchWidget::itemChanged(QTreeWidgetItem *cur, QTreeWidgetItem *)
{
    if (!cur) {
        emit refStructureReset();
        ui->toolButtonPin->setEnabled(false);
        ui->toolButtonUnpin->setEnabled(false);
    } else {
        QFileInfo fi(cur->text(0));
        emit refStructureSelected(fi.completeBaseName());
        ui->toolButtonPin->setEnabled(true);
        ui->toolButtonUnpin->setEnabled(true);
    }
}

QStringList SearchMatchWidget::getPinnedPhases()
{
    QStringList l;

    for (int i = 0; i < ui->treeWidgetPinned->topLevelItemCount(); ++i) {
        QTreeWidgetItem *it = ui->treeWidgetPinned->topLevelItem(i);
        if (it->data(FOMSOURCE, Qt::UserRole).toDouble()) {
            l.append(it->data(FILENAME, Qt::UserRole).toString());
        }
    }

    return l;
}

void SearchMatchWidget::updateQuantities(const QList<QMap<QString, QVariant> > &l, QTreeWidgetItem *mit)
{
    QMap<QString, QTreeWidgetItem*> pinnedItems;

    // store pointers to all pinned items for quick access
    for (int i = 0; i < ui->treeWidgetPinned->topLevelItemCount(); ++i) {
        QTreeWidgetItem *it = ui->treeWidgetPinned->topLevelItem(i);
        pinnedItems[it->text(1)] = it;
    }

    double gwMatch = getPhaseValue("GEWICHT", l.last());
    mit->setData(GEWICHT, Qt::UserRole, gwMatch);

    double gwSum = gwMatch;

    // loop over all pinned phases (all except the last one). Each pinned phase
    // stores its GEWICHT at the time it was refined together with the phase of "mit"
    // in a map associated with the phase name of "mit"
    for (int i = 0; i < l.size() - 1; ++i) {
        QString ph = l.at(i).value("Phase", QString()).toStringList().first();
        double gw = getPhaseValue("GEWICHT", l.at(i));

        if (pinnedItems.contains(ph)) {
            QMap<QString, QVariant> gl = pinnedItems[ph]->data(GEWICHT, Qt::UserRole).toMap();
            gl[mit->text(1)] = gw;
            pinnedItems[ph]->setData(GEWICHT, Qt::UserRole, gl);
        }

        gwSum += gw;
    }

    double quant = 0.0;
    if (!qFuzzyIsNull(gwSum)) quant = gwMatch / gwSum;

    mit->setText(GEWICHT, QString("%1").arg(100.0 * quant, 0, 'f', 2));
}

void SearchMatchWidget::showQuantities(QTreeWidgetItem *mit, int)
{
    showQuantities(mit, Q_NULLPTR);
}

void SearchMatchWidget::showQuantities(QTreeWidgetItem *mit, QTreeWidgetItem *)
{
    if (!mit) return;

    QString matched = mit->text(PHASE);
    double gwMatch = mit->data(GEWICHT, Qt::UserRole).toDouble();
    double gwSum = gwMatch;

    for (int i = 0; i < ui->treeWidgetPinned->topLevelItemCount(); ++i) {
        QMap<QString, QVariant> l = ui->treeWidgetPinned->topLevelItem(i)->data(GEWICHT, Qt::UserRole).toMap();
        gwSum += l.value(matched, 0.0).toDouble();
    }

    mit->setText(GEWICHT, QString("%1").arg(100.0 * gwMatch / gwSum, 0, 'f', 2));


    for (int i = 0; i < ui->treeWidgetPinned->topLevelItemCount(); ++i) {
        QMap<QString, QVariant> l = ui->treeWidgetPinned->topLevelItem(i)->data(GEWICHT, Qt::UserRole).toMap();
        double gwPinned = l.value(matched, 0.0).toDouble();
        ui->treeWidgetPinned->topLevelItem(i)->setText(GEWICHT, QString("%1").arg(100.0 * gwPinned / gwSum, 0, 'f', 2));
    }
}

QString SearchMatchWidget::getDeviceFile()
{
    QFileInfo fi(ui->comboBoxInstrument->currentData(Qt::UserRole).toString());
    return fi.absoluteFilePath();
}

void SearchMatchWidget::clearResults()
{
    ui->toolButtonPin->setEnabled(false);
    ui->toolButtonUnpin->setEnabled(false);
    ui->toolButtonSave->setEnabled(false);
    ui->treeWidgetPinned->clear();
    ui->treeWidgetResults->clear();
}

void SearchMatchWidget::toggleColumnHiddenResults(QAction *a)
{
    int c = a->data().toInt();
    ui->treeWidgetResults->setColumnHidden(c, !a->isChecked());
}


void SearchMatchWidget::toggleColumnHiddenPinned(QAction *a)
{
    int c = a->data().toInt();
    ui->treeWidgetPinned->setColumnHidden(c, !a->isChecked());
}

void SearchMatchWidget::save()
{
    QString fname = QFileDialog::getSaveFileName(this, tr("Save Score List"), valFile.absolutePath(), tr("CSV File (*.CSV *.csv)"));
    if (fname.isEmpty()) return;
    QStringList content("Pinned phases\nFile;Phase;Source");

    QStringList pinnedNames;
    QList<QMap<QString, QVariant> > pinnedGw;

    for (int i = 0; i < ui->treeWidgetPinned->topLevelItemCount(); ++i) {
        QTreeWidgetItem *it = ui->treeWidgetPinned->topLevelItem(i);
        QString f = it->data(FILENAME,  Qt::UserRole).toString();
        QString p = it->data(GEWICHT,   Qt::UserRole).toString();
        bool    o = it->data(FOMSOURCE, Qt::UserRole).toBool();

        pinnedNames << p;
        pinnedGw << it->data(GEWICHT, Qt::UserRole).toMap();

        content << QString("%1;%2;%3").arg(f).arg(p).arg(o ? "matched" : "project");
    }

    QStringList header;
    header << QString("\n\nMatched phases\nFile;Phase;GEWICHT;Figure of Merit");

    for (int j = 0; j < pinnedNames.size(); ++j) {
        header << QString("GEWICHT %1").arg(pinnedNames.at(j));
    }

    content << header.join(";");

    for (int i = 0; i < ui->treeWidgetResults->topLevelItemCount(); ++i) {
        QTreeWidgetItem *it = ui->treeWidgetResults->topLevelItem(i);
        QString f = it->data(FILENAME,  Qt::UserRole).toString();
        QString p = it->data(PHASE,     Qt::UserRole).toString();
        double g  = it->data(GEWICHT,   Qt::UserRole).toDouble();
        double o  = it->data(FOMSOURCE, Qt::UserRole).toDouble();

        QStringList line;
        line << f << p << QString("%1;%2").arg(g, 0, 'f', 6).arg(o, 0, 'f', 6);

        for (int j = 0; j < pinnedNames.size(); ++j) {
            line << QString("%1").arg(pinnedGw.at(j)[p].toDouble(), 0, 'f', 6);
        }

        content << line.join(";");
    }

    BgmnFileIO::writeTextFile(fname, content.join("\n"));
}

void SearchMatchWidget::changeParameterState(const QString &strFile)
{
    BgmnStrParser strParser(strFile);

    double cellLim = ui->comboBoxUnitCellRefinement->currentData(Qt::UserRole).toDouble();
    double b1Lim = ui->comboBoxCrystallinity->currentData(Qt::UserRole).toDouble();

    int cellStat = qFuzzyIsNull(cellLim) ? 0 : 1;
    int b1Stat = qFuzzyIsNull(b1Lim) ? 0 : 1;

    strParser.setRefinementState(QString("UC:A:%1").arg(cellLim, 0, 'f', 4), cellStat);
    strParser.setRefinementState(QString("UC:B:%1").arg(cellLim, 0, 'f', 4), cellStat);
    strParser.setRefinementState(QString("UC:C:%1").arg(cellLim, 0, 'f', 4), cellStat);
    strParser.setRefinementState(QString("UC:ALPHA:%1").arg(cellLim, 0, 'f', 4), cellStat);
    strParser.setRefinementState(QString("UC:BETA:%1").arg(cellLim, 0, 'f', 4), cellStat);
    strParser.setRefinementState(QString("UC:GAMMA:%1").arg(cellLim, 0, 'f', 4), cellStat);
    strParser.setRefinementState(QString("PR:B1:%1").arg(b1Lim, 0, 'f', 4), b1Stat);
    strParser.setRefinementState(QString("PR:k2:"), 0);

    strParser.writeToFile(strFile);
}
