/***************************************************************************
                          codcifretrievedialog.cpp  -  description
                             -------------------
    begin                : Thu Jan 11 19:42:15 CEST 2021
    copyright            : (C) 2021 by Nicola Doebelin
    email                : ndoebelin@gmx.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "codcifretrievedialog.h"
#include "ui_codcifretrievedialog.h"
#include "../libXrdIO/bgmnfileio.h"
#include <QRegularExpression>
#include <QString>
#include <QFileDialog>
#include <QtSql>

CodCifRetrieveDialog::CodCifRetrieveDialog(QWidget *parent) :
    QDialog(parent),
    ui(new Ui::CodCifRetrieveDialog)
{
    ui->setupUi(this);
    settings = SettingsManager::getInstance();
    ui->pushButtonSearch->setDefault(true);
    ui->pushButtonSearch->setEnabled(false);

    QStringList headers = QStringList() << "COD ID" << "Mineral" << "Formula" << "Space Group" << "Year" << "Bibliography";
    ui->tableWidgetMatches->setColumnCount(headers.size());
    ui->tableWidgetMatches->setHorizontalHeaderLabels(headers);
    setDbStatusLabel(false);

    ui->comboBoxTemperature->addItem(tr("Room temperature"));
    ui->comboBoxTemperature->addItem(tr("High temperature"));
    ui->comboBoxTemperature->addItem(tr("Low temperature"));
    ui->comboBoxTemperature->setCurrentIndex(0);

#if (QT_VERSION < QT_VERSION_CHECK(5, 12, 0))
    // older versions of Qt SQLITE3 driver don't support the REGEX extension
    ui->lineEditElementsInclude->setEnabled(false);
    ui->lineEditElementsExclude->setEnabled(false);
#endif

    initSettings();
}

CodCifRetrieveDialog::~CodCifRetrieveDialog()
{
    QSqlDatabase::database("cod").close();
    QSqlDatabase::removeDatabase("cod");
    delete ui;
}

void CodCifRetrieveDialog::accept()
{
    saveSettings();
    QDialog::accept();
}

void CodCifRetrieveDialog::initSettings()
{
    databaseMode = LOCAL_SQLITE;
    restoreGeometry(settings->value("codQueryDialog/geometry", QByteArray()).toByteArray());
    ui->tableWidgetMatches->horizontalHeader()->restoreState(settings->value("codQueryDialog/tableHeader", QByteArray()).toByteArray());

    localDbFile = settings->value("codQueryDialog/localDbFile", QString()).toString();
    ui->lineEditDbFile->setText(localDbFile);

    if (QFile::exists(localDbFile)) {
        qDebug() << QString("CodCifRetrieveDialog::initSettings(): Connecting to database %1").arg(localDbFile);
        connectDatabase();
    } else {
        qDebug() << QString("CodCifRetrieveDialog::initSettings(): Database %1 cannot be found").arg(localDbFile);
    }
}

void CodCifRetrieveDialog::saveSettings()
{
    settings->setValue("codQueryDialog/tableHeader", ui->tableWidgetMatches->horizontalHeader()->saveState());
    settings->setValue("codQueryDialog/geometry", saveGeometry());
    settings->setValue("codQueryDialog/localDbFile", ui->lineEditDbFile->text());
}

QStringList CodCifRetrieveDialog::getRecords()
{
    QStringList codids;

    for (int i = 0; i < ui->tableWidgetMatches->rowCount(); ++i) {
        if (ui->tableWidgetMatches->item(i, 0)->checkState() == Qt::Checked) {
            codids.append(ui->tableWidgetMatches->item(i, 0)->text());
        }
    }

    return codids;
}

void CodCifRetrieveDialog::connectDatabase()
{
    qApp->processEvents();

    if (QSqlDatabase::contains("cod")) {
        QSqlDatabase::removeDatabase("cod");
    }

    if (databaseMode == ONLINE_MYSQL) {
        // for direct access of the COD mysql server
        // REQUIRES QMYSQL driver to be bundled with ther application!!!!
        QSqlDatabase db = QSqlDatabase::addDatabase("QMYSQL", "cod");
        db.setHostName("sql.crystallography.net");
        db.setDatabaseName("cod");
        db.setUserName("cod_reader");
        db.setConnectOptions("MYSQL_OPT_CONNECT_TIMEOUT=10");
        ui->plainTextEditDbMessages->setPlainText("Connecting to sql.crystallography.net...");
        qDebug() << QString("CodCifRetrieveDialog::connectDatabase(): Connecting to MySQL database: %1").arg("sql.crystallography.net");
    } else {
        // for direct access of a local sqlite database
        QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE", "cod");
        db.setDatabaseName(ui->lineEditDbFile->text());
        db.setConnectOptions("QSQLITE_ENABLE_REGEXP=25;QSQLITE_OPEN_READONLY=1"); // cache size 25 is the default value
        ui->plainTextEditDbMessages->setPlainText(QString("Connecting to local database %1...").arg(localDbFile));
        qDebug() << QString("CodCifRetrieveDialog::connectDatabase(): Connecting to SQLITE database: %1").arg(localDbFile);
    }

    if (QSqlDatabase::database("cod").open()) {
        ui->pushButtonSearch->setEnabled(true);
        ui->plainTextEditDbMessages->appendPlainText("Connected");
        setDbStatusLabel(true);
        qDebug() << QString("CodCifRetrieveDialog::connectDatabase(): Opening DB successful");
    } else {
        ui->plainTextEditDbMessages->setPlainText("Error connecting to database:");
        ui->plainTextEditDbMessages->appendPlainText(QSqlDatabase::database("cod").lastError().text());
        ui->pushButtonSearch->setEnabled(false);
        setDbStatusLabel(false);
        qDebug() << QString("CodCifRetrieveDialog::connectDatabase(): Opening DB failed");
        qDebug() << QString("                                         %1").arg(QSqlDatabase::database("cod").lastError().text());
    }
}

void CodCifRetrieveDialog::disconnectDatabase()
{
    QSqlDatabase::database("cod").close();
    qDebug() << QString("CodCifRetrieveDialog::disconnectDatabase(): DB closed");
    QSqlDatabase::removeDatabase("cod");
    setDbStatusLabel(false);
}

QString CodCifRetrieveDialog::queryString()
{
    QStringList query;

    if (!ui->lineEditCodId->text().isEmpty()) {
        QStringList codIdLst = ui->lineEditCodId->text().split(QRegularExpression("[\\s,;]+"));
        QStringList subQueries;

        QRegularExpression rxCodId("(\\d{7})");

        for (int i = 0; i < codIdLst.size(); ++i) {
            QRegularExpressionMatch rmCodId = rxCodId.match(codIdLst.at(i));

            if (rmCodId.hasMatch()) {
                subQueries.append(QString("file LIKE '%1'").arg(rmCodId.captured(1)));
            }
        }

        if (subQueries.size() > 1) {
            query.append("(" + subQueries.join(" OR ") + ")");
        } else if (subQueries.size() == 1) {
            query.append(subQueries.first());
        }
    }

    if (!ui->lineEditMineralName->text().isEmpty()) query.append(QString("mineral LIKE '\%%1\%'").arg(ui->lineEditMineralName->text()));
    if (!ui->lineEditDOI->text().isEmpty()) query.append(QString("doi LIKE '\%%1\%'").arg(ui->lineEditDOI->text()));

    if (ui->spinBoxElementsMin->value() > 0) query.append(QString("nel >= %1").arg(ui->spinBoxElementsMin->value()));
    if (ui->spinBoxElementsMax->value() > 0) query.append(QString("nel <= %1").arg(ui->spinBoxElementsMax->value()));

    if (!ui->lineEditHM->text().isEmpty()) query.append(QString("sg LIKE '%1'").arg(ui->lineEditHM->text()));

    if (!ui->lineEditElementsInclude->text().isEmpty()) {
        QStringList elementsIncl = ui->lineEditElementsInclude->text().split(QRegularExpression("[\\s;,]+"));

        for (int i = 0; i < elementsIncl.size(); ++i) {
            if (databaseMode == ONLINE_MYSQL) query.append(QString("formula REGEXP ' %1[0-9 ]'").arg(elementsIncl.at(i)));
            if (databaseMode == LOCAL_SQLITE) query.append(QString("formula REGEXP '.+\\s%1[0-9\\s].+'").arg(elementsIncl.at(i)));
        }
    }

    if (!ui->lineEditElementsExclude->text().isEmpty()) {
        QStringList elementsExcl = ui->lineEditElementsExclude->text().split(QRegularExpression("[\\s;,]+"));

        for (int i = 0; i < elementsExcl.size(); ++i) {
            if (databaseMode == ONLINE_MYSQL) query.append(QString("NOT formula REGEXP ' %1[0-9 ]'").arg(elementsExcl.at(i)));
            if (databaseMode == LOCAL_SQLITE) query.append(QString("NOT formula REGEXP '.+\\s%1[0-9\\s].+'").arg(elementsExcl.at(i)));
        }
    }

    // conversion of the database from mysql to sqlite replaces NULL with \N. We need to adapt the queries accordingly

    if (ui->checkBoxRestrictTemp->isChecked()) {
        if (databaseMode == ONLINE_MYSQL) {
            if (ui->comboBoxTemperature->currentIndex() == 0) query.append("(diffrtemp between 273 and 300 or diffrtemp is null)");
            else if (ui->comboBoxTemperature->currentIndex() == 1) query.append("diffrtemp > 300");
            else if (ui->comboBoxTemperature->currentIndex() == 2) query.append("diffrtemp < 273");
        } else if (databaseMode == LOCAL_SQLITE) {
            if (ui->comboBoxTemperature->currentIndex() == 0) query.append("(diffrtemp between 273 and 300 or diffrtemp is '\\N')");
            else if (ui->comboBoxTemperature->currentIndex() == 1) query.append("(diffrtemp > 300 AND diffrtemp IS NOT '\\N')");
            else if (ui->comboBoxTemperature->currentIndex() == 2) query.append("(diffrtemp < 273 AND diffrtemp IS NOT '\\N')");
        }
    }

    if (!ui->lineEditText1->text().isEmpty()) query.append(QString("title LIKE '\%%1\%'").arg(ui->lineEditText1->text()));
    if (!ui->lineEditText2->text().isEmpty()) query.append(QString("title LIKE '\%%1\%'").arg(ui->lineEditText2->text()));

    if (!ui->lineEditAuthors1->text().isEmpty()) query.append(QString("authors LIKE '\%%1\%'").arg(ui->lineEditAuthors1->text()));
    if (!ui->lineEditAuthors2->text().isEmpty()) query.append(QString("authors LIKE '\%%1\%'").arg(ui->lineEditAuthors2->text()));

    if (!ui->lineEditJournalName->text().isEmpty()) query.append(QString("journal LIKE '\%%1\%'").arg(ui->lineEditJournalName->text()));
    if (!ui->lineEditJournalVolume->text().isEmpty()) query.append(QString("volume LIKE '\%%1\%'").arg(ui->lineEditJournalVolume->text()));

    if (ui->spinBoxJournalYearMin->value() > 0) query.append(QString("year >= %1").arg(ui->spinBoxJournalYearMin->value()));
    if (ui->spinBoxJournalYearMax->value() > 0) query.append(QString("year <= %1").arg(ui->spinBoxJournalYearMax->value()));

    return query.join(" AND ");
}

void CodCifRetrieveDialog::queryDatabase()
{
    ui->tabWidget->setEnabled(false);
    qApp->setOverrideCursor(Qt::WaitCursor);

    clearTable(false);

    QString qstr = queryString();

    if (qstr.simplified().isEmpty()) {
        qApp->restoreOverrideCursor();
        ui->tabWidget->setEnabled(true);
        return;
    }

    QString q = QString("SELECT file, mineral, formula, sg, year, text FROM data WHERE %1").arg(qstr);
    qDebug() << QString("CodCifRetrieveDialog::queryDatabase(): Executing query %1").arg(q);

    ui->plainTextEditDbMessages->appendPlainText("Querying database:");
    ui->plainTextEditDbMessages->appendPlainText(q);

    if (!QSqlDatabase::database("cod").isOpen()) {
        qApp->restoreOverrideCursor();
        ui->tabWidget->setEnabled(true);
        qDebug() << QString("CodCifRetrieveDialog::queryDatabase(): DB is closed, exiting");
        return;
    }

    qApp->processEvents();

    QSqlQuery query(QSqlDatabase::database("cod"));
    if (!query.exec(q)) {
        qApp->restoreOverrideCursor();
        ui->tabWidget->setEnabled(true);
        qDebug() << QString("CodCifRetrieveDialog::queryDatabase(): Query returned an error: %1").arg(query.lastError().text());
        ui->plainTextEditDbMessages->appendPlainText(query.lastError().text());
        return;
    }

    int n = 0;

    while (query.next()) {
        int r = ui->tableWidgetMatches->rowCount();
        ui->tableWidgetMatches->setRowCount(r+1);

        QRegularExpression rxFormula("-\\s+([^-]+)\\s+-");
        QRegularExpressionMatch rm = rxFormula.match(query.value(2).toString());
        QString formula = rm.hasMatch() ? rm.captured(1) : query.value(2).toString();
        QString mineral = query.value(1).toString();
        if (mineral == "\\N") mineral = "n.a.";

        ui->tableWidgetMatches->setItem(r, 0, new QTableWidgetItem(query.value(0).toString()));
        ui->tableWidgetMatches->setItem(r, 1, new QTableWidgetItem(mineral));
        ui->tableWidgetMatches->setItem(r, 2, new QTableWidgetItem(formula));
        ui->tableWidgetMatches->setItem(r, 3, new QTableWidgetItem(query.value(3).toString()));
        ui->tableWidgetMatches->setItem(r, 4, new QTableWidgetItem(query.value(4).toString()));
        ui->tableWidgetMatches->setItem(r, 5, new QTableWidgetItem(query.value(5).toString()));

        ui->tableWidgetMatches->item(r, 0)->setCheckState(Qt::Unchecked);
        ++n;
    }

    qDebug() << QString("CodCifRetrieveDialog::queryDatabase(): Query returned %1 matches").arg(n);

    qApp->restoreOverrideCursor();
    ui->tabWidget->setEnabled(true);
}

void CodCifRetrieveDialog::checkAll()
{
    for (int i = 0; i < ui->tableWidgetMatches->rowCount(); ++i) {
         ui->tableWidgetMatches->item(i, 0)->setCheckState(Qt::Checked);
    }
}

void CodCifRetrieveDialog::checkNone()
{
    for (int i = 0; i < ui->tableWidgetMatches->rowCount(); ++i) {
         ui->tableWidgetMatches->item(i, 0)->setCheckState(Qt::Unchecked);
    }
}

void CodCifRetrieveDialog::saveQueries()
{
    QString dir = settings->value("codQueryDialog/saveListDir", QDir::homePath()).toString();
    QString file = QFileDialog::getSaveFileName(this, tr("Save match list"), dir, tr("CSV table (*.csv *.CSV)"));
    if (file.isEmpty()) return;

    QFileInfo fi(file);
    settings->setValue("codQueryDialog/saveListDir", fi.absolutePath());

    QStringList out;

    for (int i = 0; i < ui->tableWidgetMatches->rowCount(); ++i) {
        QStringList l;

        for (int j = 0; j < ui->tableWidgetMatches->columnCount(); ++j) {
            l.append(ui->tableWidgetMatches->item(i, j)->text());
        }

        out.append(l.join(";"));
    }

    BgmnFileIO::writeTextFile(fi.absoluteFilePath(), out.join("\n"));
    qDebug() << QString("CodCifRetrieveDialog::saveQueries(): List saved to %1").arg(fi.absoluteFilePath());
}

void CodCifRetrieveDialog::selectDbFile()
{
    QFileInfo fi(ui->lineEditDbFile->text());

    QString fn = QFileDialog::getOpenFileName(this, tr("Select SqLite3 database file"),
                                              fi.absolutePath(), tr("SQLITE3 file (*.db3 *.DB3)"));

    if (fn.isEmpty()) return;

    if (!QFile::exists(fn)) {
        ui->lineEditDbFile->clear();
        return;
    }

    ui->lineEditDbFile->setText(fn);
    connectDatabase();
}

void CodCifRetrieveDialog::clearForm()
{
    ui->lineEditCodId->clear();
    ui->lineEditMineralName->clear();
    ui->lineEditElementsExclude->clear();
    ui->lineEditElementsInclude->clear();
    ui->spinBoxElementsMax->setValue(0);
    ui->spinBoxElementsMin->setValue(0);
    ui->lineEditHM->clear();
    ui->comboBoxTemperature->setCurrentIndex(0);
    ui->checkBoxRestrictTemp->setChecked(false);

    ui->lineEditText1->clear();
    ui->lineEditText2->clear();
    ui->lineEditAuthors1->clear();
    ui->lineEditAuthors2->clear();
    ui->lineEditJournalName->clear();
    ui->lineEditJournalVolume->clear();
    ui->spinBoxJournalYearMax->setValue(0);
    ui->spinBoxJournalYearMin->setValue(0);
    ui->lineEditDOI->clear();
}

void CodCifRetrieveDialog::clearTable(bool all)
{
    if (all) {
        ui->tableWidgetMatches->clearContents();
        ui->tableWidgetMatches->setRowCount(0);
        return;
    }

    // if all is false, keep the rows that are checked

    for (int i = ui->tableWidgetMatches->rowCount(); i > 0; --i) {
        int n = i - 1;

        if (ui->tableWidgetMatches->item(n, 0)->checkState() == Qt::Unchecked) {
            ui->tableWidgetMatches->removeRow(n);
        }
    }
}

void CodCifRetrieveDialog::setDbStatusLabel(bool b)
{
    if (b) {
        ui->labelDbStatus->setText("<font color=\"Green\">Database connected.</font>");
    } else {
        ui->labelDbStatus->setText("<font color=\"Red\">No database connected. Select a valid database file.</font>");
    }
}
