/***************************************************************************
                          baselinedialog.cpp  -  description
                             -------------------
    begin                : Tue Jan 23 22:00:00 CEST 2018
    copyright            : (C) 2018 by Nicola Doebelin
    email                : ndoebelin@gmx.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <QComboBox>
#include <QSpinBox>
#include <QShowEvent>
#include <QCloseEvent>
#include <QTableWidget>
#include <limits.h>
#include <QDebug>
#include "../libXrdIO/structs.h"
#include "baselinedialog.h"
#include "ui_baselinedialog.h"

BaseLineDialog::BaseLineDialog(QWidget *parent) :
    AbstractToolDialog(parent),
    ui(new Ui::BaseLineDialog)
{
    ui->setupUi(this);
    setWindowTitle(tr("Base Line"));

    ui->tableWidgetNodes->setEnabled(false);
    temporaryScan = nullptr;
    headers << tr("2theta") << tr("Intensity");

    ui->comboBoxGolotInterpolation->addItem(QString("Linear"),             QVariant(SPLINE_LINEAR));
    ui->comboBoxGolotInterpolation->addItem(QString("Catmull-Rom Spline"), QVariant(SPLINE_CATMULL_ROM));
    ui->comboBoxGolotInterpolation->addItem(QString("Cubic Spline"),       QVariant(SPLINE_CUBIC));
    ui->comboBoxGolotInterpolation->addItem(QString("Akima Spline"),       QVariant(SPLINE_AKIMA));
    ui->comboBoxGolotInterpolation->addItem(QString("Monotone Spline"),    QVariant(SPLINE_MONOTONE));

    initSettings();

    connect(ui->toolButtonAddAnchor, SIGNAL(clicked(bool)), this, SLOT(addGolotAnchor()));
    connect(ui->toolButtonRemoveAnchor, SIGNAL(clicked(bool)), this, SLOT(removeGolotAnchor()));
    connect(ui->comboBoxAlgo, SIGNAL(currentIndexChanged(int)), this, SLOT(algoChanged(int)));
    connect(ui->comboBoxScan, SIGNAL(currentIndexChanged(int)), this, SLOT(computeSnip()));

    connect(ui->comboBoxSnipWindow, SIGNAL(currentIndexChanged(int)), this, SLOT(computeSnip()));
    connect(ui->spinBoxSnipM, SIGNAL(valueChanged(int)), this, SLOT(computeSnip()));

    connect(ui->spinBoxGolotWindow, SIGNAL(valueChanged(int)), this, SLOT(computeGolot()));
    connect(ui->doubleSpinBoxGolotvinN, SIGNAL(valueChanged(double)), this, SLOT(computeGolot()));
    connect(ui->spinBoxGolotSteps, SIGNAL(valueChanged(int)), this, SLOT(computeGolot()));
    connect(ui->comboBoxGolotLast, SIGNAL(currentIndexChanged(int)), this, SLOT(computeGolot()));
    connect(ui->spinBoxGolotSmooth, SIGNAL(valueChanged(int)), this, SLOT(computeGolot()));
    connect(ui->comboBoxGolotInterpolation, SIGNAL(currentIndexChanged(int)), this, SLOT(computeGolot()));
    connect(ui->spinBoxGolotSensitivity, SIGNAL(valueChanged(int)), this, SLOT(computeGolot()));

    connect(ui->tableWidgetNodes, SIGNAL(cellChanged(int,int)), this, SLOT(generateGolot()));
    connect(ui->tableWidgetNodes, SIGNAL(currentItemChanged(QTableWidgetItem *, QTableWidgetItem *)),
            this, SLOT(highlightAnchor(QTableWidgetItem *, QTableWidgetItem *)));
}

BaseLineDialog::~BaseLineDialog()
{
    delete ui;
}

void BaseLineDialog::initSettings()
{
    ui->comboBoxAlgo->setCurrentIndex(settings->value("baseLine/algorithm", 0).toInt());
    ui->stackedWidgetParams->setCurrentIndex(ui->comboBoxAlgo->currentIndex());
    if (ui->stackedWidgetParams->currentIndex() == 1) {
        ui->tableWidgetNodes->setEnabled(true);
    }

    ui->comboBoxSnipWindow->setCurrentIndex(settings->value("baseLine/algoSnip/window", 1).toInt());
    ui->spinBoxSnipM->setValue(settings->value("baseLine/algoSnip/iterations", 60).toInt());

    ui->spinBoxGolotWindow->setValue(settings->value("baseLine/algoGolot/window", 3).toInt());
    ui->doubleSpinBoxGolotvinN->setValue(settings->value("baseLine/algoGolot/noiseMulti", 1.0).toDouble());
    ui->spinBoxGolotSteps->setValue(settings->value("baseLine/algoGolot/steps", 1).toInt());
    ui->comboBoxGolotLast->setCurrentIndex(settings->value("baseLine/algoGolot/mode", 0).toInt());
    ui->spinBoxGolotSmooth->setValue(settings->value("baseLine/algoGolot/dataSmooth", 1).toInt());
    ui->comboBoxGolotInterpolation->setCurrentIndex(settings->value("baseLine/algoGolot/interpolation", SPLINE_AKIMA).toInt());
    ui->spinBoxGolotSensitivity->setValue(settings->value("baseLine/algoGolot/sensitivity", 3).toInt());

    restoreGeometry(settings->value("baseLine/geometry", QByteArray()).toByteArray());
}

void BaseLineDialog::saveSettings()
{
    settings->setValue("baseLine/algorithm", ui->comboBoxAlgo->currentIndex());

    settings->setValue("baseLine/algoSnip/window", ui->comboBoxSnipWindow->currentIndex());
    settings->setValue("baseLine/algoSnip/iterations", ui->spinBoxSnipM->value());

    settings->setValue("baseLine/algoGolot/window", ui->spinBoxGolotWindow->value());
    settings->setValue("baseLine/algoGolot/noiseMulti", ui->doubleSpinBoxGolotvinN->value());
    settings->setValue("baseLine/algoGolot/steps", ui->spinBoxGolotSteps->value());
    settings->setValue("baseLine/algoGolot/mode", ui->comboBoxGolotLast->currentIndex());
    settings->setValue("baseLine/algoGolot/dataSmooth", ui->spinBoxGolotSmooth->value());
    settings->setValue("baseLine/algoGolot/interpolation", ui->comboBoxGolotInterpolation->currentIndex());
    settings->setValue("baseLine/algoGolot/sensitivity", ui->spinBoxGolotSensitivity->value());

    settings->setValue("baseLine/geometry", saveGeometry());
}

void BaseLineDialog::preSetProject()
{
    temporaryScan = nullptr;
}

void BaseLineDialog::postSetProject()
{
    clearAnchorPoints();
    clearTemporary();

    if (!checkBackend()) {
        clearGui();
        return;
    }

    parseScans();

    if (ui->stackedWidgetParams->currentIndex() == 0) computeSnip();
    if (ui->stackedWidgetParams->currentIndex() == 1) computeGolot();
}

void BaseLineDialog::updateView()
{
    if (!isVisible()) return;
    parseScans();
}

void BaseLineDialog::algoChanged(int i)
{
    clearAnchorPoints();
    clearTemporary();

    if (i == 0) {
        ui->tableWidgetNodes->clear();
        ui->tableWidgetNodes->setEnabled(false);
        ui->toolButtonAddAnchor->setEnabled(false);
        ui->toolButtonRemoveAnchor->setEnabled(false);
        computeSnip();
    }

    if (i == 1) {
        ui->tableWidgetNodes->setEnabled(true);
        ui->toolButtonAddAnchor->setEnabled(true);
        ui->toolButtonRemoveAnchor->setEnabled(true);
        computeGolot();
    }
}

void BaseLineDialog::computeSnip()
{
    if (!checkBackend()) return;

    int c = ui->comboBoxScan->currentIndex();
    int w = ui->comboBoxSnipWindow->currentIndex();
    int m = ui->spinBoxSnipM->value();

    if (c < 0) return;

    const Scan *scan = graphModel->getScan(ui->comboBoxScan->currentData(Qt::UserRole).toUuid());
    if (!scan) return;

    Scan bl(ScanOps::baseLineSNIP(*scan, m, w));
    bl.setName(QString("Baseline %1").arg(ui->comboBoxScan->currentText()));
    bl.setColor(QString());
    bl.setTypes(Scan::XY | Scan::SYNTHETIC | Scan::BACKGROUND | Scan::TEMPORARY);

    int idx = tempScanIndex();

    if (idx < 0) {
        graphControl->appendScan(bl, true);
        temporaryScan = graphModel->getLast();
    } else {
        graphControl->replaceScan(idx, bl, true);
        temporaryScan = graphModel->getScan(idx);
    }
}

void BaseLineDialog::computeGolot()
{
    if (!checkBackend()) return;

    int sm = ui->spinBoxGolotSmooth->value();
    int c = ui->comboBoxScan->currentIndex();
    int w = ui->spinBoxGolotWindow->value();
    double n = ui->doubleSpinBoxGolotvinN->value();
    int st = ui->spinBoxGolotSteps->value();
    int sens = ui->spinBoxGolotSensitivity->value();
    int md = ui->comboBoxGolotLast->currentIndex();

    if (c < 0) return;

    const Scan *scan = graphModel->getScan(ui->comboBoxScan->currentData(Qt::UserRole).toUuid());
    if (!scan) return;

    setTableGolot(ScanOps::baseLineGolot(*scan, sm, w, n, st, sens, md));
    generateGolot();
}

void BaseLineDialog::setTableGolot(const QVector<QPointF> &pt)
{
    blockUpdateSignals(true);

    ui->tableWidgetNodes->setColumnCount(2);
    ui->tableWidgetNodes->setRowCount(pt.size());
    ui->tableWidgetNodes->setHorizontalHeaderLabels(headers);

    for (int i = 0; i < pt.size(); ++i) {
        ui->tableWidgetNodes->setItem(i, 0, new QTableWidgetItem(QString("%1").arg(pt.at(i).x(), 0, 'f', 5)));
        ui->tableWidgetNodes->setItem(i, 1, new QTableWidgetItem(QString("%1").arg(pt.at(i).y(), 0, 'f', 2)));
    }

    blockUpdateSignals(false);
}

void BaseLineDialog::generateGolot()
{
    if (!checkBackend()) return;

    int c = ui->comboBoxScan->currentIndex();
    if (c < 0) return;

    const Scan *scan = graphModel->getScan(ui->comboBoxScan->currentData(Qt::UserRole).toUuid());
    if (!scan) return;

    setAnchorPoints();

    Scan bl = scan->clone();
    bl.pDataHkl().clear();
    bl.setName(QString("Baseline %1").arg(ui->comboBoxScan->currentText()));
    bl.setColor(QString());
    bl.setTypes(Scan::XY | Scan::SYNTHETIC | Scan::BACKGROUND | Scan::TEMPORARY);
    interpolateAnchorPoints(bl);

    graphView->setAnchorPoints(anchors);

    int idx = tempScanIndex();

    if (idx < 0) {
        graphControl->appendScan(bl, true);
        temporaryScan = graphModel->getLast();
    } else {
        graphControl->replaceScan(idx, bl, true);
        temporaryScan = graphModel->getScan(idx);
    }
}

void BaseLineDialog::setAnchorPoints()
{
    anchors.clear();

    for (int i = 0; i < ui->tableWidgetNodes->rowCount(); ++i) {
        QTableWidgetItem *ix = ui->tableWidgetNodes->item(i, 0);
        QTableWidgetItem *iy = ui->tableWidgetNodes->item(i, 1);

        if (ix && iy) {
            double dx = ix->text().toDouble();
            double dy = iy->text().toDouble();
            anchors.append(global::AnchorPoint(dx, dy, false));
        }
    }
}

void BaseLineDialog::clearAnchorPoints()
{
    if (graphView) graphView->setAnchorPoints(QVector<global::AnchorPoint>());
}

void BaseLineDialog::interpolateAnchorPoints(Scan &sc)
{
    // use a map to sort the data and eliminate duplicate x values
    // note: we must store the x value as int(double*multi), else elimination will not work reliably
    // due to rounding errors of doubles. If duplicates in x occur, the spline algorithms will crash

    QMap<int, double> data;
    double multi = 10000.0;

    for (int i = 0; i < ui->tableWidgetNodes->rowCount(); ++i) {
        QTableWidgetItem *ix = ui->tableWidgetNodes->item(i, 0);
        QTableWidgetItem *iy = ui->tableWidgetNodes->item(i, 1);

        if (ix && iy) {
            double dx = ix->text().toDouble();
            double dy = iy->text().toDouble();

            data.insert(int(multi * dx), dy);
        }
    }

    QVector<double> mX;
    QVector<double> mY;

    QMapIterator<int, double> it(data);

    while (it.hasNext()) {
        it.next();
        mX.append(double(it.key()) / multi);
        mY.append(it.value());
    }

    int ipol = ui->comboBoxGolotInterpolation->currentData(Qt::UserRole).toInt();

    if (ipol == SPLINE_LINEAR)      sc.setDataInt(ScanOps::interpolateSplineLinear(mX,     mY, sc.pDataAngle()));
    if (ipol == SPLINE_CATMULL_ROM) sc.setDataInt(ScanOps::interpolateSplineCatmullRom(mX, mY, sc.pDataAngle()));
    if (ipol == SPLINE_CUBIC)       sc.setDataInt(ScanOps::interpolateSplineCubic(mX,      mY, sc.pDataAngle()));
    if (ipol == SPLINE_AKIMA)       sc.setDataInt(ScanOps::interpolateSplineAkima(mX,      mY, sc.pDataAngle()));
    if (ipol == SPLINE_MONOTONE)    sc.setDataInt(ScanOps::interpolateSplineMonotone(mX,   mY, sc.pDataAngle()));
}

void BaseLineDialog::parseScans()
{
    bool oldState = ui->comboBoxScan->blockSignals(true);
    int n = ui->comboBoxScan->currentIndex();

    ui->comboBoxScan->clear();

    if (!graphModel) {
        ui->comboBoxScan->blockSignals(oldState);
        return;
    }

    for (int i = 0; i < graphModel->count(); ++i) {
        if (!graphModel->at(i)->hasScanData()) continue;
        if (graphModel->at(i)->isTemporary())  continue;

        ui->comboBoxScan->addItem(QString(graphModel->at(i)->name()), graphModel->at(i)->uid());
    }

    n = n < ui->comboBoxScan->count() ? n : ui->comboBoxScan->count() - 1;
    ui->comboBoxScan->setCurrentIndex(n < 0 ? 0 : n);
    ui->comboBoxScan->blockSignals(oldState);
}

void BaseLineDialog::append()
{
    keepTemporary();
    parseScans();
    computeSnip();
}

void BaseLineDialog::keepTemporary()
{
    if (!checkBackend()) return;

    clearAnchorPoints();

    if (temporaryScan) {
        temporaryScan->setTypes(Scan::XY | Scan::BACKGROUND | Scan::SYNTHETIC);
        temporaryScan = nullptr;
        graphControl->updateViews();
    }
}

void BaseLineDialog::clearTemporary()
{
    if (!checkBackend()) return;

    clearAnchorPoints();

    if (temporaryScan) {
        QUuid uid = temporaryScan->uid();
        temporaryScan = nullptr;
        graphModel->removeScan(uid);
        graphControl->updateViews();
    }
}

void BaseLineDialog::clearGui()
{
    bool oldState = ui->comboBoxScan->blockSignals(true);
    ui->comboBoxScan->clear();
    ui->comboBoxScan->blockSignals(oldState);
}

void BaseLineDialog::resetSnip()
{
    blockUpdateSignals(true);
    if (ui->comboBoxSnipWindow->count() > 1) ui->comboBoxSnipWindow->setCurrentIndex(1);
    ui->spinBoxSnipM->setValue(60);
    computeSnip();
    blockUpdateSignals(false);
}

void BaseLineDialog::resetGolot()
{
    blockUpdateSignals(true);
    if (ui->comboBoxGolotLast->count() > 0) ui->comboBoxGolotLast->setCurrentIndex(1);
    ui->spinBoxGolotSmooth->setValue(3);
    ui->spinBoxGolotSteps->setValue(20);
    ui->spinBoxGolotWindow->setValue(10);
    ui->doubleSpinBoxGolotvinN->setValue(1.0);
    ui->comboBoxGolotInterpolation->setCurrentIndex(SPLINE_AKIMA);
    computeGolot();
    blockUpdateSignals(false);
}

void BaseLineDialog::blockUpdateSignals(bool b)
{
    ui->comboBoxScan->blockSignals(b);
    ui->comboBoxAlgo->blockSignals(b);

    ui->comboBoxSnipWindow->blockSignals(b);
    ui->spinBoxSnipM->blockSignals(b);

    ui->comboBoxGolotLast->blockSignals(b);
    ui->spinBoxGolotSmooth->blockSignals(b);
    ui->spinBoxGolotSteps->blockSignals(b);
    ui->spinBoxGolotWindow->blockSignals(b);
    ui->comboBoxGolotInterpolation->blockSignals(b);

    ui->doubleSpinBoxGolotvinN->blockSignals(b);
    ui->tableWidgetNodes->blockSignals(b);
}

void BaseLineDialog::addGolotAnchor()
{
    int i = ui->tableWidgetNodes->currentRow();
    if (i < 0) i = 0;

    QTableWidgetItem *ix = ui->tableWidgetNodes->item(i, 0);
    QTableWidgetItem *iy = ui->tableWidgetNodes->item(i, 1);

    if (ix && iy) {
        QString dx = ix->text();
        QString dy = iy->text();

        ui->tableWidgetNodes->insertRow(i);

        ui->tableWidgetNodes->setItem(i, 0, new QTableWidgetItem(dx));
        ui->tableWidgetNodes->setItem(i, 1, new QTableWidgetItem(dy));

        generateGolot();
    }
}

void BaseLineDialog::removeGolotAnchor()
{
    int i = ui->tableWidgetNodes->currentRow();
    if (i < 0) return;

    ui->tableWidgetNodes->removeRow(i);
    generateGolot();
}

void BaseLineDialog::highlightAnchor(QTableWidgetItem *cur, QTableWidgetItem *)
{
    if (!cur) return;

    int r = cur->row();

    for (int i = 0; i < anchors.count(); ++i) {
        if (i == r) anchors[i].highlight = true;
        else        anchors[i].highlight = false;
    }

    graphView->setAnchorPoints(anchors);
    graphControl->updateViews();
}
