/***************************************************************************
                          browsereferencestructuresdialog.cpp  -  description
                             -------------------
    begin                : Tue 16 21:00:00 CEST 2021
    copyright            : (C) 2021 by Nicola Doebelin
    email                : ndoebelin@gmx.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "browsereferencestructuresdialog.h"
#include "ui_browsereferencestructuresdialog.h"
#include "../../../libXrdIO/functions.h"
#include "projectWidget/syntaxHighlighter/bgmnhighlighter.h"
#include "../../../libXrdIO/bgmnfileio.h"

BrowseReferenceStructuresDialog::BrowseReferenceStructuresDialog(QWidget *parent) :
    QDialog(parent),
    ui(new Ui::BrowseReferenceStructuresDialog)
{
    ui->setupUi(this);
    settings = SettingsManager::getInstance();

    initUi();
    initSettings();
    setStrDir();
}

BrowseReferenceStructuresDialog::~BrowseReferenceStructuresDialog()
{
    saveSettings();
    delete ui;
}


void BrowseReferenceStructuresDialog::initUi()
{
    refStrManager = BgmnRefStructureManager::getInstance();

    ui->labelCurrentFile->clear();
    ui->comboBoxWavelength->showKa2(true);
    ui->comboBoxWavelength->showKb(true);
    ui->comboBoxWavelength->initData();

    QString xa = "Diffraction Angle [" + QString(global::degree) + "2" + QString(global::theta) + "]";
    ui->plotWidget->xAxis->setLabel(xa);
    ui->plotWidget->yAxis->setLabel("Intensity [%]");

    QStringList tabHeader;
    tabHeader << "h" << "k" << "l" << QString("2%1 [°]").arg(global::theta) << "d [nm]" << "Intensity [\%]";
    ui->tableWidgetHkl->setColumnCount(tabHeader.size());
    ui->tableWidgetHkl->setHorizontalHeaderLabels(tabHeader);

    QStringList strHeader;
    strHeader << "File Name" << "Phase" << "Comment";
    ui->treeWidget->setColumnCount(strHeader.size());
    ui->treeWidget->setHeaderLabels(strHeader);

    ui->treeWidget->header()->setContextMenuPolicy(Qt::CustomContextMenu);
    contextMenuHeader = new QMenu(this);

    for (int i = 1; i < ui->treeWidget->header()->count(); ++i) {
        QAction *act = new QAction(ui->treeWidget->headerItem()->text(i), contextMenuHeader);
        act->setCheckable(true);
        act->setData(i);
        contextMenuHeader->addAction(act);
    }

    connect(contextMenuHeader, SIGNAL(triggered(QAction*)), this, SLOT(toggleColumnHidden(QAction*)));
    connect(ui->treeWidget->header(), SIGNAL(customContextMenuRequested(QPoint)), this, SLOT(customMenuRequested(QPoint)));
    connect(ui->treeWidget, SIGNAL(currentItemChanged(QTreeWidgetItem*,QTreeWidgetItem*)), this, SLOT(currentChanged(QTreeWidgetItem*,QTreeWidgetItem*)));
}

void BrowseReferenceStructuresDialog::initSettings()
{
    restoreGeometry(settings->value("refStrBrowser/geometry", QByteArray()).toByteArray());
    QByteArray wdt = settings->value("refStrBrowser/colWidth", QByteArray()).toByteArray();

    if (!wdt.isEmpty()) {
        ui->treeWidget->header()->restoreState(wdt);
    } else {
        QFontMetrics fm(ui->treeWidget->font());
        int w = fm.size(Qt::TextSingleLine, "Hydroxylapatite.str").width();
        ui->treeWidget->header()->resizeSection(0, 2*w);
        ui->treeWidget->header()->resizeSection(1, w);
        ui->treeWidget->header()->resizeSection(2, w/2);
    }

    sortCol = settings->value("refStrBrowser/sortColumn", ui->treeWidget->sortColumn()).toInt();
    sortOrder = (Qt::SortOrder)settings->value("refStrBrowser/sortOrder", 0).toInt();
    ui->treeWidget->sortItems(sortCol, sortOrder);
    ui->splitterVertical->restoreState(settings->value("refStrBrowser/splitterV", QByteArray()).toByteArray());
    ui->splitterHorizontal->restoreState(settings->value("refStrBrowser/splitterH", QByteArray()).toByteArray());
    ui->comboBoxWavelength->setCurrentIndex(settings->value("refStrBrowser/currentWl", 12).toInt());

    QByteArray tabHeader = settings->value("refStrBrowser/tableHeader", QByteArray()).toByteArray();
    ui->tableWidgetHkl->horizontalHeader()->restoreState(tabHeader);

    QList<QVariant> hiddenCols;
    hiddenCols.append(QVariant(2)); // hide the comment column by default
    hiddenCols = settings->value("refStrBrowser/hiddenColumns", hiddenCols).toList();

    int n = qMin(ui->treeWidget->header()->count(), contextMenuHeader->actions().count() + 1);

    for (int i = 1; i < n; ++i) {
        bool b = hiddenCols.contains(QVariant(i));
        ui->treeWidget->setColumnHidden(i, b);
        contextMenuHeader->actions().at(i - 1)->setChecked(!b);
    }

    double wl = ui->comboBoxWavelength->currentData(Qt::UserRole).toDouble();
    double utt = global::Functions::dToTwoTheta(settings->value("config/hklUpperRangeD", wl).toDouble(), wl);
    ui->plotWidget->setXLimits(0.0, utt);
    ui->plotWidget->xAxis->setRange(0.0, utt);
    ui->plotWidget->yAxis->setRange(0.0, 100.0);

    BgmnHighlighter *bhl = new BgmnHighlighter(settings->getSyntaxHighlightingMode(), this);
    bhl->setDocument(ui->plainTextEdit->document());

    QFont edFont;
    edFont.fromString(settings->value("config/editorFont", font().toString()).toString());
    ui->plainTextEdit->setFont(edFont);
}

void BrowseReferenceStructuresDialog::saveSettings()
{
    settings->setValue("refStrBrowser/geometry", saveGeometry());
    settings->setValue("refStrBrowser/sortColumn", ui->treeWidget->sortColumn());
    settings->setValue("refStrBrowser/sortOrder", (int)ui->treeWidget->header()->sortIndicatorOrder());
    settings->setValue("refStrBrowser/colWidth", ui->treeWidget->header()->saveState());
    settings->setValue("refStrBrowser/splitterV", ui->splitterVertical->saveState());
    settings->setValue("refStrBrowser/splitterH", ui->splitterHorizontal->saveState());
    settings->setValue("refStrBrowser/tableHeader", ui->tableWidgetHkl->horizontalHeader()->saveState());
    settings->setValue("refStrBrowser/currentWl", ui->comboBoxWavelength->currentIndex());

    QList<QVariant> hiddenCols;
    for (int i = 1; i < ui->treeWidget->header()->count(); ++i) {
        if (ui->treeWidget->isColumnHidden(i)) hiddenCols.append(QVariant(i));
    }

    settings->setValue("refStrBrowser/hiddenColumns", hiddenCols);
}

void BrowseReferenceStructuresDialog::customMenuRequested(QPoint p)
{
    contextMenuHeader->popup(ui->treeWidget->header()->viewport()->mapToGlobal(p));
}

void BrowseReferenceStructuresDialog::setStrDir()
{
    QStringList repos = settings->value("bgmnProject/structureDatabase", QStringList()).toStringList();
    refStrManager->createTree(ui->treeWidget, repos, false, Qt::ItemFlags(Qt::NoItemFlags | Qt::ItemIsEnabled), Qt::ItemFlags(Qt::NoItemFlags | Qt::ItemIsEnabled));

    ui->treeWidget->sortItems(0, Qt::AscendingOrder);

    for (int i = 0; i < ui->treeWidget->topLevelItemCount(); ++i) {
        ui->treeWidget->topLevelItem(i)->setExpanded(true);
    }

    for (int i = 0; i < ui->treeWidget->columnCount(); ++i) {
        if (ui->treeWidget->columnWidth(i) == 0) {
            ui->treeWidget->resizeColumnToContents(i);
        }
    }
}

void BrowseReferenceStructuresDialog::currentChanged(QTreeWidgetItem *cur, QTreeWidgetItem *)
{
    clearDataWidgets();

    if (!cur) {
        ui->plotWidget->replot();
        return;
    }

    currentFile = cur->data(0, Qt::UserRole).toString();
    updateDataWidgets();
}

void BrowseReferenceStructuresDialog::clearDataWidgets()
{
    ui->labelCurrentFile->clear();

    ui->tableWidgetHkl->clearContents();
    ui->tableWidgetHkl->setRowCount(0);

    ui->plainTextEdit->clear();

    ui->plotWidget->clearGraphs();
    ui->plotWidget->clearPlottables();
}

void BrowseReferenceStructuresDialog::updateDataWidgets()
{
    QVector<Hkl> vec = refStrManager->getReferenceLines(currentFile);
    double maxI = plotHkl(vec);
    updateTable(vec, maxI);
    updateStrEditor();
    ui->labelCurrentFile->setText(currentFile);
}

double BrowseReferenceStructuresDialog::plotHkl(const QVector<Hkl> &vec)
{
    double maxI = 0.0;
    QVector<double> xdata(hklToTwoTheta(vec, ui->comboBoxWavelength->currentData(Qt::UserRole).toDouble()));
    QVector<double> ydata(hklToIntensity(vec, maxI));

    QCPBars *hklBars = new QCPBars(ui->plotWidget->xAxis, ui->plotWidget->yAxis);
    hklBars->setData(xdata, ydata);
    hklBars->setWidthType(QCPBars::wtAbsolute);
    hklBars->setWidth(0.0);

    ui->plotWidget->replot();
    return maxI;
}

void BrowseReferenceStructuresDialog::updateTable(const QVector<Hkl> &vec, double maxI)
{
    double my = qFuzzyIsNull(maxI) ? 100.0 : maxI;
    double wl = ui->comboBoxWavelength->currentData(Qt::UserRole).toDouble();

    for (int i = 0; i < vec.size(); ++i) {
        double d = vec.at(i).position();
        double tt = global::Functions::dToTwoTheta(d, wl);
        double intens = 100 * vec.at(i).intensity() / my;
        QStringList hkl = vec.at(i).hkl().split(QRegularExpression("\\s+"));

        int r = ui->tableWidgetHkl->rowCount();
        ui->tableWidgetHkl->setRowCount(r + 1);
        ui->tableWidgetHkl->setItem(r, 0, new QTableWidgetItem(hkl.size() > 0 ? hkl.at(0) : QString()));
        ui->tableWidgetHkl->setItem(r, 1, new QTableWidgetItem(hkl.size() > 1 ? hkl.at(1) : QString()));
        ui->tableWidgetHkl->setItem(r, 2, new QTableWidgetItem(hkl.size() > 2 ? hkl.at(2) : QString()));
        ui->tableWidgetHkl->setItem(r, 3, new QTableWidgetItem(QString("%1").arg(tt, 0, 'f', 5)));
        ui->tableWidgetHkl->setItem(r, 4, new QTableWidgetItem(QString("%1").arg(d, 0, 'f', 6)));
        ui->tableWidgetHkl->setItem(r, 5, new QTableWidgetItem(QString("%1").arg(intens, 0, 'f', 2)));
    }
}

void BrowseReferenceStructuresDialog::updateStrEditor()
{
    ui->plainTextEdit->setPlainText(BgmnFileIO::readTextFile(currentFile));
}

void BrowseReferenceStructuresDialog::toggleColumnHidden(QAction *a)
{
    int c = a->data().toInt();
    ui->treeWidget->setColumnHidden(c, !a->isChecked());
}

QVector<double> BrowseReferenceStructuresDialog::hklToTwoTheta(const QVector<Hkl> &vec, double wl)
{
    QVector<double> ttvec;

    for (int i = 0; i < vec.size(); ++i) {
        ttvec.append(global::Functions::dToTwoTheta(vec.at(i).position(), wl));
    }

    return ttvec;
}

QVector<double> BrowseReferenceStructuresDialog::hklToIntensity(const QVector<Hkl> &vec, double &maxI)
{
    QVector<double> ivec;
    double imax = 0.0;

    for (int i = 0; i < vec.size(); ++i) {
        imax = qMax(imax, vec.at(i).intensity());
    }

    for (int i = 0; i < vec.size(); ++i) {
        ivec.append(100.0 * vec.at(i).intensity() / imax);
    }

    maxI = imax;
    return ivec;
}


/*
 * shows / hides items according to the filter string. does not add / remove any items
 */
void BrowseReferenceStructuresDialog::applyFilter(const QString &s)
{
    if (s.simplified().isEmpty()) {
        unhideAllItems();
        return;
    }

    QRegularExpression rx;
    QRegularExpressionMatch rm;

    QRegularExpression::PatternOptions options = rx.patternOptions();

//    if (!actionFilterCaseSensitive->isChecked()) {
        rx.setPatternOptions(options | QRegularExpression::CaseInsensitiveOption);
//    }

//    if (actionFilterRegExp->isChecked()) {
        // if regexp option is checked, we will interpret the string exactly as entered
//        rx.setPattern(s);
//    } else {
        // if the filter string doesn't contain wildcards, we wrap it in
        // wildcards, so the search string is also matched in the middle
        // of words. This feels more natural from the user perspective.
        QString filter = s.simplified();

        if (!filter.contains(QString("*"))) {
            filter = QString("*%1*").arg(filter);
        }

        filter.replace(QString("*"), QString(".*"));
        filter.replace(QString("?"), QString("."));
        rx.setPattern(QString("^%1$").arg(filter));
//    }

    if (!rx.isValid()) {
        qDebug() << QString("BgmnAddRemoveDialog::applyFilter: Invalid regular expression (%1)").arg(rx.errorString());
        return;
    }

    QString txt;

    QList<QTreeWidgetItem *> items = allStructureItems();

    for (int i = 0; i < items.size(); ++i) {
        int matches = 0;

        txt = itemText(items.at(i), 0);
        rm = rx.match(txt);
        if (rm.hasMatch()) matches++;

        txt = itemText(items.at(i), 1);
        rm = rx.match(txt);
        if (rm.hasMatch()) matches++;

        txt = itemText(items.at(i), 2);
        rm = rx.match(txt);
        if (rm.hasMatch()) matches++;

        items.at(i)->setHidden(matches == 0 ? true : false);
    }
}

void BrowseReferenceStructuresDialog::updateFilter()
{
    applyFilter(ui->lineEditFilter->text());
}

void BrowseReferenceStructuresDialog::unhideAllItems()
{
    QList<QTreeWidgetItem *> items = allStructureItems();

    for (int i = 0; i < items.size(); ++i) {
        if (items.at(i)->isHidden()) items.at(i)->setHidden(false);
    }
}

/*
 * returns a list of pointers to all file items in the structure treewidget.
 */
QList<QTreeWidgetItem *> BrowseReferenceStructuresDialog::allStructureItems()
{
    QList<QTreeWidgetItem *> items;
    QTreeWidgetItemIterator it(ui->treeWidget);

    while (*it) {
        if ((*it)->type() == TYPE_FILE) {
            items.append(*it);
        }
        ++it;
    }

    return items;
}

/*
 * returns the column text in column n of item it. If it.columnCount() is
 * < n, returns an empty string. not sure if QTreeWidgetItem::text(int) does the
 * same, but better safe than sorry.
 */
QString BrowseReferenceStructuresDialog::itemText(const QTreeWidgetItem *it, int n)
{
    int c = it->columnCount();
    if ((n >= c) || (n < 0)) return QString();
    return it->text(n);
}

void BrowseReferenceStructuresDialog::wavelengthChanged(int)
{
    clearDataWidgets();
    updateDataWidgets();
}
