/***************************************************************************
                          lutscale2d.h  -  description
                             -------------------
    begin                : Wed Jul 28 20:35:00 CEST 2021
    copyright            : (C) 2021 by Nicola Doebelin
    email                : ndoebelin@gmx.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "lutscale2d.h"
#include "math.h"
#include "../libXrdIO/functions.h"
#include <QPainter>
#include <QSet>
#include <QHash>
#include <QDebug>

LutScale2D::LutScale2D(QWidget *parent) :
    QWidget(parent)
{
    setMinimumWidth(fontMetrics().width("aaa -mmmmm"));
    _minE = 0.0;
    _maxE = 1.0;
    _scale = 0;
}

void LutScale2D::setRange(float min, float max)
{
    _maxE = max;
    _minE = min;
    updateTicks();
    update();
}

void LutScale2D::setLut(const colorMaps::Lut &l)
{
    _lut = l;
    updateTicks();
    updateGradient();
    update();
}

void LutScale2D::setColorTemp(int i)
{
    _colTemp = i;
    updateTicks();
    updateGradient();
    update();
}

void LutScale2D::setScaling(int i)
{
    _scale = i;
    updateTicks();
    updateGradient();
    update();
}

void LutScale2D::updateAll()
{
    updateTicks();
    updateGradient();
    update();
}

void LutScale2D::paintEvent(QPaintEvent *)
{
    QPainter painter(this);

    if (gradient.isNull()) {
        painter.fillRect(this->rect(), Qt::white);
        return;
    }

    painter.drawImage(0, 0, gradient);

    if      (_scale == 0) drawTicksLin(painter);
    else if (_scale == 1) drawTicksSqrt(painter);
    else if (_scale == 2) drawTicksLog(painter);
}

void LutScale2D::resizeEvent(QResizeEvent *)
{
    updateGradient();
}

void LutScale2D::updateGradient()
{
    if ((width() <= 0) || (height() <= 0)) return;

    float f = float(height()) / 256.0;
    int w = fontMetrics().width("aaa");
    QImage img(w, height(), QImage::Format_ARGB32);
    QPainter pxpainter(&img);

    int _lutSign = 1;
    int _lutOffset = 255;

    if (_lut.data.size() >= 512) {
        if (_colTemp > 0) _lutSign = -1;
    } else {
        if (_colTemp > 0) _lutSign = -1;
        else              _lutOffset = 0;
    }

    for (int i = 0; i < qMin(256, _lut.data.size()); ++i) {
        int n = _lutOffset + _lutSign * i;
        pxpainter.fillRect(QRectF(0.0, float(height()) - (i * f), float(w), -f), QBrush(QColor(_lut.data.at(n))));
    }

    pxpainter.end();
    gradient = img;
}

void LutScale2D::updateTicks()
{
    _minTicks = global::Functions::scaleAxis1(_minE, _maxE, 60);
    _majTicks = global::Functions::scaleAxis1(_minE, _maxE, 10);
}

void LutScale2D::drawTicksLin(QPainter &p)
{
    if (abs(_minE - _maxE) <= 1.0) return;

    QSet<int> minPos;
    QHash<int, QString> majPos;

    for (int i = 0; i < _minTicks.size(); ++i) {
        minPos.insert(int(0.5 + gradient.height() * (1.0 - (_minTicks.at(i) - _minE)/(_maxE - _minE))));
    }

    for (int i = 0; i < _majTicks.size(); ++i) {
       int y = int(0.5 + gradient.height() * (1.0 - (_majTicks.at(i) - _minE)/(_maxE - _minE)));
       QString s = QString("%1").arg(_majTicks.at(i), 6, 'f', 0);
       majPos.insert(y, s);
    }

    drawTicks(p, minPos, majPos);
}

void LutScale2D::drawTicksSqrt(QPainter &p)
{
    if (abs(_minE - _maxE) <= 1.0) return;

    QSet<int> minPos;
    QHash<int, QString> majPos;

    for (int i = 0; i < _minTicks.size(); ++i) {
        minPos.insert(int(0.5 + gradient.height() * (1.0 - (sqrt(_minTicks.at(i)) - sqrt(_minE))/(sqrt(_maxE) - sqrt(_minE)))));
    }

    for (int i = 0; i < _majTicks.size(); ++i) {
       int y = int(0.5 + gradient.height() * (1.0 - (sqrt(_majTicks.at(i)) - sqrt(_minE))/(sqrt(_maxE) - sqrt(_minE))));
       QString s = QString("%1").arg(_majTicks.at(i), 6, 'f', 0);
       majPos.insert(y, s);
    }

    drawTicks(p, minPos, majPos);
}

void LutScale2D::drawTicksLog(QPainter &p)
{
    if (abs(_minE - _maxE) <= 1.0) return;

    QSet<int> minPos;
    QHash<int, QString> majPos;

    double miE = _minE < 1.0 ? 0.0 : log10(_minE);
    double maE = _maxE < 1.0 ? 0.0 : log10(_maxE);

    for (int i = 0; i < _minTicks.size(); ++i) {
        double ty = _minTicks.at(i) < 1.0 ? 0.0 : log10(_minTicks.at(i));
        minPos.insert(int(0.5 + gradient.height() * (1.0 - (ty - miE)/(maE - miE))));
    }

    for (int i = 0; i < _majTicks.size(); ++i) {
        double ty = _majTicks.at(i) < 1.0 ? 0.0 : log10(_majTicks.at(i));
        QString s = QString("%1").arg(_majTicks.at(i), 6, 'f', 0);
        majPos.insert(int(0.5 + gradient.height() * (1.0 - (ty - miE)/(maE - miE))), s);
    }

    drawTicks(p, minPos, majPos);
}

void LutScale2D::drawTicks(QPainter &p, const QSet<int> &minPos, const QHash<int, QString> &majPos)
{
    int tm = fontMetrics().width("a");
    int x = gradient.width();
    QSetIterator<int> minIt(minPos);
    QHashIterator<int, QString> majIt(majPos);

    while (minIt.hasNext()) {
        int y = minIt.next();
        p.drawLine(x, y, x + tm, y);
    }

    while (majIt.hasNext()) {
        majIt.next();
        int y = majIt.key();
        p.drawLine(x, y, x + 2*tm, y);

        y += fontMetrics().ascent() / 2;
        if (y > gradient.height())           p.drawText(x + 3*tm, gradient.height(), majIt.value());
        else if (y > fontMetrics().ascent()) p.drawText(x + 3*tm, y, majIt.value());
    }
}
