/***************************************************************************
                          waterfallplotdialog.cpp  -  description
                             -------------------
    begin                : Mon Jul 26 18:26:00 CEST 2021
    copyright            : (C) 2021 by Nicola Doebelin
    email                : ndoebelin@gmx.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "ui_waterfallplotdialog.h"
#include "waterfallplotdialog.h"
#include "../libXrdIO/import/importhandler.h"
#include <QFileDialog>
#include <QDir>
#include <QFile>
#include <QVector>
#include <QDateTime>
#include <QMenu>

WaterfallPlotDialog::WaterfallPlotDialog(QWidget *parent) :
    QDialog(parent),
    ui(new Ui::WaterfallPlotDialog)
{
    ui->setupUi(this);
    settings = SettingsManager::getInstance();
    ui->plotWidget->setScanData(&_scanHeap);
    ui->plotWidget->setUidData(&_uidList);

    ui->treeWidgetScans->setColumnCount(2);
    ui->treeWidgetScans->setHeaderLabels(QStringList() << "Scan" << "Label");
    ui->treeWidgetScans->setContextMenuPolicy(Qt::CustomContextMenu);

    for (int i = 0; i < ui->treeWidgetScans->columnCount(); ++i) {
        ui->comboBoxSortColumn->addItem(ui->treeWidgetScans->headerItem()->text(i));
    }

    ui->comboBoxSortOrder->addItem(tr("Ascending"));
    ui->comboBoxSortOrder->addItem(tr("Descending"));

    ui->labelStatus->setText(QString());

    ui->comboBoxYScale->addItem(tr("Linear"));
    ui->comboBoxYScale->addItem(tr("Square root"));
    ui->comboBoxYScale->addItem(tr("Logarithm"));

    ui->comboBoxColorTemp->addItem(tr("Warm"));
    ui->comboBoxColorTemp->addItem(tr("Cold"));

    ui->spinBoxYrangeMin->setValue(0);
    ui->spinBoxYrangeMax->setValue(1);

    prevSelectionIdx = -1;

    LutGenerator lutGen(256, false);
    lutList = lutGen.getLuts();

    for (int i = 0; i < qMin(lutList.keys().size(), 5); ++i) {
        ui->comboBoxColorScale->addItem(lutList.keys().at(i));
    }

    ImportHandler iHandler;
    QMap<QString, GenericImport *> fileTypes = iHandler.importers();

    QMapIterator<QString, GenericImport *> iter(fileTypes);
    while (iter.hasNext()) {
        iter.next();
        ui->comboBoxFileTypes->addItem(iter.value()->filter(), iter.value()->extensions().first());
    }

    twContextMenu = new QMenu(this);
    twContextMenu->addAction(tr("Check all"), this, SLOT(treeWidgetCheckAll()));
    twContextMenu->addAction(tr("Uncheck all"), this, SLOT(treeWidgetUncheckAll()));
    twContextMenu->addSeparator();
    twContextMenu->addAction(tr("Check all above"), this, SLOT(treeWidgetCheckAllAbove()));
    twContextMenu->addAction(tr("Check all below"), this, SLOT(treeWidgetCheckAllBelow()));
    twContextMenu->addSeparator();
    twContextMenu->addAction(tr("Uncheck all above"), this, SLOT(treeWidgetUncheckAllAbove()));
    twContextMenu->addAction(tr("Uncheck all below"), this, SLOT(treeWidgetUncheckAllBelow()));

    initSettings();

    connect(ui->plotWidget, SIGNAL(sigMouseCoordinates(QString)), this, SLOT(mouseCoordinates(QString)));
    connect(ui->plotWidget, SIGNAL(sigXrangeWasZoomed()), this, SLOT(updateXRangeSpinboxes()));
    connect(ui->comboBoxYScale, SIGNAL(currentIndexChanged(int)), this, SLOT(yScaleChanged()));
    connect(ui->comboBoxColorScale, SIGNAL(currentIndexChanged(int)), this, SLOT(lutChanged()));
    connect(ui->comboBoxColorTemp, SIGNAL(currentIndexChanged(int)), this, SLOT(colorTempChanged()));
    connect(ui->checkBoxSmooth, SIGNAL(toggled(bool)), ui->plotWidget, SLOT(setSmoothInterpolation(bool)));
    connect(ui->checkBoxGridHorizontal, SIGNAL(toggled(bool)), ui->plotWidget, SLOT(setDrawGridHorizontal(bool)));
    connect(ui->checkBoxGridVertical, SIGNAL(toggled(bool)), ui->plotWidget, SLOT(setDrawGridVertical(bool)));
    connect(ui->lineEdit, SIGNAL(textChanged(QString)), this, SLOT(changeMonnitoringEnabled(QString)));
    connect(ui->toolButtonRegExpUpdate, SIGNAL(clicked()), this, SLOT(updateLabels()));
    connect(ui->treeWidgetScans, SIGNAL(itemChanged(QTreeWidgetItem *, int)), this, SLOT(updateScene()));
    connect(&fsWatcher, SIGNAL(directoryChanged(const QString &)), this, SLOT(dirContentChanged(const QString &)));
    connect(ui->treeWidgetScans, SIGNAL(itemSelectionChanged()), this, SLOT(itemSelectionChanged()));
    connect(ui->treeWidgetScans, SIGNAL(itemClicked(QTreeWidgetItem *, int)), this, SLOT(itemClicked(QTreeWidgetItem *, int)));
    connect(ui->treeWidgetScans, SIGNAL(customContextMenuRequested(QPoint)), this, SLOT(showContextMenu(QPoint)));
}

WaterfallPlotDialog::~WaterfallPlotDialog()
{
    delete ui;
}

void WaterfallPlotDialog::close()
{
    if (ui->toolButtonMonitor->isChecked()) {
        ui->toolButtonMonitor->setChecked(false);
        monitorDir(false);
    }

    saveSettings();
    QDialog::close();
}

void WaterfallPlotDialog::initSettings()
{
    restoreGeometry(settings->value("waterfallPlotDialog/geometry", QByteArray()).toByteArray());

    QByteArray splArr = settings->value("waterfallPlotDialog/splitter", QByteArray()).toByteArray();

    if (splArr.isEmpty()) {
        QList<int> l = QList<int>() << width() / 4 << 3 * width() / 4;
        ui->splitter->setSizes(l);
    } else {
        ui->splitter->restoreState(splArr);
    }

    QByteArray splHArr = settings->value("waterfallPlotDialog/splitterH", QByteArray()).toByteArray();

    if (splHArr.isEmpty()) {
        QList<int> l = QList<int>() << height() / 3 << 2 * height() / 3;
        ui->splitterH->setSizes(l);
    } else {
        ui->splitterH->restoreState(splHArr);
    }

    ui->comboBoxColorScale->setCurrentIndex(settings->value("waterfallPlotDialog/lut", 0).toInt());

    int sc = settings->value("waterfallPlotDialog/yAxisScaling", 0).toInt();
    ui->comboBoxYScale->setCurrentIndex(sc);
    ui->comboBoxColorTemp->setCurrentIndex(settings->value("waterfallPlotDialog/colorTemp", 0).toInt());
    ui->checkBoxSmooth->setChecked(settings->value("waterfallPlotDialog/smoothInterpolation", false).toBool());
    ui->checkBoxGridHorizontal->setChecked(settings->value("waterfallPlotDialog/drawGridH", false).toBool());
    ui->checkBoxGridVertical->setChecked(settings->value("waterfallPlotDialog/drawGridV", false).toBool());
    ui->treeWidgetScans->header()->restoreState(settings->value("waterfallPlotDialog/treeHeader", QByteArray()).toByteArray());
    ui->comboBoxSortColumn->setCurrentIndex(settings->value("waterfallPlotDialog/sortColumn", 0).toInt());
    ui->comboBoxSortOrder->setCurrentIndex(settings->value("waterfallPlotDialog/sortOrder", 0).toInt());
    ui->comboBoxFileTypes->setCurrentIndex(settings->value("waterfallPlotDialog/monitorFileType", 0).toInt());
    ui->lineEdit->setText(settings->value("waterfallPlotDialog/monitorDir", workingDir).toString());
    ui->lineEditRegExp->setText(settings->value("waterfallPlotDialog/labelRegEx", QString()).toString());
    workingDir = settings->value("config/workingdir", QDir::homePath()).toString();

    ui->plotWidget->setDrawGridHorizontal(settings->value("waterfallPlotDialog/drawGridH", false).toBool());
    ui->plotWidget->setDrawGridVertical(settings->value("waterfallPlotDialog/drawGridV", false).toBool());
    ui->plotWidget->setSmoothInterpolation(settings->value("waterfallPlotDialog/smoothInterpolation", false).toBool());

    ui->lutScaleWidget->setLut(lutList.value(ui->comboBoxColorScale->currentText()));
    ui->lutScaleWidget->setColorTemp(ui->comboBoxColorTemp->currentIndex());
    ui->lutScaleWidget->setScaling(sc);
}

void WaterfallPlotDialog::saveSettings()
{
    settings->setValue("waterfallPlotDialog/geometry", saveGeometry());
    settings->setValue("waterfallPlotDialog/splitter", ui->splitter->saveState());
    settings->setValue("waterfallPlotDialog/splitterH", ui->splitterH->saveState());
    settings->setValue("waterfallPlotDialog/treeHeader", ui->treeWidgetScans->header()->saveState());
    settings->setValue("waterfallPlotDialog/yAxisScaling", ui->comboBoxYScale->currentIndex());
    settings->setValue("waterfallPlotDialog/lut", ui->comboBoxColorScale->currentIndex());
    settings->setValue("waterfallPlotDialog/colorTemp", ui->comboBoxColorTemp->currentIndex());
    settings->setValue("waterfallPlotDialog/smoothInterpolation", ui->checkBoxSmooth->isChecked());
    settings->setValue("waterfallPlotDialog/drawGridH", ui->checkBoxGridHorizontal->isChecked());
    settings->setValue("waterfallPlotDialog/drawGridV", ui->checkBoxGridVertical->isChecked());
    settings->setValue("waterfallPlotDialog/monitorFileType", ui->comboBoxFileTypes->currentIndex());
    settings->setValue("waterfallPlotDialog/sortColumn", ui->comboBoxSortColumn->currentIndex());
    settings->setValue("waterfallPlotDialog/sortOrder", ui->comboBoxSortOrder->currentIndex());
    settings->setValue("waterfallPlotDialog/labelRegEx", ui->lineEditRegExp->text());
    if (!ui->lineEdit->text().isEmpty()) settings->setValue("waterfallPlotDialog/monitorDir", ui->lineEdit->text());
    settings->setValue("config/workingdir", workingDir);
}

void WaterfallPlotDialog::addFile()
{
    ImportHandler importHandler;
    QStringList filters;
    filters = importHandler.filters();
    filters << tr("All Files (*.*)");

    QFileDialog fdlg(this, tr("Add Scan File"), workingDir);
    fdlg.setAcceptMode(QFileDialog::AcceptOpen);
    fdlg.setNameFilters(filters);
    fdlg.setFileMode(QFileDialog::ExistingFiles);
    fdlg.setOption(QFileDialog::HideNameFilterDetails, false);

    if (curFilter.isEmpty()) curFilter = settings->value("config/lastOpenGraphFilter", "").toString();
    if (!curFilter.isEmpty()) fdlg.selectNameFilter(curFilter);

    if (fdlg.exec()) {
        QStringList fn = fdlg.selectedFiles();
        curFilter = fdlg.selectedNameFilter();

        if (fn.size()) {
            QFileInfo fi(fn.first());
            workingDir = fi.absolutePath();
        }

        for (int i = 0; i < fn.size(); ++i) {
            appendFileScans(fn.at(i), curFilter);
        }

        updateLabels();
        sortItems();
        updateScene();
    }
}

void WaterfallPlotDialog::removeFile()
{
    int n = ui->treeWidgetScans->indexOfTopLevelItem(ui->treeWidgetScans->currentItem());
    if (n < 0) return;
    removeFile(n);
}

void WaterfallPlotDialog::removeFile(int n)
{
    QTreeWidgetItem *it = ui->treeWidgetScans->takeTopLevelItem(n);
    if (!it) return;

    QUuid uid = it->data(0, Qt::UserRole).toUuid();
    _scanHeap.remove(uid);
    _fileNames.remove(uid);

    delete it;

    if (ui->treeWidgetScans->topLevelItemCount()) {
        updateScene();
    } else {
        ui->plotWidget->clearScans();
    }
}

void WaterfallPlotDialog::clearScans()
{
    ui->treeWidgetScans->clear();
    _scanHeap.clear();
    _fileNames.clear();
    _uidList.clear();
    ui->plotWidget->clearScans();
}

void WaterfallPlotDialog::saveAs()
{
    QString f = QFileDialog::getSaveFileName(this, tr("Save graph"), workingDir,
                                             QString("Pixel image (*.png *.PNG);;"
                                                     "Scalable vector graphics (*.svg *.SVG);;"
                                                     "Surfer 6 text grid (*.grd *.GRD)"));

    if (!f.isEmpty()) {
        int rasterW = settings->value("graph/rasterResolutionWidth", 1536).toInt();
        int rasterH = settings->value("graph/rasterResolutionHeight", 1024).toInt();
        ui->plotWidget->save(f, rasterW, rasterH);
    }
}

void WaterfallPlotDialog::appendFileScans(const QString &file, const QString &filter)
{
    QFileInfo fi(file);
    if (!fi.exists()) return;

    QVector<Scan> *sh = new QVector<Scan>();

    ImportHandler iHandler;
    QString uid = filter.isNull() ? iHandler.uidByFileName(fi.absoluteFilePath()) : iHandler.uidByFilter(filter);
    int n = iHandler.load(fi.absoluteFilePath(), uid, *sh, true);

    for (int i = 0; i < n; ++i) {
        QUuid uid = QUuid::createUuid();

        _scanHeap.insert(uid, sh->at(i));
        _fileNames.insert(uid, fi.absoluteFilePath());

        QTreeWidgetItem *tlIt = new QTreeWidgetItem(QStringList() << sh->at(i).name() << getLabel(sh->at(i).name()));
        tlIt->setData(0, Qt::UserRole, uid);
        tlIt->setCheckState(0, Qt::Checked);
        ui->treeWidgetScans->addTopLevelItem(tlIt);
    }

    delete sh;
}

void WaterfallPlotDialog::updateScene()
{
    updateCheckedItemList();

    ui->plotWidget->blockUpdate(true);
    ui->plotWidget->setYscale(ui->comboBoxYScale->currentIndex());
    ui->plotWidget->setLut(lutList.value(ui->comboBoxColorScale->currentText()));
    ui->plotWidget->setColorTemp(ui->comboBoxColorTemp->currentIndex());
    ui->plotWidget->updateScanData();
    ui->plotWidget->blockUpdate(false);
    ui->plotWidget->updateAll();
    ui->lutScaleWidget->updateAll();
    updateXRangeSpinboxes();
    updateYRangeSpinboxes();
}

void WaterfallPlotDialog::updateCheckedItemList()
{
    _uidList.clear();

    for (int i = 0; i < ui->treeWidgetScans->topLevelItemCount(); ++i) {
        QTreeWidgetItem *itm = ui->treeWidgetScans->topLevelItem(i);

        if (itm->checkState(0) == Qt::Unchecked) continue;

        QUuid uid = itm->data(0, Qt::UserRole).toUuid();
        if (_scanHeap.contains(uid)) _uidList.append(uid);
    }
}

void WaterfallPlotDialog::lutChanged()
{
    ui->lutScaleWidget->setLut(lutList.value(ui->comboBoxColorScale->currentText()));
    ui->plotWidget->setLut(lutList.value(ui->comboBoxColorScale->currentText()));
}

void WaterfallPlotDialog::colorTempChanged()
{
    ui->lutScaleWidget->setColorTemp(ui->comboBoxColorTemp->currentIndex());
    ui->plotWidget->setColorTemp(ui->comboBoxColorTemp->currentIndex());
}

void WaterfallPlotDialog::yScaleChanged()
{
    ui->plotWidget->setYscale(ui->comboBoxYScale->currentIndex());
    ui->lutScaleWidget->setScaling(ui->comboBoxYScale->currentIndex());
}

void WaterfallPlotDialog::updateXRangeSpinboxes()
{
    bool oldStateMin = ui->doubleSpinBoxXrangeMin->blockSignals(true);
    bool oldStateMax = ui->doubleSpinBoxXrangeMax->blockSignals(true);

    double min, lmin;
    double max, lmax;

    ui->plotWidget->getXrange(min, max);
    ui->plotWidget->getXlimits(lmin, lmax);

    ui->doubleSpinBoxXrangeMin->setValue(min);
    ui->doubleSpinBoxXrangeMax->setValue(max);

    ui->doubleSpinBoxXrangeMin->setMinimum(lmin);
    ui->doubleSpinBoxXrangeMin->setMaximum(max);
    ui->doubleSpinBoxXrangeMax->setMinimum(min);
    ui->doubleSpinBoxXrangeMax->setMaximum(lmax);

    ui->doubleSpinBoxXrangeMin->blockSignals(oldStateMin);
    ui->doubleSpinBoxXrangeMax->blockSignals(oldStateMax);
}

void WaterfallPlotDialog::updateYRangeSpinboxes()
{
    bool oldStateMin = ui->spinBoxYrangeMin->blockSignals(true);
    bool oldStateMax = ui->spinBoxYrangeMax->blockSignals(true);

    double min, lmin;
    double max, lmax;

    ui->plotWidget->getYrange(min, max);
    ui->plotWidget->getYlimits(lmin, lmax);

    ui->lutScaleWidget->setRange(min, max);

    ui->spinBoxYrangeMin->setValue(int(min));
    ui->spinBoxYrangeMax->setValue(int(max + 1.0));

    ui->spinBoxYrangeMin->setMinimum(0);
    ui->spinBoxYrangeMin->setMaximum(int(max));
    ui->spinBoxYrangeMax->setMinimum(int(min));
    ui->spinBoxYrangeMax->setMaximum(int(lmax + 1.0));

    ui->spinBoxYrangeMin->blockSignals(oldStateMin);
    ui->spinBoxYrangeMax->blockSignals(oldStateMax);
}

void WaterfallPlotDialog::xRangeChanged()
{
    ui->plotWidget->setXrange(ui->doubleSpinBoxXrangeMin->value(), ui->doubleSpinBoxXrangeMax->value());

    ui->doubleSpinBoxXrangeMin->setMaximum(ui->doubleSpinBoxXrangeMax->value() - 0.1);
    ui->doubleSpinBoxXrangeMax->setMinimum(ui->doubleSpinBoxXrangeMin->value() + 0.1);
}

void WaterfallPlotDialog::yRangeChanged()
{
    ui->lutScaleWidget->setRange(double(ui->spinBoxYrangeMin->value()), double(ui->spinBoxYrangeMax->value()));
    ui->plotWidget->setYrange(double(ui->spinBoxYrangeMin->value()), double(ui->spinBoxYrangeMax->value()));

    ui->spinBoxYrangeMin->setMaximum(ui->spinBoxYrangeMax->value() - 1);
    ui->spinBoxYrangeMax->setMinimum(ui->spinBoxYrangeMin->value() + 1);
}

void WaterfallPlotDialog::mouseCoordinates(QString s)
{
    ui->labelStatus->setText(s);
}

void WaterfallPlotDialog::resetX()
{
    ui->plotWidget->resetXrange();
    updateXRangeSpinboxes();
}

void WaterfallPlotDialog::resetY()
{
    ui->plotWidget->resetYrange();
    updateYRangeSpinboxes();
}

void WaterfallPlotDialog::selectDir()
{
    QString s = QFileDialog::getExistingDirectory(this, tr("Open directory"), workingDir);
    if (s.isEmpty()) return;
    workingDir = s;
    ui->lineEdit->setText(workingDir);
}

void WaterfallPlotDialog::monitorDir(bool b)
{
    if (b) {
        QDir dir(ui->lineEdit->text());

        if (dir.exists()) {
            fsWatcher.addPath(ui->lineEdit->text());
            dirContentChanged(ui->lineEdit->text());
            qDebug() << QString("WaterfallPlotDialog::monitorDir(): FileSystemWatcher started at %1").arg(ui->lineEdit->text());
        } else {
            qDebug() << QString("WaterfallPlotDialog::monitorDir(): Cannot start FileSystemWatcher at %1").arg(ui->lineEdit->text());
            qDebug() << QString("                                   Directory does not exist");
        }
    } else {
        fsWatcher.removePaths(fsWatcher.directories());
        qDebug() << QString("WaterfallPlotDialog::monitorDir(): FileSystemWatcher stopped");
    }
}

void WaterfallPlotDialog::dirContentChanged(const QString &s)
{
    bool doUpdate = false;
    QDir dir(s);
    QStringList files = dir.entryList(QStringList(), QDir::Files | QDir::Readable);

    for (int i = 0; i < files.size(); ++i) {
        QFileInfo fi(s + QDir::separator() + files.at(i));

        if (fi.suffix().toLower() != ui->comboBoxFileTypes->currentData(Qt::UserRole).toString().toLower()) {
            continue;
        }

        if (!_fileNames.values().contains(fi.absoluteFilePath())) {
            appendFileScans(fi.absoluteFilePath(), QString());
            doUpdate = true;
        }
    }

    removeNonexistentFiles();

    if (doUpdate) {
        sortItems();
        updateScene();
    }
}

void WaterfallPlotDialog::removeNonexistentFiles()
{
    ui->plotWidget->blockUpdate(true);

    for (int i = ui->treeWidgetScans->topLevelItemCount() - 1; i >= 0; --i) {
        QUuid uid = ui->treeWidgetScans->topLevelItem(i)->data(0, Qt::UserRole).toUuid();

        if (!QFile::exists(_fileNames.value(uid))) {
            removeFile(i);
        }
    }

    ui->plotWidget->blockUpdate(false);
    updateScene();
}

void WaterfallPlotDialog::changeMonnitoringEnabled(QString s)
{
    QFileInfo fi(s);

    ui->comboBoxFileTypes->setEnabled(fi.exists());
    ui->toolButtonMonitor->setEnabled(fi.exists());
}

void WaterfallPlotDialog::updateLabels()
{
    bool oldState = ui->treeWidgetScans->blockSignals(true);

    for (int i = 0; i < ui->treeWidgetScans->topLevelItemCount(); ++i) {
        QTreeWidgetItem *it = ui->treeWidgetScans->topLevelItem(i);        

        if (it) {
            QUuid uid = it->data(0, Qt::UserRole).toUuid();
            QString lbl = getLabel(it->text(0));

            it->setText(1, lbl);
            _scanHeap[uid].setAuxInfo("waterfallPlotLabel", lbl);
        }
    }

    ui->treeWidgetScans->blockSignals(oldState);
    updateScene();
}

QString WaterfallPlotDialog::getLabel(const QString &s)
{
    QString pattern(ui->lineEditRegExp->text());
    if (pattern.isEmpty()) return s;

    QRegularExpression rx(pattern);

    QRegularExpressionMatch rm = rx.match(s);

    if (rm.hasMatch()) {
        if (rx.captureCount() <= 0) return rm.captured(0);
        else                        return rm.captured(1);
    }

    return s;
}

void WaterfallPlotDialog::sortItems()
{
    bool oldState = ui->treeWidgetScans->blockSignals(true);

    ui->treeWidgetScans->sortItems(ui->comboBoxSortColumn->currentIndex(),
                                   ui->comboBoxSortOrder->currentIndex() ? Qt::DescendingOrder : Qt::AscendingOrder);

    ui->treeWidgetScans->blockSignals(oldState);
}

/* here we handle multiple selections */
void WaterfallPlotDialog::itemSelectionChanged()
{
    QList<QTreeWidgetItem*> itms = ui->treeWidgetScans->selectedItems();
    if (itms.size() == 1) return;

    QList<QUuid> idx;

    for (int i = 0; i < itms.size(); ++i) {
        idx.append(itms.at(i)->data(0, Qt::UserRole).toUuid());
    }

    ui->plotWidget->toggleHighlightedScans(idx);
}

/* here we handle single-item clicks */
void WaterfallPlotDialog::itemClicked(QTreeWidgetItem *it, int)
{
    if (ui->treeWidgetScans->selectedItems().size() != 1) return;
    if (it) ui->plotWidget->toggleHighlightedScans(QList<QUuid>() << it->data(0, Qt::UserRole).toUuid());
}

void WaterfallPlotDialog::itemsSorted()
{
    sortItems();
    updateScene();
}

void WaterfallPlotDialog::showContextMenu(const QPoint &p)
{
    twContextMenu->popup(ui->treeWidgetScans->viewport()->mapToGlobal(p));
}

void WaterfallPlotDialog::treeWidgetCheckAllAbove()
{
    QTreeWidgetItem *it = ui->treeWidgetScans->currentItem();

    if (it) {
        int n = ui->treeWidgetScans->indexOfTopLevelItem(it);

        for (int i = 0; i < n; ++i) {
            ui->treeWidgetScans->topLevelItem(i)->setCheckState(0, Qt::Checked);
        }
    }
}

void WaterfallPlotDialog::treeWidgetUncheckAllAbove()
{
    QTreeWidgetItem *it = ui->treeWidgetScans->currentItem();

    if (it) {
        int n = ui->treeWidgetScans->indexOfTopLevelItem(it);

        for (int i = 0; i < n; ++i) {
            ui->treeWidgetScans->topLevelItem(i)->setCheckState(0, Qt::Unchecked);
        }
    }
}

void WaterfallPlotDialog::treeWidgetCheckAllBelow()
{
    QTreeWidgetItem *it = ui->treeWidgetScans->currentItem();

    if (it) {
        int n = ui->treeWidgetScans->indexOfTopLevelItem(it);

        for (int i = n + 1; i < ui->treeWidgetScans->topLevelItemCount(); ++i) {
            ui->treeWidgetScans->topLevelItem(i)->setCheckState(0, Qt::Checked);
        }
    }
}

void WaterfallPlotDialog::treeWidgetUncheckAllBelow()
{
    QTreeWidgetItem *it = ui->treeWidgetScans->currentItem();

    if (it) {
        int n = ui->treeWidgetScans->indexOfTopLevelItem(it);

        for (int i = n + 1; i < ui->treeWidgetScans->topLevelItemCount(); ++i) {
            ui->treeWidgetScans->topLevelItem(i)->setCheckState(0, Qt::Unchecked);
        }
    }
}

void WaterfallPlotDialog::treeWidgetCheckAll()
{
    for (int i = 0; i < ui->treeWidgetScans->topLevelItemCount(); ++i) {
        ui->treeWidgetScans->topLevelItem(i)->setCheckState(0, Qt::Checked);
    }
}

void WaterfallPlotDialog::treeWidgetUncheckAll()
{
    for (int i = 0; i < ui->treeWidgetScans->topLevelItemCount(); ++i) {
        ui->treeWidgetScans->topLevelItem(i)->setCheckState(0, Qt::Unchecked);
    }
}
