/***************************************************************************
                          waterfallplotwidget.cpp  -  description
                             -------------------
    begin                : Tue Aug 03 12:57:00 CEST 2021
    copyright            : (C) 2021 by Nicola Doebelin
    email                : ndoebelin@gmx.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "waterfallplotwidget.h"
#include "../libXrdIO/scanops.h"
#include "../libXrdIO/structs.h"
#include "../libXrdIO/functions.h"
#include "../libXrdIO/bgmnfileio.h"
#include <limits>
#include <QDebug>
#include <QApplication>
#include <QFileInfo>
#include <QBuffer>
#include <QSvgGenerator>

WaterfallPlotWidget::WaterfallPlotWidget(QWidget *parent) : QFrame(parent)
{
    _scanHeap = nullptr;
    _uidList = nullptr;
    _blockUpdate = false;
    _yScale = 1;
    _lutOffset = 0;
    _lutSign = 1;
    _nPoints = 0;
    _smoothInterpol = false;
    _drawGridH = false;
    _lineBuffer = QImage();
    _scaledBuffer = QImage();
    setMouseTracking(true);
}

WaterfallPlotWidget::~WaterfallPlotWidget()
{
}

/************************* Public functions ******************************/

void WaterfallPlotWidget::updateScanData()
{
    if (!_scanHeap) return;
    if (!_uidList)  return;

    _labels.clear();

    for (int i = 0; i < _uidList->size(); ++i) {
        if (!_scanHeap->contains(_uidList->at(i))) continue;

        bool hasLbl;
        QString l = _scanHeap->value(_uidList->at(i)).auxInfo("waterfallPlotLabel", hasLbl).toString();

        if (hasLbl) {
            _labels.append(l);
        } else {
            QFileInfo fi(_scanHeap->value(_uidList->at(i)).sourceFileName());
            _labels.append(fi.fileName());
        }
    }

    if (!_blockUpdate) updateAll();
}

void WaterfallPlotWidget::clearScans()
{
    _labels.clear();
    _lineBuffer = QImage();
    _scaledBuffer = QImage();
    _nPoints = 0;

    if (!_blockUpdate) updateAll();
}

void WaterfallPlotWidget::updateAll()
{
    if (!_scanHeap) return;

    if (_scanHeap->size()) {
        qApp->setOverrideCursor(Qt::WaitCursor);
            calcDrawRegions(this, fontMetrics());
            calcLimits();
            calcXtickMarks();
            calcDisplayYlimits();
            renderImageBuffer();
            scaleDisplayBuffer();
        qApp->restoreOverrideCursor();
    }

    update();
}

void WaterfallPlotWidget::setSmoothInterpolation(bool b)
{
    _smoothInterpol = b;
    if (!_scanHeap) return;

    if (_scanHeap->size() && !_blockUpdate) {
        qApp->setOverrideCursor(Qt::WaitCursor);
            scaleDisplayBuffer();
            update();
        qApp->restoreOverrideCursor();
    }
}

void WaterfallPlotWidget::setDrawGridHorizontal(bool b)
{
    _drawGridH = b;
    if (!_scanHeap) return;

    if (_scanHeap->size() && !_blockUpdate) {
        update();
    }
}

void WaterfallPlotWidget::setDrawGridVertical(bool b)
{
    _drawGridV = b;
    if (!_scanHeap) return;

    if (_scanHeap->size() && !_blockUpdate) {
        update();
    }
}

void WaterfallPlotWidget::setYscale(int i)
{
    _yScale = i;
    if (!_scanHeap) return;

    if (_scanHeap->size() && !_blockUpdate) {
        qApp->setOverrideCursor(Qt::WaitCursor);
            calcDisplayYlimits();
            renderImageBuffer();
            scaleDisplayBuffer();
            update();
        qApp->restoreOverrideCursor();
    }
}

void WaterfallPlotWidget::setXrange(double min, double max)
{
    _wminDisp = qMax(min, _wmin);
    _wmaxDisp = qMin(max, _wmax);
    if (!_scanHeap) return;

    if (_scanHeap->size() && !_blockUpdate) {
        qApp->setOverrideCursor(Qt::WaitCursor);
            calcXtickMarks();
            scaleDisplayBuffer();
            update();
        qApp->restoreOverrideCursor();
    }

}

void WaterfallPlotWidget::setYrange(double min, double max)
{
    _iRangeMinLin = qMax(min, _imin);
    _iRangeMaxLin = qMin(max, _imax);
    if (!_scanHeap) return;

    if (_scanHeap->size() && !_blockUpdate) {
        qApp->setOverrideCursor(Qt::WaitCursor);
            calcDisplayYlimits();
            renderImageBuffer();
            scaleDisplayBuffer();
            update();
        qApp->restoreOverrideCursor();
    }
}

void WaterfallPlotWidget::getXrange(double &min, double &max)
{
    min = _wminDisp;
    max = _wmaxDisp;
}

void WaterfallPlotWidget::getYrange(double &min, double &max)
{
    min = _iRangeMinLin;
    max = _iRangeMaxLin;
}

void WaterfallPlotWidget::getXlimits(double &min, double &max)
{
    min = _wmin;
    max = _wmax;
}

void WaterfallPlotWidget::getYlimits(double &min, double &max)
{
    min = _imin;
    max = _imax;
}

void WaterfallPlotWidget::resetXrange()
{
    _wminDisp = _wmin;
    _wmaxDisp = _wmax;
    emit sigXrangeWasZoomed();

    if (!_scanHeap) return;

    if (_scanHeap->size() && !_blockUpdate) {
        qApp->setOverrideCursor(Qt::WaitCursor);
            calcXtickMarks();
            scaleDisplayBuffer();
            update();
        qApp->restoreOverrideCursor();
    }
}

void WaterfallPlotWidget::resetYrange()
{
    _iRangeMinLin = _imin < 0.0 ? _imin : 0.0;
    _iRangeMaxLin = _imax;
    if (!_scanHeap) return;

    if (_scanHeap->size() && !_blockUpdate) {
        qApp->setOverrideCursor(Qt::WaitCursor);
            calcDisplayYlimits();
            renderImageBuffer();
            scaleDisplayBuffer();
            update();
        qApp->restoreOverrideCursor();
    }
}

void WaterfallPlotWidget::setLut(const colorMaps::Lut &l)
{
    _lut = l;
    setLutOffset();
    if (!_scanHeap) return;

    if (_scanHeap->size() && !_blockUpdate) {
        qApp->setOverrideCursor(Qt::WaitCursor);
            renderImageBuffer();
            scaleDisplayBuffer();
            update();
        qApp->restoreOverrideCursor();
    }
}

void WaterfallPlotWidget::setColorTemp(int i)
{
    _colTemp = i;
    setLutOffset();
    if (!_scanHeap) return;

    if (_scanHeap->size() && !_blockUpdate) {
        qApp->setOverrideCursor(Qt::WaitCursor);
            renderImageBuffer();
            scaleDisplayBuffer();
            update();
        qApp->restoreOverrideCursor();
    }
}

void WaterfallPlotWidget::toggleHighlightedScans(const QList<QUuid> &l)
{
    if (l.size() == 1) {
        if (_hlIdx.contains(l.first())) {
            _hlIdx.clear();
        } else {
            _hlIdx = l;
        }
    } else {
        _hlIdx = l;
    }

    if (!_scanHeap) return;

    if (_scanHeap->size() && !_blockUpdate) {
        update();
    }
}

/************************* Private functions ******************************/

void WaterfallPlotWidget::setLutOffset()
{
    _lutSign = 1;
    _lutOffset = 255;

    if (_lut.data.size() >= 512) {
        if (_colTemp > 0) _lutSign = -1;
    } else {
        if (_colTemp > 0) _lutSign = -1;
        else              _lutOffset = 0;
    }
}

void WaterfallPlotWidget::calcLimits()
{
    if (!_scanHeap) return;
    if (!_uidList)  return;

    if (!_scanHeap->contains(_uidList->first())) {
        _wmin = 0.0;
        _wmax = 1.0;
        _wminDisp = 0.0;
        _wmaxDisp = 1.0;
        _imin = 0.0;
        _imax = 1.0;
        return;
    }

    _wmin = _scanHeap->value(_uidList->first()).pDataAngle().first();
    _wmax = _scanHeap->value(_uidList->first()).pDataAngle().last();
    _wminDisp = _wmin;
    _wmaxDisp = _wmax;
    _imin = std::numeric_limits<double>::max();
    _imax = 0.0;

    double twmin = _wmin;
    double twmax = _wmax;
    double timin = _imin;
    double timax = _imax;

    for (int i = 0; i < _uidList->size(); ++i) {
        if (!_scanHeap->contains(_uidList->at(i))) continue;

        ScanOps::scanMetrics(_scanHeap->value(_uidList->at(i)), twmin, twmax, timin, timax);

        _imin = qMin(_imin, timin);
        _imax = qMax(_imax, timax);
    }

    _iRangeMinLin = _imin < 0.0 ? _imin : 0.0;
    _iRangeMaxLin = _imax;
}

void WaterfallPlotWidget::calcXtickMarks()
{
    _majTicks.clear();
    _minTicks.clear();

    if (!_scanHeap) return;

    if (_scanHeap->size()) {
        _majTicks = global::Functions::scaleAxis1(_wminDisp, _wmaxDisp, 10);
        _minTicks = global::Functions::scaleAxis1(_wminDisp, _wmaxDisp, 60);
    }
}

void WaterfallPlotWidget::calcDisplayYlimits()
{
    if (_yScale == 1 ) {       // sqrt
        _iminDisp = _imin < 0.0 ? 0.0 : sqrt(_imin);
        _imaxDisp = _imax < 0.0 ? 0.0 : sqrt(_imax);
        _iRangeMinDisp = _iRangeMinLin < 0.0 ? 0.0 : sqrt(_iRangeMinLin);
        _iRangeMaxDisp = _iRangeMaxLin < 0.0 ? 0.0 : sqrt(_iRangeMaxLin);
    } else if (_yScale == 2) { // log10
        _iminDisp = _imin < 1.0 ? 0.0 : log10(_imin);
        _imaxDisp = _imax < 1.0 ? 0.0 : log10(_imax);
        _iRangeMinDisp = _iRangeMinLin < 1.0 ? 0.0 : log10(_iRangeMinLin);
        _iRangeMaxDisp = _iRangeMaxLin < 1.0 ? 0.0 : log10(_iRangeMaxLin);
    } else {                   // linear
        _iminDisp = _imin;
        _imaxDisp = _imax;
        _iRangeMinDisp = _iRangeMinLin;
        _iRangeMaxDisp = _iRangeMaxLin;
    }
}

void WaterfallPlotWidget::renderImageBuffer()
{
    if (!_scanHeap) return;
    if (!_uidList)  return;

    if (!_scanHeap->contains(_uidList->first())) {
        _nPoints = 0;
        return;
    }

    _nPoints = _scanHeap->value(_uidList->first()).size();
    _lineBuffer = QImage(_nPoints, _uidList->size(), QImage::Format_RGB888);
    _lineBuffer.fill(Qt::black);

    for (int y = 0; y < _uidList->size(); ++y) {
        if (!_scanHeap->contains(_uidList->at(y))) continue;
        renderLine(_scanHeap->value(_uidList->at(y)), _lineBuffer, y);
    }
}

void WaterfallPlotWidget::renderLine(const Scan &sc, QImage &img, int y)
{
    if (img.width() < 1) return;

    int imgSz = img.width();
    int scanSz = sc.size();
    double tarWmi = sc.minAngle();
    double tarWma = sc.maxAngle();

    for (int i = 0; i < imgSz; ++i) {
        int x = pixelToIndex(i, imgSz, _wmin, _wmax, scanSz, tarWmi, tarWma);
        if ((x < 0) || (x >= scanSz)) continue;
        img.setPixel(i, y, pixelValue(sc.intensity(x)));
    }
}

void WaterfallPlotWidget::scaleDisplayBuffer()
{
    if ((_lineBuffer.isNull() || (qFuzzyCompare(_wminDisp, _wmaxDisp)))) {
        _scaledBuffer = QImage();
        return;
    }

    int pxMin = angleToIndex(_wminDisp);
    int pxMax = angleToIndex(_wmaxDisp);
    if ((pxMin < 0) || (pxMax < 0)) return;

    QRect rc(QPoint(pxMin, 0), QPoint(pxMax, _lineBuffer.height() - 1));

    if (_smoothInterpol) {
        _scaledBuffer = _lineBuffer.copy(rc)
                                   .scaled(_imageRect.width(), _lineBuffer.height(), Qt::IgnoreAspectRatio, Qt::FastTransformation)
                                   .scaled(_imageRect.width(), _imageRect.height(), Qt::IgnoreAspectRatio, Qt::SmoothTransformation);
    } else {
        _scaledBuffer = _lineBuffer.copy(rc)
                                   .scaled(_imageRect.width(), _imageRect.height(), Qt::IgnoreAspectRatio, Qt::FastTransformation);
    }
}

void WaterfallPlotWidget::drawFileNames(QPainter &p)
{
    if (_labels.isEmpty()) return;

    double step    = double(_fileNameRect.height()) / double(_labels.size());
    double yOffset = double(_fileNameRect.top()) + 0.5 * (step + double(fontMetrics().ascent()));
    double lastY   = 0.0;

    for (int i = 0; i < _labels.size(); ++i) {
        double y = yOffset + step * double(i);

        if (y - lastY > fontMetrics().height()) {
            p.drawText(_fileNameRect.x(), int(0.5 + y), _labels.at(i));
            lastY = y;
        }
    }
}

void WaterfallPlotWidget::drawXaxis(QPainter &p)
{
    int s = fontMetrics().lineSpacing();
    QString l = QString("Diffraction Angle [%1%2%3]").arg(global::degree).arg(2).arg(global::theta);

    int tBase    = _xaxisRect.top() + s;
    int tTopMin  = _xaxisRect.top() + int(2.0 * s / 3.0);
    int tTopMaj  = _xaxisRect.top() + int(1.0 * s / 3.0);
    int lBase    = _xaxisRect.top() + 2*s;
    int nBase    = _xaxisRect.top() + 4*s;

    for (int i = 0; i < _minTicks.size(); ++i) {
        int x = _imageRect.left() + int(0.5 + _imageRect.width() * (_minTicks.at(i) - _wminDisp) / (_wmaxDisp - _wminDisp));
        p.drawLine(x, tBase, x, tTopMin);
    }

    for (int i = 0; i < _majTicks.size(); ++i) {
        int x = _imageRect.left() + int(0.5 + _imageRect.width() * (_majTicks.at(i) - _wminDisp) / (_wmaxDisp - _wminDisp));
        p.drawLine(x, tBase, x, tTopMaj);
        QString tLbl = QString("%1").arg(_majTicks.at(i), 0, 'f', 2);
        p.drawText(x - fontMetrics().width(tLbl)/2, lBase, tLbl);
    }

    p.drawLine(_xaxisRect.left(), tBase, _xaxisRect.right(), tBase);
    p.drawText(_xaxisRect.left() + _xaxisRect.width() / 2 - fontMetrics().width(l) / 2, nBase, l);
}

void WaterfallPlotWidget::drawGrid(QPainter &p)
{
    if (!_drawGridH && !_drawGridV) return;

    QPen pen(Qt::white);
    pen.setStyle(Qt::DashLine);
    p.setPen(pen);

    QPainter::CompositionMode cpm = p.compositionMode();
    p.setCompositionMode(QPainter::CompositionMode_Difference);

    if (_drawGridH) {
        double step = double(_imageRect.height()) / double(_labels.size());
        double yOffset = double(_imageRect.top()) + 0.5 * step;

        for (int i = 0; i < _labels.size(); ++i) {
            double y = yOffset + step * i;
            p.drawLine(_imageRect.x(), int(0.5 + y), _imageRect.x() + _imageRect.width(), y);
        }
    }

    if (_drawGridV) {
        for (int i = 0; i < _majTicks.size(); ++i) {
            int x = _imageRect.left() + int(0.5 + _imageRect.width() * (_majTicks.at(i) - _wminDisp) / (_wmaxDisp - _wminDisp));
            p.drawLine(x, _imageRect.top(), x, _imageRect.bottom());
        }
    }

    p.setCompositionMode(cpm);
}

void WaterfallPlotWidget::drawHighlight(QPainter &p)
{
    if (!_hlIdx.size()) return;

    QPen pen(Qt::white);
    pen.setStyle(Qt::DashLine);
    p.setPen(pen);

    QPainter::CompositionMode cpm = p.compositionMode();
    p.setCompositionMode(QPainter::CompositionMode_Difference);

    double step = double(_imageRect.height()) / double(_labels.size());
    double yOffset = double(_imageRect.top()) + 0.5 * step;
    int w = _imageRect.width() - 1;
    int h = _smoothInterpol ? 1 : int(0.5 + 0.5 * step);
    int x = _imageRect.x();

    for (int i = 0; i < _hlIdx.size(); ++i) {
        int n = _uidList->indexOf(_hlIdx.at(i));

        if (n >= 0) {
            int y = int(0.5 + yOffset + step * n);
            p.drawRect(x, y - h, w, 2 * h);
        }
    }

    p.setCompositionMode(cpm);
}

void WaterfallPlotWidget::calcDrawRegions(const QPaintDevice *pd, const QFontMetrics &fm)
{
    const QPaintDevice *canvas = pd ? pd : this;

    int margin = fm.width("M");
    int fnh = canvas->height() - 4 * fm.lineSpacing() - margin;
    int fnl = 0;

    for (int i = 0; i < _labels.size(); ++i) {
        fnl = qMax(fnl, fm.width(_labels.at(i)) + margin);
    }

    _fileNameRect = QRect(QPoint(margin,       margin), QPoint(fnl + margin,          fnh));
    _xaxisRect    = QRect(QPoint(margin + fnl, fnh),    QPoint(canvas->width() - margin, canvas->height() - margin));
    _imageRect    = QRect(QPoint(margin + fnl, margin), QPoint(canvas->width() - margin, fnh));
}

QRgb WaterfallPlotWidget::pixelValue(double v)
{
    double val = v;

    if      (_yScale == 1) val = val < 0.0 ? 0.0 : sqrt(val);
    else if (_yScale == 2) val = val < 1.0 ? 0.0 : log10(val);

    int c = _lutOffset + int(_lutSign * 255.0 * (val - _iRangeMinDisp) / (_iRangeMaxDisp - _iRangeMinDisp) + 0.5);

    if (c >= _lut.data.size()) c = _lut.data.size() - 1;
    else if (c < 0)            c = 0;

    return _lut.data.at(c);
}

/* returns the vector index in tar matching the angle at index sourcePx in source */
int WaterfallPlotWidget::pixelToIndex(int sourcePx, int sourceSz, double sourceWmi, double sourceWma, int tarSz, double tarWmi, double tarWma)
{
    double w = sourceWmi + (sourceWma - sourceWmi) * double(sourcePx) / double(sourceSz);
    return int(0.5 + double(tarSz) * (w - tarWmi) / (tarWma - tarWmi));
}

/* screen pixel to angle */
double WaterfallPlotWidget::pixelToAngle(int i)
{
    if (_imageRect.width() == 0) return 0.0;
    return _wminDisp + (_wmaxDisp - _wminDisp) * double(i - _imageRect.x()) / double(_imageRect.width());
}

int WaterfallPlotWidget::angleToIndex(double d)
{
    if (qFuzzyCompare(_wminDisp, _wmaxDisp)) return -1;
    return int(0.5 + double(_nPoints) * (d - _wmin) / (_wmax - _wmin));
}

void WaterfallPlotWidget::paintEvent(QPaintEvent *)
{
    QPainter p(this);
    p.fillRect(this->rect(), QBrush(QGuiApplication::palette().color(QPalette::Base)));

    if (!_scaledBuffer.isNull()) {
        p.drawImage(_imageRect, _scaledBuffer);
        drawFileNames(p);
        drawXaxis(p);
        drawGrid(p);
        drawHighlight(p);
    }

    p.end();
}

void WaterfallPlotWidget::resizeEvent(QResizeEvent *)
{
    if (!_scanHeap) return;

    if (_scanHeap->size()) {
        calcDrawRegions(this, fontMetrics());
        scaleDisplayBuffer();
        update();
    }
}

void WaterfallPlotWidget::mouseMoveEvent(QMouseEvent *e)
{
    if (!_imageRect.contains(e->pos())) return;

    if (e->buttons() & Qt::LeftButton) {
        double a = _wminDispClicked + (_wmaxDispClicked - _wminDispClicked) * double(e->pos().x() - _imageRect.x()) / double(_imageRect.width());
        double da = a - _clickPos;

        if (_wminDispClicked - da <  _wmin) da = _wminDispClicked - _wmin;
        if (_wmaxDispClicked - da >= _wmax) da = _wmaxDispClicked - _wmax;

        _wminDisp = _wminDispClicked - da;
        _wmaxDisp = _wmaxDispClicked - da;

        calcXtickMarks();
        scaleDisplayBuffer();
        update();
        emit sigXrangeWasZoomed();
    } else {
        QString s;

        if (_labels.size()) {
            int fnIdx = int(_labels.size() * double(e->pos().y() - _imageRect.y()) / double(_imageRect.height()));
            if (fnIdx < 0) fnIdx = 0;
            s = fnIdx >= _labels.size() ? _labels.last() : _labels.at(fnIdx);
        }

        emit sigMouseCoordinates(QString("%1: 2%2 = %3").arg(s).arg(global::theta).arg(pixelToAngle(e->pos().x()), 0, 'f', 4));
    }
}

void WaterfallPlotWidget::mousePressEvent(QMouseEvent *e)
{
    if (e->button() == Qt::LeftButton) {
        _clickPos = pixelToAngle(e->pos().x());
        _wminDispClicked = _wminDisp;
        _wmaxDispClicked = _wmaxDisp;
    } else if (e->button() == Qt::RightButton) {
        resetXrange();
    }
}

void WaterfallPlotWidget::mouseReleaseEvent(QMouseEvent *e)
{
    if (e->button() == Qt::LeftButton) {
        _clickPos = pixelToAngle(e->pos().x());
    }
}

void WaterfallPlotWidget::wheelEvent(QWheelEvent *e)
{
#if QT_VERSION >= QT_VERSION_CHECK(5, 14, 0)
    if (!_imageRect.contains(e->position().toPoint())) return;
#else
    if (!_imageRect.contains(e->pos())) return;
#endif

    if (e->modifiers().testFlag(Qt::NoModifier)) {
#if QT_VERSION >= QT_VERSION_CHECK(5, 14, 0)
        zoomStep(e->position().toPoint().x(), e->angleDelta().y());
#else
        zoomStep(e->pos().x(), e->angleDelta().y());
#endif
        e->accept();
        return;
    }
}

void WaterfallPlotWidget::zoomStep(int x, int n)
{
    double pAng = pixelToAngle(x);

    double dMin = pAng - _wminDisp;
    double dMax = _wmaxDisp - pAng;

    if (n > 0) {
        _wminDisp = qMax(_wmin, pAng - 0.9 * dMin);
        _wmaxDisp = qMin(_wmax, pAng + 0.9 * dMax);
    } else {
        _wminDisp = qMax(_wmin, pAng - 1.1 * dMin);
        _wmaxDisp = qMin(_wmax, pAng + 1.1 * dMax);
    }

    calcXtickMarks();
    scaleDisplayBuffer();
    update();
    emit sigXrangeWasZoomed();
}

void WaterfallPlotWidget::save(const QString &s, int w, int h)
{
    QFileInfo fi(s);
    if (fi.suffix().toLower() == "png") renderPixmap(fi.absoluteFilePath(), w, h);
    if (fi.suffix().toLower() == "svg") renderSvg(fi.absoluteFilePath());
    if (fi.suffix().toLower() == "grd") exportTextGrid(fi.absoluteFilePath());
}

void WaterfallPlotWidget::renderPixmap(const QString &s, int w, int h)
{
    QPixmap pix(w, h);
    calcDrawRegions(&pix, fontMetrics());

    QPainter p(&pix);
    p.setRenderHints(QPainter::Antialiasing | QPainter::TextAntialiasing | QPainter::SmoothPixmapTransform, true);
    p.fillRect(pix.rect(), QBrush(Qt::white));

    if (!_scaledBuffer.isNull()) {
        p.drawImage(_imageRect, _scaledBuffer);
        drawFileNames(p);
        drawXaxis(p);
        if (_drawGridH) drawGrid(p);
    }

    p.end();
    calcDrawRegions(this, fontMetrics());

    if (pix.save(s, "PNG")) {
        qDebug() << QString("WaterfallPlotWidget::renderPixmap: Image saved to %1").arg(s);
    } else {
        qDebug() << QString("WaterfallPlotWidget::renderPixmap: Could not save image to %1").arg(s);
    }
}

void WaterfallPlotWidget::renderSvg(const QString &s)
{
    QBuffer buf;
    buf.open(QBuffer::ReadWrite);
    QSvgGenerator svg;
    svg.setOutputDevice(&buf);

    // using an aspect ratio of sqrt(2) : 1
    double w = 27.0 * double(svg.resolution()) / 2.54;
    double h = w / sqrt(2.0);
    QRectF pageRect(0.0, 0.0, w, h);

    svg.setSize(QSize(int(w), int(h)));
    svg.setViewBox(pageRect);

    QPainter p(&svg);
    p.setRenderHints(QPainter::Antialiasing | QPainter::TextAntialiasing | QPainter::SmoothPixmapTransform, true);
    QFont ft = font();
    ft.setPointSize(12);
    p.setFont(ft);
    calcDrawRegions(&svg, p.fontMetrics());

    if (!_scaledBuffer.isNull()) {
        p.drawImage(_imageRect, _scaledBuffer);
        drawFileNames(p);
        drawXaxis(p);
        if (_drawGridH) drawGrid(p);
    }

    p.end();

    QFile f(s);

    if (f.open(QIODevice::WriteOnly | QIODevice::Truncate))     {
        QTextStream str(&f);
        buf.seek(0);
        str << buf.readAll();
        f.close();
    }

    buf.close();

    calcDrawRegions(this, fontMetrics());
}

void WaterfallPlotWidget::exportTextGrid(const QString &f)
{
    if (_nPoints <= 1) {
        qDebug() << QString("WaterfallPlotWidget::exportTextGrid(): No data available. Cannot export to Surfer 6 Text Grid file.");
        return;
    }

    QString s("DSAA\n");

    int pxMin = angleToIndex(_wminDisp);
    int pxMax = angleToIndex(_wmaxDisp);
    if ((pxMin < 0) || (pxMax < 0)) return;

    int nx = pxMax - pxMin;
    int ny = _uidList->size();

    double xlo = _wminDisp;
    double xhi = _wmaxDisp;
    double ylo = 0.0;
    double yhi = double(_uidList->size() - 1);
    double zlo = 0.0;
    double zhi = _imax;

    s += QString("%1 %2\n").arg(nx).arg(ny);
    s += QString("%1 %2\n").arg(xlo, 0, 'f', 6).arg(xhi, 0, 'f', 6);
    s += QString("%1 %2\n").arg(ylo, 0, 'f', 6).arg(yhi, 0, 'f', 6);
    s += QString("%1 %2\n").arg(zlo, 0, 'f', 6).arg(zhi, 0, 'f', 6);

    for(int i = 0; i < _uidList->size(); ++i) {
        if (!_scanHeap->contains(_uidList->at(i))) continue;

        QStringList l;

        for (int j = 0; j < nx; ++j) {
            double ang = _wminDisp + (double(j) / double(nx - 1)) * (_wmaxDisp - _wminDisp);
            double intens = _scanHeap->value(_uidList->at(i)).intensity(ang);

            l.append(QString("%1").arg(intens, 0, 'f', 2));
        }

        s += l.join(" ") + "\n";
    }

    BgmnFileIO::writeTextFile(f, s);
    qDebug() << QString("WaterfallPlotWidget::exportTextGrid(): Data exported to %1 (%2 characters in string)").arg(f).arg(s.count());
}
